// ============================================================================
// Project:	Deductive Database
// Filename:	bench.h
// Purpose:	Superclass for Standard Benchmark with load() and eval()
// Last Change:	17.05.2016
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2016 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Include File Frame:
//=============================================================================

#ifndef	BENCH_INCLUDED
#define BENCH_INCLUDED

//=============================================================================
// Used Types and Macros:
//=============================================================================

#ifndef VER_INCLUDED
#include "ver.h"
#endif

#ifndef STR_INCLUDED
#include "str.h"
#endif

#ifndef CHECK_INCLUDED
#include "check.h"
#endif


//=============================================================================
// Private Constants:
//=============================================================================

//-----------------------------------------------------------------------------
// BENCH_MAGIC: Magic number (identifies objects of this class).
//-----------------------------------------------------------------------------

static const long BENCH_MAGIC = 0x42454E0AL;	// 'BEN\n'


//=============================================================================
// Class Declaration:
//=============================================================================

/** \brief Abstract superclass for Standard Benchmarks
 *		with load() and eval() Methods.
 *  @author Stefan Brass
 *  @date   17.05.2016
 */

class bench_c {
	public:

//-----------------------------------------------------------------------------
// Constructor, Destructor:
//-----------------------------------------------------------------------------

	// Constructor:
	bench_c();

	// Destructor:
	~bench_c();

//-----------------------------------------------------------------------------
// Abstract Methods Required by this Class:
//-----------------------------------------------------------------------------

	/** Load data file with EDB facts.
	 * @filename Name of the data file.
	 * @return   true if successful, false if error detected.
	 */

	virtual bool load(str_t filename) = 0;

	// Do bottom-up evaluation.
	virtual long eval() = 0;

	// name: Name of this benchmark.
	virtual str_t name() const = 0;

	// file: List of data files.
	virtual str_t file(int i) const = 0;

	// result: List of correct results of eval method.
	virtual long result(int i) const = 0;


//-----------------------------------------------------------------------------
// Debugging Support:
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
	public:
	// Integrity check:
	str_t	check() const;

	private:
	// Magic number (identifies objects of this class for debugging).
	long	Magic;		// Must be "BENCH_MAGIC".
#endif

#ifdef	VER_DUMP
	public:
	// Display data structure:
	virtual void dump(str_t headline = STR_NULL) const;
#endif

//-----------------------------------------------------------------------------
// Copy-constructor and assignment operator are not supported for this class:
//-----------------------------------------------------------------------------

	private:

	bench_c(const bench_c& obj);		// Not implemented
	bench_c& operator=(const bench_c& obj);	// Not implemented


//=============================================================================
// End of Class:
//=============================================================================

};

//-----------------------------------------------------------------------------
// Define pointer type:
//-----------------------------------------------------------------------------

typedef bench_c *bench_t;

//-----------------------------------------------------------------------------
// Define null pointer:
//-----------------------------------------------------------------------------

#define BENCH_NULL (static_cast<bench_t>(0))

//=============================================================================
// End of Include File:
//=============================================================================

#endif

