// ============================================================================
// Project:	Deductive Database
// Filename:	bench_dblp_1.cpp
// Purpose:	Bottom-Up EValuation for DBLP Example
// Last Change:	06.10.2016
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2015-2016 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"
#include "str.h"
#include "err.h"
#include "check.h"
#include "perf.h"
#include "stack.h"
#include "atomtab.h"
#include "strtab.h"
#include "list_2.h"
#include "cur_2.h"
#include "rel_n_n.h"
#include "cur_n_n.h"
#include "load_dblp.h"
#include "out.h"

#include "bench_dblp_1.h"


//=============================================================================
// Static Methods:
//=============================================================================


//-----------------------------------------------------------------------------
// run: Load data file, do bottom-up evaluation for DBLP Example.
//-----------------------------------------------------------------------------

long bench_dblp_1_c::run(str_t filename,
			perf_t timer_load, perf_t timer_eval)
{

	// Counter for number of results:
	long num_results = 0;

	// --------------------------------------------------------------------
	// Domains:
	// --------------------------------------------------------------------

	// Make string table for first argument:
	strtab_c doc_ids("String Table for Arg.1");

	// Make atom table for second argument:
	atomtab_c attrs("Atom Table for Arg.2");

	// Construct atoms which are relevant for selection condition:
	int title  = attrs.put("title");
	int year   = attrs.put("year");
	int author = attrs.put("author");
	int month  = attrs.put("month");

	// Make string table for third argument:
	strtab_c str_values("String Table for Arg.3");

	// --------------------------------------------------------------------
	// Relations for EDB body literals:
	// --------------------------------------------------------------------

	list_2_c  att_title("ATT_TITLE");
	rel_n_n_c att_year("ATT_YEAR");
	rel_n_n_c att_author("ATT_AUTHOR");
	rel_n_n_c att_month("ATT_MONTH");

	// --------------------------------------------------------------------
	// Load data file, fill domains and EDB relations:
	// --------------------------------------------------------------------

	// Start timer for data load:
	timer_load->start();

	// Load the data:
	if(!load_dblp_c::load(filename,
			&doc_ids, &attrs, &str_values,
			title, year, author, month,
			&att_title, &att_year, &att_author, &att_month))
		return -1;

	// Stop timer for data load:
	timer_load->stop();

	//---------------------------------------------------------------------
	// eval: Do bottom-up evaluation.
	//---------------------------------------------------------------------

	// Start timer for bottom-up evaluation:
	timer_eval->start();

	// --------------------------------------------------------------------
	// Transformed program:
	//
	// 1: answer(A,B,C,D,E) :-
	//		queryrule(A,B,C,D,E).
	// 2: queryrule(A,B,C,D,E) :-
	//		att(A,title,B),
	//		att(A,year,C),
	//		att(A,author,D),
	//		att(A,month,E).
	// --------------------------------------------------------------------

	// Declaration Part:

	int v2_1 = -1;
	int v2_2 = -1;
	int v2_3 = -1;
	int v2_4 = -1;
	int v2_5 = -1;

	// att_cursor_fbf c2_1_fbf;
	// att_cursor_bbf c2_2_bbf;
	// att_cursor_bbf c2_3_bbf;
	// att_cursor_bbf c2_4_bbf;

	cur_2_c  c2_1_fbf(&att_title);	// att(A,title,B),
	cur_n_n_c c2_2_bbf(&att_year);	// att(A,year,C),
	cur_n_n_c c2_3_bbf(&att_author);// att(A,author,D),
	cur_n_n_c c2_4_bbf(&att_month);	// att(A,month,E).

	// int_stack backtrack_stack;
	// int_stack int_value_stack;
	// float_stack float_value_stack;
	// string_stack string_value_stack;

	stack_c<int> backtrack_stack("Backtrack Stack");

	// --------------------------------------------------------------------
	// Jump labels:
	// label 1: applying rule 2 resulting in symbolic fact
	//	queryrule(v2_1,v2_2,v2_3,v2_4,v2_5)
	// label 2: using symbolic fact
	//	queryrule(v2_1,v2_2,v2_3,v2_4,v2_5)
	//	with literal 1 in rule 1 resulting in symbolic fact
	//	answer(v2_1,v2_2,v2_3,v2_4,v2_5)
	// --------------------------------------------------------------------

	// Initialisation Part:
	backtrack_stack.push(1);

	// Loop Part:
	while(!backtrack_stack.is_empty()) {
		switch(backtrack_stack.pop()) {
		case 1:
			// c2_1_fbf.open("title");
			c2_1_fbf.open();
			while (c2_1_fbf.fetch()) {
				// c2_2_bbf.open(c2_1_fbf.c1(), "year");
				c2_2_bbf.open(c2_1_fbf.c1());
				while (c2_2_bbf.fetch()) {
					//c2_3_bbf.open(c2_1_fbf.c1(),
					//			"author");
					c2_3_bbf.open(c2_1_fbf.c1());
					while (c2_3_bbf.fetch()) {
						//c2_4_bbf.open(
						//  c2_1_fbf.c1(), "month");
						c2_4_bbf.open(c2_1_fbf.c1());
			    // Continuation for label 1
		case 4:
			while (c2_4_bbf.fetch()) {
				v2_1 = c2_1_fbf.c1();
				v2_2 = c2_1_fbf.c2();
				v2_3 = c2_2_bbf.c2();
				v2_4 = c2_3_bbf.c2();
				v2_5 = c2_4_bbf.c2();
				// Push backtrack label to continuation
				backtrack_stack.push(4);
				// Labels for next rules
				goto label_2;
			}

					}
				}
			}
			break;
	    
		case 2:
		label_2:
			// queryrule(v2_1,v2_2,v2_3,v2_4,v2_5)
			if(v2_1 < 0 || v2_2 < 0 || v2_3 < 0 || v2_4 < 0 ||
								v2_5 < 0)
					return -1;
			num_results++;
			out_c::answer_5s(num_results,
					doc_ids.get(v2_1),
					str_values.get(v2_2),
					str_values.get(v2_3),
					str_values.get(v2_4),
					str_values.get(v2_5));
			break;
	    
	    
		}
	}

	// Stop timer for bottom-up evaluation:
	timer_eval->stop();

	// Print table of memory pools if requested:
	out_c::mem_pools_tab();

	// Dump data structures if requested:
	if(out_c::show_dump()) {
		doc_ids.dump("String Table for Arg.1");
		attrs.dump("Atom Table for Arg.2");
		str_values.dump("String Table for Arg.3");
		att_title.dump("ATT_TITLE [list_2_c]");
		att_year.dump("ATT_YEAR [rel_n_n_c]");
		att_author.dump("ATT_AUTHOR [rel_n_n_c]");
		att_month.dump("ATT_MONTH [rel_n_n_c]");
	}

	// Print relation statistics if requested:
	if(out_c::show_rel()) {
		out_c::rel_size_headline();
		out_c::rel_size(&att_title);
		out_c::rel_size(&att_year);
		out_c::rel_size(&att_author);
		out_c::rel_size(&att_month);
		out_c::rel_size_strtab(&doc_ids);
		out_c::rel_size_atomtab(&attrs);
		out_c::rel_size_strtab(&str_values);
		out_c::rel_size_done();
	}

	// Return number of result tuples:
	return num_results;
}


//-----------------------------------------------------------------------------
// file: List of data files.
//-----------------------------------------------------------------------------

str_t bench_dblp_1_c::file(int i)
{
	switch(i) {
		case 0:
			return VER_FILE_DBLP_TEST;
		case 1:
			return VER_FILE_DBLP;
		case 2:
			return VER_FILE_DBLP_BIG;
		default:
			return STR_NULL;
	}
}


//-----------------------------------------------------------------------------
// result: List of correct results of eval method.
//-----------------------------------------------------------------------------

long bench_dblp_1_c::result(int i)
{
	switch(i) {
		case 0: return 7;
		case 1: return 4035;
		case 2: return -1;
		default: return -1;
	}
}

