// ============================================================================
// Project:	Deductive Database
// Filename:	bench_dblp_2.cpp
// Purpose:	Bottom-Up Evaluation for DBLP Example, Ver 2: Cursor interface
// Last Change:	06.10.2016
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2015-2016 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"
#include "str.h"
#include "err.h"
#include "check.h"
#include "stack.h"
#include "atomtab.h"
#include "strtab.h"
#include "list_2.h"
#include "cur_2.h"
#include "rel_n_n.h"
#include "cur_n_n.h"
#include "load_dblp.h"
#include "out.h"

#include "bench_dblp_2.h"


//=============================================================================
// Constructor, Destructor:
//=============================================================================

//-----------------------------------------------------------------------------
// Constructor (initializes subobjects):
//-----------------------------------------------------------------------------

bench_dblp_2_c::bench_dblp_2_c() :

		// Superclass:
		bench_c(),

		// Domains:
		doc_ids("String Table for Arg.1"),
		attrs("Atom Table for Arg.2"),
		str_values("String Table for Arg.3"),

		// Relations for EDB body literals:
		att_title("ATT_TITLE"),
		att_year("ATT_YEAR"),
		att_author("ATT_AUTHOR"),
		att_month("ATT_MONTH"),

		// Cursors for loops during bottom-up evaluation:
		c2_1_fbf(&att_title),
		c2_2_bbf(&att_year),
		c2_3_bbf(&att_author),
		c2_4_bbf(&att_month),

		// Backtrack stack for bottom-up evaluation:
		backtrack_stack("Backtrack Stack")

{
	// Construct atoms which are relevant for selection condition:
	title  = attrs.put("title");
	year   = attrs.put("year");
	author = attrs.put("author");
	month  = attrs.put("month");

	// Initialize variables for bottom-up evaluation:
	v2_1 = -1;
	v2_2 = -1;
	v2_3 = -1;
	v2_4 = -1;
	v2_5 = -1;

	// Set Magic number:
	CHECK_CODE(Magic = BENCH_DBLP_2_MAGIC);
}

//-----------------------------------------------------------------------------
// Destructor:
//-----------------------------------------------------------------------------

bench_dblp_2_c::~bench_dblp_2_c()
{
	// Check this object:
	CHECK_VALID("bench_dblp_2_c::destructor");

	// Make this object invalid:
	CHECK_CODE(Magic = 0);
}


//=============================================================================
// Public Methods:
//=============================================================================

//-----------------------------------------------------------------------------
// load: Load data file with EDB facts.
//-----------------------------------------------------------------------------

bool bench_dblp_2_c::load(str_t filename) {

	// Check parameter:
	CHECK(filename != STR_NULL, "bench_dblp_2_c::load: filename is null");

	// Check object:
	CHECK_VALID("bench_dblp_2_c::load");

	// Load the data:
	return load_dblp_c::load(filename,
			&doc_ids, &attrs, &str_values,
			title, year, author, month,
			&att_title, &att_year, &att_author, &att_month);
}

//-----------------------------------------------------------------------------
// eval: Do bottom-up evaluation.
//-----------------------------------------------------------------------------

long bench_dblp_2_c::eval()
{

	// number of answers found:
	long num_results = 0;

	// Open Cursor:
	open();

	// Loop while results are found:
	while(fetch()) {

		// Get results:
		str_t ans_id     = c1();
		str_t ans_title  = c2();
		str_t ans_author = c3();
		str_t ans_year   = c4();
		str_t ans_month  = c5();

		// Check whether results are ok:
		if(ans_id == STR_NULL)
			return -1;
		if(ans_title == STR_NULL)
			return -1;
		if(ans_author == STR_NULL)
			return -1;
		if(ans_year == STR_NULL)
			return -1;
		if(ans_month == STR_NULL)
			return -1;

		// Increment number of answers found:
		num_results++;

		// Print the result if below limit:
		out_c::answer_5s(num_results, ans_id, ans_title, ans_author,
						ans_year, ans_month);

	// End of fetch loop:
	}
	close();

	// Print relation statistics if requested:
	if(out_c::show_rel()) {
		out_c::rel_size_headline();
		out_c::rel_size(&att_title);
		out_c::rel_size(&att_year);
		out_c::rel_size(&att_author);
		out_c::rel_size(&att_month);
		out_c::rel_size_strtab(&doc_ids);
		out_c::rel_size_atomtab(&attrs);
		out_c::rel_size_strtab(&str_values);
		out_c::rel_size_done();
	}

	// Return number of result tuples:
	return num_results;
}


//-----------------------------------------------------------------------------
// Open cursor (note: load() must be called before this method).
//-----------------------------------------------------------------------------

void bench_dblp_2_c::open()
{
	// --------------------------------------------------------------------
	// Transformed program:
	//
	// 1: answer(A,B,C,D,E) :-
	//		queryrule(A,B,C,D,E).
	// 2: queryrule(A,B,C,D,E) :-
	//		att(A,title,B),
	//		att(A,year,C),
	//		att(A,author,D),
	//		att(A,month,E).
	// --------------------------------------------------------------------


	// --------------------------------------------------------------------
	// Jump labels:
	// label 1: applying rule 2 resulting in symbolic fact
	//	queryrule(v2_1,v2_2,v2_3,v2_4,v2_5)
	// label 2: using symbolic fact
	//	queryrule(v2_1,v2_2,v2_3,v2_4,v2_5)
	//	with literal 1 in rule 1 resulting in symbolic fact
	//	answer(v2_1,v2_2,v2_3,v2_4,v2_5)
	// --------------------------------------------------------------------

	// Initialisation Part:
	backtrack_stack.push(1);
}

//-----------------------------------------------------------------------------
// fetch: Get next tuple, returns false if there is no further tuple.
//-----------------------------------------------------------------------------

bool bench_dblp_2_c::fetch()
{

	// Loop Part:
	while(!backtrack_stack.is_empty()) {
		switch(backtrack_stack.pop()) {
		case 1:
			// c2_1_fbf.open("title");
			c2_1_fbf.open();
			while (c2_1_fbf.fetch()) {
				// c2_2_bbf.open(c2_1_fbf.c1(), "year");
				c2_2_bbf.open(c2_1_fbf.c1());
				while (c2_2_bbf.fetch()) {
					//c2_3_bbf.open(c2_1_fbf.c1(),
					//			"author");
					c2_3_bbf.open(c2_1_fbf.c1());
					while (c2_3_bbf.fetch()) {
						//c2_4_bbf.open(
						//  c2_1_fbf.c1(), "month");
						c2_4_bbf.open(c2_1_fbf.c1());
			    // Continuation for label 1
		case 4:
			while (c2_4_bbf.fetch()) {
				v2_1 = c2_1_fbf.c1();
				v2_2 = c2_1_fbf.c2();
				v2_3 = c2_2_bbf.c2();
				v2_4 = c2_3_bbf.c2();
				v2_5 = c2_4_bbf.c2();
				// Push backtrack label to continuation
				backtrack_stack.push(4);
				// Labels for next rules
				goto label_2;
			}

					}
				}
			}
			break;
	    
		case 2:
		label_2:
			// queryrule(v2_1,v2_2,v2_3,v2_4,v2_5)
			return true;
	    
	    
		}
	}

	// Backtrack stack was empty, no more answers:
	return false;
}


//-----------------------------------------------------------------------------
// file: List of data files.
//-----------------------------------------------------------------------------

str_t bench_dblp_2_c::file(int i) const
{
	switch(i) {
		case 0:
			return VER_FILE_DBLP_TEST;
		case 1:
			return VER_FILE_DBLP;
		case 2:
			return VER_FILE_DBLP_BIG;
		default:
			return STR_NULL;
	}
}


//-----------------------------------------------------------------------------
// result: List of correct results of eval method.
//-----------------------------------------------------------------------------

long bench_dblp_2_c::result(int i) const
{
	switch(i) {
		case 0: return 7;
		case 1: return 4035;
		case 2: return -1;
		default: return -1;
	}
}


//=============================================================================
// Debugging Support:
//=============================================================================


//-----------------------------------------------------------------------------
// check: Check the integrity of this object, return error message or NULL.
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
str_t	bench_dblp_2_c::check() const
{
	// Check magic number:
	if(Magic != BENCH_DBLP_2_MAGIC)
		return "wrong magic number";

	// Check sub-objects (relations for EDB predicates):
	CHECK_SUBOBJ(doc_ids, "bench_dblp_2_c::check");
	CHECK_SUBOBJ(attrs, "bench_dblp_2_c::check");
	CHECK_SUBOBJ(str_values, "bench_dblp_2_c::check");
	CHECK_SUBOBJ(att_title, "bench_dblp_2_c::check");
	CHECK_SUBOBJ(att_year, "bench_dblp_2_c::check");
	CHECK_SUBOBJ(att_author, "bench_dblp_2_c::check");
	CHECK_SUBOBJ(att_month, "bench_dblp_2_c::check");
	CHECK_SUBOBJ(c2_1_fbf, "bench_dblp_2_c::check");
	CHECK_SUBOBJ(c2_2_bbf, "bench_dblp_2_c::check");
	CHECK_SUBOBJ(c2_3_bbf, "bench_dblp_2_c::check");
	CHECK_SUBOBJ(c2_4_bbf, "bench_dblp_2_c::check");
	CHECK_SUBOBJ(backtrack_stack, "bench_dblp_2_c::check");

	// Atom IDs must be non-negative:
	if(title < 0)
		return "title is negative";
	if(year < 0)
		return "year is negative";
	if(author < 0)
		return "author is negative";
	if(month < 0)
		return "month is negative";

	// Variables are non-negative once set, but initialized to -1:
	if(v2_1 < -1)
		return "v2_1 is invalid";
	if(v2_2 < -1)
		return "v2_2 is invalid";
	if(v2_3 < -1)
		return "v2_3 is invalid";
	if(v2_4 < -1)
		return "v2_4 is invalid";
	if(v2_5 < -1)
		return "v2_5 is invalid";

	// Ok:
	return STR_NULL;
}

#endif


//-----------------------------------------------------------------------------
// dump: Show data structure.
//-----------------------------------------------------------------------------

#ifdef	VER_DUMP
void bench_dblp_2_c::dump(str_t headline) const
{
	// Headline:
	if(headline == STR_NULL)
		headline = "DBLP Benchmark Object, Ver. 2";
	err_c::dump_begin(headline);

	// Basic information:
	err_c::dump_str("ObjectSize:  ");
	err_c::dump_uint(static_cast<unsigned int>(sizeof(bench_dblp_2_c)));
	err_c::dump_str(" Bytes");
	err_c::dump_nl();
	err_c::dump_nl();

	// Show size of domains:
	err_c::dump_str("Size of Domains:  ");
	err_c::dump_nl();
	err_c::dump_str("    doc_ids:    ");
	err_c::dump_int(doc_ids.num_strings());
	err_c::dump_nl();
	err_c::dump_str("    attrs:      ");
	err_c::dump_int(attrs.num_atoms());
	err_c::dump_nl();
	err_c::dump_str("    str_values: ");
	err_c::dump_int(str_values.num_strings());
	err_c::dump_nl();

	// Show number of rows in all EDB relations:
	err_c::dump_str("Number of rows in EDB relations:  ");
	err_c::dump_nl();
	err_c::dump_str("    att_title: ");
	err_c::dump_int(att_title.num_rows());
	err_c::dump_nl();
	err_c::dump_str("    att_year: ");
	err_c::dump_int(att_year.num_rows());
	err_c::dump_nl();
	err_c::dump_str("    att_author: ");
	err_c::dump_int(att_author.num_rows());
	err_c::dump_nl();
	err_c::dump_str("    att_month: ");
	err_c::dump_int(att_month.num_rows());
	err_c::dump_nl();
	err_c::dump_nl();

	// Atom IDs:
	err_c::dump_str("Atom IDs:");
	err_c::dump_nl();
	err_c::dump_str("    title:  ");
	err_c::dump_int(title);
	err_c::dump_nl();
	err_c::dump_str("    year:   ");
	err_c::dump_int(year);
	err_c::dump_nl();
	err_c::dump_str("    author: ");
	err_c::dump_int(author);
	err_c::dump_nl();
	err_c::dump_str("    month:  ");
	err_c::dump_int(month);
	err_c::dump_nl();
	err_c::dump_nl();

	// Variables:
	err_c::dump_str("Variables:");
	err_c::dump_nl();
	err_c::dump_str("    v2_1:   ");
	err_c::dump_int(v2_1);
	err_c::dump_nl();
	err_c::dump_str("    v2_2:   ");
	err_c::dump_int(v2_2);
	err_c::dump_nl();
	err_c::dump_str("    v2_3:   ");
	err_c::dump_int(v2_3);
	err_c::dump_nl();
	err_c::dump_str("    v2_4:   ");
	err_c::dump_int(v2_4);
	err_c::dump_nl();
	err_c::dump_str("    v2_5:   ");
	err_c::dump_int(v2_5);
	err_c::dump_nl();
	err_c::dump_nl();

	// Backtrack Stack:
	err_c::dump_str("Stacks:");
	err_c::dump_nl();
	err_c::dump_str("    Backtrack Stack Size:   ");
	err_c::dump_int(backtrack_stack.num_elems());
	err_c::dump_nl();

	// Done:
	err_c::dump_end();
}
#endif

