// ============================================================================
// Project:	Deductive Database
// Filename:	bench_dblp_2.h
// Purpose:	Bottom-Up Evaluation for DBLP Example, Ver 2: Cursor interface
// Last Change:	25.05.2016
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2015-2016 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Include File Frame:
//=============================================================================

#ifndef	BENCH_DBLP_2_INCLUDED
#define BENCH_DBLP_2_INCLUDED

//=============================================================================
// Used Types and Macros:
//=============================================================================

#ifndef VER_INCLUDED
#include "ver.h"
#endif

#ifndef STR_INCLUDED
#include "str.h"
#endif

#ifndef CHECK_INCLUDED
#include "check.h"
#endif

#ifndef STACK_INCLUDED
#include "stack.h"
#endif

#ifndef ATOMTAB_INCLUDED
#include "atomtab.h"
#endif

#ifndef STRTAB_INCLUDED
#include "strtab.h"
#endif

#ifndef LIST_2_INCLUDED
#include "list_2.h"
#endif

#ifndef CUR_2_INCLUDED
#include "cur_2.h"
#endif

#ifndef REL_N_N_INCLUDED
#include "rel_n_n.h"
#endif

#ifndef CUR_N_N_INCLUDED
#include "cur_n_n.h"
#endif

#ifndef BENCH_INCLUDED
#include "bench.h"
#endif

//=============================================================================
// Private Constants:
//=============================================================================

//-----------------------------------------------------------------------------
// BENCH_DBLP_2_MAGIC: Magic number (identifies objects of this class).
//-----------------------------------------------------------------------------

static const long BENCH_DBLP_2_MAGIC = 0x444C500AL;	// 'DLP\n'


//=============================================================================
// Bottom-Up Evalutation (DBLP Benchmark) - Ver. 2: Attrib., Cursor Interface.
//=============================================================================

class bench_dblp_2_c : public bench_c {
	public:

//-----------------------------------------------------------------------------
// Constructor, Destructor:
//-----------------------------------------------------------------------------

	public:

	// Constructor:
	bench_dblp_2_c();

	// Destructor:
	~bench_dblp_2_c();

//-----------------------------------------------------------------------------
// (Object) Methods:
//-----------------------------------------------------------------------------

	// load: Load data file with EDB facts.
	bool load(str_t filename);

	// eval: Do bottom-up evaluation.
	long eval();

	// name: Name of this benchmark.
	str_t name() const {
		return "DBLP Bench., Ver.2: Rel/Var in Attrs, Cursor Interface";
	}

	// file: List of data files.
	str_t file(int i) const;

	// result: List of correct results of eval method.
	long result(int i) const;

	// Open cursor (note: load() must be called before this method).
	void open();

	// fetch: Get next tuple, returns false if there is no further tuple.
	bool fetch();

	// answer(v2_1,v2_2,v2_3,v2_4,v2_5)

	// c1: Return value for column 1 of current row in cursor.
	inline str_t c1() {
		if(v2_1 < 0)
			return STR_NULL;
		return doc_ids.get(v2_1);
	}

	// c2: Return value for column 2 of current row in cursor.
	inline str_t c2() {
		if(v2_2 < 0)
			return STR_NULL;
		return str_values.get(v2_2);
	}

	// c3: Return value for column 3 of current row in cursor.
	inline str_t c3() {
		if(v2_3 < 0)
			return STR_NULL;
		return str_values.get(v2_3);
	}

	// c4: Return value for column 4 of current row in cursor.
	inline str_t c4() {
		if(v2_4 < 0)
			return STR_NULL;
		return str_values.get(v2_4);
	}

	// c5: Return value for column 5 of current row in cursor.
	inline str_t c5() {
		if(v2_5 < 0)
			return STR_NULL;
		return str_values.get(v2_5);
	}

	// close: Close cursor.
	inline void close() {
	}

//-----------------------------------------------------------------------------
// Copy-constructor and assignment operator are not supported for this class:
//-----------------------------------------------------------------------------

	private:

	// These methods are not implemented:
	bench_dblp_2_c(const bench_dblp_2_c& obj);
	bench_dblp_2_c& operator=(const bench_dblp_2_c& obj);


//-----------------------------------------------------------------------------
// Debugging Support:
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
	// Integrity check:
	public:
	str_t	check() const;

	// Magic number (identifies objects of this class for debugging).
	private:
	long	Magic;		// Must be "BENCH_DBLP_2_MAGIC".
#endif


#ifdef	VER_DUMP
	public:
	// Display data structure:
	void dump(str_t headline = STR_NULL) const;
#endif

//-----------------------------------------------------------------------------
// Private Object Members:
//-----------------------------------------------------------------------------

	private:

//-----------------------------------------------------------------------------
// Domains:
//-----------------------------------------------------------------------------

	strtab_c doc_ids;	// String table for first argument of att.
	atomtab_c attrs;	// Atom table for second argument of att.
	strtab_c str_values;	// String table for third argument of att.


//-----------------------------------------------------------------------------
// Atoms which are relevant for selection condition:
//-----------------------------------------------------------------------------

	int title;
	int year;
	int author;
	int month;

//-----------------------------------------------------------------------------
// Relations for EDB body literals:
//-----------------------------------------------------------------------------

	list_2_c  att_title;
	rel_n_n_c att_year;
	rel_n_n_c att_author;
	rel_n_n_c att_month;

//-----------------------------------------------------------------------------
// Variables to hold values during the bottom-up evaluation:
//-----------------------------------------------------------------------------

	int v2_1;
	int v2_2;
	int v2_3;
	int v2_4;
	int v2_5;

//-----------------------------------------------------------------------------
// Cursors for loops during the bottom-up evaluation:
//-----------------------------------------------------------------------------

	cur_2_c   c2_1_fbf;	// att(A,title,B),
	cur_n_n_c c2_2_bbf;	// att(A,year,C),
	cur_n_n_c c2_3_bbf;	// att(A,author,D),
	cur_n_n_c c2_4_bbf;	// att(A,month,E).

//-----------------------------------------------------------------------------
// Backtrack stack for the bottom-up evaluation:
//-----------------------------------------------------------------------------

	stack_c<int> backtrack_stack;


//=============================================================================
// End of Class:
//=============================================================================

};


//-----------------------------------------------------------------------------
// Define pointer type:
//-----------------------------------------------------------------------------

typedef bench_dblp_2_c *bench_dblp_2_t;

//-----------------------------------------------------------------------------
// Define null pointer:
//-----------------------------------------------------------------------------

#define BENCH_DBLP_2_NULL (static_cast<bench_dblp_2_t>(0))


//=============================================================================
// End of Include File:
//=============================================================================

#endif

