// ============================================================================
// Project:	Deductive Database
// Filename:	bench_dblp_3.cpp
// Purpose:	Bottom-Up Evaluation for DBLP Example, Ver 2: Cursor interface
// Last Change:	06.10.2016
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2015-2016 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"
#include "str.h"
#include "err.h"
#include "check.h"
#include "db_dblp.h"
#include "eval_dblp.h"
#include "out.h"
#include "bench.h"

#include "bench_dblp_3.h"


//=============================================================================
// Constructor, Destructor:
//=============================================================================

//-----------------------------------------------------------------------------
// Constructor (initializes subobjects):
//-----------------------------------------------------------------------------

bench_dblp_3_c::bench_dblp_3_c() :

		// Superclass:
		bench_c(),

		// Database:
		DB(),

		// Evaluator:
		Query(&DB)
{

	// Set Magic number:
	CHECK_CODE(Magic = BENCH_DBLP_3_MAGIC);
}

//-----------------------------------------------------------------------------
// Destructor:
//-----------------------------------------------------------------------------

bench_dblp_3_c::~bench_dblp_3_c()
{
	// Check this object:
	CHECK_VALID("bench_dblp_3_c::destructor");

	// Make this object invalid:
	CHECK_CODE(Magic = 0);
}


//=============================================================================
// Public Methods:
//=============================================================================

//-----------------------------------------------------------------------------
// load: Load data file with EDB facts.
//-----------------------------------------------------------------------------

bool bench_dblp_3_c::load(str_t filename) {

	// Check parameter:
	CHECK(filename != STR_NULL, "bench_dblp_3_c::load: filename is null");

	// Check object:
	CHECK_VALID("bench_dblp_3_c::load");

	// Load the data:
	return DB.load(filename);
}

//-----------------------------------------------------------------------------
// eval: Do bottom-up evaluation.
//-----------------------------------------------------------------------------

long bench_dblp_3_c::eval()
{

	// number of answers found:
	long num_results = 0;

	// Open Cursor:
	Query.open();

	// Loop while results are found:
	while(Query.fetch()) {

		// Get results:
		str_t ans_id     = Query.c1();
		str_t ans_title  = Query.c2();
		str_t ans_author = Query.c3();
		str_t ans_year   = Query.c4();
		str_t ans_month  = Query.c5();

		// Check whether results are ok:
		if(ans_id == STR_NULL)
			return -1;
		if(ans_title == STR_NULL)
			return -1;
		if(ans_author == STR_NULL)
			return -1;
		if(ans_year == STR_NULL)
			return -1;
		if(ans_month == STR_NULL)
			return -1;

		// Increment number of answers found:
		num_results++;

		// Print the result if below limit:
		out_c::answer_5s(num_results, ans_id, ans_title, ans_author,
						ans_year, ans_month);

	// End of fetch loop:
	}
	Query.close();

	// Print relation statistics if requested:
	if(out_c::show_rel()) {
		out_c::rel_size_headline();
		out_c::rel_size(&(DB.att_title));
		out_c::rel_size(&(DB.att_year));
		out_c::rel_size(&(DB.att_author));
		out_c::rel_size(&(DB.att_month));
		out_c::rel_size_strtab(&(DB.doc_ids));
		out_c::rel_size_atomtab(&(DB.attrs));
		out_c::rel_size_strtab(&(DB.str_values));
		out_c::rel_size_done();
	}

	// Return number of result tuples:
	return num_results;
}


//-----------------------------------------------------------------------------
// file: List of data files.
//-----------------------------------------------------------------------------

str_t bench_dblp_3_c::file(int i) const
{
	switch(i) {
		case 0:
			return VER_FILE_DBLP_TEST;
		case 1:
			return VER_FILE_DBLP;
		case 2:
			return VER_FILE_DBLP_BIG;
		default:
			return STR_NULL;
	}
}


//-----------------------------------------------------------------------------
// result: List of correct results of eval method.
//-----------------------------------------------------------------------------

long bench_dblp_3_c::result(int i) const
{
	switch(i) {
		case 0: return 7;
		case 1: return 4035;
		case 2: return -1;
		default: return -1;
	}
}


//=============================================================================
// Debugging Support:
//=============================================================================


//-----------------------------------------------------------------------------
// check: Check the integrity of this object, return error message or NULL.
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
str_t	bench_dblp_3_c::check() const
{
	// Check magic number:
	if(Magic != BENCH_DBLP_3_MAGIC)
		return "wrong magic number";

	// Check sub-objects (relations for EDB predicates):
	CHECK_SUBOBJ(DB, "bench_dblp_3_c::check");
	CHECK_SUBOBJ(Query, "bench_dblp_3_c::check");

	// Ok:
	return STR_NULL;
}

#endif


//-----------------------------------------------------------------------------
// dump: Show data structure.
//-----------------------------------------------------------------------------

#ifdef	VER_DUMP
void bench_dblp_3_c::dump(str_t headline) const
{
	// Headline:
	if(headline == STR_NULL)
		headline = "DBLP Benchmark Object, Ver. 2";
	err_c::dump_begin(headline);

	// Basic information:
	err_c::dump_str("ObjectSize:  ");
	err_c::dump_uint(static_cast<unsigned int>(sizeof(bench_dblp_3_c)));
	err_c::dump_str(" Bytes");
	err_c::dump_nl();
	err_c::dump_nl();

	// Show size of domains:
	err_c::dump_str("Size of Domains:  ");
	err_c::dump_nl();
	err_c::dump_str("    doc_ids:    ");
	err_c::dump_int(DB.doc_ids.num_strings());
	err_c::dump_nl();
	err_c::dump_str("    attrs:      ");
	err_c::dump_int(DB.attrs.num_atoms());
	err_c::dump_nl();
	err_c::dump_str("    str_values: ");
	err_c::dump_int(DB.str_values.num_strings());
	err_c::dump_nl();

	// Show number of rows in all EDB relations:
	err_c::dump_str("Number of rows in EDB relations:  ");
	err_c::dump_nl();
	err_c::dump_str("    att_title: ");
	err_c::dump_int(DB.att_title.num_rows());
	err_c::dump_nl();
	err_c::dump_str("    att_year: ");
	err_c::dump_int(DB.att_year.num_rows());
	err_c::dump_nl();
	err_c::dump_str("    att_author: ");
	err_c::dump_int(DB.att_author.num_rows());
	err_c::dump_nl();
	err_c::dump_str("    att_month: ");
	err_c::dump_int(DB.att_month.num_rows());
	err_c::dump_nl();
	err_c::dump_nl();

	// Atom IDs:
	err_c::dump_str("Atom IDs:");
	err_c::dump_nl();
	err_c::dump_str("    title:  ");
	err_c::dump_int(DB.title);
	err_c::dump_nl();
	err_c::dump_str("    year:   ");
	err_c::dump_int(DB.year);
	err_c::dump_nl();
	err_c::dump_str("    author: ");
	err_c::dump_int(DB.author);
	err_c::dump_nl();
	err_c::dump_str("    month:  ");
	err_c::dump_int(DB.month);
	err_c::dump_nl();
	err_c::dump_nl();

	// Done:
	err_c::dump_end();
}
#endif

