// ============================================================================
// Project:	Deductive Database
// Filename:	bench_j1axy_3.cpp
// Purpose:	Bottom-Up Evaluation for Join1, a(X,Y), Ver. C - static method
// Last Change:	20.05.2016
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2015-2016 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"
#include "str.h"
#include "err.h"
#include "check.h"
#include "idiv.h"
#include "perf.h"
#include "stack.h"
#include "list_2.h"
#include "cur_2.h"
#include "rel_n_n.h"
#include "cur_n_n.h"
#include "pred.h"
#include "load.h"
#include "out.h"

#include "bench_j1axy_3.h"


//=============================================================================
// Public Methods:
//=============================================================================

//-----------------------------------------------------------------------------
// run: Load data file, do bottom-up evaluation.
//-----------------------------------------------------------------------------


long bench_j1axy_3_c::run(str_t filename,
			perf_t timer_load, perf_t timer_eval) {

	// Check parameter:
	CHECK(filename != STR_NULL, "bench_j1axy_3_c::run: filename is null");

	// Data storage for EDB relations:
	rel_n_n_c c2_bf("c2_bf");
	list_2_c c3_ff("c3_ff");
	rel_n_n_c c4_bf("c4_bf");
	list_2_c d1_ff("d1_ff");
	rel_n_n_c d2_bf("d2_bf");

	// Start timer for data load:
	timer_load->start();

	// Loader object (will contain data about predicates):
	load_c loader;

	// Predicates (input to loader):
	pred_c c2_p("c2", ARGTYPE_INT, ARGTYPE_INT);
	pred_c c3_p("c3", ARGTYPE_INT, ARGTYPE_INT);
	pred_c c4_p("c4", ARGTYPE_INT, ARGTYPE_INT);
	pred_c d1_p("d1", ARGTYPE_INT, ARGTYPE_INT);
	pred_c d2_p("d2", ARGTYPE_INT, ARGTYPE_INT);

	// Assign relations to predicates:
	c2_p.add_rel(&c2_bf);
	c3_p.add_rel(&c3_ff);
	c4_p.add_rel(&c4_bf);
	d1_p.add_rel(&d1_ff);
	d2_p.add_rel(&d2_bf);

	// Add predicate definitions to loader:
	loader.add(&c2_p);
	loader.add(&c3_p);
	loader.add(&c4_p);
	loader.add(&d1_p);
	loader.add(&d2_p);

	// Now do the load:
	if(!loader.load(filename))
		return -1;

	// Stop timer for data load:
	timer_load->stop();

	//---------------------------------------------------------------------
	// eval: Do bottom-up evaluation.
	//---------------------------------------------------------------------

	// Input program:
	//
	// 1: answer(A,B) :- a(A,B).
	// 2: a(A,B) :- b1(A,C), b2(C,B).
	// 3: b1(A,B) :- c1(A,C), c2(C,B).
	// 4: b2(A,B) :- c3(A,C), c4(C,B).
	// 5: c1(A,B) :- d1(A,C), d2(C,B).


	// Start timer for bottom-up evaluation:
	timer_eval->start();

	// Relation for storing derived b1 tuples (for rule with two IDB lit):
	rel_n_n_c b1_fb("b1_fb");
		// Note that columns 1 and 2 are exchanged,
		// because c1 is always input to the mapping, and c2 output.

	// Variables for storing data values (columns of derived tuples):
	int v5_1 = -1;
	int v5_2 = -1;
	int v3_2 = -1;
	int v4_1 = -1;
	int v4_2 = -1;
	int v2_1 = -1;

	// Cursors for looping over relations:
	cur_2_c d1_r5_cur(&d1_ff);
	cur_n_n_c d2_r5_cur(&d2_bf);
	cur_n_n_c c2_r3_cur(&c2_bf);
	cur_2_c c3_r4_cur(&c3_ff);
	cur_n_n_c c4_r4_cur(&c4_bf);
	cur_n_n_c b1_r2_cur(&b1_fb);

	// number of answers found:
	long num_results = 0;

	// --------------------------------------------------------------------
	// Jump labels:
	// label 1: applying rule 5
	//		5: c1(A,B) :- d1(A,C), d2(C,B).
	//		-> c1(v5_1,v5_2)
	// label 2: using c1(v5_1,v5_2) with literal 1 in rule 3
	//		3: b1(A,B) :- c1(A,C), c2(C,B).
	//		-> b1(v5_1,v3_2)
	// label 3: applying rule 4
	//		4: b2(A,B) :- c3(A,C), c4(C,B).
	//		-> b2(v4_1,v4_2)
	// label 4: using b2(v4_1,v4_2) with literal 2 in rule 2
	//		2: a(A,B) :- b1(A,C), b2(C,B).
	//		-> a(v2_1,v4_2)
	// label 5: using a(v2_1,v4_2) with literal 1 in rule 1
	//		1: answer(A,B) :- a(A,B).
	//		-> answer(v2_1,v4_2)
	// --------------------------------------------------------------------



	// Start Rule 5: c1(A,B) :- d1(A,C), d2(C,B).
	// Derives  c1(v5_1,v5_2)
	// label_1:
	d1_r5_cur.open();
	while (d1_r5_cur.fetch()) {
		v5_1 = d1_r5_cur.c1();
		d2_r5_cur.open(d1_r5_cur.c2());
		// label_7:
		while (d2_r5_cur.fetch()) {
			v5_2 = d2_r5_cur.c2();
			// Push backtrack label to continuation
			//backtrack_stack.push(PUSH7);
			// goto label_2;
	
			// Using c1(v5_1,v5_2) for literal 1 in
			// rule 3:  b1(A,B) :- c1(A,C), c2(C,B).
			// Derives  b1(v5_1,v3_2)
			// label_2:
			c2_r3_cur.open(v5_2);
			// label_8:
			while (c2_r3_cur.fetch()) {
				v3_2 = c2_r3_cur.c2();
				// Push backtrack label to continuation
				// backtrack_stack.push(8);
				//
				// Derived b1(v5_1,v3_2)
				b1_fb.insert(v3_2, v5_1);
			}
		}
	}


	// Start rule 4:   b2(A,B) :- c3(A,C), c4(C,B).
	// Derives:        b2(v4_1,v4_2)
	// label_3:
	c3_r4_cur.open();
	while (c3_r4_cur.fetch()) {
		v4_1 = c3_r4_cur.c1();
		c4_r4_cur.open(c3_r4_cur.c2());
		// Continuation for label 3
		// label_9:
		while (c4_r4_cur.fetch()) {
			v4_2 = c4_r4_cur.c2();
			// Push backtrack label to continuation
			//backtrack_stack.push(PUSH9);
			// Labels for next rules
			//goto label_4;
			// Using b2(v4_1,v4_2) with literal 2 in rule 2
			//           a(A,B) :- b1(A,C), b2(C,B).
			// Derives:  a(v2_1,v4_2)
			// label_4:
			b1_r2_cur.open(v4_1);
			// Continuation for label 4
			//label_10:
			while (b1_r2_cur.fetch()) {
				v2_1 = b1_r2_cur.c2(); // Really column 1!
				// Push backtrack label to continuation
				//backtrack_stack.push(PUSH10);
				// Labels for next rules
				//goto label_5;
				// Using a(v2_1,v4_2) with literal 1 in rule 1
				//             answer(A,B) :- a(A,B).
				// Derives:    answer(v2_1,v4_2)
				//label_5:

				// answer(v2_1,v4_2)
				if(v2_1 < 0 || v4_2 < 0)
					return -1;
				num_results++;
				out_c::answer_2(num_results, v2_1, v4_2);
			}
		}
	}

	// Stop timer for bottom-up evaluation:
	timer_eval->stop();

	// Print table of memory pools if requested:
	out_c::mem_pools_tab();

	// Print relation statistics if requested:
	if(out_c::show_rel()) {
		out_c::rel_size_headline();
		out_c::rel_size_nonstd("c2_bf", c2_bf.num_rows(),
					c2_bf.num_pages());
		out_c::rel_size(&c3_ff);
		out_c::rel_size_nonstd("c4_bf", c4_bf.num_rows(),
					c4_bf.num_pages());
		out_c::rel_size(&d1_ff);
		out_c::rel_size_nonstd("d2_bf", d2_bf.num_rows(),
					d2_bf.num_pages());
		out_c::rel_size_nonstd("b1_fb", b1_fb.num_rows(),
					b1_fb.num_pages());
		out_c::rel_size_done();
	}
	// Return number of result tuples:
	return num_results;
}


//-----------------------------------------------------------------------------
// file: List of data files.
//-----------------------------------------------------------------------------

str_t bench_j1axy_3_c::file(int i)
{
	switch(i) {
		case 0:
			return VER_FILE_JOIN1_TEST;
		case 1:
			return VER_FILE_JOIN1_10K;
		case 2:
			return VER_FILE_JOIN1_50K;
		case 3:
			return VER_FILE_JOIN1_250K;
		default:
			return STR_NULL;
	}
}


//-----------------------------------------------------------------------------
// result: List of correct results of eval method.
//-----------------------------------------------------------------------------

long bench_j1axy_3_c::result(int i)
{
	switch(i) {
		case 0: return 4;
		case 1: return -1;
		case 2: return -1;
		case 3: return -1;
		default: return -1;
	}
}

