// ============================================================================
// Project:	Deductive Database
// Filename:	bench_j1axy_8.cpp
// Purpose:	Benchmark Join1, a(X,Y), Ver. 8: functions, bitmap dupcheck
// Last Change:	26.08.2016
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2015 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"
#include "str.h"
#include "err.h"
#include "check.h"
//#include "idiv.h"
//#include "perf.h"
#include "stack.h"
#include "list_2.h"
#include "cur_2.h"
#include "rel_n_n.h"
#include "cur_n_n.h"
#include "pred.h"
#include "load.h"
#include "out.h"

#include "bench_j1axy_8.h"

// a(X, Y)  :- b1(X, Z), b2(Z, Y).
// b1(X, Y) :- c1(X, Z), c2(Z, Y).
// b2(X, Y) :- c3(X, Z), c4(Z, Y).
// c1(X, Y) :- d1(X, Z), d2(Z, Y).
//
// This means that a(X, Y) holds iff
// there is a d1-edge from X to Z1,
// then a d2-edge from Z1 to Z2,
// then a c2-edge from Z2 to Z3,
// then a c3-edge from Z3 to Z4,
// and finally a c4-edge from Z4 to Y.
//
// The file data/d1000_test.P contains
//
// d1(10,20).  - From 10 and 11 we reach 20
// d1(11,20).
// d1(197,523).
// d1(478,837).
// d2(20,30).  - From 20 we get to 30 and 31
// d2(20,31).
// d2(109,92).
// d2(382,205).
// c2(30,40).  - From 30 and 31 we reach 40
// c2(31,40).
// c2(912,311).
// c2(977,164).
// c3(10,50).
// c3(40,51).  - From 40 to 51
// c3(177,787).
// c3(38,785).
// c4(50,60).
// c4(51,61).  - From 51 to 61
// c4(792,629).
// c4(501,660).
//
// So there are two answers (10, 61) and (11, 61).
// Without duplicate elimination, we would compute each two times.


//=============================================================================
// Constructor, Destructor:
//=============================================================================

//-----------------------------------------------------------------------------
// Constructor (initializes subobjects):
//-----------------------------------------------------------------------------

bench_j1axy_8_c::bench_j1axy_8_c() :

		// Superclass:
		bench_c(),

		// Data structures for relations:
		c2_bf("c2_bf"),
		c3_ff("c3_ff"),
		c4_bf("c4_bf"),
		d1_ff("d1_ff"),
		d2_bf("d2_bf"),

		// Relation for storing derived b1 tuples (two IDB lit):
		b1_fb("b1_fb"),
		// Note that columns 1 and 2 are exchanged,
		// because c1 is always input to the mapping, and c2 output.

		// Sets for duplicate checks:
		b1_bb("b1_bb"),
		b2_bb("b2_bb"),
		c1_bb("c1_bb"),
		a_bb("a_bb"),

		// Cursors for loops over relations:
		c4_1_ff(&d1_ff),
		c4_2_bf(&d2_bf),
		c3_2_bf(&c2_bf),
		c5_1_ff(&c3_ff),
		c5_2_bf(&c4_bf),
		c2_1_fb(&b1_fb),

		// Result:
		result_list("answer_ff"),
		result_cursor(&result_list)
{
	// Set Magic number:
	CHECK_CODE(Magic = BENCH_J1AXY_8_MAGIC);

	// Initialize variables of push method:
	//v_2_1 = -1;
	//v_3_2 = -1;
	//v_4_1 = -1;
	//v_4_2 = -1;
	//v_5_1 = -1;
	//v_5_2 = -1;

	// Initialize attribute for the call depth for trace output:
	OUT_TRACE_DEPTH_INIT;

	// No error occurred so far:
	Error = false;
}

//-----------------------------------------------------------------------------
// Destructor:
//-----------------------------------------------------------------------------

bench_j1axy_8_c::~bench_j1axy_8_c()
{
	// Check this object:
	CHECK_VALID("bench_j1axy_8_c::destructor");

	// Make this object invalid:
	CHECK_CODE(Magic = 0);
}


//=============================================================================
// Public Methods:
//=============================================================================

//-----------------------------------------------------------------------------
// load: Load data file with EDB facts.
//-----------------------------------------------------------------------------


bool bench_j1axy_8_c::load(str_t filename) {

	// Check parameter:
	CHECK(filename != STR_NULL, "bench_j1axy_8_c::load: filename is null");

	// Check object:
	CHECK_VALID("bench_j1axy_8_c::load");

	// Loader object (will contain data about predicates):
	load_c loader;

	// Predicates (input to loader):
	pred_c c2_p("c2", ARGTYPE_INT, ARGTYPE_INT);
	pred_c c3_p("c3", ARGTYPE_INT, ARGTYPE_INT);
	pred_c c4_p("c4", ARGTYPE_INT, ARGTYPE_INT);
	pred_c d1_p("d1", ARGTYPE_INT, ARGTYPE_INT);
	pred_c d2_p("d2", ARGTYPE_INT, ARGTYPE_INT);

	// Assign relations to predicates:
	c2_p.add_rel(&c2_bf);
	c3_p.add_rel(&c3_ff);
	c4_p.add_rel(&c4_bf);
	d1_p.add_rel(&d1_ff);
	d2_p.add_rel(&d2_bf);

	// Add predicate definitions to loader:
	loader.add(&c2_p);
	loader.add(&c3_p);
	loader.add(&c4_p);
	loader.add(&d1_p);
	loader.add(&d2_p);

	// Now do the load:
	return loader.load(filename);
}


//-----------------------------------------------------------------------------
// eval: Do bottom-up evaluation.
//-----------------------------------------------------------------------------

long bench_j1axy_8_c::eval()
{
	// Input program:
	//
	// 1: answer(A,B) :- a(A,B).
	// 2: a(A,B) :- b1(A,C), b2(C,B).
	// 3: b1(A,B) :- c1(A,C), c2(C,B).
	// 4: b2(A,B) :- c3(A,C), c4(C,B).
	// 5: c1(A,B) :- d1(A,C), d2(C,B).

	// number of answers found:
	long num_results = 0;

	// Open Cursor:
	open();

	// Loop while results are found:
	while(fetch()) {

		// Get results:
		int ans_1 = c1();
		int ans_2 = c2();

		// Check whether results are ok:
		if(ans_1 < 0)
			return -1;
		if(ans_2 < 0)
			return -1;

		// Increment number of answers found:
		num_results++;

		// Print the result if below limit:
		out_c::answer_2(num_results, ans_1, ans_2);

	// End of fetch loop:
	}

	// Check for errors:
	if(error())
		return -1;

	// Close cursor:
	close();

	// Print relation statistics if requested:
	if(out_c::show_rel()) {
		out_c::rel_size_headline();
		out_c::rel_size_nonstd("c2_bf", c2_bf.num_rows(),
					c2_bf.num_pages());
		out_c::rel_size(&c3_ff);
		out_c::rel_size_nonstd("c4_bf", c4_bf.num_rows(),
					c4_bf.num_pages());
		out_c::rel_size(&d1_ff);
		out_c::rel_size_nonstd("d2_bf", d2_bf.num_rows(),
					d2_bf.num_pages());
		out_c::rel_size_nonstd("b1_fb", b1_fb.num_rows(),
					b1_fb.num_pages());
		out_c::rel_size(&b1_bb);
		out_c::rel_size(&b2_bb);
		out_c::rel_size(&c1_bb);
		out_c::rel_size(&a_bb);
		out_c::rel_size(&result_list);
		out_c::rel_size_done();
	}

	// Return number of result tuples:
	return num_results;
}


//-----------------------------------------------------------------------------
// file: List of data files.
//-----------------------------------------------------------------------------

str_t bench_j1axy_8_c::file(int i) const
{
	switch(i) {
		case 0:
			return VER_FILE_JOIN1_TEST;
		case 1:
			return VER_FILE_JOIN1_10K;
		case 2:
			return VER_FILE_JOIN1_50K;
		case 3:
			return VER_FILE_JOIN1_250K;
		default:
			return STR_NULL;
	}
}


//-----------------------------------------------------------------------------
// result: List of correct results of the eval method.
//-----------------------------------------------------------------------------

long bench_j1axy_8_c::result(int i) const
{
	switch(i) {
		case 0: return 2;
		case 1: return 1000000L;
		case 2: return 1000000L;
		case 3: return 1000000L;
		default: return -1;
	}
}


//-----------------------------------------------------------------------------
// Open cursor (note: load() must be called before this method).
//-----------------------------------------------------------------------------

void bench_j1axy_8_c::open()
{
	OUT_TRACE_DEPTH;

	// --------------------------------------------------------------------
	// Transformed program:
	//
	// 1: answer(A,B) :- a(A,B).
	// 2: a(A,B) :- b1(A,C), b2(C,B).
	// 3: b1(A,B) :- c1(A,C), c2(C,B).
	// 4: c1(A,B) :- d1(A,C), d2(C,B).
	// 5: b2(A,B) :- c3(A,C), c4(C,B).
	// --------------------------------------------------------------------


	// Start Rule 4: c1(A,B) :- d1(A,C), d2(C,B).
	OUT_TRACE_CALL_0("start");
	start_rule_4();

	// Start rule 5: b2(A,B) :- c3(A,C), c4(C,B).
	OUT_TRACE_CALL_0("start");
	start_rule_5();

	// Open cursor over result:
	result_cursor.open();
}

//=============================================================================
// Debugging Support:
//=============================================================================


//-----------------------------------------------------------------------------
// check: Check the integrity of this object, return error message or NULL.
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
str_t	bench_j1axy_8_c::check() const
{
	// Check magic number:
	if(Magic != BENCH_J1AXY_8_MAGIC)
		return "wrong magic number";

	// Check sub-objects (relations for EDB predicates):
	CHECK_SUBOBJ(c2_bf, "bench_j1axy_8_c::check");
	CHECK_SUBOBJ(c3_ff, "bench_j1axy_8_c::check");
	CHECK_SUBOBJ(c4_bf, "bench_j1axy_8_c::check");
	CHECK_SUBOBJ(d1_ff, "bench_j1axy_8_c::check");
	CHECK_SUBOBJ(d2_bf, "bench_j1axy_8_c::check");

	// Check sub-objects (relation for storing derived b1 tuples):
	CHECK_SUBOBJ(b1_fb, "bench_j1axy_8_c::check");

	// Check sub-objects (sets for duplicate checks):
	CHECK_SUBOBJ(b1_bb, "bench_j1axy_8_c::check");
	CHECK_SUBOBJ(b2_bb, "bench_j1axy_8_c::check");
	CHECK_SUBOBJ(c1_bb, "bench_j1axy_8_c::check");
	CHECK_SUBOBJ(a_bb, "bench_j1axy_8_c::check");

	// Check sub-objects (cursors for looping over relations):
	CHECK_SUBOBJ(c4_1_ff, "bench_j1axy_8_c::check");
	CHECK_SUBOBJ(c4_2_bf, "bench_j1axy_8_c::check");
	CHECK_SUBOBJ(c3_2_bf, "bench_j1axy_8_c::check");
	CHECK_SUBOBJ(c5_1_ff, "bench_j1axy_8_c::check");
	CHECK_SUBOBJ(c5_2_bf, "bench_j1axy_8_c::check");
	CHECK_SUBOBJ(c2_1_fb, "bench_j1axy_8_c::check");

	// Ok:
	return 0;
}

#endif


//-----------------------------------------------------------------------------
// dump: Show data structure.
//-----------------------------------------------------------------------------

#ifdef	VER_DUMP
void bench_j1axy_8_c::dump(str_t headline) const
{
	// Headline:
	if(headline == STR_NULL)
		headline = "Benchmark Join1 a(X,Y), Ver. 8";
	err_c::dump_begin(headline);

	// Basic information:
	err_c::dump_str("ObjectSize:  ");
	err_c::dump_uint(static_cast<unsigned int>(sizeof(bench_j1axy_8_c)));
	err_c::dump_str(" Bytes");
	err_c::dump_nl();
	err_c::dump_nl();

	// Show number of rows in all EDB relations:
	err_c::dump_str("Number of rows in EDB relations:");
	err_c::dump_nl();
	err_c::dump_str("    c2_bf: ");
	err_c::dump_int(c2_bf.num_rows());
	err_c::dump_nl();
	err_c::dump_str("    c3_ff: ");
	err_c::dump_int(c3_ff.num_rows());
	err_c::dump_nl();
	err_c::dump_str("    c4_bf: ");
	err_c::dump_int(c4_bf.num_rows());
	err_c::dump_nl();
	err_c::dump_str("    d1_ff: ");
	err_c::dump_int(d1_ff.num_rows());
	err_c::dump_nl();
	err_c::dump_str("    d2_bf: ");
	err_c::dump_int(d2_bf.num_rows());
	err_c::dump_nl();
	err_c::dump_nl();

	err_c::dump_str("Number of rows in IDB relation:");
	err_c::dump_nl();
	err_c::dump_str("    b1_fb: ");
	err_c::dump_int(b1_fb.num_rows());
	err_c::dump_nl();
	err_c::dump_nl();

	err_c::dump_str("Number of rows in duplicate check sets (IDB):");
	err_c::dump_nl();
	err_c::dump_str("    b1_bb: ");
	err_c::dump_int(b1_bb.num_rows());
	err_c::dump_nl();
	err_c::dump_str("    b2_bb: ");
	err_c::dump_int(b2_bb.num_rows());
	err_c::dump_nl();
	err_c::dump_str("    c1_bb: ");
	err_c::dump_int(c1_bb.num_rows());
	err_c::dump_nl();
	err_c::dump_str("    a_bb: ");
	err_c::dump_int(a_bb.num_rows());
	err_c::dump_nl();
	err_c::dump_nl();

	err_c::dump_str("Number of rows in Result:");
	err_c::dump_nl();
	err_c::dump_str("    result_list: ");
	err_c::dump_int(result_list.num_rows());
	err_c::dump_nl();
	err_c::dump_nl();

	err_c::dump_str("Error checking:");
	err_c::dump_nl();
	err_c::dump_str("    Error: ");
	err_c::dump_bool(Error);
	err_c::dump_nl();

	// Done:
	err_c::dump_end();
}
#endif
