// ============================================================================
// Project:	Deductive Database
// Filename:	bench_tcbf_1.cpp
// Purpose:	Bottom-Up Evaluation for Transitive Closure (bf) Example
// Last Change:	21.05.2016
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2015-2016 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Local Version Settings:
//=============================================================================

//#define EVAL_DUP_HASHSIZE	10001
//#define EVAL_DUP_HASHSIZE	500001
#define EVAL_DUP_HASHSIZE	5003

//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"
#include "str.h"
#include "err.h"
#include "check.h"
#include "idiv.h"
#include "perf.h"
#include "stack.h"
#include "rel_n_n.h"
#include "cur_n_n.h"
#include "rel_ii.h"
#include "cur_ii.h"
#include "dup_i.h"
#include "load_tcbf.h"
#include "out.h"

#include "bench_tcbf_1.h"

//=============================================================================
// Static Methods:
//=============================================================================


//-----------------------------------------------------------------------------
// run: Load data file, do bottom-up evaluation for transitive closure (bf).
//-----------------------------------------------------------------------------


long bench_tcbf_1_c::run(str_t filename,
			perf_t timer_load, perf_t timer_eval)
{

	// Counter for number of results:
	long num_results = 0;

	// --------------------------------------------------------------------
	// Relations for EDB body literals:
	// --------------------------------------------------------------------

	rel_n_n_c par_bf("par_bf");

	// --------------------------------------------------------------------
	// Relations for duplicate check (IDB predicates):
	// --------------------------------------------------------------------

	// This is not really necessary for termination, but eliminates duplic.
	dup_i_c p0_b("p0_b (Duplicate Check)", EVAL_DUP_HASHSIZE);
	if(!p0_b.init_ok())
		return -1;

	// This predicate is recursive, so duplicate elimination is important:
	dup_i_c p1_b("p1_b (Duplicate Check)", EVAL_DUP_HASHSIZE);
	if(!p1_b.init_ok())
		return -1;

	// --------------------------------------------------------------------
	// Load data file, fill domains and EDB relations:
	// --------------------------------------------------------------------

	// Start timer for data load:
	timer_load->start();

	// Load the data:
	if(!load_tcbf_c::load(filename, &par_bf))
		return -1;

	// Stop timer for data load:
	timer_load->stop();

	//---------------------------------------------------------------------
	// eval: Do bottom-up evaluation.
	//---------------------------------------------------------------------

	// Start timer for bottom-up evaluation:
	timer_eval->start();

	// --------------------------------------------------------------------
	// Transformed program:
	//
	// 3: p0(A) :- p1(B), par(B,A).
	// 4: p1(A) :- p1(B), par(B,A).
	// 5: tc(1,A) :- p0(A).
	// 1: p0(A) :- par(1,A).
	// 2: p1(A) :- par(1,A).
	// --------------------------------------------------------------------

	// --------------------------------------------------------------------
	// Declaration Part:
	// --------------------------------------------------------------------

	int v1_1 = -1;
	int v2_1 = -1;
	int v3_1 = -1;
	int v4_1 = -1;

	// par_cursor_bf c1_1_bf;
	// par_cursor_bf c2_1_bf;
	// par_cursor_bf c3_2_bf;
	// par_cursor_bf c4_2_bf;
	cur_n_n_c c1_1_bf(&par_bf);
	cur_n_n_c c2_1_bf(&par_bf);
	cur_n_n_c c3_2_bf(&par_bf);
	cur_n_n_c c4_2_bf(&par_bf);

	// int_stack backtrack_stack;
	// int_stack int_value_stack;
	stack_c<int> backtrack_stack("backtrack_stack");
	stack_c<int> int_value_stack("int_value_stack");
	stack_c<int> cursor_stack("cursor_stack");

	// float_stack float_value_stack;
	// string_stack string_value_stack;

	// --------------------------------------------------------------------
	// Jump labels:
	// label 1: applying rule 1
	// 		resulting in symbolic fact p0(v1_1)
	// label 2: using symbolic fact p0(v1_1) with literal 1 in rule 5
	// 		resulting in symbolic fact tc(1,v1_1)
	// label 3: applying rule 2
	// 		resulting in symbolic fact p1(v2_1)
	// label 4: using symbolic fact p1(v2_1) with literal 1 in rule 3
	// 		resulting in symbolic fact p0(v3_1)
	// label 5: using symbolic fact p0(v3_1) with literal 1 in rule 5
	// 		resulting in symbolic fact tc(1,v3_1)
	// label 6: using symbolic fact p1(v2_1) with literal 1 in rule 4
	// 		resulting in symbolic fact p1(v4_1)
	// label 7: using symbolic fact p1(v4_1) with literal 1 in rule 3
	// 		resulting in symbolic fact p0(v3_1)
	// label 8: using symbolic fact p1(v4_1) with literal 1 in rule 4
	// 		resulting in symbolic fact p1(v4_1)
	// --------------------------------------------------------------------


	// Initialisation Part:
	if(!backtrack_stack.push(3))
		return -1;
	if(!backtrack_stack.push(1))
		return -1;

	// Loop Part:
	while(!backtrack_stack.is_empty()) {
		switch(backtrack_stack.pop()) {
		case 1:
			c1_1_bf.open(1);
			// Continuation for label 1
		case 10:
			while (c1_1_bf.fetch()) {

				v1_1 = c1_1_bf.c2();

				// Derived fact: p0(v1_1)

				// Duplicate check (not really necessary):
				int dup = p0_b.dup_check(v1_1);
				if(dup < 0)
					return -1;
				if(dup > 0)
					continue;

				// Push backtrack label to continuation
				if(!backtrack_stack.push(10))
					return -1;
				// Labels for next rules
				goto label_2;
			}
			break;
		case 2: label_2:

			// Derived fact: tc(1,v1_1)
			if(v1_1 < 0)
				return -1;
			num_results++;
			out_c::answer_1(num_results, v1_1);

			break;
		case 3:
			c2_1_bf.open(1);
			// Continuation for label 3
		case 12:
			while (c2_1_bf.fetch()) {

				v2_1 = c2_1_bf.c2();

				// Derived: p1(v2_1)

				// Duplicate check (not really necessary):
				int dup = p1_b.dup_check(v2_1);
				if(dup < 0)
					return -1;
				if(dup > 0)
					continue;
				// Push backtrack label to continuation
				if(!backtrack_stack.push(12))
					return -1;
				// Labels for next rules
				if(!backtrack_stack.push(6))
					return -1;
				goto label_4;
			}
			break;
		case 4: label_4:

			c3_2_bf.open(v2_1);
			// Continuation for label 4
		case 13:
			while (c3_2_bf.fetch()) {

				v3_1 = c3_2_bf.c2();

				// Derived fact: p0(v3_1)

				// Duplicate check (not really necessary):
				int dup = p0_b.dup_check(v3_1);
				if(dup < 0)
					return -1;
				if(dup > 0)
					continue;

				// Push backtrack label to continuation
				if(!backtrack_stack.push(13))
					return -1;
				// Labels for next rules
				goto label_5;
			}
			break;
		case 5: label_5:

			// Derived fact: tc(1,v3_1)
			if(v3_1 < 0)
				return -1;
			num_results++;
			out_c::answer_1(num_results, v3_1);
 
			break;
		case 6: //label_6:

			c4_2_bf.open(v2_1);
		// Continuation for label 6
		case 15:
			while (c4_2_bf.fetch()) {

				v4_1 = c4_2_bf.c2();

				// Derived: p1(v4_1)

				// Duplicate check:
				int dup = p1_b.dup_check(v4_1);
				if(dup < 0)
					return -1;
				if(dup > 0)
					continue;

				// Push backtrack label to continuation
				if(!backtrack_stack.push(15))
					return -1;
				// Labels for next rules
				if(!backtrack_stack.push(8))
					return -1;
				goto label_7;
			}
			break;
		case 7: label_7:
 
			c3_2_bf.open(v4_1);
		// Continuation for label 7
		case 16:
			while (c3_2_bf.fetch()) {

				v3_1 = c3_2_bf.c2();

				// Derived fact: p0(v3_1)

				// Duplicate check (not really necessary):
				int dup = p0_b.dup_check(v3_1);
				if(dup < 0)
					return -1;
				if(dup > 0)
					continue;

				// Push backtrack label to continuation
				if(!backtrack_stack.push(16))
					return -1;
				// Labels for next rules
				goto label_5;
			}
			break;
		case 8: //label_8:
			//recursive
			c4_2_bf.push(&cursor_stack);
			int_value_stack.push(v4_1);
			// Push backtrack label to restore values
			backtrack_stack.push(26);
 
			c4_2_bf.open(v4_1);
		// Continuation for label 8
		case 17:
			while (c4_2_bf.fetch()) {

				v4_1 = c4_2_bf.c2();

				// Derived: p1(v4_1)

				// Duplicate check:
				int dup = p1_b.dup_check(v4_1);
				if(dup < 0)
					return -1;
				if(dup > 0)
					continue;

				// Push backtrack label to continuation
				if(!backtrack_stack.push(17))
					return -1;
				// Labels for next rules
				if(!backtrack_stack.push(8))
					return -1;
				goto label_7;
			}
			break;
		// Label to restore values of label 8
		case 26:
			v4_1 = int_value_stack.pop();
			c4_2_bf.pop(&cursor_stack);
			break;

		}
	}


	// Stop timer for bottom-up evaluation:
	timer_eval->stop();

	// Debug output - Details for Hash Tables:
	int num_rows     = p0_b.num_rows();
	int num_pages    = p0_b.num_pages();
	int hash_size    = p0_b.hash_size();
	int entries_used = p0_b.entries_used();

	out_c::debug_msg(      "Duplicate check table p0_b:\n");
	out_c::debug_ival(     "    Number of rows:         ", num_rows);
	out_c::debug_mempages( "    Number of memory pages: ", num_pages);
	out_c::debug_ival(     "    Hash table size:        ", hash_size);
	out_c::debug_hashusage("    Entries Used:           ", entries_used,
								hash_size);
	out_c::debug_hashchain("    Average Chain Length:   ", entries_used,
								num_rows);
	out_c::debug_nl();

	num_rows     = p1_b.num_rows();
	num_pages    = p1_b.num_pages();
	hash_size    = p1_b.hash_size();
	entries_used = p1_b.entries_used();

	out_c::debug_msg(      "Duplicate check table p1_b:\n");
	out_c::debug_ival(     "    Number of rows:         ", num_rows);
	out_c::debug_mempages( "    Number of memory pages: ", num_pages);
	out_c::debug_ival(     "    Hash table size:        ", hash_size);
	out_c::debug_hashusage("    Entries Used:           ", entries_used,
								hash_size);
	out_c::debug_hashchain("    Average Chain Length:   ", entries_used,
								num_rows);
	out_c::debug_nl();

	// Print table of memory pools if requested:
	out_c::mem_pools_tab();

	// Print relation statistics if requested:
	if(out_c::show_rel()) {
		out_c::rel_size_headline();
		out_c::rel_size_nonstd("par_bf", par_bf.num_rows(),
					par_bf.num_pages());
		out_c::rel_size_nonstd("p0_b", p0_b.num_rows(),
					p0_b.num_pages());
		out_c::rel_size_nonstd("p1_b", p1_b.num_rows(),
					p1_b.num_pages());
		out_c::rel_size_done();
	}

	// Return number of result tuples:
	return num_results;
}

//-----------------------------------------------------------------------------
// file: List of data files.
//-----------------------------------------------------------------------------

str_t bench_tcbf_1_c::file(int i)
{
	switch(i) {
		case 0:
			return VER_FILE_TC_TEST;
		case 1:
			return VER_FILE_TC_50K;
		case 2:
			return VER_FILE_TC_50K_CYC;
		case 3:
			return VER_FILE_TC_250K;
		case 4:
			return VER_FILE_TC_250K_CYC;
		case 5:
			return VER_FILE_TC_500K;
		case 6:
			return VER_FILE_TC_500K_CYC;
		case 7:
			return VER_FILE_TC_500K_D2;
		case 8:
			return VER_FILE_TC_500K_D2_CYC;
		case 9:
			return VER_FILE_TC_1M_D2;
		case 10:
			return VER_FILE_TC_1M_D2_CYC;
		default:
			return STR_NULL;
	}
}


//-----------------------------------------------------------------------------
// result: List of correct results of eval method.
//-----------------------------------------------------------------------------

long bench_tcbf_1_c::result(int i)
{
	switch(i) {
		case 0: return 5;
			// tc(1,2). tc(1,3). tc(1,20). tc(1,21). tc(1,22).
		case 1: return -1;
		case 2: return -1;
		case 3: return -1;
		case 4: return -1;
		case 5: return -1;
		case 6: return -1;
		case 7: return -1;
		case 8: return -1;
		case 9: return -1;
		case 10: return -1;
		default: return -1;
	}
}

