// ============================================================================
// Project:	Deductive Database
// Filename:	bench_tcbf_2.cpp
// Purpose:	Bottom-Up Evaluation for tc(1,Y), Ver 2: Cursor interface
// Last Change:	21.09.2016
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2015-2016 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"
#include "str.h"
#include "err.h"
#include "check.h"
#include "stack.h"
#include "dup_i.h"
#include "rel_n_n.h"
#include "cur_n_n.h"
#include "pred.h"
#include "load.h"
#include "out.h"

#include "bench_tcbf_2.h"


//=============================================================================
// Local Constants:
//=============================================================================

//-----------------------------------------------------------------------------
// Hash Table Size for Duplicate Elimination:
//-----------------------------------------------------------------------------

static const int BENCH_TCBF2_HASHSIZE = 5003;

//=============================================================================
// Constructor, Destructor:
//=============================================================================

//-----------------------------------------------------------------------------
// Constructor (initializes subobjects):
//-----------------------------------------------------------------------------

bench_tcbf_2_c::bench_tcbf_2_c() :

		// Superclass:
		bench_c(),

		// Relations for EDB body literals:
		par_bf("par_bf"),

		// Relations for duplicate check (IDB predicates):
		p0_b("p0_b (DupCheck)", BENCH_TCBF2_HASHSIZE),
		p1_b("p1_b (DupCheck)", BENCH_TCBF2_HASHSIZE),

		// Cursors for loops during bottom-up evaluation:
		c1_1_bf(&par_bf),
		c2_1_bf(&par_bf),
		c3_2_bf(&par_bf),
		c4_2_bf(&par_bf),

		// Backtrack stack and value restore stack:
		backtrack_stack("Backtrack Stack"),
		int_value_stack("int_value_stack"),
		cursor_stack("cursor_stack")

{
	// Initialize variables for bottom-up evaluation:
	v1_1 = -1;
	v2_1 = -1;
	v3_1 = -1;
	v4_1 = -1;

	// Initialize variable for exit point:
	first_exit_point = true;

	// No error occurred so far:
	Error = false;

	// Check whether p0_b and p1_b were correctly initialized (got memory):
	if(!p0_b.init_ok())
		Error = true;
	if(!p1_b.init_ok())
		Error = true;

	// Set Magic number:
	CHECK_CODE(Magic = BENCH_TCBF_2_MAGIC);
}

//-----------------------------------------------------------------------------
// Destructor:
//-----------------------------------------------------------------------------

bench_tcbf_2_c::~bench_tcbf_2_c()
{
	// Check this object:
	CHECK_VALID("bench_tcbf_2_c::destructor");

	// Make this object invalid:
	CHECK_CODE(Magic = 0);
}


//=============================================================================
// Public Methods:
//=============================================================================

//-----------------------------------------------------------------------------
// load: Load data file with EDB facts.
//-----------------------------------------------------------------------------

bool bench_tcbf_2_c::load(str_t filename) {

	// Check parameter:
	CHECK(filename != STR_NULL, "bench_tcbf_2_c::load: filename is null");

	// Check object:
	CHECK_VALID("bench_tcbf_2_c::load");

	// Loader object (will contain data about predicates):
	load_c loader;

	// Predicate (input to loader):
	pred_c par_p("par", ARGTYPE_INT, ARGTYPE_INT);
	par_p.add_rel(&par_bf, BIND_BF);

	// Add predicate definition to loader:
	loader.add(&par_p);

	// Load data file:
	return loader.load(filename);
}

//-----------------------------------------------------------------------------
// eval: Do bottom-up evaluation.
//-----------------------------------------------------------------------------

long bench_tcbf_2_c::eval()
{

	// number of answers found:
	long num_results = 0;

	// Open Cursor:
	open();

	// Loop while results are found:
	while(fetch()) {

		// Get results:
		int ans_1 = c1();
		int ans_2 = c2();

		// Check whether results are ok:
		if(ans_1 != 1)
			return -1;
		if(ans_2 < 0)
			return -1;

		// Increment number of answers found:
		num_results++;

		// Print the result if below limit:
		out_c::answer_2(num_results, ans_1, ans_2);

	// End of fetch loop:
	}

	// Check for errors:
	if(error())
		return -1;

	// Close cursor:
	close();

	// Print relation statistics if requested:
	if(out_c::show_rel()) {
		out_c::rel_size_headline();
		out_c::rel_size(&par_bf);
		out_c::rel_size_nonstd("p0_b", p0_b.num_rows(),
						p0_b.num_pages());
		out_c::rel_size_nonstd("p1_b", p1_b.num_rows(),
						p1_b.num_pages());
		out_c::rel_size_done();
	}

	// Return number of result tuples:
	return num_results;
}


//-----------------------------------------------------------------------------
// Open cursor (note: load() must be called before this method).
//-----------------------------------------------------------------------------

void bench_tcbf_2_c::open()
{

	// --------------------------------------------------------------------
	// Transformed program:
	//
	// 3: p0(A) :- p1(B), par(B,A).
	// 4: p1(A) :- p1(B), par(B,A).
	// 5: tc(1,A) :- p0(A).
	// 1: p0(A) :- par(1,A).
	// 2: p1(A) :- par(1,A).
	// --------------------------------------------------------------------

	// --------------------------------------------------------------------
	// Jump labels:
	// label 1: applying rule 1
	// 		resulting in symbolic fact p0(v1_1)
	// label 2: using symbolic fact p0(v1_1) with literal 1 in rule 5
	// 		resulting in symbolic fact tc(1,v1_1)
	// label 3: applying rule 2
	// 		resulting in symbolic fact p1(v2_1)
	// label 4: using symbolic fact p1(v2_1) with literal 1 in rule 3
	// 		resulting in symbolic fact p0(v3_1)
	// label 5: using symbolic fact p0(v3_1) with literal 1 in rule 5
	// 		resulting in symbolic fact tc(1,v3_1)
	// label 6: using symbolic fact p1(v2_1) with literal 1 in rule 4
	// 		resulting in symbolic fact p1(v4_1)
	// label 7: using symbolic fact p1(v4_1) with literal 1 in rule 3
	// 		resulting in symbolic fact p0(v3_1)
	// label 8: using symbolic fact p1(v4_1) with literal 1 in rule 4
	// 		resulting in symbolic fact p1(v4_1)
	// --------------------------------------------------------------------


	// Initialisation Part:
	if(!backtrack_stack.push(3)) {
		Error = true;
		return;
	}
	if(!backtrack_stack.push(1)) {
		Error = true;
		return;
	}

}

//-----------------------------------------------------------------------------
// fetch: Get next tuple, returns false if there is no further tuple.
//-----------------------------------------------------------------------------

bool bench_tcbf_2_c::fetch()
{
	// Check for errors:
	if(Error)
		return false;

	// Loop Part:
	while(!backtrack_stack.is_empty()) {
		switch(backtrack_stack.pop()) {

		// label 1: applying rule 1: p0(A) :- par(1,A).
		// resulting in symbolic fact p0(v1_1)
		case 1:
			OUT_TRACE_START(1, "p0(A) :- par(1,A).", "EDB:par");

			c1_1_bf.open(1);

		// Continuation for label 1:
		case 10:
			OUT_TRACE_CONT(1, "p0(A) :- par(1,A).", "EDB:par");

			while (c1_1_bf.fetch()) {

				v1_1 = c1_1_bf.c2();
				OUT_TRACE_ASSIGN("v1_1", v1_1);

				// Derived fact: p0(v1_1)

				// Duplicate check (not really necessary):
				int dup = p0_b.dup_check(v1_1);
				if(dup < 0) {
					Error = true;
					return false;
				}
				if(dup > 0)
					continue;

				// Push backtrack label to continuation
				if(!backtrack_stack.push(10)) {
					Error = true;
					return false;
				}

				// Labels for next rules
				goto label_2;
			}
			break;

		// label 2: using symbolic fact p0(v1_1)
		// with literal 1 in rule 5:  tc(1,A) :- p0(A).
		// resulting in symbolic fact tc(1,v1_1)
		case 2: label_2:

			// Derived fact: tc(1,v1_1)
			OUT_TRACE_ANSWER(2, "tc(1,A) :- p0(A).", "p0(v1_1)");

			// tc(1,v1_1):
			first_exit_point = true;
			return true;

			// break;

		// label 3: applying rule 2: p1(A) :- par(1,A).
		// resulting in symbolic fact p1(v2_1)
		case 3:
			OUT_TRACE_START(3, "p1(A) :- par(1,A).", "EDB:par");

			c2_1_bf.open(1);

		// Continuation for label 3:
		case 12:
			OUT_TRACE_CONT(12, "p1(A) :- par(1,A).", "EDB:par");

			while (c2_1_bf.fetch()) {

				v2_1 = c2_1_bf.c2();
				OUT_TRACE_ASSIGN("v2_1", v2_1);

				// Derived: p1(v2_1)

				// Duplicate check (not really necessary):
				int dup = p1_b.dup_check(v2_1);
				if(dup < 0) {
					Error = true;
					return false;
				}
				if(dup > 0)
					continue;

				// Push backtrack label to continuation
				if(!backtrack_stack.push(12)) {
					Error = true;
					return false;
				}

				// Labels for next rules
				if(!backtrack_stack.push(6)) {
					Error = true;
					return false;
				}
				goto label_4;
			}
			break;

		// label 4: using symbolic fact p1(v2_1)
		// with literal 1 in rule 3: p0(A) :- p1(B), par(B,A).
		// resulting in symbolic fact p0(v3_1)
		case 4: label_4:
			OUT_TRACE_START(4, "p0(A) :- p1(B), par(B,A).",
							"p0(v3_1)");

			c3_2_bf.open(v2_1);

		// Continuation for label 4
		case 13:
			OUT_TRACE_CONT(13, "p0(A) :- p1(B), par(B,A).",
							"p0(v3_1)");
			while (c3_2_bf.fetch()) {

				v3_1 = c3_2_bf.c2();
				OUT_TRACE_ASSIGN("v3_1", v3_1);

				// Derived fact: p0(v3_1)

				// Duplicate check (not really necessary):
				int dup = p0_b.dup_check(v3_1);
				if(dup < 0) {
					Error = true;
					return false;
				}
				if(dup > 0)
					continue;

				// Push backtrack label to continuation
				if(!backtrack_stack.push(13)) {
					Error = true;
					return false;
				}

				// Labels for next rules
				goto label_5;
			}
			break;

		// label 5: using symbolic fact p0(v3_1)
		// with literal 1 in rule 5: tc(1,A) :- p0(A).
		// resulting in symbolic fact tc(1,v3_1)
		case 5: label_5:
			OUT_TRACE_ANSWER(5, "tc(1,A) :- p0(A).", "p0(v3_1)");

			// Derived fact: tc(1,v3_1)
			first_exit_point = false;
			return true;
 
			// break;

		// label 6: using symbolic fact p1(v2_1)
		// with literal 1 in rule 4: p1(A) :- p1(B), par(B,A).
		// resulting in symbolic fact p1(v4_1)
		case 6: //label_6:
			OUT_TRACE_START(6, "p1(A) :- p1(B), par(B,A).",
							"p0(v4_1)");

			c4_2_bf.open(v2_1);

		// Continuation for label 6
		case 15:
			OUT_TRACE_CONT(15, "p1(A) :- p1(B), par(B,A).",
							"p0(v4_1)");

			while (c4_2_bf.fetch()) {

				v4_1 = c4_2_bf.c2();
				OUT_TRACE_ASSIGN("v4_1", v4_1);

				// Derived: p1(v4_1)

				// Duplicate check:
				int dup = p1_b.dup_check(v4_1);
				if(dup < 0) {
					Error = true;
					return false;
				}
				if(dup > 0)
					continue;

				// Push backtrack label to continuation
				if(!backtrack_stack.push(15)) {
					Error = true;
					return false;
				}

				// Labels for next rules
				if(!backtrack_stack.push(8)) {
					Error = true;
					return false;
				}
				goto label_7;
			}
			break;

		// label 7: using symbolic fact p1(v4_1)
		// with literal 1 in rule 3: p0(A) :- p1(B), par(B,A).
		// resulting in symbolic fact p0(v3_1)
		case 7: label_7:
			OUT_TRACE_START(7, "p0(A) :- p1(B), par(B,A).",
							"p1(v4_1)");
 
			c3_2_bf.open(v4_1);

		// Continuation for label 7
		case 16:
			OUT_TRACE_CONT(16, "p0(A) :- p1(B), par(B,A).",
							"p1(v4_1)");

			while (c3_2_bf.fetch()) {

				v3_1 = c3_2_bf.c2();
				OUT_TRACE_ASSIGN("v3_1", v3_1);

				// Derived fact: p0(v3_1)

				// Duplicate check (not really necessary):
				int dup = p0_b.dup_check(v3_1);
				if(dup < 0) {
					Error = true;
					return false;
				}
				if(dup > 0)
					continue;

				// Push backtrack label to continuation
				if(!backtrack_stack.push(16)) {
					Error = true;
					return false;
				}

				// Labels for next rules
				goto label_5;
			}
			break;

		// label 8: using symbolic fact p1(v4_1)
		// with literal 1 in rule 4: p1(A) :- p1(B), par(B,A).
		// resulting in symbolic fact p1(v4_1)
		case 8: //label_8:
			//recursive
			OUT_TRACE_START_REC(8, "p1(A) :- p1(B), par(B,A).",
							"p1(v4_1)");
			c4_2_bf.push(&cursor_stack);
			int_value_stack.push(v4_1);
			// Push backtrack label to restore values
			if(!backtrack_stack.push(26)) {
				Error = true;
				return false;
			}
 
			c4_2_bf.open(v4_1);

		// Continuation for label 8
		case 17:
			OUT_TRACE_CONT(17, "p1(A) :- p1(B), par(B,A).",
							"p1(v4_1)");
			while (c4_2_bf.fetch()) {

				v4_1 = c4_2_bf.c2();
				OUT_TRACE_ASSIGN("v4_1", v4_1);

				// Derived: p1(v4_1)

				// Duplicate check:
				int dup = p1_b.dup_check(v4_1);
				if(dup < 0) {
					Error = true;
					return false;
				}
				if(dup > 0)
					continue;

				// Push backtrack label to continuation
				if(!backtrack_stack.push(17)) {
					Error = true;
					return false;
				}

				// Labels for next rules
				if(!backtrack_stack.push(8)) {
					Error = true;
					return false;
				}
				goto label_7;
			}
			break;

		// Label to restore values of label 8
		case 26:
			OUT_TRACE_RESTORE(26, "p1(A) :- p1(B), par(B,A).",
							"p1(v4_1)");
			v4_1 = int_value_stack.pop();
			OUT_TRACE_POP_VAR("v4_1", v4_1);
			c4_2_bf.pop(&cursor_stack);
			break;

		}
	}

	// Backtrack stack was empty, no more answers:
	return false;
}


//-----------------------------------------------------------------------------
// file: List of data files.
//-----------------------------------------------------------------------------

str_t bench_tcbf_2_c::file(int i) const
{
	switch(i) {
		case 0:
			return VER_FILE_TC_TEST;
		case 1:
			return VER_FILE_TC_50K;
		case 2:
			return VER_FILE_TC_50K_CYC;
		case 3:
			return VER_FILE_TC_250K;
		case 4:
			return VER_FILE_TC_250K_CYC;
		case 5:
			return VER_FILE_TC_500K;
		case 6:
			return VER_FILE_TC_500K_CYC;
		case 7:
			return VER_FILE_TC_500K_D2;
		case 8:
			return VER_FILE_TC_500K_D2_CYC;
		case 9:
			return VER_FILE_TC_1M_D2;
		case 10:
			return VER_FILE_TC_1M_D2_CYC;
		default:
			return STR_NULL;
	}
}


//-----------------------------------------------------------------------------
// result: List of correct results of eval method.
//-----------------------------------------------------------------------------

long bench_tcbf_2_c::result(int i) const
{
	switch(i) {
		case 0: return 5;
			// tc(1,2). tc(1,3). tc(1,20). tc(1,21). tc(1,22).
		case 1: return -1;
		case 2: return -1;
		case 3: return -1;
		case 4: return -1;
		case 5: return -1;
		case 6: return -1;
		case 7: return -1;
		case 8: return -1;
		case 9: return -1;
		case 10: return -1;
		default: return -1;
	}
}


//=============================================================================
// Debugging Support:
//=============================================================================


//-----------------------------------------------------------------------------
// check: Check the integrity of this object, return error message or NULL.
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
str_t	bench_tcbf_2_c::check() const
{
	// Check magic number:
	if(Magic != BENCH_TCBF_2_MAGIC)
		return "wrong magic number";

	// Check sub-objects (relations for EDB predicates):
	CHECK_SUBOBJ(par_bf, "bench_tcbf_2_c::check");
	CHECK_SUBOBJ(p0_b, "bench_tcbf_2_c::check");
	CHECK_SUBOBJ(p1_b, "bench_tcbf_2_c::check");
	CHECK_SUBOBJ(c1_1_bf, "bench_tcbf_2_c::check");
	CHECK_SUBOBJ(c2_1_bf, "bench_tcbf_2_c::check");
	CHECK_SUBOBJ(c3_2_bf, "bench_tcbf_2_c::check");
	CHECK_SUBOBJ(c4_2_bf, "bench_tcbf_2_c::check");
	CHECK_SUBOBJ(backtrack_stack, "bench_tcbf_2_c::check");
	CHECK_SUBOBJ(int_value_stack, "bench_tcbf_2_c::check");
	CHECK_SUBOBJ(cursor_stack, "bench_tcbf_2_c::check");

	// Variables are non-negative once set, but initialized to -1:
	if(v1_1 < -1)
		return "v1_1 is invalid";
	if(v2_1 < -1)
		return "v2_1 is invalid";
	if(v3_1 < -1)
		return "v3_1 is invalid";
	if(v4_1 < -1)
		return "v4_1 is invalid";

	// Ok:
	return STR_NULL;
}

#endif


//-----------------------------------------------------------------------------
// dump: Show data structure.
//-----------------------------------------------------------------------------

#ifdef	VER_DUMP
void bench_tcbf_2_c::dump(str_t headline) const
{
	// Headline:
	if(headline == STR_NULL)
		headline = "tc(X,Y) Benchmark Object, Ver. 2";
	err_c::dump_begin(headline);

	// Basic information:
	err_c::dump_str("ObjectSize:  ");
	err_c::dump_uint(static_cast<unsigned int>(sizeof(bench_tcbf_2_c)));
	err_c::dump_str(" Bytes");
	err_c::dump_nl();
	err_c::dump_nl();

	// Show number of rows in all EDB relations:
	err_c::dump_str("Number of rows in EDB relations:  ");
	err_c::dump_nl();
	err_c::dump_str("    par_bf: ");
	err_c::dump_int(par_bf.num_rows());
	err_c::dump_nl();
	err_c::dump_nl();

	// Show number of rows in IDB relations (duplicate check):
	err_c::dump_str("Number of rows in IDB relations:");
	err_c::dump_nl();
	err_c::dump_str("    p0_b:   ");
	err_c::dump_int(p0_b.num_rows());
	err_c::dump_nl();
	err_c::dump_str("    p1_b:   ");
	err_c::dump_int(p1_b.num_rows());
	err_c::dump_nl();
	err_c::dump_nl();

	// Variables:
	err_c::dump_str("Variables:");
	err_c::dump_nl();
	err_c::dump_str("    v1_1:   ");
	err_c::dump_int(v1_1);
	err_c::dump_nl();
	err_c::dump_str("    v2_1:   ");
	err_c::dump_int(v2_1);
	err_c::dump_nl();
	err_c::dump_str("    v3_1:   ");
	err_c::dump_int(v3_1);
	err_c::dump_nl();
	err_c::dump_str("    v4_1:   ");
	err_c::dump_int(v4_1);
	err_c::dump_nl();
	err_c::dump_str("    first_exit_point: ");
	err_c::dump_bool(first_exit_point);
	err_c::dump_nl();
	err_c::dump_str("    Error:   ");
	err_c::dump_bool(Error);
	err_c::dump_nl();
	err_c::dump_nl();

	// Backtrack Stack:
	err_c::dump_str("Stacks:");
	err_c::dump_nl();
	err_c::dump_str("    Backtrack Stack Size:   ");
	err_c::dump_int(backtrack_stack.num_elems());
	err_c::dump_nl();
	err_c::dump_str("    Int Value Stack Size:   ");
	err_c::dump_int(int_value_stack.num_elems());
	err_c::dump_nl();
	err_c::dump_str("    Cursor    Stack Size:   ");
	err_c::dump_int(cursor_stack.num_elems());
	err_c::dump_nl();
	err_c::dump_nl();

	// Done:
	err_c::dump_end();
}
#endif

