// ============================================================================
// Project:	Deductive Database
// Filename:	bench_tcbf_4.cpp
// Purpose:	Bottom-Up Evaluation for tc(1,X), Ver 4: Functions
// Last Change:	23.09.2016
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2015-2016 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"
#include "str.h"
#include "err.h"
#include "check.h"
#include "list_2.h"
#include "cur_2.h"
#include "dup_i.h"
#include "rel_n_n.h"
#include "cur_n_n.h"
#include "pred.h"
#include "load.h"
#include "out.h"

#include "bench_tcbf_4.h"


//=============================================================================
// Local Constants:
//=============================================================================

//-----------------------------------------------------------------------------
// Hash Table Size for Duplicate Elimination:
//-----------------------------------------------------------------------------

static const int BENCH_TCBF2_HASHSIZE = 5003;

//=============================================================================
// Constructor, Destructor:
//=============================================================================

//-----------------------------------------------------------------------------
// Constructor (initializes subobjects):
//-----------------------------------------------------------------------------

bench_tcbf_4_c::bench_tcbf_4_c() :

		// Superclass:
		bench_c(),

		// Relations for EDB body literals:
		par_bf("par_bf"),

		// Relations for duplicate check (IDB predicates):
		p0_b("p0_b (DupCheck)", BENCH_TCBF2_HASHSIZE),
		p1_b("p1_b (DupCheck)", BENCH_TCBF2_HASHSIZE),

		// Result:
		result_list("tc_ff"),
		result_cursor(&result_list)

{
	// No error occurred so far:
	Error = false;

	// Check whether p0_b and p1_b were correctly initialized (got memory):
	if(!p0_b.init_ok())
		Error = true;
	if(!p1_b.init_ok())
		Error = true;

	// Set Magic number:
	CHECK_CODE(Magic = BENCH_TCBF_4_MAGIC);
}

//-----------------------------------------------------------------------------
// Destructor:
//-----------------------------------------------------------------------------

bench_tcbf_4_c::~bench_tcbf_4_c()
{
	// Check this object:
	CHECK_VALID("bench_tcbf_4_c::destructor");

	// Make this object invalid:
	CHECK_CODE(Magic = 0);
}


//=============================================================================
// Public Methods:
//=============================================================================

//-----------------------------------------------------------------------------
// load: Load data file with EDB facts.
//-----------------------------------------------------------------------------

bool bench_tcbf_4_c::load(str_t filename) {

	// Check parameter:
	CHECK(filename != STR_NULL, "bench_tcbf_4_c::load: filename is null");

	// Check object:
	CHECK_VALID("bench_tcbf_4_c::load");

	// Loader object (will contain data about predicates):
	load_c loader;

	// Predicate (input to loader):
	pred_c par_p("par", ARGTYPE_INT, ARGTYPE_INT);
	par_p.add_rel(&par_bf, BIND_BF);

	// Add predicate definition to loader:
	loader.add(&par_p);

	// Load data file:
	return loader.load(filename);
}

//-----------------------------------------------------------------------------
// eval: Do bottom-up evaluation and read result over cursor interface.
//-----------------------------------------------------------------------------

long bench_tcbf_4_c::eval()
{

	// number of answers found:
	long num_results = 0;

	// Open Cursor (this does the real work):
	open();

	// Loop while results are found:
	while(fetch()) {

		// Get results:
		int ans_1 = c1();
		int ans_2 = c2();

		// Check whether results are ok:
		if(ans_1 != 1)
			return -1;
		if(ans_2 < 0)
			return -1;

		// Increment number of answers found:
		num_results++;

		// Print the result if below limit:
		out_c::answer_2(num_results, ans_1, ans_2);

	// End of fetch loop:
	}

	// Check for errors:
	if(error())
		return -1;

	// Close cursor:
	close();

	// Print relation statistics if requested:
	if(out_c::show_rel()) {
		out_c::rel_size_headline();
		out_c::rel_size(&par_bf);
		out_c::rel_size_nonstd("p0_b", p0_b.num_rows(),
						p0_b.num_pages());
		out_c::rel_size(&result_list);
		out_c::rel_size_done();
	}

	// Return number of result tuples:
	return num_results;
}


//-----------------------------------------------------------------------------
// Open cursor (note: load() must be called before this method).
//-----------------------------------------------------------------------------

void bench_tcbf_4_c::open()
{

	// --------------------------------------------------------------------
	// Transformed program:
	//
	// 1: p0(A) :- par(1,A).
	// 2: p1(A) :- par(1,A).
	// 3: p0(A) :- p1(B), par(B,A).
	// 4: p1(A) :- p1(B), par(B,A).
	// 5: tc(1,A) :- p0(A).
	// --------------------------------------------------------------------


	OUT_TRACE_CALL_0("start");
	start();

	// Open cursor over result:
	result_cursor.open();
}


//-----------------------------------------------------------------------------
// start: Compute derived facts for rules with only EDB body literals.
//-----------------------------------------------------------------------------

inline void bench_tcbf_4_c::start()
{
	OUT_TRACE_DEPTH;

	// Execute start rules:
	rule_1();
	rule_2();
}

//-----------------------------------------------------------------------------
// rule_1:  p0(A) :- par(1,A).  Start Rule.
//-----------------------------------------------------------------------------

inline void bench_tcbf_4_c::rule_1()
{
	OUT_TRACE_DEPTH;

	// Start Rule 1: p0(A) :- par(1,A).
	OUT_TRACE_RULE_APP_0("1: p0(A) :- par(1,A).", "start");

	// Cursors for looping over relations:
	cur_n_n_c lit1_par_bf(&par_bf);

	// Loop over par(1, A):
	lit1_par_bf.open(1);
	while (lit1_par_bf.fetch()) {
		int A = lit1_par_bf.c2();

		// Derived: p0(A)
		OUT_TRACE_CALL_1("p0", A);
		derived_p0(A);
	}
	//lit1_par_bf.close();
}


//-----------------------------------------------------------------------------
// rule_2:  p1(A) :- par(1,A).  Start Rule.
//-----------------------------------------------------------------------------

inline void bench_tcbf_4_c::rule_2()
{
	OUT_TRACE_DEPTH;

	// Start Rule 2: p1(A) :- par(1,A).
	OUT_TRACE_RULE_APP_0("2: p1(A) :- par(1,A).", "start");

	// Cursors for looping over relations:
	cur_n_n_c lit1_par_bf(&par_bf);

	// Loop over par(1, A):
	lit1_par_bf.open(1);
	while (lit1_par_bf.fetch()) {
		int A = lit1_par_bf.c2();

		// Derived: p1(A)
		OUT_TRACE_CALL_1("p1", A);
		derived_p1(A);
	}
	//lit1_par_bf.close();
}


//-----------------------------------------------------------------------------
// rule_3:  p0(A) :- p1(B), par(B,A).  Given: p1(B).
//-----------------------------------------------------------------------------

inline void bench_tcbf_4_c::rule_3(int B)
{
	OUT_TRACE_DEPTH;

	// 3: p0(A) :- p1(B), par(B,A).
	OUT_TRACE_RULE_APP_1("3: p0(A) :- p1(B), par(B,A).", "p1", B);

	// Cursors for looping over relations:
	cur_n_n_c lit2_par_bf(&par_bf);

	// Loop over par(A, B):
	lit2_par_bf.open(B);
	while (lit2_par_bf.fetch()) {
 
		int A = lit2_par_bf.c2();

		// Use derived fact:
		OUT_TRACE_CALL_1("p0", A);
		derived_p0(A);
	}
	//lit2_par_bf.close();
}


//-----------------------------------------------------------------------------
// rule_4:  p1(A) :- p1(B), par(B,A).  Given: p1(B).
//-----------------------------------------------------------------------------

inline void bench_tcbf_4_c::rule_4(int B)
{
	OUT_TRACE_DEPTH;

	// 4: p1(A) :- p1(B), par(B,A).
	OUT_TRACE_RULE_APP_1("4: p1(A) :- p1(B), par(B,A).", "p1", B);

	// Cursors for looping over relations:
	cur_n_n_c lit2_par_bf(&par_bf);

	// Loop over par(A, B):
	lit2_par_bf.open(B);
	while(lit2_par_bf.fetch()) {
 
		int A = lit2_par_bf.c2();

		// Use derived fact:
		OUT_TRACE_CALL_1("p1", A);
		derived_p1(A);
	}
	//lit2_par_bf.close();
}

//-----------------------------------------------------------------------------
// rule_5:  tc(1,A) :- p0(A).  Given: p0(A).
//-----------------------------------------------------------------------------

inline void bench_tcbf_4_c::rule_5(int A)
{
	OUT_TRACE_DEPTH;

	// 5: tc(1,A) :- p0(A).
	OUT_TRACE_RULE_APP_1("1: tc(1,A) :- p0(A).", "p0", A);

	// tc(1, A) is derived:
	OUT_TRACE_CALL_2("tc", 1, A);
	derived_tc(1, A);
}

//-----------------------------------------------------------------------------
// derived_p0: Function to call when a p0-tuple was derived.
//-----------------------------------------------------------------------------

inline void bench_tcbf_4_c::derived_p0(int A)
{
	OUT_TRACE_DEPTH;

	// Duplicate check (new interface):
	//if(!p0_b.insert(A))
	//	continue;

	// Duplicate check (old interface):
	int dup = p0_b.dup_check(A);
	if(dup < 0) {
		Error = true;
		return;
	}
	if(dup > 0)
		return;

	// 5: tc(1,A) :- p0(A).
	rule_5(A);

}


//-----------------------------------------------------------------------------
// derived_p1: Function to call when a p1-tuple was derived.
//-----------------------------------------------------------------------------

inline void bench_tcbf_4_c::derived_p1(int A)
{
	OUT_TRACE_DEPTH;

	// Duplicate check (new interface):
	//if(!p1_b.insert(A))
	//	continue;

	// Duplicate check (old interface):
	int dup = p1_b.dup_check(A);
	if(dup < 0) {
		Error = true;
		return;
	}
	if(dup > 0)
		return;

	// 3: p0(A) :- p1(B), par(B,A).
	rule_3(A);

	// 4: p1(A) :- p1(B), par(B,A).
	rule_4(A);
}


//-----------------------------------------------------------------------------
// derived_tc: Function to call when a tc-tuple was derived.
//-----------------------------------------------------------------------------

inline void bench_tcbf_4_c::derived_tc(int A, int B)
{
	OUT_TRACE_DEPTH;

	// No duplicate check necessary, because the only rule is
	// 5: tc(1,A) :- p0(A).
	// and there is a duplicate check for p0.

	// The tuple is stored in the result set:
	if(!result_list.insert(A, B))
		Error = true;
}

//-----------------------------------------------------------------------------
// file: List of data files.
//-----------------------------------------------------------------------------

str_t bench_tcbf_4_c::file(int i) const
{
	switch(i) {
		case 0:
			return VER_FILE_TC_TEST;
		case 1:
			return VER_FILE_TC_50K;
		case 2:
			return VER_FILE_TC_50K_CYC;
		case 3:
			return VER_FILE_TC_250K;
		case 4:
			return VER_FILE_TC_250K_CYC;
		case 5:
			return VER_FILE_TC_500K;
		case 6:
			return VER_FILE_TC_500K_CYC;
		case 7:
			return VER_FILE_TC_500K_D2;
		case 8:
			return VER_FILE_TC_500K_D2_CYC;
		case 9:
			return VER_FILE_TC_1M_D2;
		case 10:
			return VER_FILE_TC_1M_D2_CYC;
		default:
			return STR_NULL;
	}
}


//-----------------------------------------------------------------------------
// result: List of correct results of eval method.
//-----------------------------------------------------------------------------

long bench_tcbf_4_c::result(int i) const
{
	switch(i) {
		case 0: return 5;
			// tc(1,2). tc(1,3). tc(1,20). tc(1,21). tc(1,22).
		case 1: return -1;
		case 2: return -1;
		case 3: return -1;
		case 4: return -1;
		case 5: return -1;
		case 6: return -1;
		case 7: return -1;
		case 8: return -1;
		case 9: return -1;
		case 10: return -1;
		default: return -1;
	}
}


//=============================================================================
// Debugging Support:
//=============================================================================


//-----------------------------------------------------------------------------
// check: Check the integrity of this object, return error message or NULL.
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
str_t	bench_tcbf_4_c::check() const
{
	// Check magic number:
	if(Magic != BENCH_TCBF_4_MAGIC)
		return "wrong magic number";

	// Check sub-objects:
	CHECK_SUBOBJ(par_bf, "bench_tcbf_4_c::check");
	CHECK_SUBOBJ(p0_b, "bench_tcbf_4_c::check");
	CHECK_SUBOBJ(p1_b, "bench_tcbf_4_c::check");
	CHECK_SUBOBJ(result_list, "bench_tcbf_4_c::check");
	CHECK_SUBOBJ(result_cursor, "bench_tcbf_4_c::check");

	// Ok:
	return STR_NULL;
}

#endif


//-----------------------------------------------------------------------------
// dump: Show data structure.
//-----------------------------------------------------------------------------

#ifdef	VER_DUMP
void bench_tcbf_4_c::dump(str_t headline) const
{
	// Headline:
	if(headline == STR_NULL)
		headline = "tc(X,Y) Benchmark Object, Ver. 2";
	err_c::dump_begin(headline);

	// Basic information:
	err_c::dump_str("ObjectSize:  ");
	err_c::dump_uint(static_cast<unsigned int>(sizeof(bench_tcbf_4_c)));
	err_c::dump_str(" Bytes");
	err_c::dump_nl();
	err_c::dump_nl();

	// Show number of rows in all EDB relations:
	err_c::dump_str("Number of rows in EDB relations:  ");
	err_c::dump_nl();
	err_c::dump_str("    par_bf: ");
	err_c::dump_int(par_bf.num_rows());
	err_c::dump_nl();
	err_c::dump_nl();

	// Show number of rows in IDB relations (duplicate check):
	err_c::dump_str("Number of rows in IDB relations:");
	err_c::dump_nl();
	err_c::dump_str("    p0_b:   ");
	err_c::dump_int(p0_b.num_rows());
	err_c::dump_nl();
	err_c::dump_str("    p1_b:   ");
	err_c::dump_int(p1_b.num_rows());
	err_c::dump_nl();
	err_c::dump_nl();

	// Show number of rows in the result:
	err_c::dump_str("Number of rows in Result:");
	err_c::dump_nl();
	err_c::dump_str("    result_list: ");
	err_c::dump_int(result_list.num_rows());
	err_c::dump_nl();
	err_c::dump_nl();

	// Error status:
	err_c::dump_str("Error checking:");
	err_c::dump_nl();
	err_c::dump_str("    Error: ");
	err_c::dump_bool(Error);
	err_c::dump_nl();

	// Done:
	err_c::dump_end();
}
#endif

