// ============================================================================
// Project:	Deductive Database
// Filename:	bench_tcbf_4.h
// Purpose:	Bottom-Up Evaluation for tc(1,X), Ver 4: Functions
// Last Change:	23.09.2016
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2015-2016 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Include File Frame:
//=============================================================================

#ifndef	BENCH_TCBF_4_INCLUDED
#define BENCH_TCBF_4_INCLUDED

//=============================================================================
// Used Types and Macros:
//=============================================================================

#ifndef VER_INCLUDED
#include "ver.h"
#endif

#ifndef STR_INCLUDED
#include "str.h"
#endif

#ifndef CHECK_INCLUDED
#include "check.h"
#endif

#ifndef LIST_2_INCLUDED
#include "list_2.h"
#endif

#ifndef CUR_2_INCLUDED
#include "cur_2.h"
#endif

#ifndef DUP_I_INCLUDED
#include "dup_i.h"
#endif

#ifndef REL_N_N_INCLUDED
#include "rel_n_n.h"
#endif

#ifndef BENCH_INCLUDED
#include "bench.h"
#endif

//=============================================================================
// Private Constants:
//=============================================================================

//-----------------------------------------------------------------------------
// BENCH_TCBF_4_MAGIC: Magic number (identifies objects of this class).
//-----------------------------------------------------------------------------

static const long BENCH_TCBF_4_MAGIC = 0x5442340AL;	// 'TB4\n'


//=============================================================================
// Bottom-Up Evalutation (DBLP Benchmark) - Ver. 4: Functions
//=============================================================================

class bench_tcbf_4_c : public bench_c {
	public:

//-----------------------------------------------------------------------------
// Constructor, Destructor:
//-----------------------------------------------------------------------------

	public:

	// Constructor:
	bench_tcbf_4_c();

	// Destructor:
	~bench_tcbf_4_c();

//-----------------------------------------------------------------------------
// (Object) Methods:
//-----------------------------------------------------------------------------

	// load: Load data file with EDB facts.
	bool load(str_t filename);

	// eval: Do bottom-up evaluation and read result over cursor interface.
	long eval();

	// name: Name of this benchmark.
	str_t name() const {
		return
		"tc(1,X) Bench., Ver.4: Functions, Cursor Interface";
	}

	// file: List of data files.
	str_t file(int i) const;

	// result: List of correct results of eval method.
	long result(int i) const;

	// Open cursor (note: load() must be called before this method).
	void open();

	// fetch: Get next tuple, returns false if there is no further tuple.
	inline bool fetch() {
		return result_cursor.fetch();
	}

	// c1: Return value for column 1 of current row in cursor.
	inline int c1() {
		return result_cursor.c1();
	}

	// c2: Return value for column 2 of current row in cursor.
	inline int c2() {
		return result_cursor.c2();
	}

	// error: Did an error occur?
	inline bool error() {
		if(par_bf.mem_err())
			return true;
		if(result_list.mem_err())
			return true;
		return Error;
	}

	// close: Close cursor.
	inline void close() {
		result_cursor.close();
	}

//-----------------------------------------------------------------------------
// Copy-constructor and assignment operator are not supported for this class:
//-----------------------------------------------------------------------------

	private:

	// These methods are not implemented:
	bench_tcbf_4_c(const bench_tcbf_4_c& obj);
	bench_tcbf_4_c& operator=(const bench_tcbf_4_c& obj);


//-----------------------------------------------------------------------------
// Debugging Support:
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
	// Integrity check:
	public:
	str_t	check() const;

	// Magic number (identifies objects of this class for debugging).
	private:
	long	Magic;		// Must be "BENCH_TCBF_4_MAGIC".
#endif


#ifdef	VER_DUMP
	public:
	// Display data structure:
	void dump(str_t headline = STR_NULL) const;
#endif

//-----------------------------------------------------------------------------
// Private Object Members:
//-----------------------------------------------------------------------------

	private:

//-----------------------------------------------------------------------------
// Relations for EDB body literals:
//-----------------------------------------------------------------------------

	rel_n_n_c par_bf;

//-----------------------------------------------------------------------------
// Relations for duplicate check (IDB predicates):
//-----------------------------------------------------------------------------

	dup_i_c p0_b;
	dup_i_c p1_b;

//-----------------------------------------------------------------------------
// Result list and cursor:
//-----------------------------------------------------------------------------

	list_2_c result_list;
	cur_2_c  result_cursor;

//-----------------------------------------------------------------------------
// Call Depth for Trace Output:
//-----------------------------------------------------------------------------

	OUT_TRACE_DEPTH_DECL;

//-----------------------------------------------------------------------------
// Did an error occur?
//-----------------------------------------------------------------------------

	bool Error;

//-----------------------------------------------------------------------------
// Auxiliary functions:
//-----------------------------------------------------------------------------

	private:

	// start: Compute derived facts for rules with only EDB body literals.
	inline void start();

	// derived_p0: Function to call when a p0-tuple was derived.
	inline void derived_p0(int A);

	// derived_p1: Function to call when a p1-tuple was derived.
	inline void derived_p1(int A);

	// derived_tc: Function to call when a tc-tuple was derived.
	inline void derived_tc(int A, int B);

	// rule_1:  p0(A) :- par(1,A).  Start Rule.
	inline void rule_1();

	// rule_2:  p1(A) :- par(1,A).  Start Rule.
	inline void rule_2();

	// rule_3:  p0(A) :- p1(B), par(B,A).  Given: p1(B).
	inline void rule_3(int B);

	// rule_4:  p1(A) :- p1(B), par(B,A).  Given: p1(B).
	inline void rule_4(int B);

	// rule_5:  tc(1,A) :- p0(A).  Given: p0(A).
	inline void rule_5(int A);

//=============================================================================
// End of Class:
//=============================================================================

};


//=============================================================================
// End of Include File:
//=============================================================================

#endif

