// ============================================================================
// Project:	Deductive Database
// Filename:	bench_tcff_1.cpp
// Purpose:	Bottom-Up Evaluation for Transitive Closure (ff) Example
// Last Change:	21.05.2016
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2015-2016 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Local Version Settings:
//=============================================================================

//#define EVAL_DUP_HASHSIZE	10001
#define EVAL_DUP_HASHSIZE	500001

//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"
#include "str.h"
#include "err.h"
#include "check.h"
#include "perf.h"
#include "stack.h"
#include "rel_n_n.h"
#include "cur_n_n.h"
#include "rel_ii.h"
#include "cur_ii.h"
#include "dup_2i.h"
#include "load_tc.h"
#include "out.h"

#include "bench_tcff_1.h"


//=============================================================================
// Static Methods:
//=============================================================================


//-----------------------------------------------------------------------------
// run: Load data file, do bottom-up evaluation for transitive closure (ff).
//-----------------------------------------------------------------------------


long bench_tcff_1_c::run(str_t filename,
			perf_t timer_load, perf_t timer_eval)
{
	// Counter for number of results:
	long num_results = 0;

	// --------------------------------------------------------------------
	// Relations for EDB body literals:
	// --------------------------------------------------------------------

	rel_ii_c  par_ff("par_ff");
	rel_n_n_c par_fb("par_fb");

	// --------------------------------------------------------------------
	// Relations for duplicate check (IDB predicates):
	// --------------------------------------------------------------------

	dup_2i_c tc_bb("tc_bb (Duplicate Check)", EVAL_DUP_HASHSIZE);
	if(!tc_bb.init_ok())
		return -1;

	// --------------------------------------------------------------------
	// Load data file, fill domains and EDB relations:
	// --------------------------------------------------------------------


	// Start timer for data load:
	timer_load->start();

	// Load the data:
	if(!load_tc_c::load(filename, &par_ff, &par_fb))
		return -1;

	// Stop timer for data load:
	timer_load->stop();

	//---------------------------------------------------------------------
	// Now do bottom-up evaluation.
	//---------------------------------------------------------------------

	// Start timer for bottom-up evaluation:
	timer_eval->start();

	// --------------------------------------------------------------------
	// Transformed program:
	//
	// 1: answer(A,B) :- tc(A,B).
	// 2: tc(A,B) :- par(A,B).
	// 3: tc(A,B) :- par(A,C), tc(C,B).
	// --------------------------------------------------------------------

	// --------------------------------------------------------------------
	// Declaration Part:
	// --------------------------------------------------------------------

	int v2_1 = -1;
	int v2_2 = -1;
	int v3_1 = -1;

	// par_cursor_ff c2_1_ff;
	// par_cursor_fb c3_1_fb;
	cur_ii_c  c2_1_ff(&par_ff);
	cur_n_n_c c3_1_fb(&par_fb);

	stack_c<int> backtrack_stack("backtrack_stack");
	stack_c<int> int_value_stack("int_value_stack");
	// float_stack float_value_stack;
	// string_stack string_value_stack;

	// --------------------------------------------------------------------
	// Jump labels:
	// label 1: applying rule 2 resulting in symbolic fact tc(v2_1,v2_2)
	// label 2: using symbolic fact tc(v2_1,v2_2) with literal 1 in rule 1
	//		resulting in symbolic fact answer(v2_1,v2_2)
	// label 3: using symbolic fact tc(v2_1,v2_2) with literal 2 in rule 3
	//		resulting in symbolic fact tc(v3_1,v2_2)
	// label 4: using symbolic fact tc(v3_1,v2_2) with literal 1 in rule 1
	//		resulting in symbolic fact answer(v3_1,v2_2)
	// label 5: using symbolic fact tc(v3_1,v2_2) with literal 2 in rule 3
	//		resulting in symbolic fact tc(v3_1,v2_2)
	// --------------------------------------------------------------------


	// Initialisation Part:
	if(!backtrack_stack.push(1))
		return -1;

	// Loop Part:
	while(!backtrack_stack.is_empty()) {
		switch(backtrack_stack.pop()) {
		case 1:
			c2_1_ff.open();
		// Continuation for label 1
		case 7:
			while (c2_1_ff.fetch()) {
            
				v2_1 = c2_1_ff.c1();
				v2_2 = c2_1_ff.c2();
				// Derived tc(v2_1,v2_2)

				// Duplicate check:
				int dup = tc_bb.dup_check(v2_1, v2_2);
				if(dup < 0)
					return -1;
				if(dup > 0)
					continue;

				// Push backtrack label to continuation
				if(!backtrack_stack.push(7))
					return -1;
				// Labels for next rules
				if(!backtrack_stack.push(3))
					return -1;
				goto label_2;
			}
			break;
		case 2: label_2:

			// answer(v2_1,v2_2):
			if(v2_1 < 0 || v2_2 < 0)
				return -1;

			num_results++;
			out_c::answer_2(num_results, v2_1, v2_2);

			break;
		case 3: // label_3:
 
			c3_1_fb.open(v2_1);
		// Continuation for label 3
		case 9:
			while (c3_1_fb.fetch()) {
 
				v3_1 = c3_1_fb.c2(); // Really c1
				// Derived: tc(v3_1,v2_2)

				// Duplicate check:
				int dup = tc_bb.dup_check(v3_1, v2_2);
				if(dup < 0)
					return -1;
				if(dup > 0)
					continue;

				// Push backtrack label to continuation
				if(!backtrack_stack.push(9))
					return -1;
				// Labels for next rules
				if(!backtrack_stack.push(5))
					return -1;
				goto label_4;
			}
			break;
		case 4: label_4:

			// answer(v3_1,v2_2):
			if(v3_1 < 0 || v2_2 < 0)
				return -1;

			num_results++;
			out_c::answer_2(num_results, v3_1, v2_2);
 
			break;
		case 5: // label_5:
			//recursive
			if(!c3_1_fb.push(&int_value_stack))
				return -1;
			if(!int_value_stack.push(v3_1))
				return -1;
			// Push backtrack label to restore values
			if(!backtrack_stack.push(17))
				return -1;
 
			c3_1_fb.open(v3_1);
		// Continuation for label 5
		case 11:
			while (c3_1_fb.fetch()) {

				v3_1 = c3_1_fb.c2(); // Really c1
				// Derived: tc(v3_1,v2_2)

				// Duplicate check:
				int dup = tc_bb.dup_check(v3_1, v2_2);
				if(dup < 0)
					return -1;
				if(dup > 0)
					continue;

				// Push backtrack label to continuation
				if(!backtrack_stack.push(11))
					return -1;
				// Labels for next rules
				if(!backtrack_stack.push(5))
					return -1;
				goto label_4;
			}
			break;
		// Label to restore values of label 5
		case 17:
			v3_1 = int_value_stack.pop();
			c3_1_fb.pop(&int_value_stack);
			break;
 
		}
	}

	// Stop timer for bottom-up evaluation:
	timer_eval->stop();

	// Print table of memory pools if requested:
	out_c::mem_pools_tab();

	// Print relation statistics if requested:
	if(out_c::show_rel()) {
		out_c::rel_size_headline();
		out_c::rel_size_nonstd("par_ff", par_ff.num_rows(),
					par_ff.num_pages());
		out_c::rel_size_nonstd("par_fb", par_fb.num_rows(),
					par_fb.num_pages());
		out_c::rel_size_nonstd("tc_bb", tc_bb.num_rows(),
					tc_bb.num_pages());
		out_c::rel_size_done();
	}

	// Debug output for duplicate elimination table:
	int num_rows     = tc_bb.num_rows();
	int num_pages    = tc_bb.num_pages();
	int hash_size    = tc_bb.hash_size();
	int entries_used = tc_bb.entries_used();

	out_c::debug_msg(      "Duplicate check table tc_bb:\n");
	out_c::debug_ival(     "    Number of rows:         ", num_rows);
	out_c::debug_mempages( "    Number of memory pages: ", num_pages);
	out_c::debug_ival(     "    Hash table size:        ", hash_size);
	out_c::debug_hashusage("    Entries Used:           ", entries_used,
								hash_size);
	out_c::debug_hashchain("    Average Chain Length:   ", entries_used,
								num_rows);
	out_c::debug_nl();

	// Return number of result tuples:
	return num_results;
}


//-----------------------------------------------------------------------------
// file: List of data files.
//-----------------------------------------------------------------------------

str_t bench_tcff_1_c::file(int i)
{
	switch(i) {
		case 0:
			return VER_FILE_TC_TEST;
		case 1:
			return VER_FILE_TC_50K;
		case 2:
			return VER_FILE_TC_50K_CYC;
		case 3:
			return VER_FILE_TC_250K;
		case 4:
			return VER_FILE_TC_250K_CYC;
		case 5:
			return VER_FILE_TC_500K;
		case 6:
			return VER_FILE_TC_500K_CYC;
		case 7:
			return VER_FILE_TC_500K_D2;
		case 8:
			return VER_FILE_TC_500K_D2_CYC;
		case 9:
			return VER_FILE_TC_1M_D2;
		case 10:
			return VER_FILE_TC_1M_D2_CYC;
		default:
			return STR_NULL;
	}
}


//-----------------------------------------------------------------------------
// result: List of correct results of eval method.
//-----------------------------------------------------------------------------

long bench_tcff_1_c::result(int i)
{
	switch(i) {
		case 0: return 20;
			// tc(1,2). tc(1,3). tc(1,20). tc(1,21). tc(1,22).
			// tc(2,3). tc(2,20). tc(2,21). tc(2,22).
			// tc(4,5). tc(4,6).
			// tc(5,6).
			// tc(10,2). tc(10,3). tc(10,20). tc(10,21). tc(10,22).
			// tc(20,21). tc(20,22).
			// tc(21,22).
		case 1: return -1;
		case 2: return -1;
		case 3: return -1;
		case 4: return -1;
		case 5: return -1;
		case 6: return -1;
		case 7: return -1;
		case 8: return -1;
		case 9: return -1;
		case 10: return -1;
		default: return -1;
	}
}

