// ============================================================================
// Project:	Deductive Database
// Filename:	bench_tcff_2.cpp
// Purpose:	Bottom-Up Evaluation for tc(X,Y), Ver 2: Cursor interface
// Last Change:	26.05.2016
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2015-2016 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"
#include "str.h"
#include "err.h"
#include "check.h"
#include "stack.h"
#include "list_2.h"
#include "cur_2.h"
#include "dup_2i.h"
#include "rel_n_n.h"
#include "cur_n_n.h"
#include "pred.h"
#include "load.h"
#include "out.h"

#include "bench_tcff_2.h"


//=============================================================================
// Local Constants:
//=============================================================================

//-----------------------------------------------------------------------------
// Hash Table Size for Duplicate Elimination:
//-----------------------------------------------------------------------------

static const int BENCH_TCFF2_HASHSIZE = 500001;

//=============================================================================
// Constructor, Destructor:
//=============================================================================

//-----------------------------------------------------------------------------
// Constructor (initializes subobjects):
//-----------------------------------------------------------------------------

bench_tcff_2_c::bench_tcff_2_c() :

		// Superclass:
		bench_c(),

		// Relations for EDB body literals:
		par_ff("par_ff"),
		par_fb("par_fb"),

		// Relations for duplicate check (IDB predicates):
		tc_bb("tc_bb (DupCheck)", BENCH_TCFF2_HASHSIZE),

		// Cursors for loops during bottom-up evaluation:
		c2_1_ff(&par_ff),
		c3_1_fb(&par_fb),

		// Backtrack stack and value restore stack:
		backtrack_stack("Backtrack Stack"),
		int_value_stack("int_value_stack")

{
	// Initialize variables for bottom-up evaluation:
	v2_1 = -1;
	v2_2 = -1;
	v3_1 = -1;

	// Initialize variable for exit point:
	first_exit_point = true;

	// No error occurred so far:
	Error = false;

	// Check whether tc_bb was correctly initialized (got memory):
	if(!tc_bb.init_ok())
		Error = true;

	// Set Magic number:
	CHECK_CODE(Magic = BENCH_TCFF_2_MAGIC);
}

//-----------------------------------------------------------------------------
// Destructor:
//-----------------------------------------------------------------------------

bench_tcff_2_c::~bench_tcff_2_c()
{
	// Check this object:
	CHECK_VALID("bench_tcff_2_c::destructor");

	// Make this object invalid:
	CHECK_CODE(Magic = 0);
}


//=============================================================================
// Public Methods:
//=============================================================================

//-----------------------------------------------------------------------------
// load: Load data file with EDB facts.
//-----------------------------------------------------------------------------

bool bench_tcff_2_c::load(str_t filename) {

	// Check parameter:
	CHECK(filename != STR_NULL, "bench_tcff_2_c::load: filename is null");

	// Check object:
	CHECK_VALID("bench_tcff_2_c::load");

	// Loader object (will contain data about predicates):
	load_c loader;

	// Predicate (input to loader):
	pred_c par_p("par", ARGTYPE_INT, ARGTYPE_INT);
	par_p.add_rel(&par_ff);
	par_p.add_rel(&par_fb, BIND_FB);

	// Add predicate definition to loader:
	loader.add(&par_p);

	// Load data file:
	return loader.load(filename);
}

//-----------------------------------------------------------------------------
// eval: Do bottom-up evaluation.
//-----------------------------------------------------------------------------

long bench_tcff_2_c::eval()
{

	// number of answers found:
	long num_results = 0;

	// Open Cursor:
	open();

	// Loop while results are found:
	while(fetch()) {

		// Get results:
		int ans_1 = c1();
		int ans_2 = c2();

		// Check whether results are ok:
		if(ans_1 < 0)
			return -1;
		if(ans_2 < 0)
			return -1;

		// Increment number of answers found:
		num_results++;

		// Print the result if below limit:
		out_c::answer_2(num_results, ans_1, ans_2);

	// End of fetch loop:
	}

	// Check for errors:
	if(error())
		return -1;

	// Close cursor:
	close();

	// Print relation statistics if requested:
	if(out_c::show_rel()) {
		out_c::rel_size_headline();
		out_c::rel_size(&par_ff);
		out_c::rel_size(&par_fb);
		out_c::rel_size_nonstd("tc_bb", tc_bb.num_rows(),
						tc_bb.num_pages());
		out_c::rel_size_done();
	}

	// Return number of result tuples:
	return num_results;
}


//-----------------------------------------------------------------------------
// Open cursor (note: load() must be called before this method).
//-----------------------------------------------------------------------------

void bench_tcff_2_c::open()
{
	// --------------------------------------------------------------------
	// Transformed program:
	//
	// 1: answer(A,B) :- tc(A,B).
	// 2: tc(A,B) :- par(A,B).
	// 3: tc(A,B) :- par(A,C), tc(C,B).
	// --------------------------------------------------------------------

	// --------------------------------------------------------------------
	// Jump labels:
	// label 1: applying rule 2 resulting in symbolic fact tc(v2_1,v2_2)
	// label 2: using symbolic fact tc(v2_1,v2_2) with literal 1 in rule 1
	//		resulting in symbolic fact answer(v2_1,v2_2)
	// label 3: using symbolic fact tc(v2_1,v2_2) with literal 2 in rule 3
	//		resulting in symbolic fact tc(v3_1,v2_2)
	// label 4: using symbolic fact tc(v3_1,v2_2) with literal 1 in rule 1
	//		resulting in symbolic fact answer(v3_1,v2_2)
	// label 5: using symbolic fact tc(v3_1,v2_2) with literal 2 in rule 3
	//		resulting in symbolic fact tc(v3_1,v2_2)
	// --------------------------------------------------------------------


	// Initialisation Part:
	if(!backtrack_stack.push(1))
		Error = true;

}

//-----------------------------------------------------------------------------
// fetch: Get next tuple, returns false if there is no further tuple.
//-----------------------------------------------------------------------------

bool bench_tcff_2_c::fetch()
{

	// Loop Part:
	while(!backtrack_stack.is_empty()) {
		OUT_TRACE_ITERATION(backtrack_stack.num_elems());
		switch(backtrack_stack.pop()) {
		case 1:
			OUT_TRACE_START(1, "tc(X,Y) :- par(X,Y).", "EDB:par");
			c2_1_ff.open();
		// Continuation for label 1
		case 7:
			OUT_TRACE_CONT(1, "tc(X,Y) :- par(X,Y).", "EDB:par");
			while (c2_1_ff.fetch()) {
            
				v2_1 = c2_1_ff.c1();
				OUT_TRACE_ASSIGN("v2_1", v2_1);
				v2_2 = c2_1_ff.c2();
				OUT_TRACE_ASSIGN("v2_2", v2_2);
				// Derived tc(v2_1,v2_2)

				// Duplicate check:
				int dup = tc_bb.dup_check(v2_1, v2_2);
				if(dup < 0) {
					Error = true;
					return false;
				}
				if(dup > 0)
					continue;

				// Push backtrack label to continuation
				if(!backtrack_stack.push(7)) {
					Error = true;
					return false;
				}
				// Labels for next rules
				if(!backtrack_stack.push(3)) {
					Error = true;
					return false;
				}
				goto label_2;
			}
			break;
		case 2: label_2:
			OUT_TRACE_ANSWER(2, "answer(X,Y) :- tc(X,Y).",
						"tc(v2_1,v2_2)");

			// answer(v2_1,v2_2):
			first_exit_point = true;
			return true;

			//break;
		case 3: // label_3:
			OUT_TRACE_START(3, "tc(X,Z) :- par(X,Y), tc(Y,Z).",
						"tc(v2_1,v2_2)");
 
			c3_1_fb.open(v2_1);
		// Continuation for label 3
		case 9:
			OUT_TRACE_CONT(9, "tc(X,Z) :- par(X,Y), tc(Y,Z).",
						"tc(v2_1,v2_2)");
			while (c3_1_fb.fetch()) {
 
				v3_1 = c3_1_fb.c2(); // Really c1
				OUT_TRACE_ASSIGN("v3_1", v3_1);
				// Derived: tc(v3_1,v2_2)

				// Duplicate check:
				int dup = tc_bb.dup_check(v3_1, v2_2);
				if(dup < 0) {
					Error = true;
					return false;
				}
				if(dup > 0)
					continue;

				// Push backtrack label to continuation
				if(!backtrack_stack.push(9)) {
					Error = true;
					return false;
				}
				// Labels for next rules
				if(!backtrack_stack.push(5)) {
					Error = true;
					return false;
				}
				goto label_4;
			}
			break;
		case 4: label_4:
			OUT_TRACE_ANSWER(4, "answer(X,Y) :- tc(X,Y).",
						"tc(v3_1,v2_2)");

			// answer(v3_1,v2_2):
			first_exit_point = false;
			return true;
 
			//break;
		case 5: // label_5:
			//recursive
			OUT_TRACE_START_REC(5, "tc(X,Z) :- par(X,Y), tc(Y,Z).",
						"tc(v3_1,v2_2)");
			if(!c3_1_fb.push(&int_value_stack)) {
				Error = true;
				return false;
			}
			if(!int_value_stack.push(v3_1)) {
				Error = true;
				return false;
			}
			// Push backtrack label to restore values
			if(!backtrack_stack.push(17)) {
				Error = true;
				return false;
			}
 
			c3_1_fb.open(v3_1);
		// Continuation for label 5
		case 11:
			OUT_TRACE_CONT(11, "tc(X,Z) :- par(X,Y), tc(Y,Z).",
						"tc(v3_1,v2_2)");
			while (c3_1_fb.fetch()) {

				v3_1 = c3_1_fb.c2(); // Really c1
				OUT_TRACE_ASSIGN("v3_1", v3_1);
				// Derived: tc(v3_1,v2_2)

				// Duplicate check:
				int dup = tc_bb.dup_check(v3_1, v2_2);
				if(dup < 0) {
					Error = true;
					return false;
				}
				if(dup > 0)
					continue;

				// Push backtrack label to continuation
				if(!backtrack_stack.push(11)) {
					Error = true;
					return false;
				}
				// Labels for next rules
				if(!backtrack_stack.push(5)) {
					Error = true;
					return false;
				}
				goto label_4;
			}
			break;
		// Label to restore values of label 5
		case 17:
			OUT_TRACE_RESTORE(17, "tc(X,Z) :- par(X,Y), tc(Y,Z).",
						"tc(v3_1,v2_2)");
			v3_1 = int_value_stack.pop();
			OUT_TRACE_POP_VAR("v3_1", v3_1);
			c3_1_fb.pop(&int_value_stack);
			break;
 
		}
	}

	// Backtrack stack was empty, no more answers:
	return false;
}


//-----------------------------------------------------------------------------
// file: List of data files.
//-----------------------------------------------------------------------------

str_t bench_tcff_2_c::file(int i) const
{
	switch(i) {
		case 0:
			return VER_FILE_TC_TEST;
		case 1:
			return VER_FILE_TC_50K;
		case 2:
			return VER_FILE_TC_50K_CYC;
		case 3:
			return VER_FILE_TC_250K;
		case 4:
			return VER_FILE_TC_250K_CYC;
		case 5:
			return VER_FILE_TC_500K;
		case 6:
			return VER_FILE_TC_500K_CYC;
		case 7:
			return VER_FILE_TC_500K_D2;
		case 8:
			return VER_FILE_TC_500K_D2_CYC;
		case 9:
			return VER_FILE_TC_1M_D2;
		case 10:
			return VER_FILE_TC_1M_D2_CYC;
		default:
			return STR_NULL;
	}
}


//-----------------------------------------------------------------------------
// result: List of correct results of eval method.
//-----------------------------------------------------------------------------

long bench_tcff_2_c::result(int i) const
{
	switch(i) {
		case 0: return 20;
			// tc(1,2). tc(1,3). tc(1,20). tc(1,21). tc(1,22).
			// tc(2,3). tc(2,20). tc(2,21). tc(2,22).
			// tc(4,5). tc(4,6).
			// tc(5,6).
			// tc(10,2). tc(10,3). tc(10,20). tc(10,21). tc(10,22).
			// tc(20,21). tc(20,22).
			// tc(21,22).
		case 1: return -1;
		case 2: return -1;
		case 3: return -1;
		case 4: return -1;
		case 5: return -1;
		case 6: return -1;
		case 7: return -1;
		case 8: return -1;
		case 9: return -1;
		case 10: return -1;
		default: return -1;
	}
}


//=============================================================================
// Debugging Support:
//=============================================================================


//-----------------------------------------------------------------------------
// check: Check the integrity of this object, return error message or NULL.
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
str_t	bench_tcff_2_c::check() const
{
	// Check magic number:
	if(Magic != BENCH_TCFF_2_MAGIC)
		return "wrong magic number";

	// Check sub-objects (relations for EDB predicates):
	CHECK_SUBOBJ(par_ff, "bench_tcff_2_c::check");
	CHECK_SUBOBJ(par_fb, "bench_tcff_2_c::check");
	CHECK_SUBOBJ(tc_bb, "bench_tcff_2_c::check");
	CHECK_SUBOBJ(c2_1_ff, "bench_tcff_2_c::check");
	CHECK_SUBOBJ(c3_1_fb, "bench_tcff_2_c::check");
	CHECK_SUBOBJ(backtrack_stack, "bench_tcff_2_c::check");
	CHECK_SUBOBJ(int_value_stack, "bench_tcff_2_c::check");

	// Variables are non-negative once set, but initialized to -1:
	if(v2_1 < -1)
		return "v2_1 is invalid";
	if(v2_2 < -1)
		return "v2_2 is invalid";
	if(v3_1 < -1)
		return "v3_1 is invalid";

	// Ok:
	return STR_NULL;
}

#endif


//-----------------------------------------------------------------------------
// dump: Show data structure.
//-----------------------------------------------------------------------------

#ifdef	VER_DUMP
void bench_tcff_2_c::dump(str_t headline) const
{
	// Headline:
	if(headline == STR_NULL)
		headline = "tc(X,Y) Benchmark Object, Ver. 2";
	err_c::dump_begin(headline);

	// Basic information:
	err_c::dump_str("ObjectSize:  ");
	err_c::dump_uint(static_cast<unsigned int>(sizeof(bench_tcff_2_c)));
	err_c::dump_str(" Bytes");
	err_c::dump_nl();
	err_c::dump_nl();

	// Show number of rows in all EDB relations:
	err_c::dump_str("Number of rows in EDB relations:  ");
	err_c::dump_nl();
	err_c::dump_str("    par_ff: ");
	err_c::dump_int(par_ff.num_rows());
	err_c::dump_nl();
	err_c::dump_str("    par_fb: ");
	err_c::dump_int(par_fb.num_rows());
	err_c::dump_nl();
	err_c::dump_nl();

	// Show number of rows in IDB relations (duplicate check):
	err_c::dump_str("Number of rows in IDB relations:");
	err_c::dump_nl();
	err_c::dump_str("    tc_bb:  ");
	err_c::dump_int(tc_bb.num_rows());
	err_c::dump_nl();
	err_c::dump_nl();

	// Variables:
	err_c::dump_str("Variables:");
	err_c::dump_nl();
	err_c::dump_str("    v2_1:   ");
	err_c::dump_int(v2_1);
	err_c::dump_nl();
	err_c::dump_str("    v2_2:   ");
	err_c::dump_int(v2_2);
	err_c::dump_nl();
	err_c::dump_str("    v3_1:   ");
	err_c::dump_int(v3_1);
	err_c::dump_nl();
	err_c::dump_str("    first_exit_point: ");
	err_c::dump_bool(first_exit_point);
	err_c::dump_nl();
	err_c::dump_str("    Error:   ");
	err_c::dump_bool(Error);
	err_c::dump_nl();
	err_c::dump_nl();

	// Backtrack Stack:
	err_c::dump_str("Stacks:");
	err_c::dump_nl();
	err_c::dump_str("    Backtrack Stack Size:   ");
	err_c::dump_int(backtrack_stack.num_elems());
	err_c::dump_nl();
	err_c::dump_str("    Int Value Stack Size:   ");
	err_c::dump_int(int_value_stack.num_elems());
	err_c::dump_nl();

	// Done:
	err_c::dump_end();
}
#endif

