// ============================================================================
// Project:	Deductive Database
// Filename:	bench_tcff_3.cpp
// Purpose:	Bottom-Up Evaluation for tc(X,Y), Ver 3: Seminaive Evaluation
// Last Change:	15.06.2016
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2015-2016 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"
#include "str.h"
#include "err.h"
#include "check.h"
#include "stack.h"
#include "list_2.h"
#include "cur_2.h"
#include "dup_2i.h"
#include "rel_n_n.h"
#include "cur_n_n.h"
#include "pred.h"
#include "load.h"
#include "out.h"

#include "bench_tcff_3.h"


//=============================================================================
// Local Constants:
//=============================================================================

//-----------------------------------------------------------------------------
// Hash Table Size for Duplicate Elimination:
//-----------------------------------------------------------------------------

static const int BENCH_TCFF2_HASHSIZE = 500001;

//=============================================================================
// Constructor, Destructor:
//=============================================================================

//-----------------------------------------------------------------------------
// Constructor (initializes subobjects):
//-----------------------------------------------------------------------------

bench_tcff_3_c::bench_tcff_3_c() :

		// Superclass:
		bench_c(),

		// Relations for EDB body literals:
		par_ff("par_ff"),
		par_fb("par_fb"),

		// Relations for duplicate check (IDB predicates):
		tc_bb("tc_bb (DupCheck)", BENCH_TCFF2_HASHSIZE),

		// Relation to store the result:
		tc_ff("tc_ff (Result)"),

		// Cursors for result:
		result_cur(&tc_ff)
{
	// No error occurred so far:
	Error = false;

	// Check whether tc_bb was correctly initialized (got memory):
	if(!tc_bb.init_ok())
		Error = true;

	// Set Magic number:
	CHECK_CODE(Magic = BENCH_TCFF_3_MAGIC);
}

//-----------------------------------------------------------------------------
// Destructor:
//-----------------------------------------------------------------------------

bench_tcff_3_c::~bench_tcff_3_c()
{
	// Check this object:
	CHECK_VALID("bench_tcff_3_c::destructor");

	// Make this object invalid:
	CHECK_CODE(Magic = 0);
}


//=============================================================================
// Public Methods:
//=============================================================================

//-----------------------------------------------------------------------------
// load: Load data file with EDB facts.
//-----------------------------------------------------------------------------

bool bench_tcff_3_c::load(str_t filename) {

	// Check parameter:
	CHECK(filename != STR_NULL, "bench_tcff_3_c::load: filename is null");

	// Check object:
	CHECK_VALID("bench_tcff_3_c::load");

	// Loader object (will contain data about predicates):
	load_c loader;

	// Predicate (input to loader):
	pred_c par_p("par", ARGTYPE_INT, ARGTYPE_INT);
	par_p.add_rel(&par_ff);
	par_p.add_rel(&par_fb, BIND_FB);

	// Add predicate definition to loader:
	loader.add(&par_p);

	// Load data file:
	return loader.load(filename);
}

//-----------------------------------------------------------------------------
// eval: Do bottom-up evaluation.
//-----------------------------------------------------------------------------

long bench_tcff_3_c::eval()
{

	// number of answers found:
	long num_results = 0;

	// Open Cursor:
	open();

	// Loop while results are found:
	while(fetch()) {

		// Get results:
		int ans_1 = c1();
		int ans_2 = c2();

		// Check whether results are ok:
		if(ans_1 < 0)
			return -1;
		if(ans_2 < 0)
			return -1;

		// Increment number of answers found:
		num_results++;

		// Print the result if below limit:
		out_c::answer_2(num_results, ans_1, ans_2);

	// End of fetch loop:
	}

	// Check for errors:
	if(error())
		return -1;

	// Close cursor:
	close();

	// Print relation statistics if requested:
	if(out_c::show_rel()) {
		out_c::rel_size_headline();
		out_c::rel_size(&par_ff);
		out_c::rel_size(&par_fb);
		out_c::rel_size(&tc_ff);
		out_c::rel_size_nonstd("tc_bb", tc_bb.num_rows(),
						tc_bb.num_pages());
		out_c::rel_size_done();
	}

	// Return number of result tuples:
	return num_results;
}


//-----------------------------------------------------------------------------
// Open cursor (note: load() must be called before this method).
//-----------------------------------------------------------------------------

void bench_tcff_3_c::open()
{
	// This does the seminaive bottom-up evaluation.
	bool changed = false;

	// Rule 1: tc(X,Y) :- par(X,Y).
	cur_2_c par_ff_cur(&par_ff);
	par_ff_cur.open();
	while(par_ff_cur.fetch()) {
		int X = par_ff_cur.c1();
		int Y = par_ff_cur.c2();
		// Duplicate check:
		int dup = tc_bb.dup_check(X, Y);
		if(dup < 0) {
			Error = true;
			return;
		}
		if(dup == 0) {
			// No duplicate, insert into result:
			tc_ff.insert(X, Y);
			changed = true;
		}
	}
	par_ff_cur.close();

	// Rule 2: tc(X,Y) :- par(X,Z), tc(Z,Y).
	// Seminaive iteration:
	cur_2_c tc_ff_cur(&tc_ff);
	cur_n_n_c par_fb_cur(&par_fb);
	int start_pos = 0;
	while(changed) {
		changed = false;

		// Open cursor over new tc facts:
		tc_ff_cur.open(start_pos);
		start_pos = tc_ff.num_rows();

		// Loop over new tc facts:
		while(tc_ff_cur.fetch()) {
			int Z = tc_ff_cur.c1();
			int Y = tc_ff_cur.c2();

			// Loop over matching facts par(X,Z):
			par_fb_cur.open(Z);
			while(par_fb_cur.fetch()) {
				int X = par_fb_cur.c2();

				// Duplicate check:
				int dup = tc_bb.dup_check(X, Y);
				if(dup < 0) {
					Error = true;
					return;
				}
				if(dup == 0) {
					// No duplicate, insert into result:
					tc_ff.insert(X, Y);
					changed = true;
				}
			}
			// cur_n_n_c has no close operation:
			//par_fb_cur.close();
		}
		tc_ff_cur.close();

	// End of seminaive iteration:
	}

	// Now open the cursor over the result:
	result_cur.open();
}

//-----------------------------------------------------------------------------
// fetch: Get next tuple, returns false if there is no further tuple.
//-----------------------------------------------------------------------------

bool bench_tcff_3_c::fetch()
{
	// Delegate to result cursor:
	return result_cur.fetch();
}


//-----------------------------------------------------------------------------
// file: List of data files.
//-----------------------------------------------------------------------------

str_t bench_tcff_3_c::file(int i) const
{
	switch(i) {
		case 0:
			return VER_FILE_TC_TEST;
		case 1:
			return VER_FILE_TC_50K;
		case 2:
			return VER_FILE_TC_50K_CYC;
		case 3:
			return VER_FILE_TC_250K;
		case 4:
			return VER_FILE_TC_250K_CYC;
		case 5:
			return VER_FILE_TC_500K;
		case 6:
			return VER_FILE_TC_500K_CYC;
		case 7:
			return VER_FILE_TC_500K_D2;
		case 8:
			return VER_FILE_TC_500K_D2_CYC;
		case 9:
			return VER_FILE_TC_1M_D2;
		case 10:
			return VER_FILE_TC_1M_D2_CYC;
		default:
			return STR_NULL;
	}
}


//-----------------------------------------------------------------------------
// result: List of correct results of eval method.
//-----------------------------------------------------------------------------

long bench_tcff_3_c::result(int i) const
{
	switch(i) {
		case 0: return 20;
			// tc(1,2). tc(1,3). tc(1,20). tc(1,21). tc(1,22).
			// tc(2,3). tc(2,20). tc(2,21). tc(2,22).
			// tc(4,5). tc(4,6).
			// tc(5,6).
			// tc(10,2). tc(10,3). tc(10,20). tc(10,21). tc(10,22).
			// tc(20,21). tc(20,22).
			// tc(21,22).
		case 1: return -1;
		case 2: return -1;
		case 3: return -1;
		case 4: return -1;
		case 5: return -1;
		case 6: return -1;
		case 7: return -1;
		case 8: return -1;
		case 9: return -1;
		case 10: return -1;
		default: return -1;
	}
}


//=============================================================================
// Debugging Support:
//=============================================================================


//-----------------------------------------------------------------------------
// check: Check the integrity of this object, return error message or NULL.
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
str_t	bench_tcff_3_c::check() const
{
	// Check magic number:
	if(Magic != BENCH_TCFF_3_MAGIC)
		return "wrong magic number";

	// Check sub-objects (relations for EDB predicates):
	CHECK_SUBOBJ(par_ff, "bench_tcff_3_c::check");
	CHECK_SUBOBJ(par_fb, "bench_tcff_3_c::check");
	CHECK_SUBOBJ(tc_bb, "bench_tcff_3_c::check");
	CHECK_SUBOBJ(tc_ff, "bench_tcff_3_c::check");
	CHECK_SUBOBJ(result_cur, "bench_tcff_3_c::check");

	// Ok:
	return STR_NULL;
}

#endif


//-----------------------------------------------------------------------------
// dump: Show data structure.
//-----------------------------------------------------------------------------

#ifdef	VER_DUMP
void bench_tcff_3_c::dump(str_t headline) const
{
	// Headline:
	if(headline == STR_NULL)
		headline = "tc(X,Y) Benchmark Object, Ver. 2";
	err_c::dump_begin(headline);

	// Basic information:
	err_c::dump_str("ObjectSize:  ");
	err_c::dump_uint(static_cast<unsigned int>(sizeof(bench_tcff_3_c)));
	err_c::dump_str(" Bytes");
	err_c::dump_nl();
	err_c::dump_nl();

	// Show number of rows in all EDB relations:
	err_c::dump_str("Number of rows in EDB relations:  ");
	err_c::dump_nl();
	err_c::dump_str("    par_ff: ");
	err_c::dump_int(par_ff.num_rows());
	err_c::dump_nl();
	err_c::dump_str("    par_fb: ");
	err_c::dump_int(par_fb.num_rows());
	err_c::dump_nl();
	err_c::dump_nl();

	// Show number of rows in IDB relations (duplicate check):
	err_c::dump_str("Number of rows in IDB relations:");
	err_c::dump_nl();
	err_c::dump_str("    tc_bb:  ");
	err_c::dump_int(tc_bb.num_rows());
	err_c::dump_nl();
	err_c::dump_str("    tc_ff:  ");
	err_c::dump_int(tc_ff.num_rows());
	err_c::dump_nl();
	err_c::dump_nl();

	// Done:
	err_c::dump_end();
}
#endif

