// ============================================================================
// Project:	Deductive Database
// Filename:	cur_ii.h
// Purpose:	Cursor for accessing rel_ii (full table scan).
// Last Change:	16.11.2015
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2010-2015 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Include File Frame:
//=============================================================================

#ifndef	CUR_II_INCLUDED
#define CUR_II_INCLUDED

//=============================================================================
// Used Types and Macros:
//=============================================================================

#ifndef VER_INCLUDED
#include "ver.h"
#endif

#ifndef STR_INCLUDED
#include "str.h"
#endif

#ifndef CHECK_INCLUDED
#include "check.h"
#endif

#ifndef REL_II_INCLUDED
#include "rel_ii.h"
#endif


//=============================================================================
// Private Constants:
//=============================================================================

//-----------------------------------------------------------------------------
// CUR_II_MAGIC: Magic number (identifies objects of this class).
//-----------------------------------------------------------------------------

static const long CUR_II_MAGIC = 0x434E4E0AL;	// 'RNN\n'

//=============================================================================
// Class for Cursor to Relation with two non-negative integer columns:
//=============================================================================

class cur_ii_c {
	public:

//-----------------------------------------------------------------------------
// Constructor, Destructor:
//-----------------------------------------------------------------------------

	// Constructor:
	cur_ii_c(rel_ii_t rel);

	// Destructor:
	~cur_ii_c();

//-----------------------------------------------------------------------------
// Methods:
//-----------------------------------------------------------------------------

	// open: Open cursor over entire relation rel_ii (full table scan).
	void open() {
		// Check this object:
		CHECK_VALID("cur_ii_c::open");

		// Initialize position before first tuple:
		Pos = -2;
	}

	// fetch: Get next tuple, returns false if there is no further tuple.
	inline bool fetch() {
		// Check this object:
		CHECK_VALID("cur_ii_c::fetch");

		// After increase by 2, Pos must be a valid TupleArr index:
		if(Pos >= Rel->TupleArr.length() - 3)
			return false;

		// Otherwise simply increase position by 2
		// (each tuple uses 2 entries):
		Pos += 2;
		return true;
	}

	// c1: Return value of column 1 of current tuple.
	inline int c1() {
		// Check this object:
		CHECK_VALID("cur_ii_c::c1");

		// Check state:
		CHECK(Pos >= 0, "cur_ii_c::c1: cursor not open");
		CHECK(Pos < Rel->TupleArr.length() - 1,
				"cur_ii_c::c1: fetch beyond end");

		// Return current value for column c1:
		return Rel->TupleArr.get(Pos);
	}

	// c2: Return value of column 2 of current tuple.
	inline int c2() {
		// Check this object:
		CHECK_VALID("cur_ii_c::c2");

		// Check state:
		CHECK(Pos >= 0, "cur_ii_c::c2: cursor not open");
		CHECK(Pos < Rel->TupleArr.length() - 1,
				"cur_ii_c::c2: fetch beyond end");

		// Return current value for column c2:
		return Rel->TupleArr.get(Pos + 1);
	}


//-----------------------------------------------------------------------------
// Debugging Support:
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
	public:
	// Integrity check:
	str_t	check() const;

	private:
	// Magic number (identifies objects of this class for debugging).
	long	Magic;		// Must be "CUR_II_MAGIC".
#endif

#ifdef	VER_DUMP
	public:
	// Display data structure:
	void dump(str_t headline = STR_NULL) const;
#endif

//-----------------------------------------------------------------------------
// Copy-constructor and assignment operator are not supported for this class:
//-----------------------------------------------------------------------------

	private:

	cur_ii_c(const cur_ii_c& obj);		// Not implemented
	cur_ii_c& operator=(const cur_ii_c& obj);	// Not implemented

//-----------------------------------------------------------------------------
// Private Object Members:
//-----------------------------------------------------------------------------

	private:

	// Relation, over which this cursor runs:
	rel_ii_t Rel;

	// Position in tuple array (2 * number of current tuple).
	int Pos;

//=============================================================================
// End of Class:
//=============================================================================

};

//-----------------------------------------------------------------------------
// Define pointer type:
//-----------------------------------------------------------------------------

typedef cur_ii_c *cur_ii_t;

//-----------------------------------------------------------------------------
// Define null pointer:
//-----------------------------------------------------------------------------

#define CUR_II_NULL (static_cast<cur_ii_t>(0))

//=============================================================================
// End of Include File:
//=============================================================================

#endif

