// ============================================================================
// Project:	Deductive Database
// Filename:	cur_n_n.cpp
// Purpose:	Cursor for accessing cur_n_n with given first argument.
// Last Change:	09.09.2015
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2010-2015 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"
#include "str.h"
#ifdef	VER_DUMP
#include "err.h"
#endif
#include "check.h"
#include "rel_n_n.h"

#include "cur_n_n.h"


//=============================================================================
// Constructor and Destructor:
//=============================================================================


//-----------------------------------------------------------------------------
// Constructor:
//-----------------------------------------------------------------------------

cur_n_n_c::cur_n_n_c(rel_n_n_t rel)
{
	// Set Magic number:
	CHECK_CODE(Magic = CUR_N_N_MAGIC);

	// Initialize attributes:
	Rel = rel;
	C2  = -1;
	Pos = -1;
}

//-----------------------------------------------------------------------------
// Destructor:
//-----------------------------------------------------------------------------

cur_n_n_c::~cur_n_n_c()
{
	// Check this object:
	CHECK_VALID("cur_n_n_c::destructor");

	// Make this object invalid:
	CHECK_CODE(Magic = 0);
}


//-----------------------------------------------------------------------------
// open: Open cursor over the value set for column c2, given c1.
//-----------------------------------------------------------------------------

void cur_n_n_c::open(int c1)
{
	// Check this object:
	CHECK_VALID("cur_n_n_c::open");

	// Check parameters:
	CHECK(c1 >= 0, "cur_n_n_c::open: c1 is negative");

	// Store index in cursor:
	Pos = c1;
}

//-----------------------------------------------------------------------------
// fetch: Get first/next tuple, returns false if there is no further tuple.
//-----------------------------------------------------------------------------

bool cur_n_n_c::fetch()
{
	// Check this object:
	CHECK_VALID("cur_n_n_c::fetch");

	// Position -1 means that there is no further tuple:
	if(Pos == -1) {
		C2 = -1;
		return false;
	}

	// If the position is non-negative, we must look into the Index array:
	if(Pos >= 0) {
		// If array too short, there is no tuple for this c1 value:
		if(Rel->IndexArr.length() <= Pos) {
			C2 = -1;
			Pos = -1;
			return false;
		}

		// Get entry in index array:
		int val = Rel->IndexArr.get(Pos);

		// Value found. There is only one value.
		if(val >= 0) {
			C2 = val;
			Pos = -1;
			return true;
		}

		// There is no value if the entry is -1:
		if(val == -1) {
			C2 = -1;
			Pos = -1;
			return false;
		}

		//If the value is negative, we must look at linked list:
		Pos = val;
	}

	// Indexes -2 and smaller mean to fetch value from linked list:
	int value_arr_pos = -Pos - 2;
	CHECK(value_arr_pos >= 0, "cur_n_n_c::fetch: Negative ValueArr index");
	CHECK(value_arr_pos < Rel->ValueArr.length() - 1,
			"cur_n_n_c::fetch: ValueArr index too large");
	C2 = Rel->ValueArr.get(value_arr_pos);
	Pos = Rel->ValueArr.get(value_arr_pos + 1);
	return true;
}


//=============================================================================
// Debugging Support:
//=============================================================================

//-----------------------------------------------------------------------------
// check: Check the integrity of this object, return error message or NULL.
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
str_t	cur_n_n_c::check() const
{
	// Check magic number:
	if(Magic != CUR_N_N_MAGIC)
		return "wrong magic number";

	// Check pointer to relation:
	if(Rel == REL_N_N_NULL)
		return "cursor with null relation";

	// Column values are >= 0, -1 means "none":
	if(C2 < -1)
		return "invalid value for column 2";

	// Ok:
	return 0;
}
#endif


//-----------------------------------------------------------------------------
// dump: Show data structure.
//-----------------------------------------------------------------------------

#ifdef	VER_DUMP
void	cur_n_n_c::dump(str_t headline) const
{
	// Headline:
	if(headline == STR_NULL)
		headline =
		"Cursor for rel_n_n (Relation with two natural number cols)";
	err_c::dump_begin(headline);

	// Relation name:
	err_c::dump_str("Relation:     '");
	err_c::dump_str(Rel->name());
	err_c::dump_str("'");
	err_c::dump_nl();

	// Value for column 2:
	err_c::dump_str("Col.2 Value:   ");
	err_c::dump_int(C2);
	if(C2 == -1)
		err_c::dump_str("  [None]");
	err_c::dump_nl();

	// Next position:
	err_c::dump_str("Next Position: ");
	err_c::dump_int(Pos);
	if(Pos >= 0)
		err_c::dump_str("  [IndexArr Index]");
	else if(Pos == -1)
		err_c::dump_str("  [End of List]");
	else if(Pos < -1) {
		err_c::dump_str("  [ValueArr Index: ");
		err_c::dump_int(-Pos - 2);
		err_c::dump_str("]");
	}
	err_c::dump_nl();

	// Size of this object:
	err_c::dump_str("Object Size:   ");
	err_c::dump_uint(static_cast<unsigned int>(sizeof(cur_n_n_c)));
	err_c::dump_str(" Bytes");
	err_c::dump_nl();
	err_c::dump_nl();

	// Done:
	err_c::dump_end();
}
#endif

