// ============================================================================
// Project:	Deductive Database
// Filename:	cur_n_n.h
// Purpose:	Cursor for accessing rel_n_n with given first argument.
// Last Change:	16.11.2015
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2010-2015 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Include File Frame:
//=============================================================================

#ifndef	CUR_N_N_INCLUDED
#define CUR_N_N_INCLUDED

//=============================================================================
// Used Types and Macros:
//=============================================================================

#ifndef VER_INCLUDED
#include "ver.h"
#endif

#ifndef STR_INCLUDED
#include "str.h"
#endif

#ifndef CHECK_INCLUDED
#include "check.h"
#endif

#ifndef STACK_INCLUDED
#include "stack.h"
#endif

#ifndef REL_N_N_INCLUDED
#include "rel_n_n.h"
#endif


//=============================================================================
// Private Constants:
//=============================================================================

//-----------------------------------------------------------------------------
// CUR_N_N_MAGIC: Magic number (identifies objects of this class).
//-----------------------------------------------------------------------------

static const long CUR_N_N_MAGIC = 0x434E4E0AL;	// 'RNN\n'

//=============================================================================
// Class for Cursor to Relation with two non-negative integer columns:
//=============================================================================

class cur_n_n_c {
	public:

//-----------------------------------------------------------------------------
// Constructor, Destructor:
//-----------------------------------------------------------------------------

	// Constructor:
	cur_n_n_c(rel_n_n_t rel);

	// Destructor:
	~cur_n_n_c();

//-----------------------------------------------------------------------------
// Methods:
//-----------------------------------------------------------------------------

	// open: Open cursor over the value set for column c2, given c1.
	void open(int c1);

	// fetch: Get next tuple, returns false if there is no further tuple.
	bool fetch();

	// c2: Return value of column c2 to which the cursor points.
	int c2() {
		// Check this object:
		CHECK_VALID("cur_n_n_c::c2");

		// Check state:
		CHECK(C2 >= 0, "cur_n_n_c::c2: fetch beyond end");

		// Return current value for column c2:
		return C2;
	}


	// push: Save the cursor state on a stack.
	bool push(stack_c<int> *stack) {
		// Check this object:
		CHECK_VALID("cur_n_n_c::push");

		// Save C2 and Pos:
		if(!stack->push(C2))
			return false;
		if(!stack->push(Pos))
			return false;

		// Ok:
		return true;
	}

	// pop: Restore the cursor state from a stack.
	void pop(stack_c<int> *stack) {
		// Check this object:
		CHECK_VALID("cur_n_n_c::pop");

		// Restore C2 and Pos (in opposite order):
		Pos = stack->pop();
		C2  = stack->pop();
	}

//-----------------------------------------------------------------------------
// Debugging Support:
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
	public:
	// Integrity check:
	str_t	check() const;

	private:
	// Magic number (identifies objects of this class for debugging).
	long	Magic;		// Must be "CUR_N_N_MAGIC".
#endif

#ifdef	VER_DUMP
	public:
	// Display data structure:
	void dump(str_t headline = STR_NULL) const;
#endif

//-----------------------------------------------------------------------------
// Copy-constructor and assignment operator are not supported for this class:
//-----------------------------------------------------------------------------

	private:

	cur_n_n_c(const cur_n_n_c& obj);		// Not implemented
	cur_n_n_c& operator=(const cur_n_n_c& obj);	// Not implemented

//-----------------------------------------------------------------------------
// Private Object Members:
//-----------------------------------------------------------------------------

	private:

	// Relation, over which this cursor runs:
	rel_n_n_t Rel;

	// Current value of column c2:
	int C2;

	// Position in relation index / value list for next value:
	int Pos;

//=============================================================================
// End of Class:
//=============================================================================

};

//-----------------------------------------------------------------------------
// Define pointer type:
//-----------------------------------------------------------------------------

typedef cur_n_n_c *cur_n_n_t;

//-----------------------------------------------------------------------------
// Define null pointer:
//-----------------------------------------------------------------------------

#define CUR_N_N_NULL (static_cast<cur_n_n_t>(0))

//=============================================================================
// End of Include File:
//=============================================================================

#endif

