// ============================================================================
// Project:	Deductive Database
// Filename:	db_dblp.cpp
// Purpose:	Bottom-Up Evaluation for DBLP Example, Ver 3: Database
// Last Change:	11.06.2016
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2015-2016 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"
#include "str.h"
#include "err.h"
#include "check.h"
#include "atomtab.h"
#include "strtab.h"
#include "list_2.h"
#include "rel_n_n.h"
#include "load_dblp.h"
#include "out.h"

#include "db_dblp.h"


//=============================================================================
// Constructor, Destructor:
//=============================================================================

//-----------------------------------------------------------------------------
// Constructor (initializes subobjects):
//-----------------------------------------------------------------------------

db_dblp_c::db_dblp_c() :

		// Domains:
		doc_ids("String Table for Arg.1"),
		attrs("Atom Table for Arg.2"),
		str_values("String Table for Arg.3"),

		// Relations for EDB body literals:
		att_title("ATT_TITLE"),
		att_year("ATT_YEAR"),
		att_author("ATT_AUTHOR"),
		att_month("ATT_MONTH")

{
	// Construct atoms which are relevant for selection condition:
	title  = attrs.put("title");
	year   = attrs.put("year");
	author = attrs.put("author");
	month  = attrs.put("month");

	// Set Magic number:
	CHECK_CODE(Magic = DB_DBLP_MAGIC);
}

//-----------------------------------------------------------------------------
// Destructor:
//-----------------------------------------------------------------------------

db_dblp_c::~db_dblp_c()
{
	// Check this object:
	CHECK_VALID("db_dblp_c::destructor");

	// Make this object invalid:
	CHECK_CODE(Magic = 0);
}


//=============================================================================
// Public Methods:
//=============================================================================

//-----------------------------------------------------------------------------
// load: Load data file with EDB facts.
//-----------------------------------------------------------------------------

bool db_dblp_c::load(str_t filename) {

	// Check parameter:
	CHECK(filename != STR_NULL, "db_dblp_c::load: filename is null");

	// Check object:
	CHECK_VALID("db_dblp_c::load");

	// Load the data:
	return load_dblp_c::load(filename,
			&doc_ids, &attrs, &str_values,
			title, year, author, month,
			&att_title, &att_year, &att_author, &att_month);
}


//=============================================================================
// Debugging Support:
//=============================================================================


//-----------------------------------------------------------------------------
// check: Check the integrity of this object, return error message or NULL.
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
str_t	db_dblp_c::check() const
{
	// Check magic number:
	if(Magic != DB_DBLP_MAGIC)
		return "wrong magic number";

	// Check sub-objects (relations for EDB predicates):
	CHECK_SUBOBJ(doc_ids, "db_dblp_c::check");
	CHECK_SUBOBJ(attrs, "db_dblp_c::check");
	CHECK_SUBOBJ(str_values, "db_dblp_c::check");
	CHECK_SUBOBJ(att_title, "db_dblp_c::check");
	CHECK_SUBOBJ(att_year, "db_dblp_c::check");
	CHECK_SUBOBJ(att_author, "db_dblp_c::check");
	CHECK_SUBOBJ(att_month, "db_dblp_c::check");

	// Atom IDs must be non-negative:
	if(title < 0)
		return "title is negative";
	if(year < 0)
		return "year is negative";
	if(author < 0)
		return "author is negative";
	if(month < 0)
		return "month is negative";

	// Ok:
	return STR_NULL;
}

#endif


//-----------------------------------------------------------------------------
// dump: Show data structure.
//-----------------------------------------------------------------------------

#ifdef	VER_DUMP
void db_dblp_c::dump(str_t headline) const
{
	// Headline:
	if(headline == STR_NULL)
		headline = "DBLP Benchmark Object, Ver. 2";
	err_c::dump_begin(headline);

	// Basic information:
	err_c::dump_str("ObjectSize:  ");
	err_c::dump_uint(static_cast<unsigned int>(sizeof(db_dblp_c)));
	err_c::dump_str(" Bytes");
	err_c::dump_nl();
	err_c::dump_nl();

	// Show size of domains:
	err_c::dump_str("Size of Domains:  ");
	err_c::dump_nl();
	err_c::dump_str("    doc_ids:    ");
	err_c::dump_int(doc_ids.num_strings());
	err_c::dump_nl();
	err_c::dump_str("    attrs:      ");
	err_c::dump_int(attrs.num_atoms());
	err_c::dump_nl();
	err_c::dump_str("    str_values: ");
	err_c::dump_int(str_values.num_strings());
	err_c::dump_nl();

	// Show number of rows in all EDB relations:
	err_c::dump_str("Number of rows in EDB relations:  ");
	err_c::dump_nl();
	err_c::dump_str("    att_title: ");
	err_c::dump_int(att_title.num_rows());
	err_c::dump_nl();
	err_c::dump_str("    att_year: ");
	err_c::dump_int(att_year.num_rows());
	err_c::dump_nl();
	err_c::dump_str("    att_author: ");
	err_c::dump_int(att_author.num_rows());
	err_c::dump_nl();
	err_c::dump_str("    att_month: ");
	err_c::dump_int(att_month.num_rows());
	err_c::dump_nl();
	err_c::dump_nl();

	// Atom IDs:
	err_c::dump_str("Atom IDs:");
	err_c::dump_nl();
	err_c::dump_str("    title:  ");
	err_c::dump_int(title);
	err_c::dump_nl();
	err_c::dump_str("    year:   ");
	err_c::dump_int(year);
	err_c::dump_nl();
	err_c::dump_str("    author: ");
	err_c::dump_int(author);
	err_c::dump_nl();
	err_c::dump_str("    month:  ");
	err_c::dump_int(month);
	err_c::dump_nl();
	err_c::dump_nl();

	// Done:
	err_c::dump_end();
}
#endif

