// ============================================================================
// Project:	Deductive Database
// Filename:	dup_i.cpp
// Purpose:	Duplicate Check for Rows with One Integer Column
// Last Change:	21.09.2015
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2010-2015 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"
#include "str.h"
#include "err.h"
#include "check.h"

#include "dup_i.h"


//=============================================================================
// Private Class for the Rows/Tuples:
//=============================================================================

class dup_i_row_c {
	public:
		int c1;
		dup_i_row_c *next;
};

typedef dup_i_row_c *dup_i_row_t;

#define	DUP_I_ROW_NULL	(static_cast<dup_i_row_t>(0))


//=============================================================================
// Constructor and Destructor:
//=============================================================================


//-----------------------------------------------------------------------------
// Constructor:
//-----------------------------------------------------------------------------

dup_i_c::dup_i_c(str_t tab_name, int hashtab_size) : MemPool(tab_name)
{
	// Set Magic number:
	CHECK_CODE(Magic = DUP_I_MAGIC);

	// Check parameters:
	CHECK(tab_name != STR_NULL,
			"dup_i_c::constructor: tab_name is NULL");
	CHECK(hashtab_size > 0,
			"dup_i_c::constructor: hashtab_size is not positive");

	// Initialize attributes (which are not nested objects):
	Name        = tab_name;
	NumRows     = 0;
	EntriesUsed = 0;
	InitOk      = true;
	MemPage     = PAGE_NULL;
	MemPtr      = DUP_I_ROW_NULL;
	MemFree     = 0;
	HashSize    = hashtab_size;
	HashPages   = (hashtab_size + (PAGE_SIZE(dup_i_row_t)-1)) /
			PAGE_SIZE(dup_i_row_t);
	CHECK(HashPages > 0,
			"dup_i_c::constructor: HashPages is not positive");

	// Handle limit of hashtab_size:
	if(HashPages > DUP_I_MAXPAGES) {
		err_c::hashtab_limit(tab_name, hashtab_size, HashPages,
					DUP_I_MAXPAGES);
		HashPages = DUP_I_MAXPAGES;
		HashSize = DUP_I_MAXPAGES * PAGE_SIZE(dup_i_row_t);
	}

	// Initialize page array (not really necessary):
	int n = DUP_I_MAXPAGES;
	dup_i_row_t **p = HashTab;
	while(n-- > 0)
		*p++ = static_cast<dup_i_row_t*>(0);

	// Initialize Hash table:
	for(int i = 0; i < HashPages; i++) {
		page_t page = MemPool.alloc_page();
		if(page == PAGE_NULL) {
			InitOk = false;
			return;
		}
		dup_i_row_t *ptr = reinterpret_cast<dup_i_row_t *>(page);
		HashTab[i] = ptr;
		n = PAGE_SIZE(dup_i_row_t);
		while(n-- > 0)
			*ptr++ = DUP_I_ROW_NULL;
	}
}

//-----------------------------------------------------------------------------
// Destructor:
//-----------------------------------------------------------------------------

dup_i_c::~dup_i_c()
{
	// Check this object:
	CHECK_VALID("dup_i_c::destructor");

	// Make this object invalid:
	CHECK_CODE(Magic = 0);
}

//-----------------------------------------------------------------------------
// dup_check: Search tuple in hash table, insert if not found.
//-----------------------------------------------------------------------------

		// Result 1: existing tuple
		// Result 0: new tuple
		// Result -1: Error (memory allocation failed)

int dup_i_c::dup_check(int c1)
{
	// Check this object:
	CHECK_VALID("dup_i_c::dup_check");

	// Check object status:
	CHECK(InitOk, "dup_i_c::dup_check: Object initialization failed");

	// Search in hash table. First compute hash function:
	int hash_val = c1;
	if(hash_val < 0)
		hash_val = -(hash_val+1);
	CHECK(hash_val >= 0, "dup_i_c::dup_check: hashval is negative");
	hash_val %= HashSize;

	// Get start of linked list:
	dup_i_row_t *hashtab_entry = HashTab[hash_val / PAGE_SIZE(dup_i_row_t)];
	hashtab_entry += (hash_val % PAGE_SIZE(dup_i_row_t));

	// Now search the linked list of atoms in that entry:
	dup_i_row_t row;
	for(row = *hashtab_entry; row != DUP_I_ROW_NULL; row = row->next) {
		if(row->c1 == c1)
			return 1;
	}

	// If we arrive here, we must construct a new row.

	// Get new memory page if current page has not enough space:
	if(MemFree <= 0) {
		// Allocate memory page, handle failure:
		page_t page = MemPool.alloc_page();
		if(page == PAGE_NULL)
			return -1;

		// Make new page available:
		MemPage = page;
		MemPtr  = reinterpret_cast<dup_i_row_t>(MemPage);
		MemFree = PAGE_SIZE(dup_i_row_c);
	}

	// Create row object:
	CHECK(MemFree > 0, "dup_i_c::dup_check: Page too small");
	row = MemPtr++;
	MemFree--;
	row->c1 = c1;
	row->next = *hashtab_entry;
	if(*hashtab_entry == DUP_I_ROW_NULL)
		EntriesUsed++;
	*hashtab_entry = row;

	// Increase number of rows:
	NumRows++;

	// The row was new (not a duplicate):
	return 0;
}


//=============================================================================
// Debugging Support:
//=============================================================================

//-----------------------------------------------------------------------------
// check: Check the integrity of this object, return error message or NULL.
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
str_t	dup_i_c::check() const
{
	// Check magic number:
	if(Magic != DUP_I_MAGIC)
		return "wrong magic number";

	// The table name should not be null:
	if(Name == STR_NULL)
		return "table name is null";

	// Check hash table size in entries and pages:
	if(HashSize <= 0)
		return "HashSize is not positive";
	if(HashPages <= 0)
		return "HashPages is not positive";
	if(HashPages > DUP_I_MAXPAGES)
		return "HashPages is greater than limit";
	if(HashPages != (HashSize + (PAGE_SIZE(dup_i_row_t)-1)) /
			PAGE_SIZE(dup_i_row_t))
		return "HashPages is inconsistent with HashSize";

	// MemPtr must point into CurrPage:
	if(MemPage != PAGE_NULL &&
		MemPtr < reinterpret_cast<dup_i_row_t>(MemPage))
		return "MemPtr outside bounds (too small)";
	if(MemPage != PAGE_NULL &&
		reinterpret_cast<char*>(MemPtr) >
			reinterpret_cast<char*>(MemPage) + VER_PAGESIZE)
		return "MemPtr outside bounds (too large)";

	// MemPtr can be computed from MemFree:
	if(MemPage == PAGE_NULL && MemFree != 0)
		return "MemPage is null, but MemFree != 0";
	if(MemPage != PAGE_NULL &&
		MemPtr != reinterpret_cast<dup_i_row_t>(MemPage) +
				(PAGE_SIZE(dup_i_row_c) - MemFree))
		return "MemFree is inconsistent with MemPtr";

	// Check sub-objects:
	CHECK_SUBOBJ(MemPool, "dup_i_c::check");

	// Ok:
	return 0;
}
#endif


//-----------------------------------------------------------------------------
// dump: Show data structure.
//-----------------------------------------------------------------------------

#ifdef	VER_DUMP
void	dup_i_c::dump(str_t headline) const
{
	// Headline:
	if(headline == STR_NULL)
		headline =
		"Hash Table fo Duplicate Check - Relation with one int column";
	err_c::dump_begin(headline);

	// Basic information:
	err_c::dump_str("Name:       '");
	err_c::dump_str(Name);
	err_c::dump_str("'");
	err_c::dump_nl();

	err_c::dump_str("NumRows:     ");
	err_c::dump_int(NumRows);
	err_c::dump_nl();

	err_c::dump_str("HashSize:    ");
	err_c::dump_int(HashSize);
	err_c::dump_str(" [");
	err_c::dump_int(HashPages);
	err_c::dump_str(" Pages = ");
	err_c::dump_int(PAGE_KB(HashPages));
	err_c::dump_str(" KB]");
	err_c::dump_nl();

	err_c::dump_str("EntriesUsed: ");
	err_c::dump_int(EntriesUsed);
	err_c::dump_str(" [");
	int percent = (EntriesUsed * 100 + (HashSize/2)) / HashSize;
	err_c::dump_int(percent);
	err_c::dump_str("%]");
	err_c::dump_nl();

	err_c::dump_str("Memory:      ");
	int pages = MemPool.num_alloc_pages();
	err_c::dump_int(pages);
	err_c::dump_str(" Pages [");
	err_c::dump_int(PAGE_KB(pages));
	err_c::dump_str(" KB]");
	err_c::dump_nl();

	err_c::dump_str("RowSize:     ");
	err_c::dump_uint(static_cast<unsigned int>(sizeof(dup_i_row_c)));
	err_c::dump_str(" Bytes");
	err_c::dump_nl();

	err_c::dump_str("ObjectSize:  ");
	err_c::dump_uint(static_cast<unsigned int>(sizeof(dup_i_c)));
	err_c::dump_str(" Bytes");
	err_c::dump_str(" [includes page array of size MAXPAGES = ");
	err_c::dump_int(DUP_I_MAXPAGES);
	err_c::dump_str("]");
	err_c::dump_nl();

	err_c::dump_nl();

	// Show entire hash table with chains of entries:
	err_c::dump_str("Hash Table:");
	err_c::dump_nl();
	for(int i = 0; i < HashPages; i++) {
		err_c::dump_str("  Page ");
		err_c::dump_str(str_fill(i, 4));
		err_c::dump_int(i);
		err_c::dump_str(":");
		if(HashTab[i] == static_cast<dup_i_row_t*>(0)) {
			err_c::dump_str("*** NULL ***");
			err_c::dump_nl();
			continue;
		}
		err_c::dump_nl();
		for(int j = 0; j < PAGE_SIZE(dup_i_row_t); j++) {
			dup_i_row_t row = HashTab[i][j];
			if(row != DUP_I_ROW_NULL) {
				err_c::dump_str("    ");
				err_c::dump_str(str_fill(j, 4));
				err_c::dump_int(j);
				err_c::dump_str(":");
				for(; row != DUP_I_ROW_NULL; row = row->next){
					err_c::dump_str(" (");
					err_c::dump_int(row->c1);
					err_c::dump_str(")");
				}
				err_c::dump_nl();
			}
		}
		err_c::dump_nl();
	}

	// Done:
	err_c::dump_end();
}
#endif

