// ============================================================================
// Project:	Deductive Database
// Filename:	dup_i.h
// Purpose:	Duplicate Check for Rows with One Integer Column
// Last Change:	21.09.2016
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2010-2015 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Include File Frame:
//=============================================================================

#ifndef	DUP_I_INCLUDED
#define DUP_I_INCLUDED

//=============================================================================
// Used Types and Macros:
//=============================================================================

#ifndef VER_INCLUDED
#include "ver.h"
#endif

#ifndef STR_INCLUDED
#include "str.h"
#endif

#ifndef CHECK_INCLUDED
#include "check.h"
#endif

#ifndef MPOOL_INCLUDED
#include "mpool.h"
#endif


//=============================================================================
// Private Constants:
//=============================================================================

//-----------------------------------------------------------------------------
// DUP_I_MAGIC: Magic number (identifies objects of this class).
//-----------------------------------------------------------------------------

static const long DUP_I_MAGIC = 0x4450490AL;	// 'DPI\n'

//-----------------------------------------------------------------------------
// DUP_I_MAXPAGES: Maximal hash table size in memory pages.
//-----------------------------------------------------------------------------

static const int DUP_I_MAXPAGES = 1000;

	// Additional pages will be used for entries in the hash table.
	// The number of memory pages used for this purpose is unlimited.

//=============================================================================
// Forward Declaration for Private Class:
//=============================================================================

class dup_i_row_c;

//=============================================================================
// Class for Table with One Integer Column - Binding Pattern "b":
//=============================================================================

class dup_i_c {
	public:

//-----------------------------------------------------------------------------
// Constructor, Destructor:
//-----------------------------------------------------------------------------

	// Constructor:
	dup_i_c(str_t tab_name, int hashtab_size);

	// Destructor:
	~dup_i_c();

//-----------------------------------------------------------------------------
// (Object) Methods:
//-----------------------------------------------------------------------------

	// dup_check: Search tuple in hash table, insert if not found.
	int dup_check(int c1);
		// Result 1: existing tuple
		// Result 0: new tuple
		// Result -1: Error (memory allocation failed)

	// name: Get name of this table/relation.
	str_t name() const {
		CHECK_VALID("dup_i_c::name");
		return Name;
	}

	// init_ok: Did we get enough memory pages in the constructor?
	bool init_ok() const {
		CHECK_VALID("dup_i_c::init_ok");
		return InitOk;
	}

	// num_rows: Get number of tuples (rows) in this table.
	int num_rows() const {
		CHECK_VALID("dup_i_c::num_rows");
		return NumRows;
	}

	// hash_size: Size of the hash table.
	int hash_size() const {
		CHECK_VALID("dup_i_c::hash_size");
		return HashSize;
	}

	// entries_used: Number of hash table entries that are used (not NULL).
	int entries_used() const {
		CHECK_VALID("dup_i_c::entries_used");
		return EntriesUsed;
	}

	// num_pages: Get number of memory pages used for this relation.
	int num_pages() const {
		CHECK_VALID("dup_i_c::num_pages");
		return MemPool.num_alloc_pages();
	}

//-----------------------------------------------------------------------------
// Debugging Support:
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
	public:
	// Integrity check:
	str_t	check() const;

	private:
	// Magic number (identifies objects of this class for debugging).
	long	Magic;		// Must be "DUP_I_MAGIC".
#endif

#ifdef	VER_DUMP
	public:
	// Display data structure:
	void dump(str_t headline = STR_NULL) const;
#endif

//-----------------------------------------------------------------------------
// Copy-constructor and assignment operator are not supported for this class:
//-----------------------------------------------------------------------------

	private:

	dup_i_c(const dup_i_c& obj);		// Not implemented
	dup_i_c& operator=(const dup_i_c& obj);	// Not implemented

//-----------------------------------------------------------------------------
// Private Object Members:
//-----------------------------------------------------------------------------

	private:

	// Memory Pool for Allocating Storage Pages:
	mpool_c MemPool;

	// Name of this table/relation:
	str_t Name;

	// Did we get enough memory in the constructor?
	bool InitOk;

	// Current number of rows/tuples:
	int NumRows;

	// Hash table size (in number of entries):
	int HashSize;

	// Hash table size (in memory pages):
	int HashPages;

	// Number of used entries in hash table:
	int EntriesUsed;

	// Current Memory Page for Storing Tuple Objects:
	page_t MemPage;

	// Pointer to next free place in current memory page:
	dup_i_row_c *MemPtr;

	// Number of free places in current page:
	int MemFree;

	// Hash table:
	dup_i_row_c **HashTab[DUP_I_MAXPAGES];

//=============================================================================
// End of Class:
//=============================================================================

};

//-----------------------------------------------------------------------------
// Define pointer type:
//-----------------------------------------------------------------------------

typedef dup_i_c *dup_i_t;

//-----------------------------------------------------------------------------
// Define null pointer:
//-----------------------------------------------------------------------------

#define DUP_I_NULL (static_cast<dup_i_t>(0))

//=============================================================================
// End of Include File:
//=============================================================================

#endif

