// ============================================================================
// Project:	Deductive Database
// Filename:	err.cpp
// Purpose:	Handling of error and warning messages
// Last Change:	27.08.2016
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2010-2016 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"
#include "str.h"
#include "check.h"
#include "err.h"

// #pragma GCC diagnostic ignored "-Wunreachable-code"
#include <iostream>
#include <fstream>

//=============================================================================
// Local Variables of this Module:
//=============================================================================

//-----------------------------------------------------------------------------
// File: The output stream for the alert file.
//-----------------------------------------------------------------------------

static std::ofstream Alert_File;


//=============================================================================
// Exported Functions (Public Static Members):
//=============================================================================


//=============================================================================
// Possibility to query last error messages:
//=============================================================================

//-----------------------------------------------------------------------------
// reset(): Clear buffer of error messages.
//-----------------------------------------------------------------------------

void err_c::reset()
{
	NumErrs = 0;
}

//-----------------------------------------------------------------------------
// ok(): No errors after last reset().
//-----------------------------------------------------------------------------

bool err_c::ok()
{
	return (NumErrs == 0);
}

//-----------------------------------------------------------------------------
// num_errs(): Number of errors after last reset().
//-----------------------------------------------------------------------------

int err_c::num_errs()
{
	return NumErrs;
}

//-----------------------------------------------------------------------------
// msg(): Return n-th/first error message after last reset().
//-----------------------------------------------------------------------------

str_t err_c::msg(int n)
{
	// The parameter must be non-negative:
	CHECK(n >= 0, "err_c::msg: Error index is negative");

	// The parameter must be smaller than the number of messages:
	CHECK(n < NumErrs, "err_c::msg: Error index too large");

	// Return error message:
	return ErrMsg[n];
}

//=============================================================================
// Automatic printing of error messages (Alert-File and/or stderr):
//=============================================================================

//-----------------------------------------------------------------------------
// open_alert: Open the alert file.
//-----------------------------------------------------------------------------

bool err_c::open_alert(str_t filename)
{
	// The file must be closed:
	CHECK(!Alert_File.is_open(),
		"err_c::open_alert: File is already open");

	// Open the file:
	Alert_File.open(filename, std::ios::out|std::ios::app);

	// Return true if file open succeeded:
	return Alert_File.good() && Alert_File.is_open();
}

//-----------------------------------------------------------------------------
// close_alert: Close the alert file.
//-----------------------------------------------------------------------------

bool err_c::close_alert()
{
	// Close alert file:
	Alert_File.close();

	//Check if succeeded:
	return Alert_File.good();
}

//-----------------------------------------------------------------------------
// use_stderr: Define whether error messages are printed to stderr.
//-----------------------------------------------------------------------------

void err_c::use_stderr(bool b)
{
	UseStderr = b;
}

//=============================================================================
// Errors detected in mem.cpp:
//=============================================================================

//-----------------------------------------------------------------------------
// malloc_failed: Failure to allocate memory chunk.
//-----------------------------------------------------------------------------

void err_c::malloc_failed(unsigned int chunk_size, int num_chunks)
{
	msg_begin();
	msg_str("Could not allocate memory chunk of ");
	msg_uint(chunk_size);
	msg_str(" bytes ");
	msg_str("(after allocation of ");
	msg_int(num_chunks);
	msg_str(" chunks)");
	msg_output();
}


//=============================================================================
// Errors detected in stack.h:
//=============================================================================

//-----------------------------------------------------------------------------
// stack_overflow: Fixed size array was not sufficient.
//-----------------------------------------------------------------------------

void err_c::stack_overflow(str_t name, int curr_limit)
{
	msg_begin();
	msg_str("The stack '");
	msg_str(name);
	msg_str("' was configured too small. ");
	msg_str("Current limit: ");
	msg_int(curr_limit);
	msg_output();
}


//=============================================================================
// Errors detected in flexarr.h:
//=============================================================================

//-----------------------------------------------------------------------------
// flexarr_capacity_limit: Flexible array cannot be further extended.
//-----------------------------------------------------------------------------

void err_c::flexarr_capacity_limit(str_t name, int curr_length)
{
	msg_begin();
	msg_str("The flexible array '");
	msg_str(name);
	msg_str("' cannot be extended further, ");
	msg_str("current length: ");
	msg_int(curr_length);
	msg_output();
}


//=============================================================================
// Errors detected in list.h:
//=============================================================================

//-----------------------------------------------------------------------------
// list_capacity_limit: List cannot be further extended.
//-----------------------------------------------------------------------------

void err_c::list_capacity_limit(str_t name, int curr_length)
{
	msg_begin();
	msg_str("The relation (list) '");
	msg_str(name);
	msg_str("' cannot be extended further, ");
	msg_str("current length: ");
	msg_int(curr_length);
	msg_output();
}


//=============================================================================
// Errors detected in dup_2i.cpp:
//=============================================================================

//-----------------------------------------------------------------------------
// hashtab_limit: Requested hash table size needs too many pages.
//-----------------------------------------------------------------------------

void err_c::hashtab_limit(str_t tab_name, int hash_size,
					int hash_pages, int max_pages)
{
	msg_begin();
	msg_str("Requested size ");
	msg_int(hash_size);
        msg_str(" for hash table '");
	msg_str(tab_name);
	msg_str("' too big, ");
	msg_str("would need ");
	msg_int(hash_pages);
	msg_str(" pages, limit: ");
	msg_int(max_pages);
	msg_output();
}

//=============================================================================
// Errors detected in pred.cpp:
//=============================================================================

//-----------------------------------------------------------------------------
// pred_arity_too_large: Requested arity of EDB predicate > MAX.
//-----------------------------------------------------------------------------

void err_c::pred_arity_too_large(str_t name, int arity, int limit)
{
	msg_begin();
	msg_str("Requested arity ");
	msg_int(arity);
        msg_str(" for EDB predicate '");
	msg_str(name);
	msg_str("' too big, limit: ");
	msg_int(limit);
	msg_output();
}

//-----------------------------------------------------------------------------
// pred_too_many_rels: Too many relations for storing one EDB pred.
//-----------------------------------------------------------------------------

void err_c::pred_too_many_rels(str_t name, int limit)
{
	msg_begin();
	msg_str("Too many relations for storing facts of ");
        msg_str("EDB predicate '");
	msg_str(name);
	msg_str("', limit: ");
	msg_int(limit);
	msg_output();
}

//=============================================================================
// Errors detected in in_file.cpp:
//=============================================================================

//-----------------------------------------------------------------------------
// cgi_security: Invalid input file (CGI access not allowed).
//-----------------------------------------------------------------------------

void err_c::cgi_security(str_t filename)
{
	msg_begin();
	msg_str("Invalid input file '");
	msg_str(filename);
	msg_str("' (CGI access not allowed)");
	msg_output();
}

//-----------------------------------------------------------------------------
// open_failed: Error opening an input file.
//-----------------------------------------------------------------------------

void err_c::open_failed(str_t filename, str_t msg)
{
	msg_begin();
	msg_str("Error opening file '");
	msg_str(filename);
	msg_str("': ");
	msg_str(msg);
	msg_output();
}


//-----------------------------------------------------------------------------
// close_failed: Error closing an input file.
//-----------------------------------------------------------------------------

void err_c::close_failed(str_t filename, str_t msg)
{
	msg_begin();
	msg_str("Error closing file '");
	msg_str(filename);
	msg_str("': ");
	msg_str(msg);
	msg_output();
}

//-----------------------------------------------------------------------------
// read_failed: Error reading from an input file.
//-----------------------------------------------------------------------------

void err_c::read_failed(str_t filename, str_t msg)
{
	msg_begin();
	msg_str("Error reading from file '");
	msg_str(filename);
	msg_str("': ");
	msg_str(msg);
	msg_output();
}


//=============================================================================
// Errors detected in in_cgi.cpp (Fatal Error Messages For CGI-Programs):
//=============================================================================

//-----------------------------------------------------------------------------
// bad_request_method: CGI environment variable REQUEST_METHOD incorrect.
//-----------------------------------------------------------------------------

void	err_c::bad_request_method()
{
	msg_begin();
	msg_str("CGI error: ");
	msg_str("Environment variable REQUEST_METHOD ");
	msg_str("undefined or incorrect");
	msg_output();
}

//-----------------------------------------------------------------------------
// bad_content_type: CGI environment variable CONTENT_TYPE incorrect.
//-----------------------------------------------------------------------------

void	err_c::bad_content_type()
{
	msg_begin();
	msg_str("CGI error: ");
	msg_str("Environment variable CONTENT_TYPE ");
	msg_str("undefined or incorrect");
	msg_output();
}

//-----------------------------------------------------------------------------
// bad_content_length: CGI environment variable CONTENT_LENGTH incorrect.
//-----------------------------------------------------------------------------

void	err_c::bad_content_length()
{
	msg_begin();
	msg_str("CGI error: ");
	msg_str("Environment variable CONTENT_LENGTH ");
	msg_str("undefined or incorrect");
	msg_output();
}

//-----------------------------------------------------------------------------
// bad_cgi_hex: Bad CGI hexadecimal character encoding.
//-----------------------------------------------------------------------------

void	err_c::bad_cgi_hex()
{
	msg_begin();
	msg_str("CGI error: ");
	msg_str("Percent sign was not followed by two hex digits");
	msg_output();
}

//-----------------------------------------------------------------------------
// cgi_premature_eof: Less than CONTENT_LENGTH characters in input.
//-----------------------------------------------------------------------------

void	err_c::cgi_premature_eof()
{
	msg_begin();
	msg_str("CGI error: ");
	msg_str("Less than CONTENT_LENGTH input characters");
	msg_output();
}

//-----------------------------------------------------------------------------
// cgi_null_char: Null character in CGI input.
//-----------------------------------------------------------------------------

void	err_c::cgi_null_char()
{
	msg_begin();
	msg_str("CGI error: ");
	msg_str("Cannot process null characters in the input");
	msg_output();
}

//-----------------------------------------------------------------------------
// cgi_field_not_terminated: Field name did not end in "=".
//-----------------------------------------------------------------------------

void	err_c::cgi_field_not_terminated()
{
	msg_begin();
	msg_str("CGI error: ");
	msg_str("Field name did not end in `='");
	msg_output();
}

//-----------------------------------------------------------------------------
// cgi_duplicate_program: More than one "program" field.
//-----------------------------------------------------------------------------

void	err_c::cgi_duplicate_program()
{
	msg_begin();
	msg_str("CGI error: ");
	msg_str("More than one \"program\" field");
	msg_output();
}

//-----------------------------------------------------------------------------
// cgi_no_program: No "program" field.
//-----------------------------------------------------------------------------

void	err_c::cgi_no_program()
{
	msg_begin();
	msg_str("CGI error: ");
	msg_str("No \"program\" field");
	msg_output();
}


//=============================================================================
// Errors detected in lex.cpp:
//=============================================================================


//-----------------------------------------------------------------------------
// null_char: Input file contains null characters (probably binary file).
//-----------------------------------------------------------------------------

void	err_c::null_char(str_t filename, int lineno)
{
	msg_begin();
	if(filename != STR_NULL) {
		msg_str("'");
		msg_str(filename);
		msg_str("', ");
	}
	msg_str("line ");
	msg_int(lineno);
	msg_str(": ");
	msg_str("Line contains NUL (probably binary file)");
	msg_output();
}

//-----------------------------------------------------------------------------
// line_too_long: Line too long (maybe not a text file).
//-----------------------------------------------------------------------------

void	err_c::line_too_long(str_t filename, int lineno)
{
	msg_begin();
	if(filename != STR_NULL) {
		msg_str("'");
		msg_str(filename);
		msg_str("', ");
	}
	msg_str("line ");
	msg_int(lineno);
	msg_str(": ");
	msg_str("Line too long (probably not a text file)");
	msg_output();
}


//=============================================================================
// Functions for Syntax Error Messages (includes input line):
//=============================================================================


//-----------------------------------------------------------------------------
// syntax_begin: Print error message, begin to show input line (<pre>).
//-----------------------------------------------------------------------------

void	err_c::syntax_begin(str_t filename, int line, str_t msg)
{
	msg_begin();
	msg_str("Syntax error in ");
	if(filename != STR_NULL) {
		msg_str("'");
		msg_str(filename);
		msg_str("', ");
	}
	msg_str("line ");
	msg_int(line);
	msg_str(": ");
	msg_str(msg);
	msg_str(".\n");
#ifdef	VER_CGI
	msg_str("<PRE>\n");
#endif
}

//-----------------------------------------------------------------------------
// syntax_input: Print character in input line.
//-----------------------------------------------------------------------------

void	err_c::syntax_input(char c)
{
	msg_char(c);
}

//-----------------------------------------------------------------------------
// syntax_newline: Print newline in <pre> mode.
//-----------------------------------------------------------------------------

void	err_c::syntax_newline()
{
	msg_char('\n');
}

//-----------------------------------------------------------------------------
// syntax_eof: Show EOF in input line.
//-----------------------------------------------------------------------------

void	err_c::syntax_eof()
{
	msg_str("(end of file)");
}

//-----------------------------------------------------------------------------
// syntax_end: End <pre> mode.
//-----------------------------------------------------------------------------

void	err_c::syntax_end()
{
	msg_char('\n');
#ifdef VER_CGI
	msg_str("</PRE>");
#endif
	msg_output();
}

//=============================================================================
// Errors detected in load.cpp:
//=============================================================================


//-----------------------------------------------------------------------------
// store_fact_failed: Storing loaded fact in relations failed.
//-----------------------------------------------------------------------------

void err_c::store_fact_failed(str_t filename, int line, str_t pred_name)
{
	msg_begin();
	msg_str("Storing loaded fact for predicate '");
	msg_str(pred_name);
	msg_str("' failed (");
	if(filename != STR_NULL) {
		msg_str("file '");
		msg_str(filename);
		msg_str("', ");
	}
	msg_str("line ");
	msg_int(line);
	msg_str(")");
	msg_str(".\n");
	msg_output();
}


//=============================================================================
// Errors detected in test.cpp:
//=============================================================================

//-----------------------------------------------------------------------------
// test_has_no_arg: Test was called with argument != 0, but has no arg.
//-----------------------------------------------------------------------------

void err_c::test_has_no_arg(str_t test_id)
{
	msg_begin();
	msg_str("Test '");
	msg_str(test_id);
	msg_str("' was called with argument != 0,");
	msg_str("but has no argument.");
	msg_output();
}

//-----------------------------------------------------------------------------
// invalid_test_arg: Test was called with invalid argument.
//-----------------------------------------------------------------------------

void err_c::invalid_test_arg(str_t test_id, int arg)
{
	msg_begin();
	msg_str("Test '");
	msg_str(test_id);
	msg_str("' was called argument ");
	msg_int(arg);
	msg_str(", which it does not understand.");
	msg_output();
}


//-----------------------------------------------------------------------------
// wine_bench_disabled: Wine bench was called, but not compiled in.
//-----------------------------------------------------------------------------

void err_c::wine_bench_disabled()
{
	msg_begin();
	msg_str("This version was compiled without the ");
	msg_str("Wine Ontology Benchmark.");
	msg_output();
}

//=============================================================================
// Errors detected in main.cpp:
//=============================================================================

//-----------------------------------------------------------------------------
// option_missing: Single '-' in command line.
//-----------------------------------------------------------------------------

void err_c::option_missing()
{
	msg_begin();
	msg_str("'-' on the command line must be followed by an option.");
	msg_output();
}

//-----------------------------------------------------------------------------
// unknown_option: '-XYZ' on command line, where XYZ is not an option.
//-----------------------------------------------------------------------------

void err_c::unknown_option(str_t arg)
{
	msg_begin();
	msg_str("Unknown option: '");
	msg_str(arg);
	msg_str("'.");
	msg_output();
}

//-----------------------------------------------------------------------------
// benchmark_with_debug_code: Option -b was used with debug version.
//-----------------------------------------------------------------------------

void err_c::benchmark_with_debug_code()
{
	msg_begin();
	msg_str("Warning: Benchmark (-b) with debug code compiled in!");
	msg_output();
}

//-----------------------------------------------------------------------------
// debug_output_disabled: Option -d was used, without debug code.
//-----------------------------------------------------------------------------

void err_c::debug_output_disabled()
{
	msg_begin();
	msg_str("Warning: Debug output (-d), but not all debug code compiled!");
	msg_output();
}


//=============================================================================
// Functions for Dumping Data:
//=============================================================================

//-----------------------------------------------------------------------------
// Begin dump section in alert file:
//-----------------------------------------------------------------------------

void err_c::dump_begin(str_t headline)
{
	// Check parameter:
	CHECK(headline != 0, "err_c::dump_begin: headline is NULL");

	// Print headline in "----" block:
	dump_output("\n\n");
	dump_output("----------------------------------------");
	dump_output("---------------------------------------\n");
	dump_output(headline);
	dump_output("\n");
	dump_output("----------------------------------------");
	dump_output("---------------------------------------\n");
	dump_output("\n");
}

//-----------------------------------------------------------------------------
// Print a string:
//-----------------------------------------------------------------------------

void err_c::dump_str(str_t str)
{
	if(str == 0)
		dump_output("<NULL>");
	else
		dump_output(str);
}

//-----------------------------------------------------------------------------
// Print a line end:
//-----------------------------------------------------------------------------

void err_c::dump_nl()
{
	dump_output("\n");
}

//-----------------------------------------------------------------------------
// Print a character:
//-----------------------------------------------------------------------------

void err_c::dump_char(char c)
{
	char buf[2];
	buf[0] = c;
	buf[1] = '\0';
	dump_output(buf);
}


//-----------------------------------------------------------------------------
// Print an integer in decimal notation:
//-----------------------------------------------------------------------------

void err_c::dump_int(int n)
{
	char buf[STR_INT_LEN];

	// Do the converstion. str_int returns number of characters written:
	int num_chars = str_int(buf, STR_INT_LEN, n);
	CHECK(num_chars < STR_INT_LEN,
		"err_c::dump_int: str_int caused buffer overflow");

	// Print result:
	if(num_chars < 1)
		dump_output("<Error when converting int>");
	else
		dump_output(buf);
}

//-----------------------------------------------------------------------------
// Print an unsigned integer in decimal notation:
//-----------------------------------------------------------------------------

void err_c::dump_uint(unsigned int n)
{
	char buf[STR_INT_LEN];

	// Do the converstion. str_int returns number of characters written:
	int num_chars = str_uint(buf, STR_INT_LEN, n);
	CHECK(num_chars < STR_INT_LEN,
		"err_c::dump_uint: str_uint caused buffer overflow");

	// Print result:
	if(num_chars < 1)
		dump_output("<Error when converting unsigned int>");
	else
		dump_output(buf);
}

//-----------------------------------------------------------------------------
// Print a boolean value:
//-----------------------------------------------------------------------------

void err_c::dump_bool(bool b)
{
	if(b)
		dump_output("true");
	else
		dump_output("false");
}

//-----------------------------------------------------------------------------
// Print pointer in hexadecimal notation:
//-----------------------------------------------------------------------------

void err_c::dump_ptr(const void *p)
{
	char buf[STR_PTR_LEN];

	// Do the converstion. str_int returns number of characters written:
	int num_chars = str_ptr(buf, STR_PTR_LEN, p);
	CHECK(num_chars < STR_PTR_LEN,
		"err_c::dump_ptr: str_ptr caused buffer overflow");

	// Print result:
	if(num_chars < 1)
		dump_output("<Error when converting pointer>");
	else
		dump_output(buf);
}

//-----------------------------------------------------------------------------
// End dump section in alert file:
//-----------------------------------------------------------------------------

void err_c::dump_end()
{
	// Print message:
	dump_output("\n(end of dump)\n");

	// Flush data:
	dump_flush();
}

//=============================================================================
// Auxillary Functions:
//=============================================================================

//-----------------------------------------------------------------------------
// msg_begin: Begin construction of message in buffer.
//-----------------------------------------------------------------------------

void err_c::msg_begin()
{
	BufPtr = Buf;
	BufFree = ERR_BUFSIZE - 1;
}

//-----------------------------------------------------------------------------
// msg_str: Add/Append string to message buffer.
//-----------------------------------------------------------------------------

void err_c::msg_str(str_t str)
{
	while(*str && --BufFree > 0) {
		// Note that BufFree might become 0 here, although there is
		// still one byte free for the final '\0'.
		CHECK(BufPtr >= Buf, "err_c::msg_str: BufPtr < Buf");
		CHECK(BufPtr < Buf+(ERR_BUFSIZE-1),
			"err_c::msg_str: BufPtr out of Buf");

		*BufPtr++ = *str++;
	}
}

//-----------------------------------------------------------------------------
// msg_char: Add/Append character to message buffer.
//-----------------------------------------------------------------------------

void err_c::msg_char(char c)
{
	if(--BufFree > 0) {
		// Note that BufFree might become 0 here, although there is
		// still one byte free for the final '\0'.
		CHECK(BufPtr >= Buf, "err_c::msg_char: BufPtr < Buf");
		CHECK(BufPtr < Buf+(ERR_BUFSIZE-1),
			"err_c::msg_char: BufPtr out of Buf");

		*BufPtr++ = c;
	}
}

//-----------------------------------------------------------------------------
// msg_int: Add/Append int to message buffer.
//-----------------------------------------------------------------------------

void err_c::msg_int(int i)
{
	// Check buffer pointer:
	CHECK(BufPtr >= Buf, "err_c::msg_int: BufPtr < Buf");
	CHECK(BufPtr < Buf+(ERR_BUFSIZE-1),
			"err_c::msg_int: BufPtr out of Buf");

	// str_int does the work:
	if(BufFree > 1) {
		int n = str_int(BufPtr, BufFree-1, i);
		CHECK(n < BufFree, "err_c::msg_int: bad result of str_int");
		BufFree -= n;
		BufPtr += n;
	}
}

//-----------------------------------------------------------------------------
// msg_uint: Add/Append unsigned int to message buffer.
//-----------------------------------------------------------------------------

void err_c::msg_uint(unsigned int i)
{
	// Check buffer pointer:
	CHECK(BufPtr >= Buf, "err_c::msg_uint: BufPtr < Buf");
	CHECK(BufPtr < Buf+(ERR_BUFSIZE-1),
			"err_c::msg_uint: BufPtr out of Buf");

	// str_uint does the work:
	if(BufFree > 1) {
		int n = str_uint(BufPtr, BufFree-1, i);
		CHECK(n < BufFree, "err_c::msg_uint: bad result of str_int");
		BufFree -= n;
		BufPtr += n;
	}
}

//-----------------------------------------------------------------------------
// msg_output: Output contents of message buffer.
//-----------------------------------------------------------------------------

void err_c::msg_output()
{
	// Check buffer pointer:
	CHECK(BufPtr >= Buf, "err_c::msg_output: BufPtr < Buf");
	CHECK(BufPtr < Buf+(ERR_BUFSIZE-1),
			"err_c::msg_output: BufPtr out of Buf");

	// Terminate string in buffer:
	*BufPtr = '\0';

	// Store Message in error message buffer if there is still space:
	if(NumErrs < ERR_MAXERRS - 1) {
		str_cpy(ErrMsg[NumErrs++], ERR_BUFSIZE, Buf);
	}

	// Print message to alert file if open:
	if(Alert_File.is_open()) {
		Alert_File << Buf << ".\n";
		Alert_File << std::flush;
	}

	// Print message to standard error channel if requested:
	if(UseStderr) {
		std::cerr << "\n" << Buf << ".\n";
		std::cerr << std::flush;
	}
}

//-----------------------------------------------------------------------------
// dump_output: Output a string as part of a data structure dump.
//-----------------------------------------------------------------------------

void err_c::dump_output(str_t str)
{
	// Check buffer pointer:
	CHECK(str != 0, "err_c::dump_output: str is NULL");

	// Print string to alert file if open:
	if(Alert_File.is_open())
		Alert_File << str;

	// Print message to standard error channel if requested:
	if(UseStderr)
		std::cerr << str;
}

//-----------------------------------------------------------------------------
// dump_flush: Make sure that dump data is really written.
//-----------------------------------------------------------------------------

void err_c::dump_flush()
{
	// Flush alert file if output was written there:
	if(Alert_File.is_open())
		Alert_File << std::flush;

	// Flush standard error channel if output was written there:
	if(UseStderr)
		std::cerr << std::flush;
}

//=============================================================================
// Class Attributes:
//=============================================================================


//-----------------------------------------------------------------------------
// UseStderr: Defines whether error messages are printed to stderr.
//-----------------------------------------------------------------------------

bool err_c::UseStderr;

//-----------------------------------------------------------------------------
// Buf: Buffer for current message.
//-----------------------------------------------------------------------------

char err_c::Buf[ERR_BUFSIZE];

//-----------------------------------------------------------------------------
// BufPtr: Next free position in buffer.
//-----------------------------------------------------------------------------

char *err_c::BufPtr;

//-----------------------------------------------------------------------------
// BufFree: Number of free characters in buffer.
//-----------------------------------------------------------------------------

int err_c::BufFree;


//-----------------------------------------------------------------------------
// NumErrs: Number of error messages after last reset().
//-----------------------------------------------------------------------------

int err_c::NumErrs = 0;

//-----------------------------------------------------------------------------
// ErrMsg: Array of last error messages.
//-----------------------------------------------------------------------------

char err_c::ErrMsg[ERR_MAXERRS][ERR_BUFSIZE];

