// ============================================================================
// Project:	Deductive Database
// Filename:	err.h
// Purpose:	Handling of error and warning messages
// Last Change:	06.10.2016
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2010-2016 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Include File Frame:
//=============================================================================

#ifndef	ERR_INCLUDED
#define ERR_INCLUDED

//=============================================================================
// Used Types and Macros:
//=============================================================================

#ifndef VER_INCLUDED
#include "ver.h"
#endif

#ifndef STR_INCLUDED
#include "str.h"
#endif

#ifndef CHECK_INCLUDED
#include "check.h"
#endif

// Forward declaration, in this way we do not need to include <iostream>:
class ofstream;

//=============================================================================
// Private Constants:
//=============================================================================

//-----------------------------------------------------------------------------
// ERR_BUFSIZE: Size of buffer for a single error message.
//-----------------------------------------------------------------------------

static const int ERR_BUFSIZE = 512;
	// Note: For syntax errors this contains the error message,
	// plus the input line plus the spaces for marking the error position.
	// I.e. this must be large enough to hold three full lines.

//-----------------------------------------------------------------------------
// ERR_MAXERRS: Maximal number of error messages kept in memory.
//-----------------------------------------------------------------------------

static const int ERR_MAXERRS = 10;

//=============================================================================
// Module for Error Message Handling (Class with only static members):
//=============================================================================

class err_c {
	public:


//-----------------------------------------------------------------------------
// Static Methods:
//-----------------------------------------------------------------------------

	//---------------------------------------------------------------------
	// Possibility to query last error messages:
	//---------------------------------------------------------------------

	// reset(): Clear buffer of error messages.
	static void reset();

	// ok(): No errors after last reset().
	static bool ok();

	// num_errs(): Number of errors after last reset().
	static int num_errs();

	// msg(): Return n-th/first error message after last reset().
	static str_t msg(int n = 0);

	//---------------------------------------------------------------------
	// Automatic printing of error messages (Alert-File and/or stderr):
	//---------------------------------------------------------------------

	// open_alert: Open the alert file.
	static bool open_alert(str_t filename);

	// close: Close the alert file.
	static bool close_alert();

	// use_stderr: Define whether error messages are printed to stderr:
	static void use_stderr(bool b);

	//---------------------------------------------------------------------
	// Errors detected in mem.cpp:
	//---------------------------------------------------------------------

	// malloc_failed: Failure to allocate memory chunk.
	static void malloc_failed(unsigned int chunk_size, int num_chunks);

	//---------------------------------------------------------------------
	// Errors detected in stack.h:
	//---------------------------------------------------------------------

	// stack_overflow: Fixed size array was not sufficient.
	static void stack_overflow(str_t name, int curr_limit);

	//---------------------------------------------------------------------
	// Errors detected in flexarr.h:
	//---------------------------------------------------------------------

	// flexarr_capacity_limit: Flexible array cannot be further extended.
	static void flexarr_capacity_limit(str_t name, int curr_length);

	//---------------------------------------------------------------------
	// Errors detected in list.h:
	//---------------------------------------------------------------------

	// list_capacity_limit: List cannot be further extended.
	static void list_capacity_limit(str_t name, int curr_length);

	//---------------------------------------------------------------------
	// Errors detected in dup_2i.cpp:
	//---------------------------------------------------------------------

	static void hashtab_limit(str_t tab_name, int hash_size,
					int hash_pages, int max_pages);

	//---------------------------------------------------------------------
	// Errors detected in pred.cpp:
	//---------------------------------------------------------------------

	// pred_arity_too_large: Requested arity of EDB predicate > MAX.
	static void pred_arity_too_large(str_t name, int arity, int limit);

	// pred_too_many_rels: Too many relations for storing one EDB pred.
	static void pred_too_many_rels(str_t name, int limit);

	//---------------------------------------------------------------------
	// Errors detected in in_file.cpp:
	//---------------------------------------------------------------------

	// cgi_security: Invalid input file (CGI access not allowed).
	static void cgi_security(str_t filename);

	// open_failed: Error opening an input file.
	static void open_failed(str_t filename, str_t msg);

	// close_failed: Error closing an input file.
	static void close_failed(str_t filename, str_t msg);

	// read_failed: Error reading from an input file.
	static void read_failed(str_t filename, str_t msg);

	//---------------------------------------------------------------------
	// Errors detected in in_cgi.cpp:
	//---------------------------------------------------------------------

	// bad_request_method: CGI env. variable REQUEST_METHOD incorrect.
	static void bad_request_method();

	// bad_content_type: CGI environment variable CONTENT_TYPE incorrect.
	static void bad_content_type();

	// bad_content_length: CGI env. variable CONTENT_LENGTH incorrect.
	static void bad_content_length();

	// bad_cgi_hex: Bad CGI hexadecimal character encoding.
	static void bad_cgi_hex();

	// cgi_premature_eof: Less than CONTENT_LENGTH characters in input.
	static void cgi_premature_eof();

	// cgi_null_char: Null character in CGI input.
	static void cgi_null_char();

	// cgi_field_not_terminated: Field name did not end in "=".
	static void cgi_field_not_terminated();

	// cgi_duplicate_program: More than one "program" field.
	static void cgi_duplicate_program();

	// cgi_no_program: No "program" field.
	static void cgi_no_program();

	//---------------------------------------------------------------------
	// Errors detected in lex.cpp:
	//---------------------------------------------------------------------

	// null_char: Input file contains null characters (binary file?).
	static void null_char(str_t filename, int lineno);

	// line_too_long: Line too long (maybe not a text file).
	static void line_too_long(str_t filename, int lineno);

	//---------------------------------------------------------------------
	// Functions for Syntax Error Messages (includes input line):
	//---------------------------------------------------------------------

	// syntax_begin: Print error message, begin to show input line (<pre>).
	static void syntax_begin(str_t filename, int line, str_t msg);

	// syntax_input: Print character in input line.
	static void syntax_input(char c);

	// syntax_newline: Print newline in <pre> mode.
	static void syntax_newline();

	// syntax_eof: Show EOF in input line.
	static void syntax_eof();

	// syntax_end: End <pre> mode.
	static void syntax_end();

	//---------------------------------------------------------------------
	// Errors detected in load.cpp:
	//---------------------------------------------------------------------

	// store_fact_failed: Storing loaded fact in relations failed.
	static void store_fact_failed(str_t filename, int line_no,
					str_t pred_name);

	//---------------------------------------------------------------------
	// Errors detected in test.cpp:
	//---------------------------------------------------------------------

	// test_has_no_arg: Test was called with argument != 0, but has no arg.
	static void test_has_no_arg(str_t test_id);

	// invalid_test_arg: Test was called with invalid argument.
	static void invalid_test_arg(str_t test_id, int arg);

	// wine_bench_disabled: Wine bench was called, but not compiled in.
	static void wine_bench_disabled();

	//---------------------------------------------------------------------
	// Errors detected in main.cpp:
	//---------------------------------------------------------------------

	// option_missing: Single '-' in command line.
	static void option_missing();

	// unknown_option: '-XYZ' on command line, where XYZ is not an option.
	static void unknown_option(str_t arg);

	// benchmark_with_debug_code: Option -b was used with debug version.
	static void benchmark_with_debug_code();

	// debug_output_disabled: Option -d was used, without debug code.
	static void debug_output_disabled();

	//---------------------------------------------------------------------
	// Functions for Dumping Data:
	//---------------------------------------------------------------------

	// Begin dump section in alert file:
	static void dump_begin(str_t headline);

	// Print a string:
	static void dump_str(str_t str);

	// Print a line end:
	static void dump_nl();

	// Print a character:
	static void dump_char(char c);

	// Print an integer in decimal notation:
	static void dump_int(int n);

	// Print an unsigned integer in decimal notation:
	static void dump_uint(unsigned int n);

	// Print a boolean value:
	static void dump_bool(bool b);

	// Print pointer in hexadecimal notation:
	static void dump_ptr(const void *p);

	// End dump section in alert file:
	static void dump_end();

//-----------------------------------------------------------------------------
// Auxillary Static Methods:
//-----------------------------------------------------------------------------

	private:

	// msg_begin: Begin construction of message in buffer.
	static void msg_begin();

	// msg_str: Add/Append string to message buffer.
	static void msg_str(str_t str);

	// msg_char: Add/Append character to message buffer.
	static void msg_char(char c);

	// msg_int: Add/Append int to message buffer.
	static void msg_int(int i);

	// msg_uint: Add/Append unsigned int to message buffer.
	static void msg_uint(unsigned int i);

	// msg_output: Output contents of message buffer.
	static void msg_output();

	// dump_output: Output a string as part of a data structure dump.
	static void dump_output(str_t str);

	// dump_flush: Make sure that dump data is really written.
	static void dump_flush();

//-----------------------------------------------------------------------------
// Class Variables:
//-----------------------------------------------------------------------------

	private:

	// UseStderr: Defines whether error messages are printed to stderr.
	static bool UseStderr;

	// Buf: Buffer for current message.
	static char Buf[ERR_BUFSIZE];

	// BufPtr: Next free position in buffer.
	static char *BufPtr;

	// BufFree: Number of free characters in buffer.
	static int BufFree;

	// NumErrs: Number of error messages after last reset().
	static int NumErrs;

	// ErrMsg: Array of last error messages.
	static char ErrMsg[ERR_MAXERRS][ERR_BUFSIZE];

//-----------------------------------------------------------------------------
// There are no objects of this class, therefore no constructor is available:
//-----------------------------------------------------------------------------

	private:

	// Constructor:
	err_c();				// Not implemented

	// Destructor:
	~err_c();				// Not implemented

	// Copy-Constructor:
	err_c(const err_c& obj);		// Not implemented

	// Assignment operator:
	err_c& operator=(const err_c& obj);	// Not implemented


//-----------------------------------------------------------------------------
// End of Class:
//-----------------------------------------------------------------------------

};


//=============================================================================
// End of Include File:
//=============================================================================

#endif

