// ============================================================================
// Project:	Deductive Database
// Filename:	eval_dblp.cpp
// Purpose:	Bottom-Up Evaluation for DBLP Example, Ver 3: EValuator
// Last Change:	11.06.2016
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2015-2016 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"
#include "str.h"
#include "err.h"
#include "check.h"
#include "stack.h"
#include "db_dblp.h"
#include "cur_2.h"
#include "cur_n_n.h"
#include "out.h"

#include "eval_dblp.h"


//=============================================================================
// Constructor, Destructor:
//=============================================================================

//-----------------------------------------------------------------------------
// Constructor (initializes subobjects):
//-----------------------------------------------------------------------------

eval_dblp_c::eval_dblp_c(db_dblp_t db) :

		// Cursors for loops during bottom-up evaluation:
		c2_1_fbf(&(db->att_title)),
		c2_2_bbf(&(db->att_year)),
		c2_3_bbf(&(db->att_author)),
		c2_4_bbf(&(db->att_month)),

		// Backtrack stack for bottom-up evaluation:
		backtrack_stack("Backtrack Stack")

{
	// Remember database:
	DB = db;

	// Initialize variables for bottom-up evaluation:
	v2_1 = -1;
	v2_2 = -1;
	v2_3 = -1;
	v2_4 = -1;
	v2_5 = -1;

	// Set Magic number:
	CHECK_CODE(Magic = EVAL_DBLP_MAGIC);
}

//-----------------------------------------------------------------------------
// Destructor:
//-----------------------------------------------------------------------------

eval_dblp_c::~eval_dblp_c()
{
	// Check this object:
	CHECK_VALID("eval_dblp_c::destructor");

	// Make this object invalid:
	CHECK_CODE(Magic = 0);
}


//=============================================================================
// Public Methods:
//=============================================================================


//-----------------------------------------------------------------------------
// Open cursor (note: load() must be called before this method).
//-----------------------------------------------------------------------------

void eval_dblp_c::open()
{
	// --------------------------------------------------------------------
	// Transformed program:
	//
	// 1: answer(A,B,C,D,E) :-
	//		queryrule(A,B,C,D,E).
	// 2: queryrule(A,B,C,D,E) :-
	//		att(A,title,B),
	//		att(A,year,C),
	//		att(A,author,D),
	//		att(A,month,E).
	// --------------------------------------------------------------------


	// --------------------------------------------------------------------
	// Jump labels:
	// label 1: applying rule 2 resulting in symbolic fact
	//	queryrule(v2_1,v2_2,v2_3,v2_4,v2_5)
	// label 2: using symbolic fact
	//	queryrule(v2_1,v2_2,v2_3,v2_4,v2_5)
	//	with literal 1 in rule 1 resulting in symbolic fact
	//	answer(v2_1,v2_2,v2_3,v2_4,v2_5)
	// --------------------------------------------------------------------

	// Initialisation Part:
	backtrack_stack.push(1);
}

//-----------------------------------------------------------------------------
// fetch: Get next tuple, returns false if there is no further tuple.
//-----------------------------------------------------------------------------

bool eval_dblp_c::fetch()
{

	// Loop Part:
	while(!backtrack_stack.is_empty()) {
		switch(backtrack_stack.pop()) {
		case 1:
			// c2_1_fbf.open("title");
			c2_1_fbf.open();
			while (c2_1_fbf.fetch()) {
				// c2_2_bbf.open(c2_1_fbf.c1(), "year");
				c2_2_bbf.open(c2_1_fbf.c1());
				while (c2_2_bbf.fetch()) {
					//c2_3_bbf.open(c2_1_fbf.c1(),
					//			"author");
					c2_3_bbf.open(c2_1_fbf.c1());
					while (c2_3_bbf.fetch()) {
						//c2_4_bbf.open(
						//  c2_1_fbf.c1(), "month");
						c2_4_bbf.open(c2_1_fbf.c1());
			    // Continuation for label 1
		case 4:
			while (c2_4_bbf.fetch()) {
				v2_1 = c2_1_fbf.c1();
				v2_2 = c2_1_fbf.c2();
				v2_3 = c2_2_bbf.c2();
				v2_4 = c2_3_bbf.c2();
				v2_5 = c2_4_bbf.c2();
				// Push backtrack label to continuation
				backtrack_stack.push(4);
				// Labels for next rules
				goto label_2;
			}

					}
				}
			}
			break;
	    
		case 2:
		label_2:
			// queryrule(v2_1,v2_2,v2_3,v2_4,v2_5)
			return true;
	    
	    
		}
	}

	// Backtrack stack was empty, no more answers:
	return false;
}


//=============================================================================
// Debugging Support:
//=============================================================================


//-----------------------------------------------------------------------------
// check: Check the integrity of this object, return error message or NULL.
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
str_t	eval_dblp_c::check() const
{
	// Check magic number:
	if(Magic != EVAL_DBLP_MAGIC)
		return "wrong magic number";

	// Check sub-objects (relations for EDB predicates):
	CHECK_SUBOBJ(c2_1_fbf, "eval_dblp_c::check");
	CHECK_SUBOBJ(c2_2_bbf, "eval_dblp_c::check");
	CHECK_SUBOBJ(c2_3_bbf, "eval_dblp_c::check");
	CHECK_SUBOBJ(c2_4_bbf, "eval_dblp_c::check");
	CHECK_SUBOBJ(backtrack_stack, "eval_dblp_c::check");

	// Variables are non-negative once set, but initialized to -1:
	if(v2_1 < -1)
		return "v2_1 is invalid";
	if(v2_2 < -1)
		return "v2_2 is invalid";
	if(v2_3 < -1)
		return "v2_3 is invalid";
	if(v2_4 < -1)
		return "v2_4 is invalid";
	if(v2_5 < -1)
		return "v2_5 is invalid";

	// Ok:
	return STR_NULL;
}

#endif


//-----------------------------------------------------------------------------
// dump: Show data structure.
//-----------------------------------------------------------------------------

#ifdef	VER_DUMP
void eval_dblp_c::dump(str_t headline) const
{
	// Headline:
	if(headline == STR_NULL)
		headline = "DBLP Benchmark Object, Ver. 2";
	err_c::dump_begin(headline);

	// Basic information:
	err_c::dump_str("ObjectSize:  ");
	err_c::dump_uint(static_cast<unsigned int>(sizeof(eval_dblp_c)));
	err_c::dump_str(" Bytes");
	err_c::dump_nl();
	err_c::dump_nl();

	// Variables:
	err_c::dump_str("Variables:");
	err_c::dump_nl();
	err_c::dump_str("    v2_1:   ");
	err_c::dump_int(v2_1);
	err_c::dump_nl();
	err_c::dump_str("    v2_2:   ");
	err_c::dump_int(v2_2);
	err_c::dump_nl();
	err_c::dump_str("    v2_3:   ");
	err_c::dump_int(v2_3);
	err_c::dump_nl();
	err_c::dump_str("    v2_4:   ");
	err_c::dump_int(v2_4);
	err_c::dump_nl();
	err_c::dump_str("    v2_5:   ");
	err_c::dump_int(v2_5);
	err_c::dump_nl();
	err_c::dump_nl();

	// Backtrack Stack:
	err_c::dump_str("Stacks:");
	err_c::dump_nl();
	err_c::dump_str("    Backtrack Stack Size:   ");
	err_c::dump_int(backtrack_stack.num_elems());
	err_c::dump_nl();

	// Done:
	err_c::dump_end();
}
#endif

