// ============================================================================
// Project:	Deductive Database
// Filename:	in_file.cpp
// Purpose:	Get Input From a File
// Last Change:	27.08.2015
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 1998-2015 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"
#include "err.h"

#include "in_file.h"

#ifndef VER_IOSTREAM
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/uio.h>
#include <unistd.h>
#include <fcntl.h>
#endif

#include <string.h>
#include <errno.h>
//extern int errno;

#ifdef VER_CGI
#include <stdlib.h>
#endif

//=============================================================================
// Exported Functions:
//=============================================================================

//-----------------------------------------------------------------------------
// Constructor:
//-----------------------------------------------------------------------------

in_file_c::in_file_c(str_t fname)
{
	// Check parameter:
	CHECK(fname != STR_NULL, "in_file_c::constructor: filename is NULL");

#ifdef	VER_CGI
	// In CGI-Mode, we allow only filenames consisting of up to 8 letters
	// or digits, optionally followed by ".txt".
	// The file names "core" and "slp" are explicitly excluded.
	str_t p = fname;
	if(p[0]=='c' && p[1]=='o' && p[2]=='r' && p[3]=='e' && p[4]=='\0') {
		err_c::cgi_security(fname);
		exit(4);
	}
	if(p[0]=='s' && p[1]=='l' && p[2]=='p' && p[3]=='\0') {
		err_c::cgi_security(fname);
		exit(4);
	}
	for(int i = 0; *p; p++, i++) {
		if(*p == '.' && p[1] == 't' && p[2] == 'x' && p[3] == 't')
			break;
		if(!str_contains(*p, IN_FILE_CGI_FILENAME_CHARS) || i>=8) {
			err_c::cgi_security(fname);
			exit(4);
		}
	}
#endif

	// Copy filename:
	str_cpy(Filename, IN_FILE_NAMESIZE+1, fname);

#ifdef	VER_CGI
	// In CGI-Mode, prepend example directory to specified file:
	char real_filename[IN_FILE_NAMESIZE+1];
	str_cpy(real_filename, IN_FILE_NAMESIZE+1, VER_CGI_EXDIR);
	str_cat(real_filename, IN_FILE_NAMESIZE+1, fname);
	fname = real_filename;
#endif

	// Try to open file:
#ifdef VER_IOSTREAM
	File.open(fname, std::ios::in|std::ios::binary);
#else
	Fd = ::open(fname, O_RDONLY, 0);
#endif

	// If unsuccessful, print error message:
	if(!is_open()) {
		char syserr[STR_ERRMSG_LEN];
		str_error(errno, syserr, STR_ERRMSG_LEN);
		err_c::open_failed(fname, syserr);
	}
}

//-----------------------------------------------------------------------------
// Destructor:
//-----------------------------------------------------------------------------

in_file_c::~in_file_c()
{
	// Close file if still open:
	if(is_open())
		(void) close();
}

//-----------------------------------------------------------------------------
// close: Close input file.
//-----------------------------------------------------------------------------

bool	in_file_c::close()
{
	int retval;

	// The file must be open:
	CHECK(is_open(), "in_file_c::close: file not open");

#ifdef VER_IOSTREAM

	// Try to close the file and check for errors:
	// Note that fail-bit might still be set from last read,
	// which was less than the buffer size, but otherwise ok.
	// Therefore we first reset the error bits if at eof:
	if(File.eof() && File.fail())
		File.clear();
	File.close();
	retval = File.fail() || File.bad();
#else

	// Try to close the file and make the file descriptor invalid:
	retval = ::close(Fd);
	Fd = -1;
#endif

	// If unsuccessful, print error message:
	if(retval != 0) {
		char syserr[STR_ERRMSG_LEN];
		str_error(errno, syserr, STR_ERRMSG_LEN);
		err_c::close_failed(Filename, syserr);
		return(false);
	}

	// Close was successful:
	return(true);
}

//-----------------------------------------------------------------------------
// read: Read characters from input file into buffer.
//-----------------------------------------------------------------------------

int	in_file_c::read(char *buf, int size)
{
	int	bytes;

	// File must be open:
	CHECK(is_open(), "in_file_c::read: file not open");

	// Try to read from file:
#ifdef VER_IOSTREAM
	File.read(buf, size);
	// Note: At least in Microsoft Visual C++, the fail bit will be set
	// if the requested bytes "size" is more than the remaining characters
	// in the file. Therefore we must test first for the eof condition.
	// However, in this way some real errors might be not detected
	// (treated like EOF).
	if(File.eof())
		bytes = static_cast<int>(File.gcount());
	else if(File.fail() || File.bad())
		bytes = -1;
	else
		bytes = static_cast<int>(File.gcount());
#else
	bytes = (int) ::read(Fd, buf, (size_t) size);
#endif

	// Check for error:
	if(bytes < 0) {
		char syserr[STR_ERRMSG_LEN];
		str_error(errno, syserr, STR_ERRMSG_LEN);
		err_c::read_failed(Filename, syserr);
		return(-1);
	}

	// Return number of bytes read:
	return(bytes);
}

