// ============================================================================
// Project:	Deductive Database
// Filename:	in_kbd.cpp
// Purpose:	Get Input Interactively (From Keyboard)
// Last Change:	27.08.2015
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 1997-2015 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"

#include <iostream>
#include <errno.h>
//extern int errno;

#include "str.h"
#include "err.h"
#include "ch.h"

#include "in_kbd.h"

//=============================================================================
// Exported Functions:
//=============================================================================

//-----------------------------------------------------------------------------
// Constructor:
//-----------------------------------------------------------------------------

in_kbd_c::in_kbd_c()
{
	// Remember that input is open and no EOF was seen yet:
	Open	= true;
	AtEOF	= false;
}

//-----------------------------------------------------------------------------
// Destructor:
//-----------------------------------------------------------------------------

in_kbd_c::~in_kbd_c()
{
	// Nothing to do
}

//-----------------------------------------------------------------------------
// close: Close input explicitly.
//-----------------------------------------------------------------------------

bool	in_kbd_c::close()
{
	// Remember that input is close. Don't actually close stdin:
	Open = false;

	// This is always successful:
	return true;
}

//-----------------------------------------------------------------------------
// read: Read characters from input file into buffer.
//-----------------------------------------------------------------------------

int	in_kbd_c::read(register char *buf, register int size)
{
	int c;
	int n;

	// Check object state and parameters:
	CHECK(Open, "in_kbd_c::read: Input was closed");
	CHECK(buf != 0, "in_kbd_c::read: buf is NIL");
	CHECK(size > 0, "in_kbd_c::read: size <= 0");
	CHECK(size >= 80,
		"in_kbd_c::read: small buffer sizes not yet implemented");

	// Don't read anything if already at EOF:
	if(AtEOF)
		return 0;

	// Print a prompt to tell the user that the program waits for input:
	std::cout << VER_PROMPT;
	std::cout.flush();

	// Get one line of input:
	c = 0;
	n = 0;
	while(c!='\n' //!ch_c::is_newline(c)
		&&  --size>=0 && !AtEOF) {

		// Get character from standard input:
		c = std::cin.get();

		// Test for end of file and errors. Assign character:
		if(std::cin.eof())
			AtEOF = true;
		else if(std::cin.fail() || std::cin.bad()) {
			*buf = 0;
			char syserr[STR_ERRMSG_LEN];
			str_error(errno, syserr, STR_ERRMSG_LEN);
			err_c::read_failed("<stdin>", syserr);
			return(-1);
		}
		else {
			*buf++ = (char) c;
			n++;
		}
	}

	// Return number of characters read:
	return(n);
}

