// ============================================================================
// Project:	Deductive Database
// Filename:	load.cpp
// Purpose:	Parse and load data.
// Last Change:	10.11.2015
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2015 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"
#include "str.h"
#include "err.h"
#include "check.h"
#include "in_file.h"
#include "syn.h"
#include "lex.h"

#include "load.h"


//=============================================================================
// Constructor and Destructor:
//=============================================================================


//-----------------------------------------------------------------------------
// Constructor:
//-----------------------------------------------------------------------------

load_c::load_c()
{
	// Initialize hash table for predicate data:
	int n = LOAD_HASHSIZE;
	pred_t *p = HashTab;
	while(n-- > 0)
		*p++ = PRED_NULL;

	// Set Magic number:
	CHECK_CODE(Magic = LOAD_MAGIC);
}

//-----------------------------------------------------------------------------
// Destructor:
//-----------------------------------------------------------------------------

load_c::~load_c()
{
	// Check this object:
	CHECK_VALID("load_c::destructor");

	// Make this object invalid:
	CHECK_CODE(Magic = 0);
}

//=============================================================================
// Auxiliary Methods:
//=============================================================================

//-----------------------------------------------------------------------------
// hash: Compute hash function for predicate name.
//-----------------------------------------------------------------------------

inline int load_c::hash(str_t name_start, str_t name_end)
{
	str_t name_p = name_start;
	int hash_val = *name_p;
	if(name_p != name_end) {
		hash_val ^= (*name_p++ << 7);
		int shift = 13;
		while(name_p != name_end) {
			hash_val ^= (*name_p++ << shift);
			shift += 5;
			if(shift > static_cast<int>((sizeof(int)-1) * 8))
				shift -= (sizeof(int)-1) * 6;
		}
	}
	if(hash_val < 0)
		hash_val = -(hash_val + 1);
	CHECK(hash_val >= 0, "atomtab_c::put: Hash value is negative");
	hash_val %= LOAD_HASHSIZE;
	return hash_val;
}

//-----------------------------------------------------------------------------
// find_pred: Find predicate data for given predicate name.
//-----------------------------------------------------------------------------

inline pred_t load_c::find_pred(str_t name_start, str_t name_end)
{
	// First compute the hash value:
	int hash_val = hash(name_start, name_end);

	// Now search the linked list of predicate objects in that entry:
	for(pred_t pred = HashTab[hash_val]; pred != PRED_NULL;
						pred = pred->hash_next()) {
		str_t pred_name_p = pred->name();
		str_t name_p = name_start;
		while(name_p < name_end && *pred_name_p == *name_p) {
			pred_name_p++;
			name_p++;
		}
		if(name_p == name_end && *pred_name_p == '\0')
			return pred;
	}
	return PRED_NULL;
}


//=============================================================================
// Public Methods:
//=============================================================================


//-----------------------------------------------------------------------------
// add: Define predicate.
//-----------------------------------------------------------------------------

void load_c::add(pred_t predicate)
{
	// Check this object:
	CHECK_VALID("load_c::add");

	// Check parameters:
	CHECK_PAR(predicate, "load_c::add");

	// Compute hash function of predicate name:
	str_t name_p = predicate->name();
	int hash_val = hash(name_p, name_p + str_len(name_p));

	// Make sure that there is no predicate with this name already:
	for(pred_t pred = HashTab[hash_val]; pred != PRED_NULL;
						pred = pred->hash_next()) {
		CHECK(!str_eq(predicate->name(), pred->name()),
				"load_c::add: Predicate name already used");
		if(str_eq(predicate->name(), pred->name()))
			return;
	}

	// Connect new predicate object to hash chain:
	predicate->set_next(HashTab[hash_val]);
	HashTab[hash_val] = predicate;
}

//-----------------------------------------------------------------------------
// load: Load data.
//-----------------------------------------------------------------------------

bool load_c::load(str_t filename)
{
	// Open file:
	in_file_c file(filename);
	if(!file.is_open()) {
		// Error message was printed in constructor.
		return false;
	}

	// Create Scanner Object:
	lex_c scanner(&file);
		// This reads already first token.

	// Array for storing fact arguments:
	int args[VER_MAX_PRED_ARGS];

	// Main loop (over Datalog facts in file):
	while(scanner.tok_type() != TOK_EOF) {
		// Predicate:
		if(scanner.tok_type() != TOK_ATOM) {
			scanner.show_err(SYN_PRED_EXPECTED);
			return false;
		}

		// Is the predicate known/declared?
		pred_t pred = find_pred(scanner.tok_ptr(), scanner.tok_end());
		if(pred == PRED_NULL) {
			scanner.show_err(SYN_UNKNOWN_PRED);
			return false;
		}

		// Left parenthesis '(':
		scanner.next();
		if(scanner.tok_type() != TOK_LPAREN) {
			scanner.show_err(SYN_LPAREN_EXPECTED);
			return false;
		}

		// Loop over predicate arguments:
		int arity = pred->arity();
		for(int i = 0; i < arity; i++) {

			// Get argument:
			scanner.next();

			// Check and process argument:
			int arg = 0;
			strtab_t dom = STRTAB_NULL;
			switch(pred->argtype(i)) {
			case ARGTYPE_INT:
				// Argument must be integer:
				if(scanner.tok_type() != TOK_INT) {
					scanner.show_err(SYN_ARG_MUST_BE_INT);
					return false;
				}
				arg = scanner.tok_int();
				break;
			case ARGTYPE_ATOM:
				// Argument must be atom:
				if(scanner.tok_type() != TOK_ATOM) {
					scanner.show_err(SYN_ARG_MUST_BE_ATOM);
					return false;
				}
				dom = pred->arg_dom(i);
				if(dom != STRTAB_NULL) {
					arg = dom->put(scanner.tok_ptr(),
						scanner.tok_end());
					if(arg < 0)
						return false;
				}
				break;
			case ARGTYPE_STR:
				// Argument must be string:
				if(scanner.tok_type() != TOK_STR) {
					scanner.show_err(SYN_ARG_MUST_BE_STR);
					return false;
				}
				dom = pred->arg_dom(i);
				if(dom != STRTAB_NULL) {
					arg = dom->put(scanner.tok_ptr(),
						scanner.tok_end());
					if(arg < 0)
						return false;
				}
				break;
			case ARGTYPE_ATOM_STR:
				// Argument must be an atom or a string:
				if(scanner.tok_type() != TOK_ATOM &&
						scanner.tok_type() != TOK_STR){
					scanner.show_err(
						SYN_ARG_MUST_BE_ATOM_OR_STR);
					return false;
				}
				dom = pred->arg_dom(i);
				if(dom != STRTAB_NULL) {
					arg = dom->put(scanner.tok_ptr(),
						scanner.tok_end());
					if(arg < 0)
						return false;
				}
				break;
			default:
				CHECK_IMPOSSIBLE(
					"load_c::load: unknown argument type");
				scanner.show_err(SYN_INVALID_ARG_TYPE);
				return false;
			}

			// Store argument:
			CHECK(i < VER_MAX_PRED_ARGS,
				"load_c::load: arity > MAX_PRED_ARGS");
			args[i] = arg;

			// If not last argument: Comma to separate arguments:
			if(i < arity - 1) {
				scanner.next();
				if(scanner.tok_type() != TOK_COMMA) {
					scanner.show_err(SYN_COMMA_EXPECTED);
					return false;
				}
			}
		}

		// Store fact in relations:
		if(!pred->store_fact(args)) {
			err_c::store_fact_failed(filename, scanner.line_no(),
						pred->name());
			return false;
		}

		// Right parenthesis ')':
		scanner.next();
		if(scanner.tok_type() != TOK_RPAREN) {
			scanner.show_err(SYN_RPAREN_EXPECTED);
			return false;
		}

		// Full Stop '.':
		scanner.next();
		if(scanner.tok_type() != TOK_FULLSTOP) {
			scanner.show_err(SYN_FULLSTOP_EXPECTED);
			return false;
		}

		// Insert tuple (all tuples are relevant):

		// Read next token:
		scanner.next();

	} // End of Loop Over Facts


#ifdef LOAD_STATISTICS
	print_statistics(par_ff, par_fb);
#endif

	// Read the entire file without error:
	return true;
}


//=============================================================================
// Debugging Support:
//=============================================================================

//-----------------------------------------------------------------------------
// check: Check the integrity of this object, return error message or NULL.
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
str_t	load_c::check() const
{
	// Check magic number:
	if(Magic != LOAD_MAGIC)
		return "wrong magic number";

	// Ok:
	return 0;
}
#endif


//-----------------------------------------------------------------------------
// dump: Show data structure.
//-----------------------------------------------------------------------------

#ifdef	VER_DUMP
void	load_c::dump(str_t headline) const
{
	// Headline:
	if(headline == STR_NULL)
		headline = "Load Data: Predicate Table";
	err_c::dump_begin(headline);

	// Compute number of entries used in the hash table:
	int entries_used = 0;
	for(int i = 0; i < LOAD_HASHSIZE; i++) {
		if(HashTab[i] != PRED_NULL)
			entries_used++;
	}
	int percent = (entries_used * 100 + (LOAD_HASHSIZE/2)) /
				LOAD_HASHSIZE;

	// Basic information:
	err_c::dump_str("HashTabSize:  ");
	err_c::dump_int(LOAD_HASHSIZE);
	err_c::dump_nl();
	err_c::dump_str("EntriesUsed:  ");
	err_c::dump_int(entries_used);
	err_c::dump_str(" [");
	err_c::dump_int(percent);
	err_c::dump_str("%]");
	err_c::dump_nl();
	err_c::dump_str("Load ObjSize: ");
	err_c::dump_uint(static_cast<unsigned int>(sizeof(load_c)));
	err_c::dump_str(" Bytes");
	err_c::dump_nl();
	err_c::dump_nl();

	// Show entire hash table with chains of entries:
	err_c::dump_str("Hash Table:");
	err_c::dump_nl();
	for(int i = 0; i < LOAD_HASHSIZE; i++) {
		if(HashTab[i] != PRED_NULL) {
			err_c::dump_str("  ");
			err_c::dump_str(str_fill(i, 3));
			err_c::dump_int(i);
			err_c::dump_str(":");
			for(pred_t pred = HashTab[i]; pred != PRED_NULL;
						pred = pred->hash_next()) {
				err_c::dump_str(" '");
				err_c::dump_str(pred->name());
				err_c::dump_str("'");
			}
			err_c::dump_nl();
		}
	}

	// Done:
	err_c::dump_end();
}
#endif

