// ============================================================================
// Project:	Deductive Database
// Filename:	load_dblp.cpp
// Purpose:	Parse and load data from file "dblp.data" (Datalog facts).
// Last Change:	16.11.2015
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2015 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"
#include "str.h"
#include "err.h"
#include "check.h"
#include "in_file.h"
#include "syn.h"
#include "lex.h"
#include "atom.h"
#include "atomtab.h"
#include "strtab.h"
#include "list_2.h"
#include "rel_n_n.h"

#include "load_dblp.h"

#ifdef LOAD_STATISTICS
#include <iostream>
#endif


//=============================================================================
// Static Methods:
//=============================================================================


//-----------------------------------------------------------------------------
// Load data:
//-----------------------------------------------------------------------------

bool load_dblp_c::load(
		// Name of data file to load:
		str_t filename,
		// Domains:
		strtab_t doc_ids, atomtab_t attrs, strtab_t str_values,
		// Constants:
		int title, int year, int author, int month,
		// Relations for EDB-literals:
		list_2_t  att_title,  rel_n_n_t att_year,
		rel_n_n_t att_author, rel_n_n_t att_month)
{
	register str_t p;

	// Open file:
	in_file_c file(filename);
	if(!file.is_open()) {
		// Error message was printed in constructor.
		return false;
	}

	// Create Scanner Object:
	lex_c scanner(&file);
		// This reads already first token.

	// Variables for fact arguments:
	int arg1;
	int arg2;
	int arg3;

	// Main loop (over Datalog facts in file):
	while(scanner.tok_type() != TOK_EOF) {
		// Predicate:
		if(scanner.tok_type() != TOK_ATOM) {
			scanner.show_err(SYN_PRED_EXPECTED);
			return false;
		}
		p = scanner.tok_ptr();

		// There is only one predicate in the data file: 'att'
		if((scanner.tok_end() - p) != 3 ||
			((p[0] != 'a') | (p[1] != 't') | (p[2] != 't'))) {
			scanner.show_err(SYN_UNKNOWN_PRED);
			return false;
		}

		// Left parenthesis '(':
		scanner.next();
		if(scanner.tok_type() != TOK_LPAREN) {
			scanner.show_err(SYN_LPAREN_EXPECTED);
			return false;
		}

		// First argument: ID (must be a string):
		scanner.next();
		if(scanner.tok_type() != TOK_STR) {
			scanner.show_err(SYN_ARG_MUST_BE_STR);
			return false;
		}
		arg1 = doc_ids->put(scanner.tok_ptr(), scanner.tok_end());
		if(arg1 < 0)
			return false;

		// Comma to separate arguments:
		scanner.next();
		if(scanner.tok_type() != TOK_COMMA) {
			scanner.show_err(SYN_COMMA_EXPECTED);
			return false;
		}

		// Second argument: Attribute (must be an atom):
		scanner.next();
		if(scanner.tok_type() != TOK_ATOM) {
			scanner.show_err(SYN_ARG_MUST_BE_ATOM);
			return false;
		}
		arg2 = attrs->put(scanner.tok_ptr(), scanner.tok_end());
		if(arg2 < 0)
			return false;

		// Comma to separate arguments:
		scanner.next();
		if(scanner.tok_type() != TOK_COMMA) {
			scanner.show_err(SYN_COMMA_EXPECTED);
			return false;
		}

		// Third argument: Value (must be a string):
		scanner.next();
		if(scanner.tok_type() != TOK_STR) {
			scanner.show_err(SYN_ARG_MUST_BE_STR);
			return false;
		}
		arg3 = str_values->put(scanner.tok_ptr(), scanner.tok_end());
		if(arg3 < 0)
			return false;

		// Right parenthesis ')':
		scanner.next();
		if(scanner.tok_type() != TOK_RPAREN) {
			scanner.show_err(SYN_RPAREN_EXPECTED);
			return false;
		}

		// Full Stop '.':
		scanner.next();
		if(scanner.tok_type() != TOK_FULLSTOP) {
			scanner.show_err(SYN_FULLSTOP_EXPECTED);
			return false;
		}

		// Insert tuple if relevant:
		if(arg2 == title)
			att_title->insert(arg1, arg3);
		if(arg2 == year)
			att_year->insert(arg1, arg3);
		if(arg2 == author)
			att_author->insert(arg1, arg3);
		if(arg2 == month)
			att_month->insert(arg1, arg3);

		// Read next token:
		scanner.next();

	} // End of Loop Over Facts

	// Read the entire file without error:
	return true;
}

