// ============================================================================
// Project:	Deductive Database
// Filename:	main.cpp
// Purpose:	Main program
// Last Change:	23.05.2016
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2011-2016 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"
#include "str.h"
#include "idiv.h"
#include "err.h"
#include "perf.h"
//#include "file.h"
//#include "buf.h"
#include "page.h"
#include "mpool.h"
#include "in_file.h"
#include "ch.h"
#include "tok.h"
#include "lex.h"
#include "strtab.h"
#include "atomtab.h"
#include "row_2.h"
#include "list.h"
#include "cur_list.h"
#include "set_2.h"
#include "set_2e.h"
#include "rel_ii.h"
#include "cur_ii.h"
#include "rel_n_n.h"
#include "cur_n_n.h"
#include "dup_2i.h"
#include "pred.h"
#include "load.h"
#include "load_dblp.h"
#include "load_tc.h"
#include "load_tcbf.h"
#include "out.h"
#include "test.h"

// #pragma GCC diagnostic ignored "-Wunreachable-code"
#include <iostream>
#include <fstream>
#include <iomanip>
using namespace std;

//=============================================================================
// Select a default test sequence (if no command line arguments):
//=============================================================================

//#define MAIN_DEFAULT_TESTS "j1axy_6"
//#define MAIN_DEFAULT_TESTS "j1axy_7"
// #define MAIN_DEFAULT_TESTS "-d j1axy_8"

//#define MAIN_DEFAULT_TESTS "j1axy_1_10k"
//#define MAIN_DEFAULT_TESTS "j1axy_2_10k"
//#define MAIN_DEFAULT_TESTS "j1axy_3_10k"
//#define MAIN_DEFAULT_TESTS "j1axy_4_10k"
//#define MAIN_DEFAULT_TESTS "j1axy_5_10k"
//#define MAIN_DEFAULT_TESTS "j1axy_6_10k"
//#define MAIN_DEFAULT_TESTS "j1axy_7_10k"
//#define MAIN_DEFAULT_TESTS "j1axy_8_10k"
//#define MAIN_DEFAULT_TESTS "j1axy_9"

//#define MAIN_DEFAULT_TESTS "j1axy_6_10k eval_j1_10f"
//#define MAIN_DEFAULT_TESTS "eval_wine1"
//#define MAIN_DEFAULT_TESTS "load_wine"
//#define MAIN_DEFAULT_TESTS "eval_j1_10d"
//#define MAIN_DEFAULT_TESTS "eval_j1 eval_j1_d eval_j1_e"
//#define MAIN_DEFAULT_TESTS "eval_j1_10"
//#define MAIN_DEFAULT_TESTS "load_j1_10"
//#define MAIN_DEFAULT_TESTS "set2"
//#define MAIN_DEFAULT_TESTS "dup3 dup2 set2"
//#define MAIN_DEFAULT_TESTS "list1 list2"
//#define MAIN_DEFAULT_TESTS "tcbf0 tcbf1"
//
//#define MAIN_DEFAULT_TESTS "tcff_4_test"
//#define MAIN_DEFAULT_TESTS "tcbf_2_test"
//#define MAIN_DEFAULT_TESTS "tcbf_4_test"
//#define MAIN_DEFAULT_TESTS "set_e1"
//#define MAIN_DEFAULT_TESTS "set_1e_test"
#define MAIN_DEFAULT_TESTS "set_1d_m set_1e_m"

//=============================================================================
// Constants:
//=============================================================================

// Data File for DBLP Benchmark: 125.434 KB (~122 MB)
#define MAIN_TESTFILE "data/dblp.data"

// Data Files for TC Benchmark (transitive closure):
//	Small Version (250.000 facts, 3.3 MB, without cycles):
#define MAIN_FILE_TC1 "data/tc_d1000_parsize250000_xsb_nocyc.P"
// #define MAIN_FILE_TC1 "par.P"

// Data File for Wine Ontology Benchmark:
#define MAIN_FILE_WINE "data/wine_data.P"

// Data files for the Join1 Benchmarks:
#define MAIN_FILE_JOIN1_TEST "data/d1000_test.P";
#define MAIN_FILE_JOIN1_10K "data/d1000_relsize10000_xsb_cyc.P"
#define MAIN_FILE_JOIN1_50K "data/d1000_relsize50000_xsb_cyc.P"
#define MAIN_FILE_JOIN1_250K "data/d1000_relsize250000_xsb_cyc.P"

// Show XSB Times for comparison. Makes only sense on machine inf-ts:
#define MAIN_SHOW_XSB_TIME

//=============================================================================
// Macros for dumping data structures (objects):
//=============================================================================

//-----------------------------------------------------------------------------
// Optional dump of an object:
//-----------------------------------------------------------------------------

#ifdef VER_DEBUG
#define MAIN_OPT_DUMP(SHOW_DUMP, OBJ) \
	if(SHOW_DUMP) (OBJ).dump()
#else
#define MAIN_OPT_DUMP(SHOW_DUMP, OBJ) \
	if(SHOW_DUMP) cout << "Data structure dump " << \
			"is available only in debug mode\n"
#endif

//-----------------------------------------------------------------------------
// Error was detected (test failed), give more information:
//-----------------------------------------------------------------------------

#ifdef VER_DEBUG
#define MAIN_ERR_DUMP(OBJ) \
	(OBJ).dump()
#else
#define MAIN_ERR_DUMP(OBJ) \
	((void)0)
#endif

//=============================================================================
// Note:
// ----
//	All test functions are called with a single integer argument
//	(some functions can perform different tests, e.g. tests with different
//	numbers of elements).
//	If the test function does not use this argument, it should call
//	arg_unused(arg)
//	which checks that the passed value is 0 and voids a compiler warning
//	for the unused argument.
//
//	All test functions should return a bool value (true if the test is
//	passed and false otherwise).
//
//	The test functions must be registered in the list at the end of this
//	file. The list is at the end because otherwise we would need forward
//	declarations for the test functions.
//	All test functions have an external name (a string) which can
//	be used as a command line argument to identify the function plus
//	a concrete argument value.
//	Therefore, an entry in the list contains this name, the function,
//	and the argument value (0 if unused).
//=============================================================================


//-----------------------------------------------------------------------------
// Structure with information about a test:
//-----------------------------------------------------------------------------

struct main_test_s {
	str_t Name;
	bool (*Fun)(int arg);
	int   Arg;
	main_test_s(str_t name, bool (*fun)(int arg), int arg) {
		Name = name;
		Fun  = fun;
		Arg  = arg;
	}
};

typedef main_test_s *main_test_t;

//-----------------------------------------------------------------------------
// Function to call if argument is not used:
//-----------------------------------------------------------------------------

static void arg_unused(int arg)
{
	if(arg != 0)
		cerr << "Test was called with unused arg != 0\n";
}

//-----------------------------------------------------------------------------
// Function to print help:
//-----------------------------------------------------------------------------

static void main_print_help()
{
	out_c::help_line(
	"ydb - Test Implementation of the Push Method for Bottom-Up Evaluation"
	);
	out_c::help_line("");
	out_c::help_line("Usage: ./ydb <Options> <Tests>");
	out_c::help_line("Options are:");
	out_c::help_line(
	"-aN: Show the first N answers of the benchmark query.");
	out_c::help_line(
	"-a:  Do not show answers (default are the first 10 answers).");
	out_c::help_line(
	"-d:  Print additional debug output (includes data structure dumps).");
	out_c::help_line("-l:  List all tests.");
	out_c::help_line("-m:  Show memory pools.");
	out_c::help_line("-r:  List relations.");
	out_c::help_line(
	"-s:  Print only summary line for test/benchmark (\"silent\").");
	out_c::help_line(
   "-b:  For benchmarks: Similar to -s, no newline at end, msg in debug ver.");
	out_c::help_line("");
	out_c::help_line("To see a list of all tests, call \"./ydb -l\"");
	out_c::help_line("");
	out_c::help_line("The benchmark summary line contains the test ID, ");
	out_c::help_line(
		"the load time (in ms), the eval time and total time, ");
	out_c::help_line("and finally the status/result (OK or otherwise).");
	out_c::help_line(
	"\"Warn\" usually means that wallclock time and real time differ much."
	);
#ifdef VER_DEBUG
	out_c::help_line("");
	out_c::help_line(
		"This version has the checking of assertions compiled in,");
	out_c::help_line("and is not good for benchmarks.");
#endif
}


//-----------------------------------------------------------------------------
// Function to excute a test:
//-----------------------------------------------------------------------------

static int main_num_tests();
static main_test_t main_tests();

static void main_exec_test(str_t name)
{
	// Handle Options:
	if(name[0] == '-') {
		int i = 1;
		switch(name[i++]) {
		case '\0':
			err_c::option_missing();
			return;
		case 'a':
			// See answers:
			{ // Without own block g++ complains
			  // main.cpp:214:8: error: jump to case label
			  // [-fpermissive]
			  //   case 's':
			  //        ^
			  // main.cpp:204:9: error:
			  // crosses initialization of long int limit
			if(name[i] == '\0') {
				// Default limit for answers:
				out_c::set_answer_limit(0);
				break;
			}
			long limit = 0;
			int d;
			while(name[i] != 0 &&
				(d = ch_c::digit_val(name[i])) >= 0 && d < 10){
				limit = limit * 10 + d;
				i++;
			}
			if(name[i] == '\0')
				out_c::set_answer_limit(limit);
			}
			break;
		case 'b':
			// For benchmarks:
			// Silent mode (with only summary line):
			out_c::set_verbose(false);
			out_c::set_answer_limit(0);
			out_c::set_mem_pools(false);
			out_c::set_summary(true, false);
#ifdef VER_DEBUG
			err_c::benchmark_with_debug_code();
#else
#ifdef VER_DEBUG_OUT_ENABLED
			err_c::benchmark_with_debug_code();
#endif
#endif
			break;
		case 'd':
			// Debug Mode:
			out_c::set_debug(true);
			out_c::set_mem_pools(true);
			out_c::set_dump(true);
#ifndef VER_DEBUG_OUT_ENABLED
			err_c::debug_output_disabled();
#endif
			break;
		case 'h':
			// Print help:
			main_print_help();
			break;
		case 'l':
			// Print list of all tests:
			test_c::print_list();
			break;
		case 'm':
			// Show memory pools:
			out_c::set_mem_pools(true);
			break;
		case 'r':
			// Show relations:
			out_c::set_rel(true);
			break;
		case 's':
			// Silent mode (with only summary line):
			out_c::set_verbose(false);
			out_c::set_answer_limit(0);
			out_c::set_mem_pools(false);
			out_c::set_summary(true, true);
			break;
		default:
			// Unknown Option:
			err_c::unknown_option(name);
			return;
		}

		// Currently, we understand only one-character options:
		if(name[i] != 0) {
			err_c::unknown_option(name);
			return;
		}

		// We are done with this argument:
		return;
	}

	// Search in the new test class:
	bool found = false;
	test_t test_obj = test_c::get_test(name);
	if(test_obj != TEST_NULL) {
		found = true;
		(void) test_obj->run();
		return;
	}

	// Normal Test - Print headline:
	cout << "\n";
	cout << "----------------------------------------" <<
		"---------------------------------------\n";
	cout << "--- TEST \"" << name << "\"\n";
	cout << "----------------------------------------" <<
		"---------------------------------------\n";
	cout << "\n";

	// Get test list:
	int num_tests = main_num_tests();
	main_test_t test = main_tests();

	// Search and execute test:
	for(int i = 0; i < num_tests; i++) {
		if(str_eq(test->Name, name)) {
			found = true;
			bool ok = (test->Fun)(test->Arg);
			if(!ok) {
				cerr << "\n";
				cerr << "Test '" << name << "' failed!\n";
			}
		}
		test++;
	}

	// No test was found with this name:
	if(!found)
		cerr << "There is no test named '" << name << "'!\n";

	// Some extra space after the test output:
	cout << "\n";
}

//=============================================================================
// Exported Functions:
//=============================================================================

//-----------------------------------------------------------------------------
// main: Main program.
//-----------------------------------------------------------------------------

int main(int argc, char *argv[])
{
	// Open alert file:
	if(!err_c::open_alert(VER_ALERT_FILE)) {
		cerr << "Alert file could not be opened!\n";
		return 1;
	}
	err_c::use_stderr(true);

	// Initialize character classes:
	ch_c::init();

	// Initialize buffer manager:
	//buf_c::init();

	// If there are command line arguments: execute these tests:
	if(argc > 1) {
		for(int i = 1; i < argc; i++)
			main_exec_test(argv[i]);
	}

	// Otherwise execute default test sequence:
	else {
		static const int MAX_TEST_NAME_LEN = 100;
		char name_arr[MAX_TEST_NAME_LEN+1];
		char *name_p = name_arr;
		for(str_t tests_p = MAIN_DEFAULT_TESTS; *tests_p; tests_p++) {
			if(*tests_p != ' ')
				*name_p++ = *tests_p;
			else {
				if(name_p != name_arr) {
					*name_p = '\0';
					main_exec_test(name_arr);
				}
				name_p = name_arr;
			}
		}
		if(name_p != name_arr) {
			*name_p = '\0';
			main_exec_test(name_arr);
		}
	}

	// Check that all memory pools were deleted:
	mpool_t mpool = mpool_c::first_pool();
	if(mpool != MPOOL_NULL) {
		cout << "There are still memory pools:\n";
		for( ; mpool != MPOOL_NULL; mpool = mpool->next_pool())
			cout << "    '" << mpool->name() << "'\n";
		cout << "This should not happen (memory leak)!";
	}

	// Check for errors:
	if(!err_c::ok()) {
		cout << "\n";
		int n = err_c::num_errs();
		cout << n << " Errors detected:\n";
		for(int i = 0; i < n; i++)
			cout << i+1 << ". " << err_c::msg(i) << "\n";
		cout << "\n";
	}

	// Close alert file:
	if(!err_c::close_alert()) {
		cerr << "Error closing alert file!\n";
		return 1;
	}

#ifdef VER_PRESS_RETURN
	//char c_return;
	cout << "\n(Please press Return/Enter to finish)\n";
	//while(cin.get(c_return) && c_return != '\n' && c_return != '\r')
	//	{}
	for(char c = ' '; !cin.fail() && !cin.eof() && c != '\n'; )
		cin.get(c);
#endif
	// Terminate with success:
	return 0;
}


//=============================================================================
// Local Functions (Auxiliary Functions):
//=============================================================================

//-----------------------------------------------------------------------------
// main_mem_statistics: Print memory pool statistics.
//-----------------------------------------------------------------------------

void main_mem_statistics(str_t headline)
{
	static const int name_field_width = 22;
	static const int page_field_width = 7;
	static const int kb_field_width = 7;

	int total_pages = 0;

	if(headline != STR_NULL)
		cout << headline << "\n";
	mpool_t mpool = mpool_c::first_pool();
	if(mpool == MPOOL_NULL) {
		cout << "    No memory pools exist.\n";
		return;
	}
	for( ; mpool != MPOOL_NULL; mpool = mpool->next_pool()) {
		str_t name = mpool->name();
		int pages = mpool->num_alloc_pages();
		int kb = PAGE_KB(pages);

		cout << "    " << name << ": " <<
			str_indent(name_field_width-str_len(name)) <<
			str_fill(pages, page_field_width) << pages <<
			" Pages " <<
			str_fill(kb, kb_field_width) << "[" << kb << " KB]" <<
			"\n";
		total_pages += pages;
	}
	if(total_pages > 0) {
		cout << "    [" <<
			(static_cast<long>(total_pages) * VER_PAGESIZE)/1024 <<
			" KB in total]\n";
	}
}

//=============================================================================
// Tests of Basic File Reading, Lexical Scanner:
//=============================================================================


//-----------------------------------------------------------------------------
// Read a large file with standard libary character by character
//-----------------------------------------------------------------------------

static bool main_std_fileread(int arg)
{
	// Argument is not used for this test:
	arg_unused(arg);

	// Print test headline:
	cout << "Reading '" << MAIN_TESTFILE << "' character by character "
		<< "with standard library ...\n";

	// Measure time for reading DBLP data file (about 122 MB):
	perf_c timer;
	timer.start();

	// Open data file:
	ifstream input(MAIN_TESTFILE);
	if(!input) {
		cerr << "Cannot open input file: '" << MAIN_TESTFILE << "'.\n";
		return false;
	}

	// Read character by character:
	char c;
	int count = 0;
	while(input.get(c))
		count++;

	// Close data file:
	if(!input.eof()) {
		cerr << "Input file '" << MAIN_TESTFILE << "' not at EOF!\n";
		return false;
	}
	if(input.bad()) {
		cerr << "Input stream '" << MAIN_TESTFILE << "' corrupted!\n";
		return false;
	}
	input.close();

	// Print runtime:
	timer.stop();
	cout << "Time for reading " << count << " bytes "
		<< "from '" << MAIN_TESTFILE "': "
		<< timer.realtime_msec() << " ms "
		<< "(" << timer.cputime_msec() << "ms CPU)\n";
	return true;

}


//-----------------------------------------------------------------------------
// Allocate and deallocate many memory pages.
//-----------------------------------------------------------------------------

static bool main_mem_alloc(int num_pages)
{
	const int max_pages = 100000;
	page_t pages[max_pages];

	// Print test headline:
	cout << "Allocating " << num_pages << " memory pages of "
		<< VER_PAGESIZE << " bytes each ...\n";

	// Test argument:
	if(num_pages > max_pages) {
		cerr << "Invalid call to main_mem_alloc: " <<
			"num_pages (" << num_pages << ") > max_pages (" <<
			max_pages << "\n";
		return false;
	}

	// Measure time for allocating memory pages:
	perf_c timer;
	timer.start();

	// Create memory pool inside block (thus destructor is checked, too):
	{
	mpool_c pool("Testpool");

	// Print address of first pages:
	pages[0] = pool.alloc_page();
	cout << "    Address of first page:  " << pages[0] << "\n";
	if(pages[0] == PAGE_NULL) {
		cerr << "Allocation of first page failed\n";
		return false;
	}
	pages[1] = pool.alloc_page();
	cout << "    Address of second page: " << pages[1] << "\n";
	if(pages[1] == PAGE_NULL) {
		cerr << "Allocation of second page failed\n";
		return false;
	}
	if(pages[1] == pages[0]) {
		cerr << "Same page was returned in both calls\n";
		return false;
	}

	// Allocate memory pages:
	for(int i = 2; i < num_pages; i++) {
		pages[i] = pool.alloc_page();
		if(pages[i] == PAGE_NULL) {
			cerr << "Allocation of page " << i << " failed\n";
			return false;
		}
		for(int j = 0; j < i; j++) {
			if(pages[i] == pages[j]) {
				cerr << "Same page was returned in calls " <<
					i << " and " << j << "\n";
				return false;
			}
		}
	}

	// Print runtime:
	timer.stop();
	cout << "Time for allocating " << num_pages * VER_PAGESIZE
		<< " bytes: "
		<< timer.realtime_msec() << " ms "
		<< "(" << timer.cputime_msec() << "ms CPU)\n";

	// Deallocate pages:
	timer.start();
	cout << "\n";
	cout << "Destructor for memory pool is called, " <<
		"pages will be freed now.\n";
	}
	timer.stop();
	cout << "Time for deallocating " << num_pages << " pages: "
		<< timer.realtime_msec() << " ms "
		<< "(" << timer.cputime_msec() << "ms CPU)\n";

	// Ok:
	return true;
}

//-----------------------------------------------------------------------------
// Do only lexical analysis:
//-----------------------------------------------------------------------------

static bool main_lex(int arg)
{
	// Argument is not used for this test:
	arg_unused(arg);

	// Print test headline:
	cout << "Lexical analysis of '" << MAIN_TESTFILE << "' ...\n";

	// Measure time for doing lexical analysis:
	perf_c timer;
	timer.start();

	// Open file:
	in_file_c file(MAIN_TESTFILE);
	if(!file.is_open()) {
		cout << "File not open, cannot do lexical analysis.\n";
		return false;
	}

	// Do lexical analysis:
	lex_c scanner(&file);
	cout << "First token: " << tok_name(scanner.tok_type()) << "\n";
	scanner.next();
	cout << "Second token: " << tok_name(scanner.tok_type()) << "\n";
	scanner.next();
	cout << "Third token: " << tok_name(scanner.tok_type()) << "\n";
	int num_tokens = 3;
	while(scanner.tok_type() != TOK_EOF) {
		scanner.next();
		num_tokens++;
	}

	// Print runtime:
	timer.stop();
	cout << "Time for reading " << num_tokens << " tokens: "
		<< timer.realtime_msec() << " ms "
		<< "(" << timer.cputime_msec() << "ms CPU)\n";

	// Ok:
	return false;
}


//=============================================================================
// Tests of Data Structures for Storing Strings:
//=============================================================================


//-----------------------------------------------------------------------------
// Check Atom Table:
//-----------------------------------------------------------------------------

static bool main_atomtab(int arg)
{
	// Test parameters:
	int num_atoms = 1000; // Must be < 9.999. Hashtab: 251
	int lookup_cycles = 100;
	bool dump_table;
	switch(arg) {
		case 0:
			num_atoms = 1000;
			lookup_cycles = 100;
			dump_table = false;
			break;

		case 1:
			num_atoms = 1000;
			lookup_cycles = 100;
			dump_table = true;
			break;

		case 2:
			num_atoms = 100;
			lookup_cycles = 10;
			dump_table = true;
			break;

		default:
			cerr << "Invalid argument to atomtab test\n";
			return false;
	}

	// Print test headline:
	cout << "Storing " << num_atoms << " atoms of size 10 ...\n";

	// Measure time for storing atoms:
	perf_c timer;
	timer.start();

	// Make atom table:
	atomtab_c atom_tab("Atom Table");

	// Construct test atom, fixed part:
	char name[11];
	name[0]  = 'A';
	name[1]  = 'T';
	name[2]  = 'O';
	name[3]  = 'M';
	name[4]  = '_';
	name[5]  = '_';
	name[6]  = '_';
	name[7]  = '_';
	name[8]  = '_';
	name[9]  = '0';
	name[10]  = '\0';
	char *no_ptr = name + 6;
	char *end = name + 10;

	// Store atoms:
	int id = atom_tab.put(name, end);
	if(id < 0) {
		cerr << "Storing first atom '" << name << "' failed.\n";
		return false;
	}
	cout << "First Atom: '" << atom_tab.get(id) << "', Id: " <<
					id << "\n";
	for(int i = 1; i < num_atoms; i++) {
		name[6]  = '_';
		name[7]  = '_';
		name[8]  = '_';
		name[9]  = '_';
		if(i < 10)
			str_int(no_ptr+3, 2, i);
		else if(i < 100)
			str_int(no_ptr+2, 3, i);
		else if(i < 1000)
			str_int(no_ptr+1, 4, i);
		else
			str_int(no_ptr,   5, i);
		id = atom_tab.put(name, end);
		if(id < 0) {
			cerr << "Storing atom '" << name << "' failed.\n";
			return false;
		}
		if(id != i) {
			cerr << "Atom '" << name << "' " <<
				"has wrong id: " << id << ".\n";
			return false;
		}
	}
	cout << "Last Atom:  '" << atom_tab.get(id) << "', Id: " <<
					id << "\n";

	// Print runtime:
	timer.stop();
	cout << "Time for storing " << num_atoms << " atoms: "
		<< timer.realtime_msec() << " ms "
		<< "(" << timer.cputime_msec() << "ms CPU)\n";

	// Show hash table:
#ifdef VER_DEBUG
	if(dump_table)
		atom_tab.dump();
#else
	if(dump_table)
		cout << "Data structure dump " <<
			"is available only in debug mode\n";
#endif

	// Lookup existing atoms:
	timer.start();
	for(int n = 0; n < lookup_cycles; n++) {
		for(int i = 0; i < num_atoms; i++) {
			name[6]  = '_';
			name[7]  = '_';
			name[8]  = '_';
			name[9]  = '_';
			if(i < 10)
				str_int(no_ptr+3, 2, i);
			else if(i < 100)
				str_int(no_ptr+2, 3, i);
			else if(i < 1000)
				str_int(no_ptr+1, 4, i);
			else
				str_int(no_ptr,   5, i);
			id = atom_tab.put(name, end);
			if(id < 0) {
				cerr << "Lookup of atom '" << name << "' " <<
				       	"failed.\n";
				return false;
			}
			if(id != i) {
				cerr << "Atom '" << atom_tab.get(id) << "' " <<
				       	"has wrong id: "<< id << ".\n";
				return false;
			}
		}
	}
	timer.stop();
	cout << "Time for retrieving " << num_atoms << " atoms " <<
			lookup_cycles << " times: " <<
			timer.realtime_msec() << " ms " <<
			"(" << timer.cputime_msec() << "ms CPU)\n";

	// Print memory allocation statistics:
	cout << "\n";
	main_mem_statistics("Memory Pools:");

	// Ok:
	return true;
}

//-----------------------------------------------------------------------------
// Check String Table:
//-----------------------------------------------------------------------------

static bool main_strtab_1(int arg)
{
	// Argument is not used for this test:
	arg_unused(arg);

	// Print test headline:
	cout << "First Test of String Table ...\n";

	// Measure time for storing strings:
	perf_c timer;
	timer.start();

	// Make string table:
	strtab_c str_tab("String Table 1");

	// First test string:
	str_t s1 = "abcdef";
	cout << "Storing '" << s1 << "' ...";
	int i1 = str_tab.put(s1, s1+str_len(s1));
	cout << " ID: " << i1 << "\n";
	if(i1 < 0) {
		cerr << "First insert failed\n";
		return false;
	}

	// Second test string:
	str_t s2 = "abcxyz";
	cout << "Storing '" << s2 << "' ...";
	int i2 = str_tab.put(s2, s2+str_len(s2));
	cout << " ID: " << i2 << "\n";
	if(i2 < 0) {
		cerr << "Second insert failed\n";
		return false;
	}

	// Third test string:
	str_t s3 = "ab";
	cout << "Storing '" << s3 << "' ...";
	int i3 = str_tab.put(s3, s3+str_len(s3));
	cout << " ID: " << i3 << "\n";
	if(i3 < 0) {
		cerr << "Third insert failed\n";
		return false;
	}

	// Fourth test string:
	str_t s4 = "abijk";
	cout << "Storing '" << s4 << "' ...";
	int i4 = str_tab.put(s4, s4+str_len(s4));
	cout << " ID: " << i4 << "\n";
	if(i4 < 0) {
		cerr << "Fourth insert failed\n";
		return false;
	}

	// Fifth test string:
	str_t s5 = "abm";
	cout << "Storing '" << s5 << "' ...";
	int i5 = str_tab.put(s5, s5+str_len(s5));
	cout << " ID: " << i5 << "\n";
	if(i5 < 0) {
		cerr << "Fifth insert failed\n";
		return false;
	}

	// Sixth test string:
	str_t s6 = "abpqr";
	cout << "Storing '" << s6 << "' ...";
	int i6 = str_tab.put(s6, s6+str_len(s6));
	cout << " ID: " << i6 << "\n";
	if(i6 < 0) {
		cerr << "Sixth insert failed\n";
		return false;
	}

	// Seventh test string:
	str_t s7 = "!";
	cout << "Storing '" << s7 << "' ...";
	int i7 = str_tab.put(s7, s7+str_len(s7));
	cout << " ID: " << i7 << "\n";
	if(i7 < 0) {
		cerr << "Seventh insert failed\n";
		return false;
	}

	// Eighth test string:
	str_t s8 = "abpqst";
	cout << "Storing '" << s8 << "' ...";
	int i8 = str_tab.put(s8, s8+str_len(s8));
	cout << " ID: " << i8 << "\n";
	if(i8 < 0) {
		cerr << "Eighth insert failed\n";
		return false;
	}

	// Print runtime:
	timer.stop();
	cout << "Time for storing test strings: "
		<< timer.realtime_msec() << " ms "
		<< "(" << timer.cputime_msec() << "ms CPU)\n";

	// Show data structure:
#ifdef	VER_DUMP
	str_tab.dump();
	cout << "\n";
#endif

	// Retrieve all strings:
	timer.start();

	// Count number of errors:
	int e = 0;

	// First string:
	cout << "Retrieving '" << s1 << "' ...";
	int j1 = str_tab.put(s1, s1+str_len(s1));
	cout << "ID: " << j1 << ((j1 == i1)?" OK.":(e++," ***WRONG!")) << "\n";

	// Second string:
	cout << "Retrieving '" << s2 << "' ...";
	int j2 = str_tab.put(s2, s2+str_len(s2));
	cout << "ID: " << j2 << ((j2 == i2)?" OK.":(e++," ***WRONG!")) << "\n";

	// Third string:
	cout << "Retrieving '" << s3 << "' ...";
	int j3 = str_tab.put(s3, s3+str_len(s3));
	cout << "ID: " << j3 << ((j3 == i3)?" OK.":(e++," ***WRONG!")) << "\n";

	// Fourth string:
	cout << "Retrieving '" << s4 << "' ...";
	int j4 = str_tab.put(s4, s4+str_len(s4));
	cout << "ID: " << j4 << ((j4 == i4)?" OK.":(e++," ***WRONG!")) << "\n";

	// Fifth string:
	cout << "Retrieving '" << s5 << "' ...";
	int j5 = str_tab.put(s5, s5+str_len(s5));
	cout << "ID: " << j5 << ((j5 == i5)?" OK.":(e++," ***WRONG!")) << "\n";

	// Sixth string:
	cout << "Retrieving '" << s6 << "' ...";
	int j6 = str_tab.put(s6, s6+str_len(s6));
	cout << "ID: " << j6 << ((j6 == i6)?" OK.":(e++," ***WRONG!")) << "\n";

	// Seventh string:
	cout << "Retrieving '" << s7 << "' ...";
	int j7 = str_tab.put(s7, s7+str_len(s7));
	cout << "ID: " << j7 << ((j7 == i7)?" OK.":(e++," ***WRONG!")) << "\n";

	// Eighth string:
	cout << "Retrieving '" << s8 << "' ...";
	int j8 = str_tab.put(s8, s8+str_len(s8));
	cout << "ID: " << j8 << ((j8 == i8)?" OK.":(e++," ***WRONG!")) << "\n";

	// Check for errors:
	if(e == 1) {
		cerr << "*** There was one error!\n\n";
		return false;
	}
	else if(e > 1) {
		cerr << "*** There were " << e << " errors!\n\n";
		return false;
	}

	// Print time for retrieving strings:
	timer.stop();
	cout << "Time for retrieving test strings: "
		<< timer.realtime_msec() << " ms "
		<< "(" << timer.cputime_msec() << "ms CPU)\n";

	// Print memory allocation statistics:
	cout << "\n";
	main_mem_statistics("Memory Pools:");

	// Ok:
	return true;
}

//-----------------------------------------------------------------------------
// Check String Table:
//-----------------------------------------------------------------------------

static bool main_strtab_2(int num_strings)
{
	perf_c timer;

	// The constructed strings have the form ABC...(base_no+i),
	// where i runs from 1 to num_strings.
	// All these numbers should have exactly "digits" digits.
	int base_no;
	int digits;

	// Check argument:
	if(num_strings < 900) {
		digits = 3;
		base_no = 100;
	}
	else if(num_strings < 9000) {
		digits = 4;
		base_no = 1000;
	}
	else if(num_strings < 90000) {
		digits = 5;
		base_no = 10000;
	}
	else if(num_strings < 900000) {
		digits = 6;
		base_no = 100000;
	}
	else {
		cerr << "num_strings too large for test StrTab2: " <<
				num_strings << "\n";
		return false;
	}

	// Print test headline:
	cout << "Storing " << num_strings << " strings of size 20 ...\n\n";

	// Measure time for storing strings:
	timer.start();

	// Make string table:
	strtab_c str_tab("String Table");

	// Construct test string, fixed part:
	char str[21];
	str[0]  = 'A';
	str[1]  = 'B';
	str[2]  = 'C';
	str[3]  = 'D';
	str[4]  = 'E';
	str[5]  = 'F';
	str[6]  = 'G';
	str[7]  = 'H';
	str[8]  = 'I';
	str[9]  = 'J';
	str[10] = 'K';
	str[11] = 'L';
	str[12] = 'M';
	str[13] = 'N';
	str[14] = '_';
	str[15] = '_';
	str[16] = '_';
	str[17] = '_';
	str[18] = '_';
	str[19] = '_';
	char *buf = str + (20-digits);
	char *end = str + 20;

	// Store strings:
	str_int(buf, digits+1, base_no);
	str[20] = '\0';
	cout << "First string: '" << str << "'.\n";
	int str_i = str_tab.put(str, end);
	if(str_i < 0) {
		cerr << "Storing first string '" << str << "' failed.\n";
		return false;
	}
	cout << "Index of first string: " << str_i << "\n";
	for(int i = 1; i < num_strings; i++) {
		str_int(buf, digits+1, base_no + i);
		str[20] = '\0';
		str_i = str_tab.put(str, end);
		if(str_i < 0) {
			cerr << "Storing string '" << str << "' failed.\n";
			return false;
		}
	}
	cout << "Index of last string: " << str_i << "\n\n";

	// Print runtime:
	timer.stop();
	cout << "Time for storing " << num_strings << " strings: "
		<< timer.realtime_msec() << " ms "
		<< "(" << timer.cputime_msec() << "ms CPU)\n";

	// Retrieve all strings:
	cout << "\n";
	cout << "Retrieving all strings again ...\n\n";

	timer.start();
	str_int(buf, digits+1, base_no);
	str[20] = '\0';
	str_t s = str_tab.get(0);
	if(s == STR_NULL) {
		cerr << "Retrieving first string '" << str << "' failed.\n";
		return false;
	}
	if(!str_eq(s, str)) {
		cerr << "Wrong string '" << s << "' was returned " <<
			"instead of '" << str << "'.\n";
		return false;
	}
	cout << "First string: '" << s << "'\n";
	for(int i = 1; i < num_strings; i++) {
		str_int(buf, digits+1, base_no + i);
		str[20] = '\0';
		s = str_tab.get(i);
		if(s == STR_NULL || !str_eq(s, str)) {
			cerr << "Retrieving string " << i << " failed :'" <<
					s << "'.\n";
			return false;
		}
	}
	cout << "Last string:  '" << s << "'\n\n";
	timer.stop();
	cout << "Time for retrieving " << num_strings << " strings: "
		<< timer.realtime_msec() << " ms "
		<< "(" << timer.cputime_msec() << "ms CPU)\n";

	// Print memory allocation statistics:
	cout << "\n";
	main_mem_statistics("Memory Pools:");

	// Ok:
	return true;
}


//=============================================================================
// Tests of Data Structures for Storing Relations:
//=============================================================================


//-----------------------------------------------------------------------------
// Check Relation with two natural number columns, array index over 1st:
//-----------------------------------------------------------------------------

static bool main_rel_n_n_1(int arg)
{
	// Argument is not used for this test:
	arg_unused(arg);

	// Print test headline:
	cout << "First Test of Relation with two natural number columns, " <<
	       "Array Index over 1st\n";

	// Measure time for storing values:
	perf_c timer;
	timer.start();

	// Make Relation:
	rel_n_n_c rel1("test1");

	// Insert values:
	cout << "  Inserting values ...";
	cout << " [1,100]";
	if(!rel1.insert(1, 100)) {
		cerr << " Insert of [1,100] failed.\n";
		return false;
	}
	cout << " [5,501]";
	if(!rel1.insert(5, 501)) {
		cerr << " Insert of [5,501] failed.\n";
		return false;
	}
	cout << " [5,502]";
	if(!rel1.insert(5, 502)) {
		cerr << " Insert of [5,502] failed.\n";
		return false;
	}
	cout << " [3,301]";
	if(!rel1.insert(3, 301)) {
		cout << " Insert of [3,301] failed.\n";
		return false;
	}
	cout << " [3,302]";
	if(!rel1.insert(3, 302)) {
		cout << " Insert of [3,302] failed.\n";
		return false;
	}
	cout << " [5,503]";
	if(!rel1.insert(5, 503)) {
		cout << " Insert of [5,503] failed.\n";
		return false;
	}
	cout << " [10,1000]";
	if(!rel1.insert(10, 1000)) {
		cout << " Insert of [10,1000] failed.\n";
		return false;
	}
	cout << " ok.\n";
#ifdef VER_DUMP
	rel1.dump();
#endif

	// Retrieve first tuple:
	cur_n_n_c cur(&rel1);
	cout << "  Retrieving values for c1 = 1 ... ";
	cur.open(1);
	if(!cur.fetch()) {
		cerr << "First tuple not found.\n";
		return false;
	}
	int val = cur.c2();
	if(val != 100) {
		cerr << "Wrong value  " << val <<
			"returned for first tuple.\n";
		return false;
	}
	if(cur.fetch()) {
		cerr << "Too many values found for c1 = 1.\n";
		return false;
	}
	cout << " ok.\n\n";

	// Value set for c1 == 3:
	cout << "  Retrieving values for c1 = 3 ... ";
	bool found301 = false;
	bool found302 = false;
	cur.open(3);
#ifdef VER_DUMP
	cur.dump();
#endif
	if(!cur.fetch()) {
		cerr << "Tuples for c1 = 3 not found.\n";
		return false;
	}
	val = cur.c2();
	if(val == 301)
		found301 = true;
	else if(val == 302)
		found302 = true;
	else {
		cerr << "Wrong value  " << val << "returned for c1 = 3.\n";
		return false;
	}
#ifdef VER_DUMP
	cur.dump();
#endif
	if(!cur.fetch()) {
		cerr << "Second tuple for c1 = 3 not found.\n";
		return false;
	}
	val = cur.c2();
	if(val == 301)
		found301 = true;
	else if(val == 302)
		found302 = true;
	else {
		cerr << "Wrong value  " << val << "returned for c1 = 3.\n";
		return false;
	}
#ifdef VER_DUMP
	cur.dump();
#endif
	if(cur.fetch()) {
		cerr << "Too many tuples found for c1 = 3.\n";
		return false;
	}
	if(!found301) {
		cerr << "301 not found for c1 = 3.\n";
		return false;
	}
	if(!found302) {
		cerr << "302 not found for c1 = 3.\n";
		return false;
	}
	cout << "ok.\n";

	// Print time for inserting and retrieving a few rows:
	timer.stop();
	cout << "Time for first test: "
		<< timer.realtime_msec() << " ms "
		<< "(" << timer.cputime_msec() << "ms CPU)\n";

	// Print memory allocation statistics:
	cout << "\n";
	main_mem_statistics("Memory Pools:");

	// Ok:
	return true;
}

//-----------------------------------------------------------------------------
// Performance Test of Relation with two nat.num. columns, array index:
//-----------------------------------------------------------------------------

static bool main_rel_n_n_2(int arg)
{
	// Argument is not used for this test:
	arg_unused(arg);

	// Print test headline:
	cout << "Performance Test of Relation with two NatNum columns, " <<
		"Array Index over 1st\n";

	// Test size:
	const int num_c1_values = 1000; //10000;
	const int num_c2_values = 10;
	cout << "Inserting " << num_c1_values * num_c2_values <<
		" tuples with " << num_c1_values <<
		" different c1 values ...\n";

	// Measure time for storing values:
	perf_c timer;
	timer.start();

	// Make Relation:
	rel_n_n_c rel2("rel2");

	// Insert tuples:
	for(int c1 = 0; c1 < num_c1_values; c1++) {
		for(int c2 = 0; c2 < num_c2_values; c2++) {
			//cout << "[" << c1 << ", " << c2 << "]\n";
			if(!rel2.insert(c1, c2)) {
				cerr << "Insert failed for c1=" << c1 <<
					", c2=" << c2 << ".\n";
				return false;
			}
		}
	}

	// Print result:
	timer.stop();
	cout << "Time for inserting " << num_c1_values * num_c2_values <<
		" tuples " <<
		timer.realtime_msec() << " ms " <<
		"(" << timer.cputime_msec() << "ms CPU)\n";

	// Print memory allocation statistics:
	cout << "\n";
	main_mem_statistics("Memory Pools:");

	// Ok:
	return true;
}


//-----------------------------------------------------------------------------
// Test of List Class (Relation with ff binding pattern):
//-----------------------------------------------------------------------------

static bool main_list(int arg)
{
	// Test parameters:
	int num_tuples = 1;
	int lookup_cycles = 1;
	bool show_dump;
	switch(arg) {
		case 1:
			num_tuples = 1000;
			lookup_cycles = 1;
			show_dump = true;
			break;
		case 2:
			num_tuples = 1000000; // One Million
			lookup_cycles = 10;
			show_dump = true;
			break;

		default:
			cerr << "Invalid argument to list test\n";
			return false;
	}

	// Print test headline:
	cout << "Performance Test of Relation with two integer columns, " <<
		"List (Binding ff)\n";

	// Print headline:
	cout << "Inserting " << num_tuples << " tuples with two columns ...\n";

	// Measure time for storing values:
	perf_c timer;
	timer.start();

	// Make Relation:
	list_c<row_2_c> list("list");

	// Insert tuples:
	for(int i = 0; i < num_tuples; i++) {
		int c1 = i;
		int c2 = 1000 * i;
		bool ok = list.insert(row_2_c(c1, c2));
		if(!ok) {
			if(i == 0)
				cerr << "Insert failed for first tuple.\n";
			else {
				cerr << "Insert failed for tuple " << i+1 <<
						": " << "[c1=" << c1 <<
						", c2=" << c2 << "].\n";
			}
			MAIN_ERR_DUMP(list);
			return false;
		}
	}

	// Print result:
	timer.stop();
	cout << "Time for inserting " << num_tuples << " tuples " <<
		timer.realtime_msec() << " ms " <<
		"(" << timer.cputime_msec() << "ms CPU)\n";

	// Show data structure:
	MAIN_OPT_DUMP(show_dump, list);

	// Query values in list:
	if(lookup_cycles > 0) {
		// Start timer again:
		timer.start();

		// Print headline:
		cout << "Reading values in list ";
		if(lookup_cycles != 1)
			cout << lookup_cycles << " times ";
		cout << "...\n";

		// Create cursor object:
		cur_list_c<row_2_c> cursor(&list);

		// Do given number of lookup cycles (full table scans):
		for(int j = 0; j < lookup_cycles; j++) {
			cursor.open();
			int i = 0;
			while(cursor.fetch()) {
				// cout << "Tuple " << i+1 << " ...";
				int c1 = cursor.row()->c1();
				int c2 = cursor.row()->c2();
				// cout << "done.\n";

				// Check result tuple:
				if(c1 != i) {
					cerr << "Wrong value for c1 in row " <<
						i+1 << ": " << c1 <<
						" (expected: " << i << ")\n";
					MAIN_ERR_DUMP(cursor);
					return false;
				}
				if(c2 != i * 1000) {
					cerr << "Wrong value for c2 in row " <<
						i+1 << ": " << c2 <<
						" (expected: " << i*1000 <<
						")\n";
					return false;
				}
				i++;
			}

			// Check number of tuples returned:
			if(i != num_tuples) {
				cerr << "Found " << i << " tuples " <<
					"instead of " << num_tuples << ".\n";
			}
		}

		// Print time for reading relation from list:
		timer.stop();
		cout << "Time for " << lookup_cycles <<
			" full table scans through " << num_tuples <<
			" tuples " <<
			timer.realtime_msec() << " ms " <<
			"(" << timer.cputime_msec() << "ms CPU)\n";
	}

	
	// Print memory allocation statistics:
	cout << "\n";
	main_mem_statistics("Memory Pools:");

	// Ok:
	return true;
}

//-----------------------------------------------------------------------------
// Test of Set Class (Relation with bb binding pattern):
//-----------------------------------------------------------------------------

static bool main_set(int arg)
{
	// Test parameters:
	int num_rows = 1;
	int c1_start = 1;
	int c2_start = 1;
	int c1_step = 1;
	int c2_step = 1;
	int lookup_cycles = 1;
	bool print_statistics = true;
	bool show_dump = false;
	bool show_mem = false;

	// Select test setup:
	switch(arg) {
		case 1:
			num_rows = 100;
			show_dump = true;
			break;
		case 2:
			num_rows = 1000000; // One Million
			break;

		default:
			cerr << "Invalid argument to set test\n";
			return false;
	}

	// Print test headline:
	cout << "Performance Test of Relation with two integer columns, " <<
		"Set (Binding bb)\n";

	// Print headline:
	cout << "Inserting " << num_rows << " rows with two columns ...\n";

	// Measure time for storing values:
	perf_c timer;
	timer.start();

	// Make Relation:
	set_2_c set("test_set");

	// Insert tuples:
	int c1 = c1_start;
	int c2 = c2_start;
	for(int i = 0; i < num_rows; i++) {
		bool ok = set.insert(c1, c2);
		if(!ok) {
			if(i == 0)
				cerr << "Insert failed for first tuple.\n";
			else {
				cerr << "Insert failed for tuple " << i+1 <<
						": " << "[c1=" << c1 <<
						", c2=" << c2 << "].\n";
			}
			return false;
		}
		c1 += c1_step;
		c2 += c2_step;
	}

	// Check memory allocation error:
	if(set.mem_err()) {
		cerr << "Memory allocation error.\n";
		return false;
	}

	// Check number of tuples:
	int count = set.num_rows();
	if(count != num_rows) {
		cerr << "Wrong number of rows: " << count <<
				" (expected: " << num_rows << ").\n";
		return false;
	}

	// Print result:
	timer.stop();
	cout << "Time for inserting " << num_rows << " rows: " <<
		timer.realtime_msec() << " ms " <<
		"(" << timer.cputime_msec() << "ms CPU)\n";

	// Show data structure:
#ifdef VER_DEBUG
	if(show_dump)
		set.dump();
#else
	if(show_dump)
		cout << "Data structure dump " <<
			"is available only in debug mode\n";
#endif

	// Query values in set:
	if(lookup_cycles > 0) {
		// Start timer again:
		timer.start();

		// Print headline:
		cout << "Checking inserted values in set ";
		if(lookup_cycles != 1)
			cout << lookup_cycles << " times ";
		cout << "...\n";

		// Do given number of lookup cycles:
		for(int j = 0; j < lookup_cycles; j++) {
			c1 = c1_start;
			c2 = c2_start;
			for(int i = 0; i < num_rows; i++) {
				if(set.insert(c1, c2)) {
					cerr << "Row " << i+1 << " (" <<
						c1 << ", " << c2 << ") " <<
						"was not found.\n";
					return false;
				}
				c1 += c1_step;
				c2 += c2_step;
			}
		}

		// Check memory allocation error:
		if(set.mem_err()) {
			cerr << "Memory allocation error after lookups.\n";
			return false;
		}

		// Check number of tuples again:
		count = set.num_rows();
		if(count != num_rows) {
			cerr << "Wrong number of rows: " << count <<
					" (expected: " << num_rows << ").\n";
			return false;
		}

		// Print time for reading relation from set:
		timer.stop();
		cout << "Time for ";
		if(lookup_cycles == 1)
			cout << "lookup";
		else
			cout << lookup_cycles << "lookups";
		cout << " of " << num_rows << " rows: " <<
			timer.realtime_msec() << " ms " <<
			"(" << timer.cputime_msec() << "ms CPU)\n";
	}

	// Print hash table statistics:
	if(print_statistics) {
		int hash_size    = set.hash_size();
		int entries_used = set.entries_used();
		int num_pages    = set.num_pages();
		cout << "Hash table size:   " << hash_size << "\n";
		cout << "Used entries:      " << entries_used << " (" <<
			IDIV_PERCENT(entries_used, hash_size) << "%)\n";
		cout << "Avg. chain length: " <<
			IDIV_WHOLE(num_rows, entries_used) << "." <<
			IDIV_TENTH(num_rows, entries_used) << "\n";
		cout << "Used Memory Pages: " << num_pages <<
			" (" << PAGE_KB(num_pages) << ")\n";

	}

	// Print memory allocation statistics:
	if(show_mem) {
		cout << "\n";
		main_mem_statistics("Memory Pools:");
	}

	// Ok:
	return true;
}


//-----------------------------------------------------------------------------
// Test of Set Class (Relation with bb binding pattern) - Extendible Hashing:
//-----------------------------------------------------------------------------

static bool main_set_e(int arg)
{
	// Test parameters:
	int num_rows = 1;
	int c1_start = 1;
	int c2_start = 1;
	int c1_step = 1;
	int c2_step = 1;
	int lookup_cycles = 1;
	bool print_statistics = true;
	bool show_dump = false;
	bool show_mem = false;

	// Select test setup:
	switch(arg) {
		case 1:
			num_rows = 100;
			show_dump = true;
			break;
		case 2:
			num_rows = 1000000; // One Million
			break;

		default:
			cerr << "Invalid argument to set test\n";
			return false;
	}

	// Print test headline:
	cout << "Performance Test of Relation with two integer columns, " <<
		"Set (Binding bb) - E\n";

	// Print headline:
	cout << "Inserting " << num_rows << " rows with two columns ...\n";

	// Measure time for storing values:
	perf_c timer;
	timer.start();

	// Make Relation:
	set_2e_c set("test_set");

	// Insert tuples:
	int c1 = c1_start;
	int c2 = c2_start;
	for(int i = 0; i < num_rows; i++) {
		bool ok = set.insert(c1, c2);
		if(!ok) {
			if(i == 0)
				cerr << "Insert failed for first tuple.\n";
			else {
				cerr << "Insert failed for tuple " << i+1 <<
						": " << "[c1=" << c1 <<
						", c2=" << c2 << "].\n";
			}
			return false;
		}
		c1 += c1_step;
		c2 += c2_step;
	}

	// Check memory allocation error:
	if(set.mem_err()) {
		cerr << "Memory allocation error.\n";
		return false;
	}

	// Check number of tuples:
	int count = set.num_rows();
	if(count != num_rows) {
		cerr << "Wrong number of rows: " << count <<
				" (expected: " << num_rows << ").\n";
		return false;
	}

	// Print result:
	timer.stop();
	cout << "Time for inserting " << num_rows << " rows: " <<
		timer.realtime_msec() << " ms " <<
		"(" << timer.cputime_msec() << "ms CPU)\n";

	// Show data structure:
#ifdef VER_DUMP
	if(show_dump)
		set.dump();
#else
	if(show_dump)
		cout << "Data structure dump " <<
			"is available only in debug mode\n";
#endif

	// Query values in set:
	if(lookup_cycles > 0) {
		// Start timer again:
		timer.start();

		// Print headline:
		cout << "Checking inserted values in set ";
		if(lookup_cycles != 1)
			cout << lookup_cycles << " times ";
		cout << "...\n";

		// Do given number of lookup cycles:
		for(int j = 0; j < lookup_cycles; j++) {
			c1 = c1_start;
			c2 = c2_start;
			for(int i = 0; i < num_rows; i++) {
				if(set.insert(c1, c2)) {
					cerr << "Row " << i+1 << " (" <<
						c1 << ", " << c2 << ") " <<
						"was not found.\n";
					return false;
				}
				c1 += c1_step;
				c2 += c2_step;
			}
		}

		// Check memory allocation error:
		if(set.mem_err()) {
			cerr << "Memory allocation error after lookups.\n";
			return false;
		}

		// Check number of tuples again:
		count = set.num_rows();
		if(count != num_rows) {
			cerr << "Wrong number of rows: " << count <<
					" (expected: " << num_rows << ").\n";
			return false;
		}

		// Print time for reading relation from set:
		timer.stop();
		cout << "Time for ";
		if(lookup_cycles == 1)
			cout << "lookup";
		else
			cout << lookup_cycles << "lookups";
		cout << " of " << num_rows << " rows: " <<
			timer.realtime_msec() << " ms " <<
			"(" << timer.cputime_msec() << "ms CPU)\n";
	}

	// Print hash table statistics:
	if(print_statistics) {
		int hash_size    = set.hash_size();
		int num_pages    = set.num_pages();
		int buckets      = set.num_buckets();
		int overflow     = set.num_overflow();
		buckets -= overflow;
		cout << "Hash table size:   " << hash_size << "\n";
		cout << "Normal buckets:    " << buckets << " (" <<
			IDIV_PERCENT(buckets, hash_size) << "%)\n";
		cout << "Overflow buckets:  " <<
			overflow << "\n";
		cout << "Used Memory Pages: " << num_pages <<
			" (" << PAGE_KB(num_pages) << " KB)\n";

	}

	// Print memory allocation statistics:
	if(show_mem) {
		cout << "\n";
		main_mem_statistics("Memory Pools:");
	}

	// Ok:
	return true;
}

//-----------------------------------------------------------------------------
// Test of old dup_2i Class (Relation with bb binding pattern):
//-----------------------------------------------------------------------------

static bool main_dup(int arg)
{
	// Test parameters:
	int num_rows = 1;
	int c1_start = 1;
	int c2_start = 1;
	int c1_step = 1;
	int c2_step = 1;
	int lookup_cycles = 1;
	int hashtab_size = 262147;
	// int hashtab_size = 854041;
	bool print_statistics = true;
	bool show_dump = false;
	bool show_mem = false;

	// Select test setup:
	switch(arg) {
		case 1:
			num_rows = 100;
			show_dump = true;
			break;

		case 2:
			num_rows = 1000000; // One Million
			break;

		case 3:
			num_rows = 1000000; // One Million
			hashtab_size = 524309;
			break;

		default:
			cerr << "Invalid argument to dup_2i test\n";
			return false;
	}

	// Print test headline:
	cout << "Performance Test of Relation with two integer columns, " <<
		"Dup_2i (Binding bb)\n";

	// Print headline:
	cout << "Inserting " << num_rows << " rows with two columns ...\n";

	// Measure time for storing values:
	perf_c timer;
	timer.start();

	// Make Relation:
	dup_2i_c dup("test_dup_2i", hashtab_size);
	if(!dup.init_ok()) {
		cerr << "Initialization failed.\n";
		return false;
	}

	// Insert tuples:
	int c1 = c1_start;
	int c2 = c2_start;
	for(int i = 0; i < num_rows; i++) {
		int result = dup.dup_check(c1, c2);
		if(result != 0) {
			if(i == 0)
				cerr << "Insert failed for first tuple.\n";
			else {
				cerr << "Insert failed for tuple " << i+1 <<
						": " << "[c1=" << c1 <<
						", c2=" << c2 << "].\n";
			}
			return false;
		}
		c1 += c1_step;
		c2 += c2_step;
	}

	// Check number of tuples:
	int count = dup.num_rows();
	if(count != num_rows) {
		cerr << "Wrong number of rows: " << count <<
				" (expected: " << num_rows << ").\n";
		return false;
	}

	// Print result:
	timer.stop();
	cout << "Time for inserting " << num_rows << " rows: " <<
		timer.realtime_msec() << " ms " <<
		"(" << timer.cputime_msec() << "ms CPU)\n";

	// Show data structure:
#ifdef VER_DEBUG
	if(show_dump)
		dup.dump();
#else
	if(show_dump)
		cout << "Data structure dump " <<
			"is available only in debug mode\n";
#endif

	// Query values in set:
	if(lookup_cycles > 0) {
		// Start timer again:
		timer.start();

		// Print headline:
		cout << "Checking inserted values in set ";
		if(lookup_cycles != 1)
			cout << lookup_cycles << " times ";
		cout << "...\n";

		// Do given number of lookup cycles:
		for(int j = 0; j < lookup_cycles; j++) {
			c1 = c1_start;
			c2 = c2_start;
			for(int i = 0; i < num_rows; i++) {
				int result = dup.dup_check(c1, c2);
				if(result != 1) {
					cerr << "Row " << i+1 << " (" <<
						c1 << ", " << c2 << ") " <<
						"was not found.\n";
					return false;
				}
				c1 += c1_step;
				c2 += c2_step;
			}
		}

		// Check number of tuples again:
		count = dup.num_rows();
		if(count != num_rows) {
			cerr << "Wrong number of rows: " << count <<
					" (expected: " << num_rows << ").\n";
			return false;
		}

		// Print time for reading relation from set:
		timer.stop();
		cout << "Time for ";
		if(lookup_cycles == 1)
			cout << "lookup";
		else
			cout << lookup_cycles << "lookups";
		cout << " of " << num_rows << " rows: " <<
			timer.realtime_msec() << " ms " <<
			"(" << timer.cputime_msec() << "ms CPU)\n";
	}

	// Print hash table statistics:
	if(print_statistics) {
		int hash_size    = dup.hash_size();
		int entries_used = dup.entries_used();
		int num_pages    = dup.num_pages();
		cout << "Hash table size:   " << hash_size << "\n";
		cout << "Used entries:      " << entries_used << " (" <<
			IDIV_PERCENT(entries_used, hash_size) << "%)\n";
		cout << "Avg. chain length: " <<
			IDIV_WHOLE(num_rows, entries_used) << "." <<
			IDIV_TENTH(num_rows, entries_used) << "\n";
		cout << "Used Memory Pages: " << num_pages <<
			" (" << PAGE_KB(num_pages) << ")\n";

	}

	// Print memory allocation statistics:
	if(show_mem) {
		cout << "\n";
		main_mem_statistics("Memory Pools:");
	}

	// Ok:
	return true;
}

//=============================================================================
// List of all tests:
//=============================================================================


main_test_s MainTests[] = {
	main_test_s("FileRead", main_std_fileread, 0),
	main_test_s("MemAlloc", main_mem_alloc, 100000),
	main_test_s("Lex", main_lex, 0),

	main_test_s("AtomTab", main_atomtab, 0),
	main_test_s("AtomTab1", main_atomtab, 1),
	main_test_s("AtomTab2", main_atomtab, 2),
	main_test_s("StrTab1", main_strtab_1, 0),
	main_test_s("StrTab2", main_strtab_2, 100000),
	main_test_s("RelMapNN1", main_rel_n_n_1, 0),
	main_test_s("RelMapNN2", main_rel_n_n_2, 0),
	main_test_s("list1", main_list, 1),
	main_test_s("list2", main_list, 2),
	main_test_s("set1", main_set, 1),
	main_test_s("set2", main_set, 2),
	main_test_s("set_e1", main_set_e, 1),
	main_test_s("set_e2", main_set_e, 2),
	main_test_s("dup1", main_dup, 1),
	main_test_s("dup2", main_dup, 2),
	main_test_s("dup3", main_dup, 3),
};

static int main_num_tests() { return sizeof(MainTests)/sizeof(main_test_s); }
static main_test_t main_tests() { return MainTests; }

