// ============================================================================
// Project:	Deductive Database
// Filename:	mem.h
// Purpose:	Big pieces of dynamic memory, from which pages are taken
// Last Change:	08.09.2015
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2011-2015 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"
#include "err.h"
#include "check.h"

// for std::malloc
#include <cstdlib>

#ifdef VER_DEBUG
// for std::memset(ptr, byte_value, size):
#include <cstring>
#endif

#include "mem.h"

//=============================================================================
// Global Variables of this Module:
//=============================================================================


//-----------------------------------------------------------------------------
// FirstChunk: Start of linked list of all allocated memory chunks.
//-----------------------------------------------------------------------------

mem_c *mem_c::FirstChunk = MEM_NULL;

//-----------------------------------------------------------------------------
// LastChunk: Last/Current memory chunk.
//-----------------------------------------------------------------------------

mem_c *mem_c::LastChunk = MEM_NULL;

//-----------------------------------------------------------------------------
// NumChunks: Total number of allocated memory chunks.
//-----------------------------------------------------------------------------

int mem_c::NumChunks = 0;

//-----------------------------------------------------------------------------
// Linked list of free memory pages:
//-----------------------------------------------------------------------------

page_t mem_c::FreePageList = PAGE_NULL;

//-----------------------------------------------------------------------------
// NumAllocPages: Total number of allocated memory pages.
//-----------------------------------------------------------------------------

int mem_c::NumAllocPages = 0;


//=============================================================================
// Auxiliary Methods:
//=============================================================================


//-----------------------------------------------------------------------------
// init: Similar to constructor.
//-----------------------------------------------------------------------------

void mem_c::init()
{
	// Make address of memory pages aligned to page boundary:
	char *ptr = Memory;
	int offset = reinterpret_cast<VER_INT64>(ptr) % VER_PAGESIZE;

	// Pointer to first memory page:
	FirstPage = ptr;
	if(offset > 0)
		FirstPage += (VER_PAGESIZE - offset);

	// Pointer to next memory page:
	NextPage = FirstPage;

	// Number of free memory pages.
	NumFreePages = MEM_CHUNK_PAGES;

	// Linked list of all allocated memory chunks:
	NextChunk = 0;

	// Set Magic number:
	CHECK_CODE(Magic = MEM_MAGIC);
}


//=============================================================================
// Exported Class Methods:
//=============================================================================


//-----------------------------------------------------------------------------
// alloc_page: Get a new memory page.
//-----------------------------------------------------------------------------

void *mem_c::alloc_page()
{
	page_t page;

	// If there are free pages, use one of these:
	if(FreePageList != PAGE_NULL) {
		// Get page from freelist:
		page = FreePageList;
		FreePageList = *(reinterpret_cast<page_t *>(page));

		// Wipe out freelist next link in debug mode:
		CHECK_CODE(std::memset(page, 0xFF, sizeof(page_t)));
			// Rest was wiped out when page was put on free list
			// (we could check whether that has changed ...)

		// For statistics:
		NumAllocPages++;

		// Return page:
		return page;
	}

	// Allocate first memory chunk:
	if(FirstChunk == 0) {
		CHECK(LastChunk == 0,
				"mem_c::alloc_page: LastChunk not null");
		FirstChunk = (mem_c*) std::malloc(sizeof(mem_c));
		if(FirstChunk == 0) {
			err_c::malloc_failed(sizeof(mem_c), 0);
			return PAGE_NULL;
		}
		FirstChunk->init();
		LastChunk = FirstChunk;
		NumChunks = 1;
	}

	// Allocate another memory chunk if necessary:
	else if(LastChunk->NumFreePages == 0) {
		CHECK(LastChunk != 0, "mem_c::alloc_page: LastChunk null");
		mem_c *new_chunk = (mem_c*) std::malloc(sizeof(mem_c));
		if(new_chunk == 0) {
			err_c::malloc_failed(sizeof(mem_c), NumChunks);
			return PAGE_NULL;
		}
		new_chunk->init();
		LastChunk->NextChunk = new_chunk;
		LastChunk = new_chunk;
		NumChunks++;
	}

	// Cut memory page from current chunk:
	LastChunk->NumFreePages--;
	page = LastChunk->NextPage;
	LastChunk->NextPage += VER_PAGESIZE;
	NumAllocPages++;

	// Wipe out page in debug mode:
	CHECK_CODE(std::memset(page, 0xFF, VER_PAGESIZE));

	// Return memory page:
	return page;
}

//-----------------------------------------------------------------------------
// free_page: Deallocate a memory page (enter into free list).
//-----------------------------------------------------------------------------

void mem_c::free_page(page_t page)
{
	// Wipe out page in debug mode:
	CHECK_CODE(std::memset(page, 0xFF, VER_PAGESIZE));

	// Enter page into list of free pages:
	*(reinterpret_cast<page_t *>(page)) = FreePageList;
	FreePageList = page;

	// For statistics:
	NumAllocPages--;
}


//-----------------------------------------------------------------------------
// Debugging Support:
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
str_t	mem_c::check() const
{
	// Check magic number:
	if(Magic != MEM_MAGIC)
		return "wrong magic number";

	// Check global variables of this module:
	if(NumChunks < 0)
		return "NumChunks is negative";
	if(NumAllocPages < 0)
		return "NumAllocPages is negative";

	// Ok:
	return 0;
}

#endif

