// ============================================================================
// Project:	Deductive Database
// Filename:	mpool.cpp
// Purpose:	Pool of memory pages (allocation, deallocation, statistics).
// Last Change:	08.09.2015
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2011-2015 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"
#include "str.h"
#ifdef	VER_DUMP
#include "err.h"
#endif
#include "check.h"
#include "min.h"
#include "page.h"
#include "mem.h"

#include "mpool.h"

//=============================================================================
// Private Constants:
//=============================================================================

//-----------------------------------------------------------------------------
// MPOOL_PAGEDIR_NULL: Null pointer to array of memory page addresses.
//-----------------------------------------------------------------------------

#define MPOOL_PAGEDIR_NULL	(static_cast<page_t*>(0))


//=============================================================================
// Private Class Members (Static Attributes):
//=============================================================================

//-----------------------------------------------------------------------------
// FirstPool: Linked list of all memory pools (for printing statistics).
//-----------------------------------------------------------------------------

mpool_c *mpool_c::FirstPool = 0;

//-----------------------------------------------------------------------------
// LastPool: Linked list of all memory pools (for printing statistics).
//-----------------------------------------------------------------------------

mpool_c *mpool_c::LastPool = 0;


//=============================================================================
// Constructor and Destructor:
//=============================================================================

//-----------------------------------------------------------------------------
// Constructor:
//-----------------------------------------------------------------------------

mpool_c::mpool_c(str_t pool_name)
{
	// Check parameter:
	CHECK(pool_name != STR_NULL,
		"mpool_c::constructor: pool_name is null");

	// Initialize attributes:
	Name = pool_name;
	NumAllocPages = 0;
	NextPool = 0;
	PrevPool = 0;
	FirstDirPage = MPOOL_PAGEDIR_NULL;
	CurrDirPage  = MPOOL_PAGEDIR_NULL;
	CurrDirFree  = 0;
	CurrDirPtr   = MPOOL_PAGEDIR_NULL;

	// Not really necessary, but make everything defined:
	for(int i = 0; i < MPOOL_PAGEARRAY_SIZE; i++)
		PageDirArr[i] = PAGE_NULL;

	// Enter this object into doubly linked list of all memory pools:
	if(LastPool != 0) {
		// This is not the first pool:
		CHECK(FirstPool != 0,
			"mpool_c::constructor: LastPool!=0, FirstPool==0");
		CHECK(LastPool->NextPool == 0,
			"mpool_c::constructor: LastPool->NextPool != 0");
		PrevPool = LastPool;
		LastPool->NextPool = this;
		LastPool = this;
	}
	else {
		// This is the first pool:
		FirstPool = this;
		LastPool = this;
	}

	// Set Magic number:
	CHECK_CODE(Magic = MPOOL_MAGIC);
}

//-----------------------------------------------------------------------------
// Destructor:
//-----------------------------------------------------------------------------

mpool_c::~mpool_c()
{
	// Check this object:
	CHECK_VALID("mpool_c::destructor");

	// Deallocate all memory pages. First those in the local array:
	int n = MIN(NumAllocPages, MPOOL_PAGEARRAY_SIZE);
	NumAllocPages -= n;
	page_t *page = PageDirArr;
	while(n-- > 0)
		mem_c::free_page(*page++);

	// Now deallocate pages from the linked list of directory pages:
	page_t *next;
	for(page_t *dir = FirstDirPage; dir != MPOOL_PAGEDIR_NULL; dir = next){

		// Compute number of page addresses on this page:
		n = MIN(NumAllocPages, PAGE_SIZE(page_t) - 1);
		NumAllocPages -= n;

		// Deallocate pages referenced on this directory page:
		page = dir + 1; // Skip link to next page in linked list
		while(n-- > 0) {
			CHECK(*page != PAGE_NULL,
				"mpool_c::destructor: page link is null");
			mem_c::free_page(*page++);
		}

		// Get link to next directory page (before deallocation):
		next = reinterpret_cast<page_t*>(dir[0]);

		// Deallocate directory page:
		mem_c::free_page(dir);
	}

	// Remove this memory pool from the list of all memory pools:
	if(PrevPool == 0) {
		// This is the first pool:
		CHECK(FirstPool == this,
			"mpool_c::destructor: Inconsistent FirstPool");
		FirstPool = NextPool;
	}
	else
		PrevPool->NextPool = NextPool;

	if(NextPool == 0) {
		// This is the last pool:
		CHECK(LastPool == this,
			"mpool_c::destructor: Inconsistent LastPool");
		LastPool = PrevPool;
	}
	else
		NextPool->PrevPool = PrevPool;


	// Make this object invalid:
	CHECK_CODE(Magic = 0);
}


//=============================================================================
// Object Methods:
//=============================================================================


//-----------------------------------------------------------------------------
// alloc_page: Allocate a memory page.
//-----------------------------------------------------------------------------

page_t mpool_c::alloc_page()
{
	// Check this object:
	CHECK_VALID("mpool_c::alloc_page");

	// Alloc page (from mem_c, which gets large chunks from OS):
	page_t page = mem_c::alloc_page();
	if(page == PAGE_NULL)
		return PAGE_NULL;

	// Enter page into list of allocated pages:
	if(NumAllocPages < MPOOL_PAGEARRAY_SIZE)
		PageDirArr[NumAllocPages] = page;
	else if(CurrDirFree-- > 0)
		*CurrDirPtr++ = page;
	else {
		// Must get a new directory page.
		// It looks nicer if we use the just allocated page as
		// directory page, so that all allocated pages listed in
		// the directory are consecutive (if from the same chunk).
		page_t dir_page = page;
		page = mem_c::alloc_page();
		if(page == PAGE_NULL) {
			mem_c::free_page(dir_page);
			return PAGE_NULL;
		}
		if(FirstDirPage != MPOOL_PAGEDIR_NULL)
			CurrDirPage[0] = dir_page;
		else
			FirstDirPage = reinterpret_cast<page_t*>(dir_page);
		CurrDirPage   = reinterpret_cast<page_t*>(dir_page);
		CurrDirPtr    = CurrDirPage;
		*CurrDirPtr++ = PAGE_NULL;	// Link to next directory page
		*CurrDirPtr++ = page;		// First entry in this dir page
		CurrDirFree   = PAGE_SIZE(page_t) - 2;
	}

	// Increase number of allocated pages of this pool:
	NumAllocPages++;

	// Return allocated page:
	return page;
}


//=============================================================================
// Debugging Support:
//=============================================================================

//-----------------------------------------------------------------------------
// check: Check the integrity of this object, return error message or NULL.
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
str_t	mpool_c::check() const
{
	// Check magic number:
	if(Magic != MPOOL_MAGIC)
		return "wrong magic number";

	// FirstPool and LastPool cannot be null, because this object exists:
	if(FirstPool == 0)
		return "FirstPool == 0, but mpool_c object exists";
	if(LastPool == 0)
		return "LastPool == 0, but mpool_c object exists";

	// This pool must have a name:
	if(Name == 0)
		return "Pool name is null";

	// Check current number of pages allocated:
	if(NumAllocPages < 0)
		return "NumAllocPages cannot be negative";

	// PrevPool is null if and only if this is the first pool:
	if(PrevPool == 0 && FirstPool != this)
		return "PrevPool is null, but this is not the first pool";
	if(PrevPool != 0 && FirstPool == this)
		return "PrevPool is not null, but this is the first pool";

	// NextPool is null if and only if this is the last pool:
	if(NextPool == 0 && LastPool != this)
		return "NextPool is null, but this is not the last pool";
	if(NextPool != 0 && LastPool == this)
		return "NextPool is not null, but this is the last pool";

	// Finally check whether we find this pool on the linked list:
	mpool_t pool = FirstPool;
	while(pool != 0 && pool != this)
		pool = pool->NextPool;
	if(pool != this)
		return "Pool is not on the linked list of all memory pools";

	// Page directory is used only if NumAllocPages > MPOOL_PAGEARRAY_SIZE:
	if(NumAllocPages <= MPOOL_PAGEARRAY_SIZE) {
		if(FirstDirPage != MPOOL_PAGEDIR_NULL)
			return "FirstDirPage not null for small pool";
		if(CurrDirPage != MPOOL_PAGEDIR_NULL)
			return "CurrDirPage not null for small pool";
		if(CurrDirFree != 0)
			return "CurrDirFree not 0 for small pool";
		if(CurrDirPtr != MPOOL_PAGEDIR_NULL)
			return "CurrDirPtr not null for small pool";
	}
	else {
		if(FirstDirPage == MPOOL_PAGEDIR_NULL)
			return "FirstDirPage null for large pool";
		if(CurrDirPage == MPOOL_PAGEDIR_NULL)
			return "CurrDirPage null for large pool";
		if(CurrDirPtr == MPOOL_PAGEDIR_NULL)
			return "CurrDirPtr null for large pool";
		if(CurrDirPtr != CurrDirPage + (PAGE_SIZE(page_t)-CurrDirFree))
			return "CurrDirPtr inconsistent with CurrDirFree";
		if(CurrDirFree < 0)
			return "CurrDirFree is negative";
		if(CurrDirFree > PAGE_SIZE(page_t) - 2)
			return "CurrDirFree is too large (outside bounds)";
	}

	// Ok:
	return 0;
}
#endif

//-----------------------------------------------------------------------------
// dump: Show data structure.
//-----------------------------------------------------------------------------

#ifdef	VER_DUMP

void mpool_c::dump(str_t headline) const
{
	// Headline:
	if(headline == STR_NULL)
		headline = "Memory Pool";
	err_c::dump_begin(headline);

	// Basic information:
	err_c::dump_str("Name:         '");
	err_c::dump_str(Name);
	err_c::dump_str("'");
	err_c::dump_nl();
	err_c::dump_str("NumAllocPages: ");
	err_c::dump_int(NumAllocPages);
	err_c::dump_nl();
	err_c::dump_str("Page Size:     ");
	err_c::dump_int(VER_PAGESIZE);
	err_c::dump_str(" Bytes [");
	err_c::dump_int(PAGE_SIZE(page_t));
	err_c::dump_str(" page_t Pointers of ");
	err_c::dump_int(static_cast<int>(sizeof(page_t)));
	err_c::dump_str(" Bytes each]");
	err_c::dump_nl();
	err_c::dump_str("CurrDirPage:   ");
	err_c::dump_ptr(CurrDirPage);
	err_c::dump_nl();
	err_c::dump_str("CurrDirPtr:    ");
	err_c::dump_ptr(CurrDirPtr);
	err_c::dump_str(" [corresponds to Index: ");
	int ptr_i = static_cast<int>((reinterpret_cast<char*>(CurrDirPtr) -
					reinterpret_cast<char*>(CurrDirPage)) /
				sizeof(page_t));
	err_c::dump_int(ptr_i);
	err_c::dump_str("]");
	err_c::dump_nl();
	err_c::dump_str("CurrDirFree:   ");
	err_c::dump_int(CurrDirFree);
	err_c::dump_str(" [Sum with CurrDirPtr Index: ");
	err_c::dump_int(CurrDirFree + ptr_i);
	err_c::dump_str("]");
	err_c::dump_nl();
	err_c::dump_str("Object Size:   ");
	err_c::dump_uint(static_cast<unsigned int>(sizeof(mpool_c)));
	err_c::dump_str(" Bytes");
	err_c::dump_nl();
	err_c::dump_nl();

	// Page Directory Array:
	err_c::dump_str("Internal Page Directory Array:");
	err_c::dump_nl();
	int num_pages = NumAllocPages;
	int n = MIN(num_pages, MPOOL_PAGEARRAY_SIZE);
	num_pages -= n;
	for(int i = 0; i < n; i++) {
		err_c::dump_str("  ");
		err_c::dump_int(i);
		err_c::dump_str(". ");
		err_c::dump_ptr(PageDirArr[i]);
		err_c::dump_nl();
	}
	err_c::dump_nl();

	// Linked list of page directory entries:
	int page_no = 1;
	for(page_t *dir = FirstDirPage; dir != MPOOL_PAGEDIR_NULL;
				dir = reinterpret_cast<page_t*>(dir[0])) {

		// Compute number of page addresses on this page:
		n = MIN(num_pages, PAGE_SIZE(page_t) - 1);
		num_pages -= n;

		// Headline:
		err_c::dump_str("Page ");
		err_c::dump_int(page_no++);
		err_c::dump_str(" [");
		err_c::dump_int(n);
		err_c::dump_str(" page pointers, Address ");
		err_c::dump_ptr(dir);
		err_c::dump_str("]:");
		err_c::dump_nl();

		// List all pages on this directory page:
		for(int j = 0; j <= n; j++) {
			err_c::dump_str("  ");
			err_c::dump_int(j);
			err_c::dump_str(". ");
			err_c::dump_ptr(dir[j]);
			if(j == 0)
				err_c::dump_str("  [Link to next page]");
			err_c::dump_nl();
		}
	}
	err_c::dump_nl();

	// Done:
	err_c::dump_end();
}

#endif
