// ============================================================================
// Project:	Deductive Database
// Filename:	out.cpp
// Purpose:	Class for Output of Tests/Benchmarks
// Last Change:	24.05.2016
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2016 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"
#include "str.h"
#include "check.h"
#include "page.h"
#include "mpool.h"

#include "out.h"

#include <iostream>

//=============================================================================
// Local Constants for Field Widths:
//=============================================================================

static const int OUT_NAME_WIDTH    = 22;
static const int OUT_ROWS_WIDTH    = 7;
static const int OUT_ATOMS_WIDTH   = 7;
static const int OUT_STRINGS_WIDTH = 7;
static const int OUT_PAGES_WIDTH   = 7;
static const int OUT_KB_WIDTH      = 7;

//=============================================================================
// Static Methods for Printing the Output:
//=============================================================================


//-----------------------------------------------------------------------------
// print_str: Print string.
//-----------------------------------------------------------------------------

void out_c::print_str(str_t s)
{
	std::cout << s;
}

//-----------------------------------------------------------------------------
// print_int: Print integer value.
//-----------------------------------------------------------------------------

void out_c::print_int(int i)
{
	std::cout << i;
}


//-----------------------------------------------------------------------------
// print_long: Print long value.
//-----------------------------------------------------------------------------

void out_c::print_long(long n)
{
	std::cout << n;
}

//-----------------------------------------------------------------------------
// print_nl: Print newline.
//-----------------------------------------------------------------------------

void out_c::print_nl()
{
	std::cout << "\n";
}

//-----------------------------------------------------------------------------
// print_fact_1: Print fact with predicate and one integer argument.
//-----------------------------------------------------------------------------

void out_c::print_fact_1(str_t pred, int arg1)
{
	std::cout << pred << "(" << arg1 << ")\n";
}

//-----------------------------------------------------------------------------
// print_fact_2: Print fact with predicate and two integer arguments.
//-----------------------------------------------------------------------------

void out_c::print_fact_2(str_t pred, int arg1, int arg2)
{
	std::cout << pred << "(" << arg1 << ", " << arg2 << ")\n";
}

//-----------------------------------------------------------------------------
// print_fact_s: Print fact with predicate and one string argument.
//-----------------------------------------------------------------------------

void out_c::print_fact_s(str_t pred, str_t arg1)
{
	std::cout << pred << "(" << arg1 << ")\n";
}


//-----------------------------------------------------------------------------
// print_fact_5s: Print fact with predicate and five string arguments.
//-----------------------------------------------------------------------------

void out_c::print_fact_5s(str_t pred, str_t arg1, str_t arg2,
					str_t arg3, str_t arg4, str_t arg5)
{
	std::cout << pred << "(" << arg1 << ", " << arg2 << ", " <<
				arg3 << ", " << arg4 << ", " << arg5 <<")\n";
}

//-----------------------------------------------------------------------------
// print_answer_limit: Print information that there are more answers.
//-----------------------------------------------------------------------------

void out_c::print_answer_limit()
{
	std::cout << "    ... (There are more answers. " <<
			"Limit of answers to print: " <<
			(AnswerLimit - 1) << ").\n";
}

//-----------------------------------------------------------------------------
// print_answer_no: Print prefix of line showing a computed answer.
//-----------------------------------------------------------------------------

void out_c::print_answer_no(long no)
{
	static const int no_field_width = 3;
	std::cout << "    " <<
		str_fill(no, no_field_width) << no << ": ";
}

//-----------------------------------------------------------------------------
// print_summary: Print summary line.
//-----------------------------------------------------------------------------

void out_c::print_summary(str_t bench_id, long load_time, long eval_time,
						long total_time, str_t result)
{
	std::cout << bench_id;
	if(load_time != -1)
		std::cout << "\t" << load_time;
	if(eval_time != -1)
		std::cout << "\t" << eval_time;
	std::cout << "\t" << total_time << "\t" << result;
	if(Summary_NewLine)
		std::cout << "\n";
	else
		std::cout << "\t";
}

//-----------------------------------------------------------------------------
// print_rel_size: Print some data about a relation.
//-----------------------------------------------------------------------------

void out_c::print_rel_size(str_t name, long num_rows, long num_pages)
{
	// Check parameters:
	CHECK(name != STR_NULL, "out_c::print_rel_size: name is NULL");
	CHECK(num_rows >= 0, "out_c::print_rel_size: num_rows is negative");
	CHECK(num_pages >= 0, "out_c::print_rel_size: num_pages is negative");

	// Print the name of the relation:
	print_str(name);
	print_str(": ");
	print_str(str_indent(OUT_NAME_WIDTH-str_len(name)));

	// Print number of rows:
	print_str(str_fill(num_rows, OUT_ROWS_WIDTH));
	print_long(num_rows);
	print_str(" Rows    ");

	// Print memory in pages:
	print_str(str_fill(num_pages, OUT_PAGES_WIDTH));
	print_int(num_pages);
	print_str(" Pages ");

	// Print memory in kilobytes:
	long kb = PAGE_KB(num_pages);
	print_str(str_fill(kb, OUT_KB_WIDTH));
	print_str("[");
	print_long(kb);
	print_str(" KB]\n");
}

//-----------------------------------------------------------------------------
// print_rel_size_strtab: Print some data about a string table.
//-----------------------------------------------------------------------------

void out_c::print_rel_size_strtab(strtab_t tab)
{
	// Check parameters:
	CHECK_PAR(tab, "out_c::print_strtab_size");

	// Get string table data:
	str_t name      = tab->name();
	int num_strings = tab->num_strings();
	int leaf_nodes  = tab->num_leaf_nodes();
	int small_nodes = tab->num_small_nodes();
	int full_nodes  = tab->num_full_nodes();
	int node_pages  = tab->num_node_pages();
	int str_pages   = tab->num_str_pages();
	int array_pages = tab->num_array_pages();
	int num_pages   = node_pages + str_pages + array_pages;

	// Print the name of the string table:
	print_str(name);
	print_str(": ");
	print_str(str_indent(OUT_NAME_WIDTH-str_len(name)));

	// Print number of distinct strings:
	print_str(str_fill(num_strings, OUT_STRINGS_WIDTH));
	print_long(num_strings);
	print_str(" Strings ");

	// Print memory in pages:
	print_str(str_fill(num_pages, OUT_PAGES_WIDTH));
	print_int(num_pages);
	print_str(" Pages ");

	// Print memory in kilobytes:
	long kb = PAGE_KB(num_pages);
	print_str(str_fill(kb, OUT_KB_WIDTH));
	print_str("[");
	print_long(kb);
	print_str(" KB]\n");

	// In debug mode, we print further details. First, number of nodes:
	out_c::debug_ival("    Number of leaf nodes:        ", leaf_nodes);
	out_c::debug_ival("    Number of small nodes:       ", small_nodes);
	out_c::debug_ival("    Number of full nodes:        ", full_nodes);
	out_c::debug_nl();

	// Print memory sizes for different components of the string table:
	out_c::debug_mempages("    Number of pages for strings: ", str_pages);
	out_c::debug_mempages("    Number of pages for nodes:   ", node_pages);
	out_c::debug_mempages("    Number of pages for array:   ", array_pages);
	out_c::debug_nl();
}

//-----------------------------------------------------------------------------
// print_rel_size_atomtab: Print some data about an atom table.
//-----------------------------------------------------------------------------

void out_c::print_rel_size_atomtab(atomtab_t tab)
{
	// Check parameter:
	CHECK_PAR(tab, "out_c::print_rel_size_atomtab");

	// Get atom table data:
	str_t name      = tab->name();
	int   num_atoms = tab->num_atoms();
	int   num_pages = tab->num_pages();

	// Print the name of the atom table:
	print_str(name);
	print_str(": ");
	print_str(str_indent(OUT_NAME_WIDTH-str_len(name)));

	// Print number of atoms:
	print_str(str_fill(num_atoms, OUT_ATOMS_WIDTH));
	print_long(num_atoms);
	print_str(" Atoms   ");

	// Print memory in pages:
	print_str(str_fill(num_pages, OUT_PAGES_WIDTH));
	print_int(num_pages);
	print_str(" Pages ");

	// Print memory in kilobytes:
	long kb = PAGE_KB(num_pages);
	print_str(str_fill(kb, OUT_KB_WIDTH));
	print_str("[");
	print_long(kb);
	print_str(" KB]\n");
}

//-----------------------------------------------------------------------------
// print_mem_pools_tab: Print table of memory pools.
//-----------------------------------------------------------------------------

void out_c::print_mem_pools_tab(str_t headline)
{
	// Define default headline:
	if(headline == STR_NULL)
		headline = "Memory Pools:";

	// Print headline:
	print_str(headline);
	print_str("\n");

	// Get start of linked list of memory pools:
	mpool_t mpool = mpool_c::first_pool();

	// Hand case that there are no memory pools:
	if(mpool == MPOOL_NULL) {
		print_str("    No memory pools exist.\n");
		return;
	}

	// Loop over memory pools, print memory pool data, add memory pages:
	long total_pages = 0;
	for( ; mpool != MPOOL_NULL; mpool = mpool->next_pool()) {
		str_t name = mpool->name();
		int pages = mpool->num_alloc_pages();
		int kb = PAGE_KB(pages);

		// Print a line in the memory pool list:
		print_str("    ");
		print_str(name);
		print_str(": ");
		print_str(str_indent(OUT_NAME_WIDTH-str_len(name)));
		print_str(str_fill(pages, OUT_PAGES_WIDTH));
		print_int(pages);
		print_str(" Pages ");
		print_str(str_fill(kb, OUT_KB_WIDTH));
		print_str("[");
		print_int(kb);
		print_str(" KB]\n");

		total_pages += pages;
	}

	// Print total size of all memory pools:
	if(total_pages > 0) {
		long total_kb = (total_pages * VER_PAGESIZE)/1024;
		print_str("    [");
		print_long(total_kb);
		print_str(" KB in total]\n");
	}
}


//=============================================================================
// Static Variables:
//=============================================================================


//-----------------------------------------------------------------------------
// Verbose: Is verbose output switched on?
//-----------------------------------------------------------------------------

bool out_c::Verbose = true;

//-----------------------------------------------------------------------------
// AnswerOutput: Is output of answers switched on?
//-----------------------------------------------------------------------------

bool out_c::AnswerOutput = true;

//-----------------------------------------------------------------------------
// AnswerLimit: One more than the maximal number of answers to print.
//-----------------------------------------------------------------------------

long out_c::AnswerLimit = 11;

//-----------------------------------------------------------------------------
// Debug: Is debug output switched on?
//-----------------------------------------------------------------------------

bool out_c::Debug = false;

//-----------------------------------------------------------------------------
// Summary: Is the summary line switched on?
//-----------------------------------------------------------------------------

bool out_c::Summary = false;

//-----------------------------------------------------------------------------
// Summary_NewLine: Should there be a newline at the end of the summary line?
//-----------------------------------------------------------------------------

bool out_c::Summary_NewLine = false;
	// If there is no newline, the output can be easily combined with the
	// output of the time command into one line.

//-----------------------------------------------------------------------------
// Rel: Should relation statistics be shown?
//-----------------------------------------------------------------------------

bool out_c::Rel = false;

//-----------------------------------------------------------------------------
// MemPools: Should the list of all memory pools be printed?
//-----------------------------------------------------------------------------

bool out_c::MemPools = false;

//-----------------------------------------------------------------------------
// Dump: Is the data structure dump switched on?
//-----------------------------------------------------------------------------

bool out_c::Dump = false;

