// ============================================================================
// Project:	Deductive Database
// Filename:	out.h
// Purpose:	Class for Output of Tests/Benchmarks
// Last Change:	29.10.2016
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2016 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Include File Frame:
//=============================================================================

#ifndef	OUT_INCLUDED
#define OUT_INCLUDED

//=============================================================================
// Used Types and Macros:
//=============================================================================

#ifndef VER_INCLUDED
#include "ver.h"
#endif

#ifndef STR_INCLUDED
#include "str.h"
#endif

#ifndef CHECK_INCLUDED
#include "check.h"
#endif

#ifndef IDIV_INCLUDED
#include "idiv.h"
#endif

#ifndef PAGE_INCLUDED
#include "page.h"
#endif

#ifndef STRTAB_INCLUDED
#include "strtab.h"
#endif

#ifndef ATOMTAB_INCLUDED
#include "atomtab.h"
#endif

#ifndef REL_INCLUDED
#include "rel.h"
#endif


//=============================================================================
// Module for Output of Tests/Benchmarks (Class with only static members):
//=============================================================================

class out_c {
	public:


//-----------------------------------------------------------------------------
// Static Methods for Output Selection:
//-----------------------------------------------------------------------------

	// set_verbose: Set verbose mode with output while test runs.
	static void set_verbose(bool verbose) {
		Verbose = verbose;
	}

	// set_answer_limit: Set maximal number of answers to print.
	static void set_answer_limit(long max_answers) {
		if(max_answers <= 1) {
			AnswerOutput = false;
			AnswerLimit = 0;
		}
		else {
			AnswerOutput = true;
			AnswerLimit = max_answers + 1;
			// If the answer number is equal to limit,
			// a message is printed that there are more answers.
			// The answer itself is not printed.
			// Therefore the limit must be one larger than
			// the maximum number of answers to print.
		}
	}

	// set_debug: Set debug mode with additional output while test runs.
	static void set_debug(bool debug) {

		// Set debug mode:
		Debug = debug;
	}

	// set_summary: Set flag for printing summary line.
	static void set_summary(bool summary, bool newline) {
		Summary = summary;
		Summary_NewLine = newline;
	}

	// set_rel: Set flag for printing relation statistics.
	static void set_rel(bool rel) {
		Rel = rel;
	}

	// set_mem_pools: Set flag for printing memory pool list.
	static void set_mem_pools(bool mem_pools) {
		MemPools = mem_pools;
	}

	// set_dump: Set flag for printing data structure dump.
	static void set_dump(bool dump) {
		Dump = dump;
	}


//-----------------------------------------------------------------------------
// Static Methods for Querying Output Selection:
//-----------------------------------------------------------------------------

	// show_rel: Should the relation statistics be shown?
	static bool show_rel() {
		return Rel;
	}

	// show_dump: Should the data structure dump be shown?
	static bool show_dump() {
		return Dump;

		// Note: The dump is currently written via methods
		// of err_c, because it is also used if check constraints fail.
		// Therefore, it is written to cerr and the alert file
		// (as selected by methods of err_c).
		// It would be possible to print a note or headline here.
	}

//-----------------------------------------------------------------------------
// Static Methods for Verbose Output:
//-----------------------------------------------------------------------------

	// verbose_msg: Print string only in verbose mode.
	static inline void verbose_msg(str_t msg) {
		if(Verbose)
			print_str(msg);
	}

	// verbose_int: Print integer value only in verbose mode.
	static inline void verbose_int(int i) {
		if(Verbose)
			print_int(i);
	}

	// verbose_long: Print long value only in verbose mode.
	static inline void verbose_long(long n) {
		if(Verbose)
			print_long(n);
	}

	// verbose_fact_1: Print fact with predicate and one integer argument.
	static inline void verbose_fact_1(str_t pred, int arg1) {
		if(Verbose)
			print_fact_1(pred, arg1);
	}

	// verbose_fact_22: Print fact with predicate and two integer arguments.
	static inline void verbose_fact_2(str_t pred, int arg1, int arg2) {
		if(Verbose)
			print_fact_2(pred, arg1, arg2);
	}

	// verbose_nl: Print newline.
	static inline void verbose_nl() {
		if(Verbose) {
			print_nl();
		}
	}

	// verbose_hrule: Print "horizontal rule" - separating line.
	static inline void verbose_hrule() {
		if(Verbose) {
			print_str("----------------------------------------");
			print_str("---------------------------------------\n");
		}
	}

	// verbose_test_head: Print test headline.
	static inline void verbose_headline(str_t test_id) {
		if(Verbose) {
			print_str("----------------------------------------");
			print_str("---------------------------------------\n");
			print_str("--- TEST \"");
			print_str(test_id);
			print_str("\"\n");
			print_str("----------------------------------------");
			print_str("---------------------------------------\n");
		}
	}

	// verbose_time: Print time measurements.
	static inline void verbose_time(str_t name, long realtime, long cpu) {
		if(Verbose) {
			print_str(name);
			print_long(realtime);
			print_str(" ms ");
			print_str("(CPU: ");
			print_long(cpu);
			print_str(" ms)\n");
		}
	}

	// verbose_success: Print success message for test.
	static inline void verbose_success(str_t test_id) {
		if(Verbose) {
			print_str("Test '");
			print_str(test_id);
			print_str("' successfully finished.\n");
		}
	}

	// verbose_bad_ok: Test function reported ok, but error detected.
	static inline void verbose_bad_ok(str_t test_id) {
		if(Verbose) {
			print_str("Test '");
			print_str(test_id);
			print_str("' returned ok status,");
			print_str("but an error was detected.\n");
		}
	}

	// verbose_failure: Print failure message for test.
	static inline void verbose_failure(str_t test_id) {
		if(Verbose) {
			print_str("*** TEST '");
			print_str(test_id);
			print_str(" FAILED! ***\n");
		}
	}

	// verbose_benchmark: Print benchmark name.
	static inline void verbose_benchmark(str_t bench_name) {
		if(Verbose) {
			print_str("Benchmark:  '");
			print_str(bench_name);
			print_str("'\n");
		}
	}

	// verbose_input: Print input file name.
	static inline void verbose_input(str_t file_name) {
		if(Verbose) {
			print_str("Input File: '");
			print_str(file_name);
			print_str("'\n");
		}
	}

	// verbose_result: Print number of result rows.
	static inline void verbose_result(long eval_result) {
		if(Verbose) {
			print_str("Number of result rows: ");
			print_long(eval_result);
			print_nl();
		}
	}

	// verbose_cannot_check: Cannot check result, data missing.
	static inline void verbose_cannot_check() {
		if(Verbose) {
			print_str("Result cannot be checked, ");
			print_str("expected result unknown.\n");
		}
	}

	// verbose_bench_ok: Benchmark was successfully executed (result ok).
	static inline void verbose_bench_ok() {
		if(Verbose) {
			print_str("Benchmark was successfully executed.\n");
		}
	}

	// verbose_real_gt_cpu: Realtime was much bigger than cpu time.
	static inline void verbose_real_gt_cpu(long realtime, long cputime) {
		if(Verbose) {
			print_str("Warning: Wallclock time (");
			print_long(realtime);
			print_str(" ms) much more than CPU time (");
			print_long(cputime);
			print_str(" ms)\n");
		}
	}

	// verbose_cpu_gt_real: CPU time was significantly more than cpu time.
	static inline void verbose_cpu_gt_real(long realtime, long cputime) {
		if(Verbose) {
			print_str("Warning: CPU time (");
			print_long(cputime);
			print_str(" ms) is more than Wallclock time (");
			print_long(realtime);
			print_str(" ms)\n");
		}
	}

	// verbose_test_failed: Function call returned false.
	static inline void verbose_test_failed(str_t msg) {
		if(Verbose) {
			print_str("*** TEST FAILED: ");
			print_str(msg);
			print_str(" ***\n");
		}
	}

	// verbose_wrong_result: Function call returned unexpected value.
	static inline void verbose_wrong_result(str_t fun, long result,
						long expected_result) {
		if(Verbose) {
			print_str("*** TEST FAILED: ");
			print_str("Function ");
			print_str(fun);
			print_str(" returned ");
			print_long(result);
			print_str(" (expected: ");
			print_long(expected_result);
			print_str(") ***\n");
		}
	}

	// verbose_num_rows: Print relation size (number of rows).
	static inline void verbose_num_rows(int num_rows) {
		if(Verbose) {
			print_str("Number of rows:    ");
			print_int(num_rows);
			print_nl();
		}
	}

	// verbose_num_cols: Print number of columns (with bound, free args).
	static inline void verbose_num_cols(int bound_args, int free_args) {
		if(Verbose) {
			print_str("Number of columns: ");
			print_int(bound_args + free_args);
			print_str(" (bound: ");
			print_int(bound_args);
			print_str(", free: ");
			print_int(free_args);
			print_str(")");
			print_nl();
		}
	}

	// verbose_implementation: Print implementation (data structure) name.
	static inline void verbose_implementation(str_t impl_name) {
		if(Verbose) {
			print_str("Implementation:    ");
			print_str(impl_name);
			print_nl();
		}
	}

	// verbose_hashtab_size: Print hash table size.
	static inline void verbose_hashtab_size(int hash_size) {
		if(Verbose) {
			print_str("Hash table size:   ");
			print_int(hash_size);
			print_nl();
		}
	}

	// verbose_hashtab_usage: Print hash table usage.
	static inline void verbose_hashtab_usage(int hash_size,
							int entries_used){
		if(Verbose) {
			print_str("Used entries:      ");
			print_int(entries_used);
			print_str(" (");
			print_int(IDIV_PERCENT(entries_used, hash_size));
			print_str("%)");
			print_nl();
		}
	}

	// verbose_hashtab_chain: Print average hash table chain length.
	static inline void verbose_hashtab_chain(int num_rows,
				int entries_used) {
		if(Verbose) {
			print_str("Avg. chain length: ");
			if(entries_used == 0)
				print_int(0);
			else {
				print_int(IDIV_WHOLE(num_rows, entries_used));
				print_str(".");
				print_int(IDIV_TENTH(num_rows, entries_used));
			}
			print_nl();

		}
	}

	// verbose_bucket_size: Print bucket size.
	static inline void verbose_bucket_size(int num_entries, int bytes) {
		if(Verbose) {
			print_str("Bucket size:       ");
			print_int(num_entries);
			if(bytes >= 0) {
				print_str(" (");
				print_int(bytes);
				print_str(" Bytes)");
			}
			print_nl();
		}
	}

	// verbose_total_buckets: Print total number of buckets.
	static inline void verbose_total_buckets(int total_buckets) {
		if(Verbose) {
			print_str("Total buckets:     ");
			print_int(total_buckets);
			print_nl();
		}
	}

	// verbose_overflow_buckets: Print number of overflow buckets.
	static inline void verbose_overflow_buckets(int overflow_buckets) {
		if(Verbose) {
			print_str("Overflow buckets:  ");
			print_int(overflow_buckets);
			print_nl();
		}
	}

	// verbose_rows_per_bucket: Print average entries per bucket.
	static inline void verbose_rows_per_bucket(int num_rows,
					int total_buckets, int bucket_size) {
		if(Verbose && total_buckets > 0) {
			print_str("Rows per bucket:   ");
			print_int(IDIV_WHOLE(num_rows, total_buckets));
			print_str(".");
			print_int(IDIV_TENTH(num_rows, total_buckets));
			if(bucket_size > 0) {
				int rows_per_bucket =
					IDIV_WHOLE(num_rows, total_buckets)*10+
					IDIV_TENTH(num_rows, total_buckets);
				print_str(" (");
				print_int(IDIV_PERCENT(rows_per_bucket,
							bucket_size * 10));
				print_str("%)");
			}
			print_nl();
		}
	}


	// verbose_mem_pages: Print memory usage.
	static inline void verbose_mem_pages(int num_pages) {
		if(Verbose) {
			print_str("Used memory pages: ");
			print_int(num_pages);
			print_str(" (");
			print_int(PAGE_KB(num_pages));
			print_str(" KB)");
			print_nl();
		}
	}

	// verbose_obj_size: Print size of the data structure object.
	static inline void verbose_obj_size(int obj_size) {
		if(Verbose) {
			print_str("Object size:       ");
			print_int(obj_size);
			print_str(" Bytes");
			print_nl();
		}
	}

	// verbose_bytes_per_row: Print memory usage per row.
	static inline void verbose_bytes_per_row(int num_pages, int obj_size,
							int num_rows) {
		if(Verbose && num_rows > 0) {
			int bytes = PAGE_KB(num_pages) * 1024 + obj_size;
			print_str("Bytes per row:     ");
			print_int(IDIV_WHOLE(bytes, num_rows));
			print_str(".");
			print_int(IDIV_TENTH(bytes, num_rows));
			print_str(" Bytes");
			print_nl();
		}
	}

//-----------------------------------------------------------------------------
// Static Methods for Output of Answers:
//-----------------------------------------------------------------------------

	// answer_1: Print answer with one integer argument.
	static inline void answer_1(long no, int arg1) {
		if(AnswerOutput && no <= AnswerLimit) {
			if(no < AnswerLimit) {
				print_answer_no(no);
				print_fact_1("answer", arg1);
			}
			else if( AnswerLimit > 1) {
				print_answer_limit();
			}
		}
	}

	// answer_2: Print answer with 2 integer arguments.
	static inline void answer_2(long no, int arg1, int arg2) {
		if(AnswerOutput && no <= AnswerLimit) {
			if(no < AnswerLimit) {
				print_answer_no(no);
				print_fact_2("answer", arg1, arg2);
			}
			else if( AnswerLimit > 1) {
				print_answer_limit();
			}
		}
	}

	// answer_s: Print answer with one string argument.
	static inline void answer_s(long no, str_t arg1) {
		if(AnswerOutput && no <= AnswerLimit) {
			if(no < AnswerLimit) {
				print_answer_no(no);
				print_fact_s("answer", arg1);
			}
			else if( AnswerLimit > 1) {
				print_answer_limit();
			}
		}
	}

	// answer_5s: Print answer with five string arguments.
	static inline void answer_5s(long no, str_t arg1, str_t arg2,
					str_t arg3, str_t arg4, str_t arg5) {
		if(AnswerOutput && no <= AnswerLimit) {
			if(no < AnswerLimit) {
				print_answer_no(no);
				print_fact_5s("answer", arg1, arg2, arg3,
							arg4, arg5);
			}
			else if( AnswerLimit > 1) {
				print_answer_limit();
			}
		}
	}

//-----------------------------------------------------------------------------
// Static Methods for Output of Relation Statistics:
//-----------------------------------------------------------------------------

	// rel_size_headline: Print headline for relation list.
	static inline void rel_size_headline() {
		if(Rel) {
			print_nl();
			print_str("Relations and Domains:\n");
			print_str("--------------------- \n");
		}
	}

	// rel_size: Show number of rows and pages of relation:
	static inline void rel_size(rel_t rel) {
		if(Rel) {
			print_rel_size(rel->name(), rel->num_rows(),
							rel->num_pages());
		}
	}

	// rel_size_nonstd: Version for non-standard implementations.
	static inline void rel_size_nonstd(str_t name, long num_rows,
						long num_pages){
		if(Rel) {
			print_rel_size(name, num_rows, num_pages);
		}
	}

	// rel_size_strtab: Print size of a string table.
	static inline void rel_size_strtab(strtab_t tab) {
		if(Rel) {
			print_rel_size_strtab(tab);
		}
	}

	// rel_size_atomtab: Print size of an atom table.
	static inline void rel_size_atomtab(atomtab_t tab) {
		if(Rel) {
			print_rel_size_atomtab(tab);
		}
	}

	// rel_size_done: Called at the end of the relation/domain table.
	static inline void rel_size_done() {
		if(Rel) {
			print_nl();
		}
	}

//-----------------------------------------------------------------------------
// Static Method for Output of Memory Pools:
//-----------------------------------------------------------------------------

	// mem_pools_tab: Show table of memory pools.
	static inline void mem_pools_tab(str_t headline = STR_NULL) {
		if(MemPools) {
			print_mem_pools_tab(headline);
		}
	}

//-----------------------------------------------------------------------------
// Static Method for Benchmark Summary Line:
//-----------------------------------------------------------------------------

	// summary: Print summary line.
	static inline void summary(str_t bench_id, long load_time,
			long eval_time, long total_time, str_t result) {
		if(Summary)
			print_summary(bench_id, load_time, eval_time,
						total_time, result);
	}

//-----------------------------------------------------------------------------
// Static Methods for Debug Output:
//-----------------------------------------------------------------------------

	// debug_msg: Print string only in debug mode.
	static inline void debug_msg(str_t msg) {
		if(Debug)
			print_str(msg);
	}

	// debug_int: Print integer value only in debug mode.
	static inline void debug_int(int i) {
		if(Debug)
			print_int(i);
	}

	// debug_long: Print long value only in debug mode.
	static inline void debug_long(long n) {
		if(Debug)
			print_long(n);
	}

	// debug_nl: Print newline only in debug mode.
	static inline void debug_nl() {
		if(Debug)
			print_nl();
	}

	// debug_fact_1: Print fact with predicate and one integer argument.
	static inline void debug_fact_1(str_t pred, int arg1) {
		if(Debug)
			print_fact_1(pred, arg1);
	}

	// debug_fact_2: Print fact with predicate and two integer arguments.
	static inline void debug_fact_2(str_t pred, int arg1, int arg2) {
		if(Debug)
			print_fact_2(pred, arg1, arg2);
	}

	// debug_ival: Print attribute-value pair (int val) in debug mode.
	static inline void debug_ival(str_t attr, int i) {
		if(Debug) {
			print_str(attr);
			print_int(i);
			print_nl();
		}
	}

	// debug_lval: Print attribute-value pair (long val) in debug mode.
	static inline void debug_lval(str_t attr, long n) {
		if(Debug) {
			print_str(attr);
			print_int(n);
			print_nl();
		}
	}

	// debug_sval: Print attribute-value pair (string val) in debug mode.
	static inline void debug_sval(str_t attr, str_t s) {
		if(Debug) {
			print_str(attr);
			print_str("'");
			print_str(s);
			print_str("'");
			print_nl();
		}
	}

	// debug_mempages: Print number of memory pages in debug mode.
	static inline void debug_mempages(str_t attr, int num_pages) {
		if(Debug) {
			print_str(attr);
			print_int(num_pages);
			print_str(" [");
			print_int(PAGE_KB(num_pages));
			print_str(" KB]");
			print_nl();
		}
	}

	// debug_hashusage: Print number of used entries in a hash table.
	static inline void debug_hashusage(str_t attr, int entries_used,
							int hash_size) {
		if(Debug) {
			print_str(attr);
			print_int(entries_used);
			print_str(" [");
			print_int(IDIV_PERCENT(entries_used, hash_size));
			print_str("%]");
			print_nl();
		}
	}

	// debug_hashchain: Print average chain length for a hash table.
	static inline void debug_hashchain(str_t attr, int entries_used,
							int num_rows) {
		if(Debug) {
			print_str(attr);
			print_int(entries_used);
			print_int(IDIV_WHOLE(num_rows, entries_used));
			print_str(".");
			print_int(IDIV_TENTH(num_rows, entries_used));
			print_nl();
		}
	}


//-----------------------------------------------------------------------------
// Methods and Macros for printing trace information in debug mode:
//-----------------------------------------------------------------------------

	// debug_trace_iteration: Print trace information for main loop.
	static void debug_trace_iteration(int stacksize)
	{
		if(Debug) {
			print_str("NEXT ITERATION [Stacksize: ");
			print_int(stacksize);
			print_str("]");
			print_nl();
		}
	}

	// debug_trace_label: Print trace information for label reached.
	static void debug_trace_label(str_t type, int label,
					str_t rule, str_t fact_type)
	{
		if(Debug) {
			static const int LABEL_WIDTH = 3;
			print_str("    ");
			print_str(type);
			print_str("[");
			print_int(label);
			print_str("] ");
			print_str(str_fill(label, LABEL_WIDTH));
			print_str(rule);
			print_str(" <= ");
			print_str(fact_type);
			print_nl();
		}
	}

	// debug_trace_assign: Print trace information for variable assigned.
	static void debug_trace_assign(str_t var, int value)
	{
		if(Debug) {
			print_str("        ");
			print_str(var);
			print_str(" = ");
			print_int(value);
			print_nl();
		}
	}

	// debug_trace_pop_var: Print trace information for variable restored.
	static void debug_trace_pop_var(str_t var, int value)
	{
		if(Debug) {
			print_str("        ");
			print_str(var);
			print_str(" = ");
			print_int(value);
			print_str(" [Restored from Stack]");
			print_nl();
		}
	}

	static void debug_trace_rule_app(int call_depth, str_t rule,
					str_t pred, int num_args,
					int arg1, int arg2)
	{
		if(Debug) {
			while(call_depth-- > 0)
				print_str("    ");
			print_str("RULE ");
			print_str(rule);
			print_str("   <= ");
			print_str(pred);
			print_str("(");
			if(num_args > 0)
				print_int(arg1);
			if(num_args > 1) {
				print_str(", ");
				print_int(arg2);
			}
			print_str(")");
			print_nl();
		}
	}

	static void debug_trace_call(int call_depth, str_t pred, int num_args,
					int arg1, int arg2)
	{
		if(Debug) {
			while(call_depth-- > 0)
				print_str("    ");
			print_str("CALL: ");
			print_str(pred);
			print_str("(");
			if(num_args > 0)
				print_int(arg1);
			if(num_args > 1) {
				print_str(", ");
				print_int(arg2);
			}
			print_str(")");
			print_nl();
		}
	}

#ifdef	VER_DEBUG_OUT_ENABLED

// Trace-Output for Standard Algorithm (with Switch):
#define	OUT_TRACE_ITERATION(STACKSIZE) \
	(out_c::debug_trace_iteration((STACKSIZE)))
#define	OUT_TRACE_START(LABEL, RULE, FACT_TYPE) \
	(out_c::debug_trace_label("START  ", (LABEL), (RULE), (FACT_TYPE)))
#define	OUT_TRACE_START_REC(LABEL, RULE, FACT_TYPE) \
	(out_c::debug_trace_label("START_R", (LABEL), (RULE), (FACT_TYPE)))
#define	OUT_TRACE_CONT(LABEL, RULE, FACT_TYPE) \
	(out_c::debug_trace_label("CONT   ", (LABEL), (RULE), (FACT_TYPE)))
#define	OUT_TRACE_ANSWER(LABEL, RULE, FACT_TYPE) \
	(out_c::debug_trace_label("ANSWER ", (LABEL), (RULE), (FACT_TYPE)))
#define	OUT_TRACE_RESTORE(LABEL, RULE, FACT_TYPE) \
	(out_c::debug_trace_label("RESTORE", (LABEL), (RULE), (FACT_TYPE)))
#define	OUT_TRACE_ASSIGN(VAR, VALUE) \
	(out_c::debug_trace_assign((VAR), (VALUE)))
#define	OUT_TRACE_POP_VAR(VAR, VALUE) \
	(out_c::debug_trace_pop_var((VAR), (VALUE)))

// Trace-Output for Function Call Version:
#define	OUT_TRACE_DEPTH_DECL \
	int out_call_depth_global
#define	OUT_TRACE_DEPTH_INIT \
	(out_call_depth_global = 0)
#define	OUT_TRACE_DEPTH \
	int out_call_depth_local = (this->out_call_depth_global + 1)

#define	OUT_TRACE_RULE_APP_0(RULE, PRED) \
	(out_c::debug_trace_rule_app(out_call_depth_local, \
				     (RULE), (PRED), 0, 0, 0))
#define	OUT_TRACE_RULE_APP_1(RULE, PRED, ARG1) \
	(out_c::debug_trace_rule_app(out_call_depth_local, \
				     (RULE), (PRED), 1, (ARG1), 0))
#define	OUT_TRACE_RULE_APP_2(RULE, PRED, ARG1, ARG2) \
	(out_c::debug_trace_rule_app(out_call_depth_local, \
				     (RULE), (PRED), 2, (ARG1), (ARG2)))

#define	OUT_TRACE_CALL_0(PRED) \
	(this->out_call_depth_global = out_call_depth_local, \
	 	out_c::debug_trace_call(out_call_depth_local, \
			(PRED), 0, 0, 0))
#define	OUT_TRACE_CALL_1(PRED, ARG1) \
	(this->out_call_depth_global = out_call_depth_local, \
	 	out_c::debug_trace_call(out_call_depth_local, \
			(PRED), 1, (ARG1), 0))
#define	OUT_TRACE_CALL_2(PRED, ARG1, ARG2) \
	(this->out_call_depth_global = out_call_depth_local, \
	 	out_c::debug_trace_call(out_call_depth_local, \
			(PRED), 2, (ARG1), (ARG2)))

// If debug output is not enabled, trace output is not compiled in:
#else

#define	OUT_TRACE_ITERATION(STACKSIZE) \
	((void)0)
#define	OUT_TRACE_START(LABEL, RULE, FACT_TYPE) \
	((void)0)
#define	OUT_TRACE_START_REC(LABEL, RULE, FACT_TYPE) \
	((void)0)
#define	OUT_TRACE_CONT(LABEL, RULE, FACT_TYPE) \
	((void)0)
#define	OUT_TRACE_ANSWER(LABEL, RULE, FACT_TYPE) \
	((void)0)
#define	OUT_TRACE_RESTORE(LABEL, RULE, FACT_TYPE) \
	((void)0)
#define	OUT_TRACE_ASSIGN(VAR, VALUE) \
	((void)0)
#define	OUT_TRACE_POP_VAR(VAR, VALUE) \
	((void)0)

#define	OUT_TRACE_DEPTH_DECL \
	void out_depth_not_used()
	// If we simply leave it empty, gcc -pedantic gives a warning
	// that there two semicolons one after the other.
#define	OUT_TRACE_DEPTH_INIT \
	((void)0)
#define	OUT_TRACE_DEPTH \
	((void)0)

#define	OUT_TRACE_RULE_APP_0(RULE, PRED) \
	((void)0)
#define	OUT_TRACE_RULE_APP_1(RULE, PRED, ARG1) \
	((void)0)
#define	OUT_TRACE_RULE_APP_2(RULE, PRED, ARG1, ARG2) \
	((void)0)

#define	OUT_TRACE_CALL_0(PRED) \
	((void)0)
#define	OUT_TRACE_CALL_1(PRED, ARG1) \
	((void)0)
#define	OUT_TRACE_CALL_2(PRED, ARG1, ARG2) \
	((void)0)

// End of conditional compilation of debug output:
#endif

//-----------------------------------------------------------------------------
// Static Methods for Printing Help:
//-----------------------------------------------------------------------------

	// help_line: One line of the -h help output:
	static inline void help_line(str_t line) {
		print_str(line);
		print_nl();
	}

	// : One line of the -h help output:
	static inline void help_test(str_t id, str_t name) {
		static const int ID_WIDTH = 15;
		print_str(id);
		print_str(": ");
		print_str(str_indent(ID_WIDTH - str_len(id)));
		print_str(name);
		print_nl();
	}

//-----------------------------------------------------------------------------
// Auxiliary Static Methods for Output:
//-----------------------------------------------------------------------------

	private:

	// print_str: Print string.
	static void print_str(str_t msg);

	// print_int: Print integer.
	static void print_int(int i);

	// print_long: Print long value.
	static void print_long(long n);

	// print_nl: Print newline.
	static void print_nl();

	// print_fact_1: Print fact with predicate and one integer argument.
	static void print_fact_1(str_t pred, int arg1);

	// print_fact_2: Print fact with predicate and two integer arguments.
	static void print_fact_2(str_t pred, int arg1, int arg2);

	// print_fact_s: Print fact with predicate and one string argument.
	static void print_fact_s(str_t pred, str_t arg1);

	// print_fact_5s: Print fact with predicate and five string arguments.
	static void print_fact_5s(str_t pred, str_t arg1, str_t arg2,
					str_t arg3, str_t arg4, str_t arg5);

	// print_answer_limit: Print information that there are more answers.
	static void print_answer_limit();

	// print_answer_no: Print prefix of line showing a computed answer.
	static void print_answer_no(long no);

	// print_summary: Print benchmark summary line.
	static void print_summary(str_t bench_id, long load_time,
				long eval_time, long total_time, str_t result);

	// print_rel_size: Print some data about a relation.
	static void print_rel_size(str_t name, long num_rows, long num_pages);

	// print_rel_size_strtab: Print some data about a string table.
	static void print_rel_size_strtab(strtab_t tab);

	// print_rel_size_atomtab: Print some data about an atom table.
	static void print_rel_size_atomtab(atomtab_t tab);

	// print_mem_pools_tab: Print table of memory pools.
	static void print_mem_pools_tab(str_t headline);

//-----------------------------------------------------------------------------
// There are no objects of this class, therefore no constructor is available:
//-----------------------------------------------------------------------------

	private:

	// Constructor:
	out_c();				// Not implemented

	// Destructor:
	~out_c();				// Not implemented

	// Copy-Constructor:
	out_c(const out_c& obj);		// Not implemented

	// Assignment operator:
	out_c& operator=(const out_c& obj);	// Not implemented

//-----------------------------------------------------------------------------
// Private Class Variables:
//-----------------------------------------------------------------------------

	private:

	// Verbose mode:
	static bool Verbose;

	// AnswerOutput: Is output of answers switched on?
	static bool AnswerOutput;

	// AnswerLimit: One more than the maximal number of answers to print.
	static long AnswerLimit;

	// Debug mode:
	static bool Debug;

	// Flag for printing summary line:
	static bool Summary;

	// Flag for printing newline at end of summary line (else only a tab):
	static bool Summary_NewLine;
		// If there is no newline, the output can be easily combined
		// with the output of the time command in one line.

	// Flag for showing relation statistics:
	static bool Rel;

	// Flag for showing memory pool data:
	static bool MemPools;

	// Flag for showing data structure dump:
	static bool Dump;


//=============================================================================
// End of Class:
//=============================================================================

};


//=============================================================================
// End of Include File:
//=============================================================================

#endif

