// ============================================================================
// Project:	Deductive Database
// Filename:	perf.cpp
// Purpose:	Performance measurement functions, especially timer.
// Last Change:	18.08.2015
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2011-2015 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"
#include "str.h"
#include "check.h"
#include "perf.h"

//=============================================================================
// Operating System dependent Include Files:
//=============================================================================


#ifdef VER_UNIX
#include <sys/time.h>
#endif

#ifdef VER_WINDOWS
#include <windows.h>
#endif

#include <time.h>


//=============================================================================
// Global Variables of this Module:
//=============================================================================

//-----------------------------------------------------------------------------
// Perf_NumUsed: Number of initialized places in Perf_InUse
//-----------------------------------------------------------------------------

static int Perf_NumUsed = 0;

//-----------------------------------------------------------------------------
// Perf_InUse: Indexes in data arrays that are currently in use.
//-----------------------------------------------------------------------------

static bool Perf_InUse[VER_MAX_TIMERS];

//-----------------------------------------------------------------------------
// Perf_Real_Start, Perf_Real_Stop: Data for real time measurement.
//-----------------------------------------------------------------------------

#ifdef VER_UNIX
static timeval Perf_Real_Start[VER_MAX_TIMERS];
static timeval Perf_Real_Stop[VER_MAX_TIMERS];
#endif

#ifdef VER_WINDOWS
static FILETIME Perf_Real_Start[VER_MAX_TIMERS];
static FILETIME Perf_Real_Stop[VER_MAX_TIMERS];
#endif

//-----------------------------------------------------------------------------
// Perf_CPU_Start, Perf_CPU_Stop: Data for CPU time measurement.
//-----------------------------------------------------------------------------

#ifdef VER_UNIX
static clock_t Perf_CPU_Start[VER_MAX_TIMERS];
static clock_t Perf_CPU_Stop[VER_MAX_TIMERS];
#endif

#ifdef VER_WINDOWS
static FILETIME Perf_User_Start[VER_MAX_TIMERS];
static FILETIME Perf_Kernel_Start[VER_MAX_TIMERS];
static FILETIME Perf_User_Stop[VER_MAX_TIMERS];
static FILETIME Perf_Kernel_Stop[VER_MAX_TIMERS];
static boolean Perf_Error[VER_MAX_TIMERS];
#endif

//=============================================================================
// Constructor and Destructor:
//=============================================================================


//-----------------------------------------------------------------------------
// Constructor:
//-----------------------------------------------------------------------------

perf_c::perf_c()
{
	// Find free index (or set ID to -1 if none):
	if(Perf_NumUsed < VER_MAX_TIMERS) {
		ID = Perf_NumUsed;
		Perf_InUse[ID] = true;
		Perf_NumUsed++;
	}
	else {
		ID = -1;
		for(int i = 0; i < VER_MAX_TIMERS; i++) {
			if(!Perf_InUse[i]) {
				ID = i;
				Perf_InUse[ID] = true;
				Perf_NumUsed++;
			}
		}
	}

	// Set Magic number:
	CHECK_CODE(Magic = PERF_MAGIC);

	// Initialize timer data:
	if(ID >= 0) {
		start();
		stop();
	}
}

//-----------------------------------------------------------------------------
// Destructor:
//-----------------------------------------------------------------------------

perf_c::~perf_c()
{
	// Check this object:
	CHECK_VALID("perf_c::destructor");

	// Free entry in arrays:
	if(ID != -1)
		Perf_InUse[ID] = false;

	// Make this object invalid:
	CHECK_CODE(Magic = 0);
}

//-----------------------------------------------------------------------------
// start: Start time measurement.
//-----------------------------------------------------------------------------

void perf_c::start()
{
	// Check this object:
	CHECK_VALID("perf_c::start");

	// Remember current time if we got a place in the arrays:
	if(ID != -1) {

		// Current Real Time, UNIX version:
#ifdef VER_UNIX
		gettimeofday(Perf_Real_Start+ID, 0);
#endif

		// Current Real Time, Windows version:
#ifdef VER_WINDOWS
		GetSystemTimeAsFileTime(&(Perf_Real_Start[ID]));
#endif

		// Current CPU time, UNIX version:
#ifdef VER_UNIX
		Perf_CPU_Start[ID] = clock();
#endif

		// Current CPU time, Windows version:
#ifdef VER_WINDOWS
		FILETIME create_time;
		FILETIME exit_time;
		Perf_Error[ID] = false;
		if(GetProcessTimes(GetCurrentProcess(),
					&create_time, &exit_time,
					&(Perf_Kernel_Start[ID]),
					&(Perf_User_Start[ID])) == 0) {
			Perf_Error[ID] = true;
		}
#endif

	}
}

//-----------------------------------------------------------------------------
// stop: Stop time measurement.
//-----------------------------------------------------------------------------

void perf_c::stop()
{
	// Check this object:
	CHECK_VALID("perf_c::stop");

	// Remember current time if we got a place in the arrays:
	if(ID != -1) {
		// Current Real Time, UNIX version:
#ifdef VER_UNIX
		gettimeofday(Perf_Real_Stop+ID, 0);
#endif

		// Current Real Time, Windows version:
#ifdef VER_WINDOWS
		GetSystemTimeAsFileTime(&(Perf_Real_Stop[ID]));
#endif

		// Current CPU time:
#ifdef VER_UNIX
		Perf_CPU_Stop[ID] = clock();
#endif

		// Current CPU time, Windows version:
#ifdef VER_WINDOWS
		FILETIME create_time;
		FILETIME exit_time;
		if(GetProcessTimes(GetCurrentProcess(),
					&create_time, &exit_time,
					&(Perf_Kernel_Stop[ID]),
					&(Perf_User_Stop[ID])) == 0) {
			Perf_Error[ID] = true;
		}
#endif
	}
}

//-----------------------------------------------------------------------------
// realtime_msec: Time between last start() and stop() in milliseconds.
//-----------------------------------------------------------------------------

long int perf_c::realtime_msec()
{
	// Check this object:
	CHECK_VALID("perf_c::realtime_msec");

	// If it was not possible to store timing data, return 0:
	if(ID == -1)
		return 0;

	// Compute milliseconds, UNIX Version:
#ifdef VER_UNIX
	long sec = Perf_Real_Stop[ID].tv_sec - Perf_Real_Start[ID].tv_sec;
	long usec = Perf_Real_Stop[ID].tv_usec - Perf_Real_Start[ID].tv_usec;
// XXX cout << "sec=" << sec << ", usec= " << usec << "\n";
	if(usec < 0) {
		usec += 1000000;
		sec -= 1;
	}
	return sec * 1000 + (usec + 500) / 1000;
#endif

	// Compute milliseconds, Windows Version:
#ifdef VER_WINDOWS
	VER_INT64 start = Perf_Real_Start[ID].dwHighDateTime;
	start <<= 32;
	start |= Perf_Real_Start[ID].dwLowDateTime;
	VER_INT64 stop = Perf_Real_Stop[ID].dwHighDateTime;
	stop <<= 32;
	stop |= Perf_Real_Stop[ID].dwLowDateTime;

	return static_cast<long>((stop - start) / 10000);
#endif

}

//-----------------------------------------------------------------------------
// cputime_msec: CPU time between last start() and stop() in millisecs.
//-----------------------------------------------------------------------------

long int perf_c::cputime_msec()
{
	// Check this object:
	CHECK_VALID("perf_c::cputime_msec");

	// If it was not possible to store timing data, return 0:
	if(ID == -1)
		return 0;

	// Compute CPU time in milliseconds:
#ifdef VER_UNIX
	long ticks = Perf_CPU_Stop[ID] - Perf_CPU_Start[ID];
	return ticks / (CLOCKS_PER_SEC / 1000);
#endif

#ifdef VER_WINDOWS
	if(Perf_Error[ID])
		return 0;
	VER_INT64 user_stop = Perf_User_Stop[ID].dwHighDateTime;
	user_stop <<= 32;
	user_stop |= Perf_User_Stop[ID].dwLowDateTime;
	VER_INT64 user_start = Perf_User_Start[ID].dwHighDateTime;
	user_start <<= 32;
	user_start |= Perf_User_Start[ID].dwLowDateTime;

	VER_INT64 kernel_stop = Perf_Kernel_Stop[ID].dwHighDateTime;
	kernel_stop <<= 32;
	kernel_stop |= Perf_Kernel_Stop[ID].dwLowDateTime;
	VER_INT64 kernel_start = Perf_Kernel_Start[ID].dwHighDateTime;
	kernel_start <<= 32;
	kernel_start |= Perf_Kernel_Start[ID].dwLowDateTime;

	return static_cast<long>(
			((user_stop-user_start) + (kernel_stop-kernel_start))
				/ 10000); // Unit was 100ns
#endif
}


//-----------------------------------------------------------------------------
// Debugging Support:
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
str_t	perf_c::check() const
{
	// Check magic number:
	if(Magic != PERF_MAGIC)
		return "wrong magic number";

	// Check global variables of this module:
	if(Perf_NumUsed < 0)
		return "Perf_NumUsed is negative";

	// Check ID:
	if(ID < -1)
		return "ID is negative (and not -1)";
	if(ID >= VER_MAX_TIMERS)
		return "ID is too large";
	if(ID > Perf_NumUsed)
		return "ID > Perf_NumUsed";
	if(ID != -1 && !Perf_InUse[ID])
		return "Place in Arrays is not reserved";

	// Ok:
	return STR_NULL;
}

#endif

