// ============================================================================
// Project:	Deductive Database
// Filename:	rel.cpp
// Purpose:	Superclass of all data structures for relations
// Last Change:	31.10.2016
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2010-2016 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"
#include "str.h"
#include "err.h"
#include "check.h"

#include "rel.h"



//=============================================================================
// Constructor and Destructor:
//=============================================================================


//-----------------------------------------------------------------------------
// Constructor:
//-----------------------------------------------------------------------------

rel_c::rel_c(str_t tab_name, int bound, int free, str_t impl) :
		MemPool(tab_name)
{
	// Set Magic number:
	CHECK_CODE(Magic = REL_MAGIC);

	// Check parameter:
	CHECK(tab_name != STR_NULL,
			"rel_c::constructor: tab_name is NULL");
	CHECK(bound >= 0,
			"rel_c::constructor: bound is negative");
	CHECK(free >= 0,
			"rel_c::constructor: free is negative");
	CHECK(impl != STR_NULL,
			"rel_c::constructor: impl is NULL");

	// Initialize attributes (which are not nested objects):
	Name        = tab_name;
	MemErr      = false;
	NumRows     = 0;
	BoundArgs   = bound;
	FreeArgs    = free;
	ImplName    = impl;
}

//-----------------------------------------------------------------------------
// Destructor:
//-----------------------------------------------------------------------------

rel_c::~rel_c()
{
	// Check this object:
	CHECK_VALID("rel_c::destructor");

	// Make this object invalid:
	CHECK_CODE(Magic = 0);
}


//=============================================================================
// Debugging Support:
//=============================================================================

//-----------------------------------------------------------------------------
// check: Check the integrity of this object, return error message or NULL.
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
str_t	rel_c::check() const
{
	// Check magic number:
	if(Magic != REL_MAGIC)
		return "wrong magic number";

	// The table name should not be null:
	if(Name == STR_NULL)
		return "table name is null";

	// The number of rows cannot be negative:
	if(NumRows < 0)
		return "number of rows is negative";

	// The number of bound arguments cannot be negative:
	if(BoundArgs < 0)
		return "number of bound arguments is negative";

	// The number of free arguments cannot be negative:
	if(FreeArgs < 0)
		return "number of free arguments is negative";

	// The implementation data structure name cannot be null:
	if(ImplName == STR_NULL)
		return "implementation name is null";

	// Check sub-objects:
	CHECK_SUBOBJ(MemPool, "rel_c::check");

	// Ok:
	return 0;
}
#endif


//-----------------------------------------------------------------------------
// dump: Show data structure.
//-----------------------------------------------------------------------------

#ifdef	VER_DUMP
void	rel_c::dump(str_t headline) const
{
	// Headline:
	if(headline == STR_NULL)
		headline = "Relation";
	err_c::dump_begin(headline);

	// Basic information:
	err_c::dump_str("Name:       '");
	err_c::dump_str(Name);
	err_c::dump_str("'");
	err_c::dump_nl();

	err_c::dump_str("NumRows:     ");
	err_c::dump_int(NumRows);
	err_c::dump_nl();

	err_c::dump_str("BoundArgs:   ");
	err_c::dump_int(BoundArgs);
	err_c::dump_nl();

	err_c::dump_str("FreeArgs:    ");
	err_c::dump_int(FreeArgs);
	err_c::dump_nl();

	err_c::dump_str("Impl.:      '");
	err_c::dump_str(ImplName);
	err_c::dump_str("'");
	err_c::dump_nl();

	err_c::dump_str("Memory:      ");
	int pages = MemPool.num_alloc_pages();
	err_c::dump_int(pages);
	err_c::dump_str(" Pages [");
	err_c::dump_int(PAGE_KB(pages));
	err_c::dump_str(" KB]");
	err_c::dump_nl();

	err_c::dump_str("MemErr:      ");
	err_c::dump_bool(MemErr);
	err_c::dump_nl();

	// Done:
	err_c::dump_end();
}
#endif

