// ============================================================================
// Project:	Deductive Database
// Filename:	rel.h
// Purpose:	Superclass of all data structures for relations
// Last Change:	09.11.2015
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2010-2015 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Include File Frame:
//=============================================================================

#ifndef	REL_INCLUDED
#define REL_INCLUDED

//=============================================================================
// Used Types and Macros:
//=============================================================================

#ifndef VER_INCLUDED
#include "ver.h"
#endif

#ifndef STR_INCLUDED
#include "str.h"
#endif

#ifndef CHECK_INCLUDED
#include "check.h"
#endif

#ifndef MPOOL_INCLUDED
#include "mpool.h"
#endif


//=============================================================================
// Private Constants:
//=============================================================================


//-----------------------------------------------------------------------------
// REL_MAGIC: Magic number (identifies objects of this class).
//-----------------------------------------------------------------------------

static const long REL_MAGIC = 0x52454C0AL;	// 'REL\n'


//=============================================================================
// Superclass of all data structures for storing relations:
//=============================================================================

class rel_c {
	protected:

//-----------------------------------------------------------------------------
// Constructor, Destructor:
//-----------------------------------------------------------------------------

	// Constructor:
	rel_c(str_t rel_name, int bound, int free, str_t impl);

	// Destructor:
	~rel_c();

//-----------------------------------------------------------------------------
// Copy-constructor and assignment operator are not supported for this class:
//-----------------------------------------------------------------------------

	private:

	rel_c(const rel_c& obj);		// Not implemented
	rel_c& operator=(const rel_c& obj);	// Not implemented


//=============================================================================
// (Object) Methods:
//=============================================================================

	public:

//-----------------------------------------------------------------------------
// name: Get name of this table/relation.
//-----------------------------------------------------------------------------

	str_t name() const {
		CHECK_VALID("set::name");
		return Name;
	}

//-----------------------------------------------------------------------------
// mem_err: Did we get enough memory pages?
//-----------------------------------------------------------------------------

	bool mem_err() const {
		CHECK_VALID("rel_c::mem_err");
		return MemErr;
	}

//-----------------------------------------------------------------------------
// num_rows: Get number of rows (tuples) in this table.
//-----------------------------------------------------------------------------

	int num_rows() const {
		CHECK_VALID("rel_c::num_rows");
		return NumRows;
	}

//-----------------------------------------------------------------------------
// num_cols: Get number of columns (attributes) of this table.
//-----------------------------------------------------------------------------

	int num_cols() const {
		CHECK_VALID("rel_c::num_cols");
		return FreeArgs + BoundArgs;
	}


//-----------------------------------------------------------------------------
// bound_args: Number of bound (input) arguments of the relation.
//-----------------------------------------------------------------------------

	/** Number of bound (input) arguments of the relation.
	 *
	 * The sum with the number of free arguments gives the arity
	 * of the relation, i.e. the number of columns.
	 *
	 * @return	Number of arguments that must be given to access
	 * 		rows in this relation.
	 * 		0 means that this is a list
	 *		(i.e. one can do a full table scan).
	 */

	int bound_args() {

		// Check this object:
		CHECK_VALID("rel_c::bound_args");

		// Return number of bound arguments:
		return BoundArgs;
	}


//-----------------------------------------------------------------------------
// free_args: Number of free (output) arguments of the relation.
//-----------------------------------------------------------------------------


	/** Number of free (output) arguments of the relation.
	 *
	 * The sum with the number of bound arguments gives the arity
	 * of the relation, i.e. the number of columns.
	 *
	 * @return	Number of arguments for which a lookuo operation
	 * 		returns values (or really a set of rows of values).
	 * 		0 means that this is a set
	 *		(i.e. one can only do an element test).
	 */

	int free_args() {

		// Check this object:
		CHECK_VALID("rel_c::free_args");

		// Return number of free arguments:
		return FreeArgs;
	}

//-----------------------------------------------------------------------------
// impl_name: Implementation (data structure) used for this relation.
//-----------------------------------------------------------------------------


	/** Implementation (data structure) used for this relation.
	 *
	 * This method returns a short name of the data structure used
	 * for implementing this relation.
	 *
	 * @return	Implementation name, e.g. "Extendible hash table".
	 */

	str_t impl_name() {

		// Check this object:
		CHECK_VALID("rel_c::impl_name");

		// Return the implementation name.
		return ImplName;
	}

//-----------------------------------------------------------------------------
// num_pages: Get number of memory pages used for this relation.
//-----------------------------------------------------------------------------

	int num_pages() const {
		CHECK_VALID("rel_c::num_pages");
		return MemPool.num_alloc_pages();
	}


//=============================================================================
// Debugging Support:
//=============================================================================

#ifdef	VER_DEBUG
	public:
	// Integrity check:
	str_t	check() const;

	private:
	// Magic number (identifies objects of this class for debugging).
	long	Magic;		// Must be "REL_MAGIC".
#endif

#ifdef	VER_DUMP
	public:
	// Display data structure:
	void dump(str_t headline = STR_NULL) const;
#endif


//=============================================================================
// Protected Object Members:
//=============================================================================

	protected:

	// Memory Pool for Allocating Storage Pages:
	mpool_c MemPool;

	// Name of this set (table/relation):
	str_t Name;

	// Did a memory page allocation error occur?
	bool MemErr;

	// Current number of rows/tuples:
	int NumRows;

	// Number of "bound" (input) arguments:
	int BoundArgs;

	// Number of "free" (output) arguments:
	int FreeArgs;

		// If there are no bound arguments, it is a list data structure.
		// If there are no free arguments, it is a set data structure.
		// Otherwise it is a multi-map,
		// i.e. for given values for the input arguments,
		// there is a set/list of rows for the output arguments.

	// Implementation (data structure) name:
	str_t ImplName;

//=============================================================================
// End of Class:
//=============================================================================

};

//-----------------------------------------------------------------------------
// Define pointer type:
//-----------------------------------------------------------------------------

typedef rel_c *rel_t;

//-----------------------------------------------------------------------------
// Define null pointer:
//-----------------------------------------------------------------------------

#define REL_NULL (static_cast<rel_t>(0))


//=============================================================================
// End of Include File:
//=============================================================================

#endif

