// ============================================================================
// Project:	Deductive Database
// Filename:	rel_n_n.cpp
// Purpose:	Relation with two non-negative integer columns, index over 1st.
// Last Change:	31.10.2016
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2010-2016 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"
#include "str.h"
#ifdef	VER_DUMP
#include "err.h"
#endif
#include "check.h"
#include "flexarr.h"
#include "rel.h"

#include "rel_n_n.h"


//=============================================================================
// Constructor and Destructor:
//=============================================================================


//-----------------------------------------------------------------------------
// Constructor:
//-----------------------------------------------------------------------------

rel_n_n_c::rel_n_n_c(str_t rel_name) :
		rel_c(rel_name, 1, 1, "Flexible Array with Linked List"),
		IndexArr(&MemPool), ValueArr(&MemPool)
{
	// Check parameter:
	CHECK(rel_name != STR_NULL,
			"rel_n_n_c::constructor: rel_name is NULL");
	// Set Magic number:
	CHECK_CODE(Magic = REL_N_N_MAGIC);
}

//-----------------------------------------------------------------------------
// Destructor:
//-----------------------------------------------------------------------------

rel_n_n_c::~rel_n_n_c()
{
	// Check this object:
	CHECK_VALID("rel_n_n_c::destructor");

	// Make this object invalid:
	CHECK_CODE(Magic = 0);
}

//-----------------------------------------------------------------------------
// insert: Insert tuple into relation.
//-----------------------------------------------------------------------------

bool rel_n_n_c::insert(int c1, int c2)
{
	// Check this object:
	CHECK_VALID("rel_n_n_c::insert");

	// Check parameters:
	CHECK(c1 >= 0, "rel_n_n_c::insert: c1 is negative");
	CHECK(c2 >= 0, "rel_n_n_c::insert: c2 is negative");

	// Ensure that index is long enough:
	while(IndexArr.length() < c1) {
		if(IndexArr.append(-1) < 0)
			return false;
	}

	// If the index is smaller than c1, we simply append an entry with c2:
	if(IndexArr.length() == c1) {
		if(IndexArr.append(c2) < 0)
			return false;
		NumRows++;
		return true;
	}

	// When we reach this, the index was longer than c1.
	CHECK(IndexArr.length() > c1,
			"rel_n_n_c::insert: IndexArr not as long as expected");

	// If there was no map for this index before, add it:
	int entry = IndexArr.get(c1);
	if(entry == -1) {
		*(IndexArr.ptr(c1)) = c2;
		NumRows++;
		return true;
	}

	// If there was a normal entry before, we must create a linked list:
	if(entry >= 0) {
		int start = ValueArr.length();
		if(ValueArr.append(entry) < 0)
			return false;
		if(ValueArr.append(-(start+2)-2) < 0)
			return false;
		if(ValueArr.append(c2) < 0)
			return false;
		if(ValueArr.append(-1) < 0)
			return false;
		*(IndexArr.ptr(c1)) = -start-2;
		NumRows++;
		return true;
	}

	// Otherwise we must append a single node to the linked list:
	int node = ValueArr.length();
	if(ValueArr.append(c2) < 0)
		return false;
	if(ValueArr.append(entry) < 0)
		return false;
	*(IndexArr.ptr(c1)) = -node-2;
	NumRows++;
	return true;
}

//=============================================================================
// Debugging Support:
//=============================================================================

//-----------------------------------------------------------------------------
// check: Check the integrity of this object, return error message or NULL.
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
str_t	rel_n_n_c::check() const
{
	// Check magic number:
	if(Magic != REL_N_N_MAGIC)
		return "wrong magic number";

	// Check superclass:
	str_t msg = rel_c::check();
	if(msg)
		return msg;

#ifdef VER_FULLCHECK
	// Check links from index array to value list array:
	for(int i = 0; i < IndexArr.length(); i++) {
		int val = IndexArr.get(i);
		if(val < -1) {
			val = -val - 2;
			if(val >= ValueArr.length())
				return "Invalid ValueArr index in IndexArr";
		}
	}

	// Check values and links in value list array:
	for(int i = 0; i < ValueArr.length(); i++) {
		int val = ValueArr.get(i);
		if(i % 2 == 0 && val < 0)
			return "Negative value in ValueArr array";
		if(i % 2 != 0 && val >= 0)
			return "Invalid link in ValueArr array";
	}
#endif

	// Check sub-objects:
	CHECK_SUBOBJ(MemPool, "rel_n_n_c::check");
	CHECK_SUBOBJ(IndexArr, "rel_n_n_c::check");
	CHECK_SUBOBJ(ValueArr, "rel_n_n_c::check");

	// Ok:
	return 0;
}
#endif


//-----------------------------------------------------------------------------
// dump: Show data structure.
//-----------------------------------------------------------------------------

#ifdef	VER_DUMP
void	rel_n_n_c::dump(str_t headline) const
{
	// Headline:
	if(headline == STR_NULL)
		headline =
		"Relation (with two natural number columns, arrindex 1st)";
	err_c::dump_begin(headline);

	// Determine percentage used of index array entries:
	int used = 0;
	for(int i = 0; i < IndexArr.length(); i++)
		if(IndexArr.get(i) != -1)
			used++;
	if(IndexArr.length() > 0)
		used = (used * 100 + IndexArr.length()/2) / IndexArr.length();

	// Basic information:
	err_c::dump_str("Name:           '");
	err_c::dump_str(Name);
	err_c::dump_str("'");
	err_c::dump_nl();
	err_c::dump_str("NumRows:         ");
	err_c::dump_int(NumRows);
	err_c::dump_nl();
	err_c::dump_str("Memory:          ");
	int pages = MemPool.num_alloc_pages();
	err_c::dump_int(pages);
	err_c::dump_str(" Pages [");
	err_c::dump_int(PAGE_KB(pages));
	err_c::dump_str(" KB]");
	err_c::dump_nl();
	err_c::dump_str("IndexArr Length: ");
	err_c::dump_int(IndexArr.length());
	err_c::dump_str("  [Percentage used: ");
	err_c::dump_int(used);
	err_c::dump_str("%]");
	err_c::dump_nl();
	err_c::dump_str("ValueArr Length: ");
	err_c::dump_int(ValueArr.length());
	err_c::dump_str("  [");
	err_c::dump_int(ValueArr.length() / 2);
	err_c::dump_str(" List Nodes]");
	err_c::dump_nl();
	err_c::dump_str("Object Size:     ");
	err_c::dump_uint(static_cast<unsigned int>(sizeof(rel_n_n_c)));
	err_c::dump_str(" Bytes");
	err_c::dump_nl();
	err_c::dump_nl();

	// Show entire index array with chains of entries:
	err_c::dump_str("Index Array:");
	err_c::dump_nl();
	for(int i = 0; i < IndexArr.length(); i++) {
		int n = IndexArr.get(i);
		if(n != -1) {
			err_c::dump_str("  ");
			err_c::dump_str(str_fill(i, 5));
			err_c::dump_int(i);
			err_c::dump_str(": ");
			err_c::dump_str(str_fill(n, 5));
			err_c::dump_int(n);
			if(n < -1) {
				err_c::dump_str(" -> ");
				while(n != -1) {
					n = -n - 2;
					if(n >= ValueArr.length() - 1) {
						err_c::dump_str(" INVALID! ");
						break;
					}
					else {
						err_c::dump_str(" ");
						int c2 = ValueArr.get(n);
						err_c::dump_int(c2);
						n = ValueArr.get(n+1);
						if(n > -1) {
							err_c::dump_str(
								" BAD LINK: ");
							err_c::dump_int(n);
							break;
						}
					}
				}
			}
		err_c::dump_nl();
		}
	}
	err_c::dump_nl();

	// Show entire value list array:
	err_c::dump_str("Value List Array:");
	err_c::dump_nl();
	for(int i = 0; i < ValueArr.length(); i++) {
		int n = ValueArr.get(i);
		if(i % 2 == 0) {
			err_c::dump_str("  ");
			err_c::dump_str(str_fill(i, 5));
			err_c::dump_int(i);
			err_c::dump_str(": ");
			err_c::dump_str(str_fill(n, 5));
			err_c::dump_int(n);
			err_c::dump_str("  ");
		}
		else {
			err_c::dump_str("Link: ");
			err_c::dump_str(str_fill(n, 5));
			err_c::dump_int(n);
			if(n >= 0)
				err_c::dump_str(" INVALID!");
			else if(n == -1)
				err_c::dump_str(" NIL");
			else {
				n = -n - 2;
				err_c::dump_str(" -> ");
				err_c::dump_str(str_fill(n, 5));
				err_c::dump_int(n);
			}
			err_c::dump_nl();
		}
	}
	err_c::dump_nl();


	// Done:
	err_c::dump_end();
}
#endif

