// ============================================================================
// Project:	Deductive Database
// Filename:	relsize.h
// Purpose:	Size measures for relation data structures.
// Last Change:	31.10.2016
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2016 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================

/**
 * \enum  relsize_t
 * \brief Size measures for relation data structures.
 *
 * This enumeration type contains size measures for relation data structures,
 * e.g. the size of a hash table, the number of used entries in the hash table,
 * the number of nodes in a search tree, and so on.
 * These measures help to evaluate the efficiency of a data structure.
 *
 * Each data structure has its own size measures, and in general test code,
 * it should be possible to print these characteristic numbers of the specific
 * data structure.
 * Therefore, each relation implementation has a method \c size_value
 * with a parameter of this enumeration type.
 * If the measure does not apply to the data structure,
 * the method returns \c -1.
 *
 * It is to be expected that this enumeration type is extended from time
 * to time (when a new data structure is implemented).
 * Therefore, it is recommended that the method \c size_value
 * uses a \c switch
 * that handles the size measures of the specific data structure
 * and then contains a \c default that returns \c -1.
 *
 * Note that the number of rows in the relation is not contained among
 * the size measures of this type, because every relation has a method
 * \c num_rows() that returns this value.
 * This enumeration type is intended only for the data structure specific
 * size characteristics.
 */

//=============================================================================
// Include File Frame:
//=============================================================================

#ifndef	RELSIZE_INCLUDED
#define RELSIZE_INCLUDED

//=============================================================================
// Used Types and Macros:
//=============================================================================

#ifndef VER_INCLUDED
#include "ver.h"
#endif

#ifndef STR_INCLUDED
#include "str.h"
#endif

#ifndef CHECK_INCLUDED
#include "check.h"
#endif

//=============================================================================
// Exported Types:
//=============================================================================

//-----------------------------------------------------------------------------
// relsize_t: Size measures for relation data structures.
//-----------------------------------------------------------------------------

typedef enum {
	RELSIZE_OBJ_SIZE,
	RELSIZE_HASHTAB_SIZE,
	RELSIZE_ENTRIES_USED,
	RELSIZE_BUCKET_SIZE,
	RELSIZE_BUCKET_BYTES,
	RELSIZE_TOTAL_BUCKETS,
	RELSIZE_OVERFLOW_BUCKETS
} relsize_t;

//=============================================================================
// Exported Functions:
//=============================================================================

//-----------------------------------------------------------------------------
// relsize_valid: Is the parameter a valid element of the enumeration type?
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
bool	relsize_valid(relsize_t relsize);
#endif

//-----------------------------------------------------------------------------
// relsize_name: Name of size measure.
//-----------------------------------------------------------------------------

str_t	relsize_name(relsize_t relsize);


//=============================================================================
// End of File:
//=============================================================================

#endif

