// ============================================================================
// Project:	Deductive Database
// Filename:	rgen_cur1.cpp
// Purpose:	Cursor over a generated one-column relation (for tests).
// Last Change:	29.10.2016
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2016 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"
#include "str.h"
#ifdef	VER_DUMP
#include "err.h"
#endif
#include "check.h"

#include "rgen_cur1.h"


//=============================================================================
// Constructor and Destructor:
//=============================================================================


//-----------------------------------------------------------------------------
// Constructor 1 with full parameter selection:
//-----------------------------------------------------------------------------

rgen_cur1_c::rgen_cur1_c(int type, int size, int max_value)
{
	// Check parameters:
	CHECK(type == 1 || type == 2,
		"rgen_cur1_c::constructor: currently, there are only type 1,2");
	CHECK(size > 0,
		"rgen_cur1_c::constructor: size must be at least 1");
	CHECK(max_value >= 6,
		"rgen_cur1_c::constructor: max_value must be at least 6");
	CHECK(size <= max_value - 1,
		"rgen_cur1_c::constructor: max_value too small or size too big");

	// Set Magic number:
	CHECK_CODE(Magic = RGEN_CUR1_MAGIC);

	// Initialize attributes that are never changed during object lifetime:
	Type       = type;
	Size       = size;
	MaxValue   = max_value;

	// Initialize sequence generator attributes and current tuple:
	init();
}

//-----------------------------------------------------------------------------
// Constructor 2 with simplified relation ID:
//-----------------------------------------------------------------------------

rgen_cur1_c::rgen_cur1_c(int rel_id)
{
	// Check parameters:
	CHECK(rel_id >= 1 && rel_id <= 3,
		"rgen_cur1_c::constructor2: there are only relations 1-3");

	// Set Magic number:
	CHECK_CODE(Magic = RGEN_CUR1_MAGIC);

	// Set relation parameters:
	switch(rel_id) {
		case 1:
			// Simple relation with 10 tuples:
			Type     = 1;
			Size     = 10;
			MaxValue = 10;
			break;
		case 2:
			// Simple relation with one million tuples:
			Type     = 1;
			Size     = 1000000;
			MaxValue = 1000000;
			break;
		case 3:
			// Relation with mixed sequences, one million tuples:
			Type     = 2;
			Size     = 1000000;
			MaxValue = 1000000;
			break;
		default:
			CHECK_IMPOSSIBLE(
				"rgen_cur1_c::constructor2: invalid rel_id");
			Type     = 1;
			Size     = 0;
			MaxValue = 10;
			break;
	}

	// Initialize sequence generator attributes and current tuple:
	init();
}

//-----------------------------------------------------------------------------
// Destructor:
//-----------------------------------------------------------------------------

rgen_cur1_c::~rgen_cur1_c()
{
	// Check this object:
	CHECK_VALID("rgen_cur1_c::destructor");

	// Make this object invalid:
	CHECK_CODE(Magic = 0);
}


//-----------------------------------------------------------------------------
// open: Open the cursor.
//-----------------------------------------------------------------------------

void rgen_cur1_c::open()
{
	// Check this object:
	CHECK_VALID("rgen_cur1_c::open");

	// Initialize sequence generator attributes and current tuple:
	init();

	// The tuple number must be 0 (cursor open, just before first tuple):
	TupleNo = 0;
}



//=============================================================================
// Auxiliary Methods:
//=============================================================================

//-----------------------------------------------------------------------------
// init: Initialize sequence generator attributes and current tuple.
//-----------------------------------------------------------------------------

void rgen_cur1_c::init()
{
	// Check domain size:
	CHECK(MaxValue >= 6,
		"rgen_cur1_c::init: MaxValue must be at least 6");

	// Initialize attributes to represent current tuple (cursor closed):
	TupleNo    = -1;
	C1         = -1;

	// Initialize sequence generator attributes:
	BottomUp = 3;
	TopDown  = MaxValue;
	MiddleUp = MaxValue / 2;
	MiddleDown = MiddleUp - 1;

	// Make sure that TopDown == 1 mod 3:
	if(TopDown % 3 != 1)
		TopDown--;
	if(TopDown % 3 != 1)
		TopDown--;
	CHECK(TopDown % 3 == 1,
		"rgen_cur1_c::init: TopDown mod 3 != 1");
	CHECK(TopDown > 0,
		"rgen_cur1_c::init: TopDown is not > 0");

	// Make sure that MiddleUp == 2 mod 3:
	if(MiddleUp % 3 != 2)
		MiddleUp++;
	if(MiddleUp % 3 != 2)
		MiddleUp++;
	CHECK(MiddleUp % 3 == 2,
		"rgen_cur1_c::init: MiddleUp mod 3 != 2");
	CHECK(MiddleUp <= MaxValue,
		"rgen_cur1_c::init: MiddleUp is not <= MaxValue");

	// Make sure that MiddleDown == 2 mod 3:
	if(MiddleDown % 3 != 2)
		MiddleDown--;
	if(MiddleDown % 3 != 2)
		MiddleDown--;
	CHECK(MiddleDown % 3 == 2,
		"rgen_cur1_c::init: MiddleDown mod 3 != 2");
	CHECK(MiddleDown > 0,
		"rgen_cur1_c::init: MiddleDown is not > 0");
}

//=============================================================================
// Debugging Support:
//=============================================================================

//-----------------------------------------------------------------------------
// check: Check the integrity of this object, return error message or NULL.
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
str_t	rgen_cur1_c::check() const
{
	// Check magic number:
	if(Magic != RGEN_CUR1_MAGIC)
		return "wrong magic number";

	// Check parameters of sequence:
	if(Type != 1 && Type != 2)
		return "currently, there are only type 1 and 2";
	if(Size <= 0)
		return "size must be at least 1";
	if(MaxValue < 6)
		return "MaxValue must be at least 6";
	if(Size > MaxValue + 1)
		"MaxValue too small or size too big";

	// Check tuple number and data values:
	if(TupleNo < -1) {
		return "TupleNo is negative and not -1";
	}
	else if(TupleNo == -1 || TupleNo == 0) {
		if(C1 != -1)
			return "cursor closed, but C1 != -1";
	}
	else {
		if(C1 < 0)
			return "cursor open, but C1 is negative";
		if(C1 > MaxValue)
			return "C1 is too large";
		if(TupleNo > Size + 1)
			return "cursor has run behind end of relation";
	}

	// Check BottomUp (can be one step outside 1..MaxValue):
	if(BottomUp < 1)
		return "BottomUp is too small";
	if(BottomUp > MaxValue + 3)
		return "BottomUp is too large";
	if(BottomUp % 3 != 0)
		return "BottomUp mod 3 is not 0";

	// Check TopDown (can be one step outside 1..MaxValue):
	if(TopDown < -2)
		return "TopDown is too small";
	if(TopDown > MaxValue)
		return "TopDown is too large";
	if(TopDown >= 0 && TopDown % 3 != 1)
		return "TopDown mod 3 is not 1";

	// Check MiddleUp (can be one step outside 1..MaxValue):
	if(MiddleUp < 1)
		return "MiddleUp is too small";
	if(MiddleUp > MaxValue + 3)
		return "MiddleUp is too large";
	if(MiddleUp % 3 != 2)
		return "MiddleUp mod 3 is not 2";

	// Check MiddleDown (can be one step outside 1..MaxValue):
	if(MiddleDown < -2)
		return "MiddleDown is too small";
	if(MiddleDown > MaxValue)
		return "MiddleDown is too large";
	if(MiddleDown >= 0 && MiddleDown % 3 != 2)
		return "MiddleDown mod 3 is not 2";

	// MiddleUp starts from the middle upwards, MiddleDown downwards:
	if(MiddleUp <= MiddleDown)
		return "MiddleUp must always be larger than MiddleDown";

	// Ok:
	return 0;
}
#endif


//-----------------------------------------------------------------------------
// dump: Show data structure.
//-----------------------------------------------------------------------------

#ifdef	VER_DUMP
void	rgen_cur1_c::dump(str_t headline) const
{
	// Headline:
	if(headline == STR_NULL)
		headline =
		"Cusror for test data (Relation with two int number cols)";
	err_c::dump_begin(headline);

	// Relation Parameters:
	err_c::dump_str("Relation Parameters:");
	err_c::dump_nl();
	err_c::dump_str("    Type:       ");
	err_c::dump_int(Type);
	err_c::dump_nl();
	err_c::dump_str("    Size:       ");
	err_c::dump_int(Size);
	err_c::dump_nl();
	err_c::dump_str("    MaxValue:   ");
	err_c::dump_int(MaxValue);
	err_c::dump_nl();
	err_c::dump_nl();

	// Current Tuple:
	err_c::dump_str("Current Tuple:");
	err_c::dump_nl();
	err_c::dump_str("    TupleNo:    ");
	err_c::dump_int(TupleNo);
	err_c::dump_nl();
	err_c::dump_str("    C1:         ");
	err_c::dump_int(C1);
	err_c::dump_nl();
	err_c::dump_nl();

	// Sequence Generator:
	err_c::dump_str("Sequence Generator:");
	err_c::dump_nl();
	err_c::dump_str("    BottomUp:   ");
	err_c::dump_int(BottomUp);
	err_c::dump_nl();
	err_c::dump_str("    TopDown:    ");
	err_c::dump_int(TopDown);
	err_c::dump_nl();
	err_c::dump_str("    MiddleUp:   ");
	err_c::dump_int(MiddleUp);
	err_c::dump_nl();
	err_c::dump_str("    MiddleDown: ");
	err_c::dump_int(MiddleDown);
	err_c::dump_nl();
	err_c::dump_nl();

	// Size of this object:
	err_c::dump_str("Object Size:    ");
	err_c::dump_uint(static_cast<unsigned int>(sizeof(rgen_cur1_c)));
	err_c::dump_str(" Bytes");
	err_c::dump_nl();
	err_c::dump_nl();

	// Done:
	err_c::dump_end();
}
#endif

