// ============================================================================
// Project:	Deductive Database
// Filename:	row_2.h
// Purpose:	Tuple/Table Row with two integer columns.
// Last Change:	07.11.2015
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2010-2015 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Include File Frame:
//=============================================================================

#ifndef	ROW_2_INCLUDED
#define ROW_2_INCLUDED

//=============================================================================
// Used Types and Macros:
//=============================================================================

#ifndef VER_INCLUDED
#include "ver.h"
#endif

#ifndef STR_INCLUDED
#include "str.h"
#endif

#ifndef CHECK_INCLUDED
#include "check.h"
#endif


//=============================================================================
// Private Constants:
//=============================================================================

//-----------------------------------------------------------------------------
// ROW_2_PRIME: Some large prime number (needed for hash function).
//-----------------------------------------------------------------------------

static const long ROW_2_PRIME = 1073806763;

//-----------------------------------------------------------------------------
// ROW_2_MAGIC: Magic number (identifies objects of this class).
//-----------------------------------------------------------------------------

static const long ROW_2_MAGIC = 0x525F320AL;	// 'R_2\n'

//=============================================================================
// Class for Relation with two integer columns, no index:
//=============================================================================

class row_2_c {
	public:

//-----------------------------------------------------------------------------
// Constants:
//-----------------------------------------------------------------------------

	static const int NUM_COLS = 2;

//-----------------------------------------------------------------------------
// Constructor, Destructor:
//-----------------------------------------------------------------------------

	// Constructor:
	row_2_c(int col1, int col2) : C1(col1), C2(col2) {

		// Set Magic number:
		CHECK_CODE(Magic = ROW_2_MAGIC);
	}

	// Destructor:
	~row_2_c() {
		// Check this object:
		CHECK_VALID("row_2_c::destructor");

		// Make this object invalid:
		CHECK_CODE(Magic = 0);
	}

	// Copy-constructor:
	row_2_c(const row_2_c& row) : C1(row.C1), C2(row.C2) {

		// Set Magic number:
		CHECK_CODE(Magic = ROW_2_MAGIC);
	}

//-----------------------------------------------------------------------------
// (Object) Methods:
//-----------------------------------------------------------------------------

	// c1: Get value of first column.
	int c1() {
		// Check this object:
		CHECK_VALID("row_2_c::c1");

		// Return value of column 1:
		return C1;
	}

	// c2: Get value of second column.
	int c2() {
		// Check this object:
		CHECK_VALID("row_2_c::c2");

		// Return value of column 2:
		return C2;
	}

	// hash: Compute a hash value.
	int hash() {
		// Check this object:
		CHECK_VALID("row_2_c::hash");

		// Compute hash value:
		long hash_val = C1 ^ (C2 << 11) ^ (C2 << 7);
		if(hash_val < 0)
			hash_val = -(hash_val+1);
		CHECK(hash_val >= 0,
			"row_2_c::hash: hash_val is negative");
		return static_cast<int>(hash_val % ROW_2_PRIME);
	}


	// ==: equality test.
	bool operator==(const row_2_c& other) {
		// Check this object:
		CHECK_VALID("row_2_c::==");

		// Objects are equal if both columns have the same value:
		return C1 == other.C1 && C2 == other.C2;
	}

	// =: Assignment Operator.
	row_2_c& operator=(const row_2_c& other) {
		// Copy data values:
		C1 = other.C1;
		C2 = other.C2;

		// Set Magic number:
		CHECK_CODE(Magic = ROW_2_MAGIC);

		// Return this object (standard interface of assignment op):
		return *this;
	}

	// Create a string representation in a given buffer:
	void to_string(char *buf, int buf_len);
	
//-----------------------------------------------------------------------------
// Debugging Support:
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
	public:
	// Integrity check:
	str_t	check() const;

	private:
	// Magic number (identifies objects of this class for debugging).
	long	Magic;		// Must be "ROW_2_MAGIC".
#endif

#ifdef	VER_DUMP
	public:
	// Display data structure:
	void dump(str_t headline = STR_NULL) const;
#endif


//-----------------------------------------------------------------------------
// Private Object Members:
//-----------------------------------------------------------------------------

	private:

	// Column values:
	int C1;
	int C2;

//=============================================================================
// End of Class:
//=============================================================================

};

//-----------------------------------------------------------------------------
// Define pointer type:
//-----------------------------------------------------------------------------

typedef row_2_c *row_2_t;

//-----------------------------------------------------------------------------
// Define null pointer:
//-----------------------------------------------------------------------------

#define ROW_2_NULL (static_cast<row_2_t>(0))

//=============================================================================
// End of Include File:
//=============================================================================

#endif

