// ============================================================================
// Project:	Deductive Database
// Filename:	rtest.cpp
// Purpose:	Superclass for Standard Tests of Relation-like Data Structures
// Last Change:	14.10.2016
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2016 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"
#include "str.h"
#include "err.h"
#include "check.h"
#include "rel.h"

#include "rtest.h"


//=============================================================================
// Constructor and Destructor:
//=============================================================================


//-----------------------------------------------------------------------------
// Constructor:
//-----------------------------------------------------------------------------

rtest_c::rtest_c(rel_t test_rel)
{
	// Set Magic number:
	CHECK_CODE(Magic = RTEST_MAGIC);

	// Check parameters:
	CHECK_PAR(test_rel, "rtest_c::constructor");

	// Initialize attributes:
	Rel          = test_rel;
	NumRows      = 0;
	ErrorRow[0]  = '\0';
	ErrorRowPtr  = ErrorRow;
	ErrorRowFree = RTEST_ROW_SIZE - 1;
	ErrorRowCol  = 0;
}

//-----------------------------------------------------------------------------
// Destructor:
//-----------------------------------------------------------------------------

rtest_c::~rtest_c()
{
	// Check this object:
	CHECK_VALID("rtest_c::destructor");

	// Make this object invalid:
	CHECK_CODE(Magic = 0);
}


//=============================================================================
// Debugging Support:
//=============================================================================

//-----------------------------------------------------------------------------
// check: Check the integrity of this object, return error message or NULL.
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
str_t	rtest_c::check() const
{
	// Check magic number:
	if(Magic != RTEST_MAGIC)
		return "wrong magic number";

	// The number of rows cannot be negative:
	if(NumRows < 0)
		return "number of rows is negative";

	// The number of free places in the buffer must be non-negative:
	if(ErrorRowFree < 0)
		return "ErrorRowFree is negative";

	// The number of free places cannot be larger than the initial size:
	if(ErrorRowFree > RTEST_ROW_SIZE - 1)
		return "ErrorRowFree is too large";

	// The error row pointer can be computed from the free characters:
	if(ErrorRowPtr != ErrorRow + (RTEST_ROW_SIZE - 1 - ErrorRowFree))
		return "ErrorRowPtr and ErrorRowFree are inconsistent";

	// The number of column values in the buffer must be non-negative:
	if(ErrorRowCol < 0)
		return "ErrorRowCol is negative";

	// Ok:
	return 0;
}
#endif


//-----------------------------------------------------------------------------
// dump: Show data structure.
//-----------------------------------------------------------------------------

#ifdef	VER_DUMP
void	rtest_c::dump(str_t headline) const
{
	// Headline:
	if(headline == STR_NULL)
		headline = "Relation Test";
	err_c::dump_begin(headline);

	// Basic information:
	err_c::dump_str("Relation Name: '");
	err_c::dump_str(Rel->name());
	err_c::dump_str("'");
	err_c::dump_nl();

	err_c::dump_str("NumRows:        ");
	err_c::dump_int(NumRows);
	err_c::dump_nl();

	// Done:
	err_c::dump_end();
}
#endif

