// ============================================================================
// Project:	Deductive Database
// Filename:	rtest.h
// Purpose:	Superclass for Standard Tests of Relation-like Data Structures
// Last Change:	31.10.2016
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2016 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Include File Frame:
//=============================================================================

#ifndef	RTEST_INCLUDED
#define RTEST_INCLUDED

//=============================================================================
// Used Types and Macros:
//=============================================================================

#ifndef VER_INCLUDED
#include "ver.h"
#endif

#ifndef STR_INCLUDED
#include "str.h"
#endif

#ifndef CHECK_INCLUDED
#include "check.h"
#endif

#ifndef RELSIZE_INCLUDED
#include "relsize.h"
#endif

#ifndef REL_INCLUDED
#include "rel.h"
#endif


//=============================================================================
// Private Constants:
//=============================================================================

//-----------------------------------------------------------------------------
// RTEST_MAGIC: Magic number (identifies objects of this class).
//-----------------------------------------------------------------------------

static const long RTEST_MAGIC = 0x5254450AL;	// 'RTE\n'

//-----------------------------------------------------------------------------
// RTEST_ROW_SIZE: Size of the string buffer for the problematic row.
//-----------------------------------------------------------------------------

static const int RTEST_ROW_SIZE = 80;


//=============================================================================
// Class Declaration:
//=============================================================================

/** \brief Abstract superclass for Standard Tests/Benchmarks of relation-like
 *		data structures with insert() and lookup() Methods.
 *  @author Stefan Brass
 *  @date   14.10.2016
 */

class rtest_c {
	public:

//-----------------------------------------------------------------------------
// Constructor, Destructor:
//-----------------------------------------------------------------------------

	// Constructor:
	rtest_c(rel_t rel);

	// Destructor:
	~rtest_c();

//-----------------------------------------------------------------------------
// Abstract Methods Required by this Class:
//-----------------------------------------------------------------------------

	/** Insert a specified number of rows.
	 * @return	0 if successful, otherwise number i of the row where
	 * 		insertion failed.
	 */

	virtual int insert() = 0;

	/** Lookup all rows that were inserted.
	 * @return	0 if successful, otherwise number i of the row where
	 * 		the lookup failed.
	 */

	virtual int lookup() = 0;


	/** Return various data structure size measures (e.g. hash table size).
	 * @param relsize Enumeration value that selects the size measure.
	 * @return	Value of the size measure or -1 if not applicable
	 */

	virtual int size_value(relsize_t relsize) = 0;


//-----------------------------------------------------------------------------
// Methods for accessing attributes (basic test data):
//-----------------------------------------------------------------------------

	//---------------------------------------------------------------------
	/** This method returns the relation that is being tested.
	 * @return	Relation that is subject of this test.
	 */
	//---------------------------------------------------------------------

	rel_t rel() {

		// Check this object:
		CHECK_VALID("rtest_c::rel");

		// Return number of bound arguments:
		return Rel;
	}


	//---------------------------------------------------------------------
	/** Number of rows that were successfully inserted.
	 * If an error occurred, this might help in the error diagnosis.
	 * If no error occurred, this is the current size of the relation
	 * in the test.
	 * @return	Number of rows that were successfully inserted so far
	 *		into the relation to be tested.
	 */
	//---------------------------------------------------------------------

	int num_rows() {

		// Check this object:
		CHECK_VALID("rtest_c::num_rows");

		// Return number of inserted rows:
		return NumRows;
	}


	//---------------------------------------------------------------------
	/** String representation of the row that caused an error to be
	 * detected.
	 * Of course, this should be printed as part of the error message.
	 * Since this superclass should work with tests of arbitrary relations,
	 * the row can only be returned as string.
	 * Although most relations work with columns of type int,
	 * this is not required by the system.
	 *
	 * This function should only be called if an error was detected.
	 *
	 * @return	Row that caused an error
	 * 		(string to be printed as part of the error message).
	 */
	//---------------------------------------------------------------------

	str_t error_row() {

		// Check this object:
		CHECK_VALID("rtest_c::error_row");

		// There must be an error if this function is called:
		CHECK(ErrorRow[0] != '\0',
				"rtest_c::error_row: No error detected");

		// Return row at which the error was detected:
		return ErrorRow;
	}


//-----------------------------------------------------------------------------
// Methods to be called by subclasses to store data in this object:
//-----------------------------------------------------------------------------

	protected:

	//---------------------------------------------------------------------
	/** Add number of inserted rows.
	 * This function must be called when a test function inserted rows
	 * into the relation to be tested.
	 * If an error occurred during the test, this must be the number of
	 * insert operations that were successfully executed before the error
	 * was detected.
	 *
	 * @param n	Number of rows that were successfully inserted.
	 */
	//---------------------------------------------------------------------

	void inserted_rows(int n) {

		// Check this object:
		CHECK_VALID("rtest_c::inserted_rows");

		// Check parameter:
		CHECK(n >= 0,
			"rtest_c::inserted_rows: n must be non-negative");

		// Increase number of rows:
		NumRows += n;
	}

	//---------------------------------------------------------------------
	/** Start writing the buffer for the display version of the row
	 * at which an error was detected.
	 */
	//---------------------------------------------------------------------

	void error_row_open() {

		// Check this object:
		CHECK_VALID("rtest_c::error_row_open");

		// Initialize string buffer attributes:
		ErrorRowPtr  = ErrorRow;
		ErrorRowFree = RTEST_ROW_SIZE - 1;
		ErrorRowCol  = 0;

		// Write opening paranthesis:
		if(ErrorRowFree >= 1) {
			*ErrorRowPtr++ = '[';
			ErrorRowFree--;
		}

		// Ensure that string is always null-terminated:
		*ErrorRowPtr = '\0';
	}

	//---------------------------------------------------------------------
	/** Write an integer column value to the string buffer for the display
	 * version of the row that caused the error.
	 *
	 * @param val	Column value.
	 */
	//---------------------------------------------------------------------

	void error_row_int(int val) {

		// Check this object:
		CHECK_VALID("rtest_c::error_row_int");

		// If this is not the first column, write separator:
		if(ErrorRowCol > 0 && ErrorRowFree >= 2) {
			*ErrorRowPtr++ = ',';
			ErrorRowFree--;
			*ErrorRowPtr++ = ' ';
			ErrorRowFree--;
		}

		// write integer value:
		int chars = str_int(ErrorRowPtr, ErrorRowFree, val);
		ErrorRowPtr += chars;
		ErrorRowFree -= chars;
		ErrorRowCol++;

		// Ensure that string is always null-terminated:
		*ErrorRowPtr = '\0';
	}

	//---------------------------------------------------------------------
	/** Close the buffer for the display version of the row that caused
	 *  the error.
	 */
	//---------------------------------------------------------------------

	void error_row_close() {

		// Check this object:
		CHECK_VALID("rtest_c::error_row_close");

		// Write closing paranthesis:
		if(ErrorRowFree >= 1) {
			*ErrorRowPtr++ = ']';
			ErrorRowFree--;
		}

		// Ensure that string is always null-terminated:
		*ErrorRowPtr = '\0';
	}

//-----------------------------------------------------------------------------
// Debugging Support:
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
	public:
	// Integrity check:
	str_t	check() const;

	private:
	// Magic number (identifies objects of this class for debugging).
	long	Magic;		// Must be "RTEST_MAGIC".
#endif

#ifdef	VER_DUMP
	public:
	// Display data structure:
	virtual void dump(str_t headline = STR_NULL) const;
#endif

//-----------------------------------------------------------------------------
// Copy-constructor and assignment operator are not supported for this class:
//-----------------------------------------------------------------------------

	private:

	rtest_c(const rtest_c& obj);		// Not implemented
	rtest_c& operator=(const rtest_c& obj);	// Not implemented


//-----------------------------------------------------------------------------
// Attributes:
//-----------------------------------------------------------------------------

	private:

	// Relation to be tested:
	rel_t Rel;

	// Number of rows that were successfully inserted (relation size):
	int NumRows;

	// Row, for which an operation failed (if an error was detected):
	char ErrorRow[RTEST_ROW_SIZE];

	// Pointer to next free place in the buffer:
	char *ErrorRowPtr;

	// Number of free places in the buffer (- 1 for terminating '\0'):
	int ErrorRowFree;

	// Number of column values already in the buffer:
	int ErrorRowCol;

//=============================================================================
// End of Class:
//=============================================================================

};

//-----------------------------------------------------------------------------
// Define pointer type:
//-----------------------------------------------------------------------------

typedef rtest_c *rtest_t;

//-----------------------------------------------------------------------------
// Define null pointer:
//-----------------------------------------------------------------------------

#define RTEST_NULL (static_cast<rtest_t>(0))

//=============================================================================
// End of Include File:
//=============================================================================

#endif

