// ============================================================================
// Project:	Deductive Database
// Filename:	set.cpp
// Purpose:	Set of objects/tuples - relation with binding pattern b..b
// Last Change:	31.10.2016
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2010-2015 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================

// #define SET_TEST_REHASH

//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"
#include "str.h"
#include "err.h"
#include "check.h"
#include "page.h"
#include "mpool.h"
#include "relsize.h"
#include "rel.h"

#include "set.h"

#ifdef SET_TEST_REHASH
#include <iostream>
#endif

//=============================================================================
// Private Constants:
//=============================================================================

//-----------------------------------------------------------------------------
// SET_MAX_ELEM_CHARS: Size of string buffer for printing element row.
//-----------------------------------------------------------------------------

static const int SET_MAX_ELEM_CHARS = 20;

//=============================================================================
// Constructor and Destructor:
//=============================================================================


//-----------------------------------------------------------------------------
// Constructor:
//-----------------------------------------------------------------------------

template<class T> set_c<T>::set_c(str_t set_name) :
				rel_c(set_name, T::NUM_COLS, 0,
						"Dynamic Hash Table")
{
	// Set Magic number:
	CHECK_CODE(Magic = SET_MAGIC);

	// Check parameters:
	CHECK(set_name != STR_NULL,
			"set_c::constructor: tab_name is NULL");

	// Initialize attributes (which are not nested objects):
	EntriesUsed = 0;
	MemPage     = PAGE_NULL;
	MemPtr      = SET_NODE_NULL;
	MemFree     = 0;
	HashSize    = 0;
	HashPages   = 0;

	// Initialize page array (not really necessary):
	int n = SET_MAXPAGES;
	SET_NODE_T **p = HashTab;
	while(n-- > 0)
		*p++ = static_cast<SET_NODE_T*>(0);

	// Initialize Hash table (initial size: one memory page):
	page_t page = MemPool.alloc_page();
	if(page == PAGE_NULL) {
		MemErr = true;
		return;
	}
	SET_NODE_T *ptr = reinterpret_cast<SET_NODE_T *>(page);
	HashTab[0] = ptr;
	n = PAGE_SIZE(SET_NODE_T);
	while(n-- > 0)
		*ptr++ = SET_NODE_NULL;
	HashPages = 1;
	HashSize = PAGE_SIZE(SET_NODE_T);
}

//-----------------------------------------------------------------------------
// Destructor:
//-----------------------------------------------------------------------------

template<class T> set_c<T>::~set_c()
{
	// Check this object:
	CHECK_VALID("set_c::destructor");

	// Make this object invalid:
	CHECK_CODE(Magic = 0);
}

//-----------------------------------------------------------------------------
// insert: Insert an element into the set (returns true if successful).
//-----------------------------------------------------------------------------

template<class T> bool set_c<T>::insert(T elem)
{
	// Check this object:
	CHECK_VALID("set_c::insert");

	// If there was a memory allocation error, we don't try again:
	if(MemErr)
		return false;

	// Do we need to extend the hash table (if > 75% of entries are used)?
	if(EntriesUsed * 4 > HashPages * PAGE_SIZE(SET_NODE_T) * 3 &&
			HashPages * 2 <= SET_MAXPAGES) {

#ifdef SET_TEST_REHASH
		std::cout << "Rehashing '" << Name << "' from " <<
			HashPages << " Pages (EntriesUsed: " <<
			EntriesUsed << ", HashSize: " << HashSize << ")\n";
		int old_count = 0;
		for(int i = 0; i < HashPages; i++) {
			SET_NODE_T *ptr = HashTab[i];
			for(int n = PAGE_SIZE(SET_NODE_T); n-- > 0; ptr++) {
				for(SET_NODE_T node = *ptr;
						node != SET_NODE_NULL;
						node = node->Next) {
					old_count++;
					CHECK(contains(node->Elem),
					    "set_c::insert: elem not found");
					if(!contains(node->Elem))
						std::cerr <<
							"*** ELEM NOT FOUND\n";
				}
			}
		}
		CHECK(old_count == NumRows,
			"set_c::insert: NumRows inconsistent with num nodes");
		if(old_count != NumRows)
			std::cerr << "*** NUMROWS IS WRONG: " << NumRows <<
				" (counted: " << old_count << ")\n";
#endif

		// Allocate new pages for hash table (for doubling its size):
		for(int i = 0; i < HashPages; i++) {
			page_t page = MemPool.alloc_page();
			if(page == PAGE_NULL) {
				MemErr = true;
				return false;
			}
			SET_NODE_T *ptr = reinterpret_cast<SET_NODE_T *>(page);
			HashTab[HashPages+i] = ptr;
			for(int n = PAGE_SIZE(SET_NODE_T); n-- > 0; )
				*ptr++ = SET_NODE_NULL;
		}

		// Now we must rehash all entries:
		EntriesUsed = 0;
		for(int i = 0; i < HashPages; i++) {
			SET_NODE_T *low_entry  = HashTab[i];
			SET_NODE_T *high_entry = HashTab[HashPages+i];
			for(int n = PAGE_SIZE(SET_NODE_T); n-- > 0;
						low_entry++, high_entry++) {
				SET_NODE_T node = *low_entry;
				*low_entry = SET_NODE_NULL;
				*high_entry = SET_NODE_NULL;
				SET_NODE_T next_node = SET_NODE_NULL;
				for( ; node != SET_NODE_NULL; node=next_node) {
					next_node = node->Next;
					int hash_val = node->Elem.hash();
					CHECK(hash_val >= 0,
					    "set_c::insert: hash_val < 0");
					hash_val %= HashSize * 2;
					if(hash_val >= HashSize) {
						node->Next = *high_entry;
						if(*high_entry==SET_NODE_NULL)
							EntriesUsed++;
						*high_entry = node;
					}
					else {
						node->Next = *low_entry;
						if(*low_entry==SET_NODE_NULL)
							EntriesUsed++;
						*low_entry = node;
					}
				} // end of loop over old hash chain
			} // end of loop over entries in one hash table page
		} // end of loop over hash table pages
		HashSize  = 2 * HashSize;
		HashPages = 2 * HashPages;

#ifdef SET_TEST_REHASH
		int new_count = 0;
		for(int i = 0; i < HashPages; i++) {
			SET_NODE_T *ptr = HashTab[i];
			for(int n = PAGE_SIZE(SET_NODE_T); n-- > 0; ptr++) {
				for(SET_NODE_T node = *ptr;
						node != SET_NODE_NULL;
						node = node->Next) {
					new_count++;
					CHECK(contains(node->Elem),
					    "set_c::insert: elem not found");
					if(!contains(node->Elem))
						std::cerr <<
							"*** ELEM NOT FOUND\n";
				}
			}
		}
		CHECK(old_count == new_count,
			"set_c::insert: Rehashing changed number of elements");
		if(old_count != new_count)
			std::cerr << "*** REHASH FAILURE! " <<
				"Old count: " << old_count << ", " <<
				"New count: " << new_count << "\n";
		else {
			std::cout << "Rehashed '" << Name <<
				"' to " << HashPages <<
				" Pages. NumRows: " << NumRows << " Ok.\n";
		}
		if(HashSize > 0)
			std::cout << "HashSize: " << HashSize <<
				", EntriesUsed: " << EntriesUsed <<
				", Percent: " <<
				(EntriesUsed*100+(HashSize/2)) / HashSize <<
				"\n";
		if(EntriesUsed > 0)
			std::cout << "Chain Length: " << NumRows/EntriesUsed <<
				"." << (NumRows*10/EntriesUsed)%10 << "\n";
		std::cout << "\n";
#endif
	}

	// Search in hash table. First compute hash function:
	// int hash_val = c1 ^ (c2 << 11);
	int hash_val = elem.hash();
	CHECK(hash_val >= 0, "set_c::insert: hash_val is negative");
	hash_val %= HashSize;

	// Get start of linked list:
	SET_NODE_T *hashtab_entry =
				HashTab[hash_val / PAGE_SIZE(SET_NODE_T)];
	hashtab_entry += (hash_val % PAGE_SIZE(SET_NODE_T));

	// Now search the linked list of atoms in that entry:
	SET_NODE_T node;
	for(node = *hashtab_entry; node != SET_NODE_NULL; node = node->Next) {
		if(node->Elem == elem)
			return false;
	}

	// If we arrive here, we must construct a new row.

	// Get new memory page if current page has not enough space:
	if(MemFree <= 0) {
		// Allocate memory page, handle failure:
		page_t page = MemPool.alloc_page();
		if(page == PAGE_NULL) {
			MemErr = true;
			return false;
		}

		// Make new page available:
		MemPage = page;
		MemPtr  = reinterpret_cast<SET_NODE_T>(MemPage);
		MemFree = PAGE_SIZE(set_node_c<T>);
	}

	// Create row object:
	CHECK(MemFree > 0, "set_c::insert: Page too small");
	node = MemPtr++;
	MemFree--;
	node->Elem = elem;
	node->Next = *hashtab_entry;
	if(*hashtab_entry == SET_NODE_NULL)
		EntriesUsed++;
	*hashtab_entry = node;

	// Increase number of rows:
	NumRows++;

	// The row was new (not a duplicate):
	return true;
}


//=============================================================================
// Debugging Support:
//=============================================================================

//-----------------------------------------------------------------------------
// check: Check the integrity of this object, return error message or NULL.
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
template<class T> str_t	set_c<T>::check() const
{
	// Check magic number:
	if(Magic != SET_MAGIC)
		return "wrong magic number";

	// Check superclass:
	str_t msg = rel_c::check();
	if(msg)
		return msg;

	// Check hash table size in entries and pages:
	if(HashSize <= 0)
		return "HashSize is not positive";
	if(HashPages <= 0)
		return "HashPages is not positive";
	if(HashPages > SET_MAXPAGES)
		return "HashPages is greater than limit";
	if(HashPages != (HashSize + (PAGE_SIZE(SET_NODE_T)-1)) /
			PAGE_SIZE(SET_NODE_T))
		return "HashPages is inconsistent with HashSize";

	// MemPtr must point into MemPage:
	if(MemPage != PAGE_NULL &&
		MemPtr < reinterpret_cast<SET_NODE_T>(MemPage))
		return "MemPtr outside bounds (too small)";
	if(MemPage != PAGE_NULL &&
		reinterpret_cast<char*>(MemPtr) >
			reinterpret_cast<char*>(MemPage) + VER_PAGESIZE)
		return "MemPtr outside bounds (too large)";

	// MemPtr can be computed from MemFree:
	if(MemPage == PAGE_NULL && MemFree != 0)
		return "MemPage is null, but MemFree != 0";
	if(MemPage != PAGE_NULL &&
		MemPtr != reinterpret_cast<SET_NODE_T>(MemPage) +
				(PAGE_SIZE(set_node_c<T>) - MemFree))
		return "MemFree is inconsistent with MemPtr";

	// Ok:
	return 0;
}
#endif


//-----------------------------------------------------------------------------
// dump: Show data structure.
//-----------------------------------------------------------------------------

#ifdef	VER_DUMP
template<class T> void set_c<T>::dump(str_t headline) const
{
	// Headline:
	if(headline == STR_NULL)
		headline =
		"Set: Relation with binding pattern b..b for Duplicate Check";
	err_c::dump_begin(headline);

	// Basic information:
	err_c::dump_str("Name:       '");
	err_c::dump_str(Name);
	err_c::dump_str("'");
	err_c::dump_nl();

	err_c::dump_str("NumRows:     ");
	err_c::dump_int(NumRows);
	err_c::dump_nl();

	err_c::dump_str("HashSize:    ");
	err_c::dump_int(HashSize);
	err_c::dump_str(" [");
	err_c::dump_int(HashPages);
	err_c::dump_str(" Pages = ");
	err_c::dump_int(PAGE_KB(HashPages));
	err_c::dump_str(" KB]");
	err_c::dump_nl();

	err_c::dump_str("EntriesUsed: ");
	err_c::dump_int(EntriesUsed);
	err_c::dump_str(" [");
	int percent = (EntriesUsed * 100 + (HashSize/2)) / HashSize;
	err_c::dump_int(percent);
	err_c::dump_str("%]");
	err_c::dump_nl();

	err_c::dump_str("Memory:      ");
	int pages = MemPool.num_alloc_pages();
	err_c::dump_int(pages);
	err_c::dump_str(" Pages [");
	err_c::dump_int(PAGE_KB(pages));
	err_c::dump_str(" KB]");
	err_c::dump_nl();

	err_c::dump_str("NodeSize:    ");
	err_c::dump_uint(static_cast<unsigned int>(sizeof(set_node_c<T>)));
	err_c::dump_str(" Bytes");
	err_c::dump_nl();

	err_c::dump_str("ObjectSize:  ");
	err_c::dump_uint(static_cast<unsigned int>(sizeof(set_c<T>)));
	err_c::dump_str(" Bytes");
	err_c::dump_str(" [includes page array of size MAXPAGES = ");
	err_c::dump_int(SET_MAXPAGES);
	err_c::dump_str("]");
	err_c::dump_nl();

	err_c::dump_nl();

	// Show entire hash table with chains of entries:
	err_c::dump_str("Hash Table:");
	err_c::dump_nl();
	for(int i = 0; i < HashPages; i++) {
		err_c::dump_str("  Page ");
		err_c::dump_str(str_fill(i, 4));
		err_c::dump_int(i);
		err_c::dump_str(":");
		if(HashTab[i] == static_cast<SET_NODE_T*>(0)) {
			err_c::dump_str("*** NULL ***");
			err_c::dump_nl();
			continue;
		}
		err_c::dump_nl();
		for(int j = 0; j < PAGE_SIZE(SET_NODE_T); j++) {
			SET_NODE_T node = HashTab[i][j];
			if(node != SET_NODE_NULL) {
				err_c::dump_str("    ");
				err_c::dump_str(str_fill(j, 4));
				err_c::dump_int(j);
				err_c::dump_str(":");
				for(; node != SET_NODE_NULL;
							node = node->Next) {
					char buf[SET_MAX_ELEM_CHARS];
					node->Elem.to_string(buf,
							SET_MAX_ELEM_CHARS);
					err_c::dump_str(buf);
					err_c::dump_str(" ");
				}
				err_c::dump_nl();
			}
		}
		err_c::dump_nl();
	}

	// Done:
	err_c::dump_end();
}
#endif

