// ============================================================================
// Project:	Deductive Database
// Filename:	set_e.cpp
// Purpose:	Set of tuples (binding pattern b..b): Using Extendible Hashing
// Last Change:	26.09.2016
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2016 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"
#include "str.h"
#include "err.h"
#include "check.h"

#include "set_e.h"


//=============================================================================
// Private Constants:
//=============================================================================

//-----------------------------------------------------------------------------
// SET_E_MAX_ELEM_CHARS: Size of string buffer for printing element row.
//-----------------------------------------------------------------------------

static const int SET_E_MAX_ELEM_CHARS = 20;


//=============================================================================
// Constructor and Destructor:
//=============================================================================


//-----------------------------------------------------------------------------
// Constructor:
//-----------------------------------------------------------------------------

template<class T> set_e_c<T>::set_e_c(str_t rel_name) :
				rel_c(rel_name, T::NUM_COLS, 0,
						"Extendible hash table")
{
	// Set Magic number:
	CHECK_CODE(Magic = SET_E_MAGIC);

	// Check parameters:
	CHECK(rel_name != STR_NULL,
			"set_e_c::constructor: rel_name is NULL");

	// Initialize attributes (which are not nested objects):
	GlobalBits  = 0;
	BitMask     = 0;
	HashSize    = 0;
	HashPages   = 0;
	NumBuckets  = 0;
	NumOverflow = 0;
	MemPage     = PAGE_NULL;
	MemPtr      = SET_E_BUCKET_NULL;
	MemFree     = 0;

	// Initialize page array (not really necessary):
	int n = SET_E_MAXPAGES;
	SET_E_BUCKET_T **p = HashTab;
	while(n-- > 0)
		*p++ = static_cast<SET_E_BUCKET_T*>(0);

	// Initialize Hash table (initial size: one memory page):
	page_t page = MemPool.alloc_page();
	if(page == PAGE_NULL) {
		MemErr = true;
		return;
	}
	SET_E_BUCKET_T *ptr = reinterpret_cast<SET_E_BUCKET_T *>(page);
	HashTab[0] = ptr;

	// We make the initial size one page, all pointing to the same bucket:
	SET_E_BUCKET_T bucket = get_new_bucket(0);
	n = PAGE_SIZE(SET_E_BUCKET_T);
	while(n-- > 0)
		*ptr++ = bucket;
	HashPages = 1;

	// Compute initial global bits:
	HashSize = 1;
	while(HashSize * 2 <= PAGE_SIZE(SET_E_BUCKET_T)) {
		HashSize = HashSize * 2;
		GlobalBits++;
		BitMask = (BitMask << 1) | 0x1;
	}
	CHECK(HashSize == PAGE_SIZE(SET_E_BUCKET_T),
			"set_e_c::constructor: Page size is no power of 2");
}

//-----------------------------------------------------------------------------
// Destructor:
//-----------------------------------------------------------------------------

template<class T> set_e_c<T>::~set_e_c()
{
	// Check this object:
	CHECK_VALID("set_e_c::destructor");

	// Make this object invalid:
	CHECK_CODE(Magic = 0);
}

//-----------------------------------------------------------------------------
// insert: Insert an element into the set (returns true if successful).
//-----------------------------------------------------------------------------

template<class T> bool set_e_c<T>::insert(T elem)
{
	static int bit_mask[32] = {
		0x000000,
		0x000001,
		0x000003,
		0x000007,
		0x00000F,
		0x00001F,
		0x00003F,
		0x00007F,
		0x0000FF,
		0x0001FF,
		0x0003FF,
		0x0007FF,
		0x000FFF,
		0x001FFF,
		0x003FFF,
		0x007FFF,
		0x00FFFF,
		0x01FFFF,
		0x03FFFF,
		0x07FFFF,
		0x0FFFFF,
		0x1FFFFF,
		0x3FFFFF,
		0x7FFFFF,
		0xFFFFFF,
	};

	// Check this object:
	CHECK_VALID("set_e_c::insert");

	// If there was a memory allocation error, we don't try again:
	if(MemErr)
		return false;

	// Search in hash table. First compute hash function:
	int hash_val = elem.hash();
	CHECK(hash_val >= 0, "set_e_c::insert: hash_val is negative");

	// Get hash table entry:
	int n = hash_val & BitMask;
	SET_E_BUCKET_T *hashtab_entry = HashTab[n / PAGE_SIZE(SET_E_BUCKET_T)];
	hashtab_entry += (n % PAGE_SIZE(SET_E_BUCKET_T));

	// There are no null entries on the hash table:
	SET_E_BUCKET_T bucket = *hashtab_entry;
	CHECK(bucket != SET_E_BUCKET_NULL,
			"set_e_c::insert: Null entry in hash table");

	// Now search the linked list of overflow buckets in that entry:
	SET_E_BUCKET_T last = SET_E_BUCKET_NULL;
	for(; bucket != SET_E_BUCKET_NULL; bucket = bucket->Overflow) {
		T *elem_ptr = bucket->Elems;
		for(int i = 0; i < bucket->NumElems; i++) {
			if(*elem_ptr++ == elem)
				// Found, i.e. insertion not possible:
				return false;
		}

		// Simple insertion. Bucket has still space:
		if(bucket->NumElems < SET_E_BUCKETSIZE) {
			CHECK(bucket->Overflow == SET_E_BUCKET_NULL,
				"set_e_c::insert: Overflow, bucket not full");
			*elem_ptr = elem;
			bucket->NumElems++;
			NumRows++;
			return true;
		}

		// Remember the last element of the overflow chain:
		last = bucket;
	}
	bucket = last;

	// We are sure that bucket is not null here:
	CHECK(bucket != SET_E_BUCKET_NULL,
			"set_e_c::insert: bucket is null");

	// Since elem was not found, bucket is last on the chain:
	CHECK(bucket->Overflow == SET_E_BUCKET_NULL,
			"set_e_c::insert: bucket not last chain elem");

	// If it was not already inserted, the bucket is full:
	CHECK(bucket->NumElems == SET_E_BUCKETSIZE,
			"set_e_c::insert: bucket not full");

	// If this is already an overflow chain, no split is possible:
	CHECK(bucket == *hashtab_entry || bucket->LocalBits == GlobalBits,
			"set_e_c::insert: Overflow chain, but split possible");

	// Also, the hash table size has reached its maximum:
	CHECK(bucket == *hashtab_entry || HashPages * 2 > SET_E_MAXPAGES,
			"set_e_c::insert: Overflow chain, but htab can grow");

	// Let's split the bucket as often as necessary:
	for(;;) {

		// If local bits are equal to global: Double entire table:
		if(bucket->LocalBits == GlobalBits) {

			// If not possible, add overflow bucket:
			if(HashPages * 2 > SET_E_MAXPAGES) {

				// Get new bucket:
				SET_E_BUCKET_T new_bucket =
					get_new_bucket(bucket->LocalBits);
				if(new_bucket == SET_E_BUCKET_NULL)
					return false;

				// Add to overflow chain:
				bucket->Overflow = new_bucket;
				NumOverflow++;

				// Enter element:
				new_bucket->NumElems = 1;
				new_bucket->Elems[0] = elem;
				NumRows++;
				return true;
			}

			// Double table:
			for(int i = 0; i < HashPages; i++) {

				// Allocate memory page, handle failure:
				page_t page = MemPool.alloc_page();
				if(page == PAGE_NULL) {
					MemErr = true;
					return false;
				}

				// Enter new page in page table:
				SET_E_BUCKET_T *to =
					reinterpret_cast<SET_E_BUCKET_T*>(page);
				HashTab[HashPages+i] = to;

				// Copy page contents (links to buckets):
				SET_E_BUCKET_T* from = HashTab[i];
				for(int j = PAGE_SIZE(SET_E_BUCKET_T); --j>=0;)
					*to++ = *from++;
			}
			HashPages = HashPages * 2;
			HashSize  = HashSize  * 2;
			GlobalBits++;
			BitMask = (BitMask << 1) | 0x1;
		}

		// Now we are sure that LocalBits < GlobalBits:
		CHECK(bucket->LocalBits < GlobalBits,
				"set_e_c::insert: LocalBits too large");

		// Split bucket:
		int next_bit = 0x1 << (bucket->LocalBits);
		int base = hash_val & bit_mask[bucket->LocalBits];
		int dist = next_bit * 2;
		bucket->LocalBits++;
		SET_E_BUCKET_T new_bucket = get_new_bucket(bucket->LocalBits);
		if(new_bucket == SET_E_BUCKET_NULL) {
			MemErr = true;
			return false;
		}
		SET_E_BUCKET_T upper_bucket = bucket;
		SET_E_BUCKET_T lower_bucket = new_bucket;
		for(int i = base; i < HashSize; i += dist)
			set_hashtab_entry(i, lower_bucket);

		// Distribute bucket elements into upper and lower bucket:
		int num_upper = 0;
		int num_lower = 0;
		for(int i = 0; i < bucket->NumElems; i++) {
			if(bucket->Elems[i].hash() & next_bit)
				upper_bucket->Elems[num_upper++] =
					bucket->Elems[i];
			else
				lower_bucket->Elems[num_lower++] =
					bucket->Elems[i];
		}
		upper_bucket->NumElems = num_upper;
		lower_bucket->NumElems = num_lower;

		// Try again to insert:
		if(hash_val & next_bit)
			bucket = upper_bucket;
		else
			bucket = lower_bucket;
		if(bucket->NumElems < SET_E_BUCKETSIZE) {
			bucket->Elems[bucket->NumElems++] = elem;
			NumRows++;
			return true;
		}
	}
}


//=============================================================================
// Debugging Support:
//=============================================================================

//-----------------------------------------------------------------------------
// check: Check the integrity of this object, return error message or NULL.
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
template<class T> str_t	set_e_c<T>::check() const
{
	// Check magic number:
	if(Magic != SET_E_MAGIC)
		return "wrong magic number";

	// Check superclass:
	str_t msg = rel_c::check();
	if(msg)
		return msg;

	// Check hash table size in entries and pages:
	if(HashSize <= 0)
		return "HashSize is not positive";
	if(HashPages <= 0)
		return "HashPages is not positive";
	if(HashPages > SET_E_MAXPAGES)
		return "HashPages is greater than limit";
	if(HashPages != (HashSize + (PAGE_SIZE(SET_E_BUCKET_T)-1)) /
			PAGE_SIZE(SET_E_BUCKET_T))
		return "HashPages is inconsistent with HashSize";

	// MemPtr must point into MemPage:
	if(MemPage != PAGE_NULL &&
		MemPtr < reinterpret_cast<SET_E_BUCKET_T>(MemPage))
		return "MemPtr outside bounds (too small)";
	if(MemPage != PAGE_NULL &&
		reinterpret_cast<char*>(MemPtr) >
			reinterpret_cast<char*>(MemPage) + VER_PAGESIZE)
		return "MemPtr outside bounds (too large)";

	// MemPtr can be computed from MemFree:
	if(MemPage == PAGE_NULL && MemFree != 0)
		return "MemPage is null, but MemFree != 0";
	if(MemPage != PAGE_NULL &&
		MemPtr != reinterpret_cast<SET_E_BUCKET_T>(MemPage) +
				(PAGE_SIZE(set_e_bucket_c<T>) - MemFree))
		return "MemFree is inconsistent with MemPtr";

	// Ok:
	return 0;
}
#endif


//-----------------------------------------------------------------------------
// dump: Show data structure.
//-----------------------------------------------------------------------------

#ifdef	VER_DUMP
template<class T> void set_e_c<T>::dump(str_t headline) const
{
	// Headline:
	if(headline == STR_NULL)
		headline =
		"Set: Relation with binding pattern b..b for Duplicate Check";
	err_c::dump_begin(headline);

	// Basic information:
	err_c::dump_str("Name:       '");
	err_c::dump_str(Name);
	err_c::dump_str("'");
	err_c::dump_nl();

	err_c::dump_str("NumRows:     ");
	err_c::dump_int(NumRows);
	err_c::dump_nl();

	err_c::dump_str("HashSize:    ");
	err_c::dump_int(HashSize);
	err_c::dump_str(" [");
	err_c::dump_int(HashPages);
	err_c::dump_str(" Pages = ");
	err_c::dump_int(PAGE_KB(HashPages));
	err_c::dump_str(" KB]");
	err_c::dump_nl();

	err_c::dump_str("GlobalBits:  ");
	err_c::dump_int(GlobalBits);
	err_c::dump_nl();

	err_c::dump_str("BitMask:     ");
	err_c::dump_int(BitMask);
	err_c::dump_nl();

	err_c::dump_str("NumBuckets:  ");
	err_c::dump_int(NumBuckets);
	err_c::dump_nl();

	err_c::dump_str("NumOverflow: ");
	err_c::dump_int(NumOverflow);
	err_c::dump_nl();

	err_c::dump_str("Memory:      ");
	int pages = MemPool.num_alloc_pages();
	err_c::dump_int(pages);
	err_c::dump_str(" Pages [");
	err_c::dump_int(PAGE_KB(pages));
	err_c::dump_str(" KB]");
	err_c::dump_nl();

	err_c::dump_str("BucketSize:  ");
	err_c::dump_uint(static_cast<unsigned int>(sizeof(set_e_bucket_c<T>)));
	err_c::dump_str(" Bytes");
	err_c::dump_nl();

	err_c::dump_str("ObjectSize:  ");
	err_c::dump_uint(static_cast<unsigned int>(sizeof(set_e_c<T>)));
	err_c::dump_str(" Bytes");
	err_c::dump_str(" [includes page array of size MAXPAGES = ");
	err_c::dump_int(SET_E_MAXPAGES);
	err_c::dump_str("]");
	err_c::dump_nl();

	err_c::dump_nl();

	// Show entire hash table with buckets:
	err_c::dump_str("Hash Table:");
	err_c::dump_nl();
	for(int i = 0; i < HashPages; i++) {
		err_c::dump_str("  Page ");
		err_c::dump_str(str_fill(i, 4));
		err_c::dump_int(i);
		err_c::dump_str(":");
		if(HashTab[i] == static_cast<SET_E_BUCKET_T*>(0)) {
			err_c::dump_str("*** NULL ***");
			err_c::dump_nl();
			continue;
		}
		err_c::dump_nl();
		for(int j = 0; j < PAGE_SIZE(SET_E_BUCKET_T); j++) {
			SET_E_BUCKET_T bucket = HashTab[i][j];
			if(bucket != SET_E_BUCKET_NULL) {
				err_c::dump_str("    ");
				err_c::dump_str(str_fill(j, 4));
				err_c::dump_int(j);
				err_c::dump_str(" [");
				err_c::dump_int(bucket->LocalBits);
				err_c::dump_str("]:");
				for(int k = 0; k < bucket->NumElems; k++) {
					char buf[SET_E_MAX_ELEM_CHARS];
					bucket->Elems[k].to_string(buf,
							SET_E_MAX_ELEM_CHARS);
					err_c::dump_str(buf);
					err_c::dump_str(" ");
				}
				for(bucket = bucket->Overflow;
						bucket != SET_E_BUCKET_NULL;
						bucket = bucket->Overflow) {
					err_c::dump_nl();
					err_c::dump_str("*** OVERFLOW *** ");
					for(int k = 0; k < bucket->NumElems;
									k++) {
						char buf[SET_E_MAX_ELEM_CHARS];
						bucket->Elems[k].to_string(buf,
							SET_E_MAX_ELEM_CHARS);
						err_c::dump_str(buf);
						err_c::dump_str(" ");
					}
				}
				err_c::dump_nl();
			}
		}
		err_c::dump_nl();
	}

	// Done:
	err_c::dump_end();
}
#endif

