// ============================================================================
// Project:	Deductive Database
// Filename:	set_e.h
// Purpose:	Set of tuples (binding pattern b..b): Using Extendible Hashing
// Last Change:	31.10.2016
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2016 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Include File Frame:
//=============================================================================

#ifndef	SET_E_INCLUDED
#define SET_E_INCLUDED

//=============================================================================
// Used Types and Macros:
//=============================================================================

#ifndef VER_INCLUDED
#include "ver.h"
#endif

#ifndef STR_INCLUDED
#include "str.h"
#endif

#ifndef CHECK_INCLUDED
#include "check.h"
#endif

#ifndef PAGE_INCLUDED
#include "page.h"
#endif

#ifndef MPOOL_INCLUDED
#include "mpool.h"
#endif

#ifndef RELSIZE_INCLUDED
#include "relsize.h"
#endif

#ifndef REL_INCLUDED
#include "rel.h"
#endif


//=============================================================================
// Private Constants:
//=============================================================================

//-----------------------------------------------------------------------------
// SET_E_MAXPAGES: Maximal hash table size in memory pages.
//-----------------------------------------------------------------------------

static const int SET_E_MAXPAGES = 1024;

	// Since the hash table size is doubled when necessary,
	// SET_E_MAXPAGES should be a power of 2
	// (assuming that a memory page contains a power of 2 of pointers).
	// For 4 KB pages and 64 bit pointers, one page contains 512 pointers,
	// so the total size of the hash table is SET_E_MAXPAGES * 512.

	// The number of memory pages used for buckets is unlimited.
	// If unavoidable, overflow buckets are linked to a bucket.
	// So there is no hard limit on the number of entries in the hash table
	// (but the performance decreases, of course).

//-----------------------------------------------------------------------------
// SET_E_BUCKETSIZE: Number of entries that fit into one bucket.
//-----------------------------------------------------------------------------

static const int SET_E_BUCKETSIZE = 16;

//-----------------------------------------------------------------------------
// SET_E_MAGIC: Magic number (identifies objects of this class).
//-----------------------------------------------------------------------------

static const long SET_E_MAGIC = 0x5354450AL;	// 'STE\n'


//=============================================================================
// Private Class for the Buckets in the Hash Table:
//=============================================================================


template<class T> class set_e_bucket_c {
	public:
		int NumElems;
		int LocalBits;
		set_e_bucket_c<T> *Overflow;
		T Elems[SET_E_BUCKETSIZE];
};

#define SET_E_BUCKET_T		set_e_bucket_c<T>*
#define	SET_E_BUCKET_NULL	(static_cast<set_e_bucket_c<T> *>(0))


//=============================================================================
// Class Template for Sets (for Duplicate Check):
//=============================================================================

/** Implementation of a set data structure with an extendible hash table.
 *
 * This is an implementation of a set data structure, i.e. a relation that
 * is accessed with the binding pattern "bb...b" (all arguments bound).
 * The set data structure mainly offers an {@code insert} method for
 * inserting a tuple into the relation, and a {@code contains} method
 * for checking whether a given tuple is contained in the set.
 * The {@code insert} method also contains an element check:
 * It returns {@code false} if the tuple cannot be inserted because it is
 * already contained in the set, and {@code true} if the insertion was
 * successful. There is no method for deleting elements.
 *
 * This implementation uses an extendible hash table.
 *
 * Note that any reliable implementation must handle special cases like
 * several tuples returning the same hash value (more than fitting into
 * one bucket) or the hash table cannot be extended further.
 * Our implementation does have a strict upper limit on the hash table size.
 * Under bad circumstances, it uses a chain of overflow buckets.
 * Overflow buckets are used only as a last resort, when the hash table
 * has already grown to its maximal size and no further splitting of the
 * bucket is possible. Therefore, once a hash table entry does contain
 * a chain of overflow buckets, we do not need to program any further
 * splitting of it, but simply add more buckets to the chain.
 * We are out of luck in this case anyway.
 * On the other hand, we do have to split the same bucket iteratively,
 * if a single split did not yield a better distribution of the elements
 * (this should be a very uncommon case, though).
 *
 * The alternative would have been to split only once per insertion,
 * and use an overflow bucket if it does not work.
 * When the next element is inserted, we try again to split.
 * In this case there could be overflow chains that are removed
 * after another insertion. This is NOT the way we do it here.
 *
 * Note that in both cases empty buckets are possible if the split
 * does not lead to a redistribution of entries.
 * However, only a logarithmic number of them can be produced in a single
 * insertion. We cannot simply remove the empty bucket and replace it
 * by a null pointer because the information about the local bit depth
 * is important.
 *
 * Of course, all this are border cases, which occur seldom.
 * However, the code has to handle them.
 *
 */

template<class T> class set_e_c: public rel_c {
	public:

//-----------------------------------------------------------------------------
// Constructor, Destructor:
//-----------------------------------------------------------------------------

	// Constructor:
	set_e_c(str_t set_e_name);

	// Destructor:
	~set_e_c();

//-----------------------------------------------------------------------------
// Copy-constructor and assignment operator are not supported for this class:
//-----------------------------------------------------------------------------

	private:

	set_e_c(const set_e_c& obj);		// Not implemented
	set_e_c& operator=(const set_e_c& obj);	// Not implemented


//=============================================================================
// (Object) Methods:
//=============================================================================

	public:

//-----------------------------------------------------------------------------
// insert: Insert an element into the set (returns true if successful).
//-----------------------------------------------------------------------------

	// This method returns true if the element was not already in the set
	// and was successfully inserted (i.e. if it is not a duplicate).
	// It returns false if the element was already in the set (duplicate).
	// It also returns false if it was impossible to insert the element
	// because there was not enough memory.
	// By pretending that it was a duplicate, recursion at least comes
	// to an end and the program terminates.
	// One should check the method mem_err() from time to time
	// (at least at the end) to see whether this problem occurred
	// or the result is reliable.
	// A memory error is also logged in the class err_c.
	// So even if one does not check mem_err() for every set object,
	// one will notice the error if one at least checks the global error
	// log or has enabled output of error messages.

	bool insert(T elem);

//-----------------------------------------------------------------------------
// contains: Check whether an element is contained in the set.
//-----------------------------------------------------------------------------

	bool contains(T elem) const {

		// Check this object:
		CHECK_VALID("set_e_c::contains");

		// Search in hash table. First compute hash function:
		int hash_val = elem.hash();
		CHECK(hash_val >= 0, "set_e_c::contains: hash_val is negative");
		hash_val &= BitMask;

		// Get hash table entry (bucket, might contain overflow chain):
		SET_E_BUCKET_T *hashtab_entry =
				HashTab[hash_val / PAGE_SIZE(SET_E_BUCKET_T)];
		hashtab_entry += (hash_val % PAGE_SIZE(SET_E_BUCKET_T));

		// Now search the linked list of buckets in that entry:
		for(SET_E_BUCKET_T bucket = *hashtab_entry;
				bucket != SET_E_BUCKET_NULL;
				bucket = bucket->Overflow) {
			T *elem_ptr = bucket->Elems;
			for(int i = 0; i < bucket->NumElems; i++)
				if(*elem_ptr++ == elem)
					return true;
		}
		return false;
	}

//-----------------------------------------------------------------------------
// hash_size: Size of the hash table.
//-----------------------------------------------------------------------------

	int hash_size() const {
		CHECK_VALID("set_e_c::hash_size");
		return HashSize;
	}

//-----------------------------------------------------------------------------
// num_buckets: Number of distinct buckets that are used (includes overflow).
//-----------------------------------------------------------------------------

	int num_buckets() const {
		CHECK_VALID("set_e_c::num_buckets");
		return NumBuckets;
	}

//-----------------------------------------------------------------------------
// num_overflow: Number of overflow buckets that are used.
//-----------------------------------------------------------------------------

	int num_overflow() const {
		CHECK_VALID("set_e_c::num_overflow");
		return NumOverflow;
	}

//-----------------------------------------------------------------------------
// size_value: Values of data-structure specific size measures.
//-----------------------------------------------------------------------------

	int size_value(relsize_t relsize) const {
		CHECK_VALID("set_e_c::size_value");
		CHECK(relsize_valid(relsize),
				"set_e_c::size_value: invalid size measure");

		// Return value of selected size measure:
		switch(relsize) {
			case RELSIZE_OBJ_SIZE:
				return static_cast<int>(sizeof(set_e_c<T>));
			case RELSIZE_HASHTAB_SIZE:
				return HashSize;
			case RELSIZE_BUCKET_SIZE:
				return SET_E_BUCKETSIZE;
			case RELSIZE_BUCKET_BYTES:
				return static_cast<int>(
						sizeof(set_e_bucket_c<T>));
			case RELSIZE_TOTAL_BUCKETS:
				return NumBuckets;
			case RELSIZE_OVERFLOW_BUCKETS:
				return NumOverflow;
			default:
				return -1;
		}
	}

//=============================================================================
// Debugging Support:
//=============================================================================

#ifdef	VER_DEBUG
	public:
	// Integrity check:
	str_t	check() const;

	private:
	// Magic number (identifies objects of this class for debugging).
	long	Magic;		// Must be "SET_E_MAGIC".
#endif

#ifdef	VER_DUMP
	public:
	// Display data structure:
	void dump(str_t headline = STR_NULL) const;
#endif


//=============================================================================
// Private Object Members:
//=============================================================================

	private:

	// Hash table size (in number of bits used):
	int GlobalBits;

	// Corresponding bit mask for filtering part of hash value used:
	int BitMask;

	// Hash table size (in number of entries, redundant):
	int HashSize;

	// Hash table size (in memory pages):
	int HashPages;

	// Number of distinct buckets in hash table (includes overflow ones):
	int NumBuckets;

	// Number of overflow buckets:
	int NumOverflow;

	// Current Memory Page for Storing Bucket Objects:
	page_t MemPage;

	// Pointer to next free place in current memory page:
	set_e_bucket_c<T> *MemPtr;

	// Number of free places in current page:
	int MemFree;

	// Hash table:
	set_e_bucket_c<T> **HashTab[SET_E_MAXPAGES];


//=============================================================================
// Auxiliary Functions:
//=============================================================================

	private:

	//---------------------------------------------------------------------
	// Get new hashtable bucket:
	//---------------------------------------------------------------------

	inline SET_E_BUCKET_T get_new_bucket(int bits)
	{
		// Get new memory page if current page has not enough space:
		if(MemFree <= 0) {
			// Allocate memory page, handle failure:
			page_t page = MemPool.alloc_page();
			if(page == PAGE_NULL) {
				MemErr = true;
				return SET_E_BUCKET_NULL;
			}

			// Make new page available:
			MemPage = page;
			MemPtr  = reinterpret_cast<SET_E_BUCKET_T>(MemPage);
			MemFree = PAGE_SIZE(set_e_bucket_c<T>);
		}

		// Cut new hashtable bucket from memory page:
		CHECK(MemFree > 0, "set_e_c::get_new_bucket: Page too small");
		SET_E_BUCKET_T bucket = MemPtr++;
		MemFree--;

		// Increase number of buckets used (for statistics):
		NumBuckets++;

		// Initialize and return new bucket:
		bucket->NumElems = 0;
		bucket->LocalBits = bits;
		bucket->Overflow = SET_E_BUCKET_NULL;
		return bucket;
	}


	//---------------------------------------------------------------------
	// Set a hash table entry:
	//---------------------------------------------------------------------

	inline void set_hashtab_entry(int i, SET_E_BUCKET_T bucket)
	{
		// CHECK parameter:
		CHECK(i >= 0, "set_e_c::set_hashtab_entry: i is negative");
		CHECK(i < HashSize,
				"set_e_c::set_hashtab_entry: i is too large");

		// Determine pointer:
		SET_E_BUCKET_T *hashtab_entry =
			HashTab[i / PAGE_SIZE(SET_E_BUCKET_T)];
		hashtab_entry += (i % PAGE_SIZE(SET_E_BUCKET_T));

		// Set entry:
		*hashtab_entry = bucket;
	}

//=============================================================================
// End of Class Template:
//=============================================================================

};


//=============================================================================
// End of Include File:
//=============================================================================

#endif

