// ============================================================================
// Project:	Deductive Database
// Filename:	set_tt.cpp
// Purpose:	Set with two columns with range [0..1023] (ten bits)
// Last Change:	31.10.2016
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2010-2016 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"
#include "str.h"
#ifdef	VER_DUMP
#include "err.h"
#endif
#include "check.h"
#include "rel.h"

#include "set_tt.h"


//=============================================================================
// Constructor and Destructor:
//=============================================================================

//-----------------------------------------------------------------------------
// Constructor:
//-----------------------------------------------------------------------------

set_tt_c::set_tt_c(str_t rel_name) :
		rel_c(rel_name, 2, 0, "Array of Bitmaps")
{
	// Check parameter:
	CHECK(rel_name != STR_NULL,
			"set_tt_c::constructor: rel_name is NULL");

	// Initialize memory management attributes:
	MemPage     = PAGE_NULL;
	MemPtr      = SET_TT_BITMAP_NULL;
	MemFree     = 0;

	// Initialize Array to null pointers:
	set_tt_bitmap_t *p = Arr;
	for(int n = SET_TT_C1_SIZE - 1; n-- >= 0; )
		*p++ = SET_TT_BITMAP_NULL;

	// Set Magic number:
	CHECK_CODE(Magic = SET_TT_MAGIC);
}

//-----------------------------------------------------------------------------
// Destructor:
//-----------------------------------------------------------------------------

set_tt_c::~set_tt_c()
{
	// Check this object:
	CHECK_VALID("set_tt_c::destructor");

	// Make this object invalid:
	CHECK_CODE(Magic = 0);
}

//-----------------------------------------------------------------------------
// insert: Insert tuple into relation.
//-----------------------------------------------------------------------------

bool set_tt_c::insert(int c1, int c2)
{
	// Check this object:
	CHECK_VALID("set_tt_c::insert");

	// Check parameters:
	CHECK(c1 >= 0, "set_tt_c::insert: c1 is negative");
	CHECK(c2 >= 0, "set_tt_c::insert: c2 is negative");
	CHECK(c1 < SET_TT_C1_SIZE, "set_tt_c::insert: c1 is too large");
	CHECK(c2 < SET_TT_C2_SIZE, "set_tt_c::insert: c2 is too large");

	// Compute positions:
	set_tt_bitmap_t bitmap = Arr[c1];
	int index  = c2 / VER_UINT_BITS;
	int bit_no = c2 % VER_UINT_BITS;

	// Handle easy case (bitmap exists already):
	if(bitmap != SET_TT_BITMAP_NULL) {
		unsigned int *ptr = bitmap->BitMap + index;
		unsigned int  mask = 1 << bit_no;
		if(((*ptr) & mask) == 0) {
			// Row is new, was actually inserted:
			*ptr = (*ptr | mask);
			NumRows++;
			return true;
		}
		else
			return false;
	}

	// If we arrive here, we must create a new bitmap.

	// Get new memory page if current page has not enough space:
	if(MemFree <= 0) {
		// Allocate memory page, handle failure:
		page_t page = MemPool.alloc_page();
		if(page == PAGE_NULL) {
			MemErr = true;
			return false;
		}

		// Make new page available:
		MemPage = page;
		MemPtr  = reinterpret_cast<set_tt_bitmap_t>(MemPage);
		MemFree = PAGE_SIZE(set_tt_bitmap_c);
	}

	// Create new bitmap object:
	CHECK(MemFree > 0, "set_tt_c::insert: Page too small");
	bitmap = MemPtr++;
	MemFree--;

	// Initialize bitmap to 0 bits:
	unsigned int *ptr = bitmap->BitMap;
	for(int n = SET_TT_BITMAP_SIZE - 1; n-- >= 0; )
		*ptr++ = 0;

	// Set the bit for c2 value:
	ptr = bitmap->BitMap + index;
	*ptr = 1 << bit_no;

	// Store the new bitmap:
	Arr[c1] = bitmap;

	// Row was new (actually inserted):
	NumRows++;
	return true;
}

//=============================================================================
// Debugging Support:
//=============================================================================

//-----------------------------------------------------------------------------
// check: Check the integrity of this object, return error message or NULL.
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
str_t	set_tt_c::check() const
{
	// Check magic number:
	if(Magic != SET_TT_MAGIC)
		return "wrong magic number";

	// Check superclass:
	str_t msg = rel_c::check();
	if(msg)
		return msg;

	// Check sub-objects:
	CHECK_SUBOBJ(MemPool, "set_tt_c::check");

	// Ok:
	return 0;
}
#endif


//-----------------------------------------------------------------------------
// dump: Show data structure.
//-----------------------------------------------------------------------------

#ifdef	VER_DUMP
void	set_tt_c::dump(str_t headline) const
{
	// Headline:
	if(headline == STR_NULL)
		headline = "Set (with two int columns, range [0..1023])";
	err_c::dump_begin(headline);

	// Determine number of bitmaps (different entries for c1):
	int num_bitmaps = 0;
	for(int i = 0; i < SET_TT_C1_SIZE; i++)
		if(Arr[i] != SET_TT_BITMAP_NULL)
			num_bitmaps++;

	// Basic information:
	err_c::dump_str("Name:           '");
	err_c::dump_str(Name);
	err_c::dump_str("'");
	err_c::dump_nl();
	err_c::dump_str("NumRows:         ");
	err_c::dump_int(NumRows);
	err_c::dump_nl();
	err_c::dump_str("Memory:          ");
	int pages = MemPool.num_alloc_pages();
	err_c::dump_int(pages);
	err_c::dump_str(" Pages [");
	err_c::dump_int(PAGE_KB(pages));
	err_c::dump_str(" KB]");
	err_c::dump_nl();
	err_c::dump_str("Number of Bitmaps: ");
	err_c::dump_int(num_bitmaps);
	err_c::dump_nl();
	err_c::dump_str("Object Size:     ");
	err_c::dump_uint(static_cast<unsigned int>(sizeof(set_tt_c)));
	err_c::dump_str(" Bytes");
	err_c::dump_nl();
	err_c::dump_nl();

	// Done:
	err_c::dump_end();
}
#endif

