// ============================================================================
// Project:	Deductive Database
// Filename:	str.cpp
// Purpose:	The String Datatype
// Last Change:	11.09.2015
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 1998-2015 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"

#include "str.h"

#include <string.h>

//=============================================================================
// Exported Functions:
//=============================================================================

//-----------------------------------------------------------------------------
// str_len: Determine string length.
//-----------------------------------------------------------------------------

int	str_len(register str_t str)
{
	if(str == 0)
		return 0;

	register int n = 0;
	while(*str++)
		n++;
	return n;
}

//-----------------------------------------------------------------------------
// str_cpy: Secure string copy.
//-----------------------------------------------------------------------------

void	str_cpy(register char* buf, int len, register str_t str)
{
	register int i;

	// Avoid crashes:
	if(buf == 0)
		return;
	else if(str == 0) {
		if(len >= 6) {
			*buf++ = '<';
			*buf++ = 'N';
			*buf++ = 'I';
			*buf++ = 'L';
			*buf++ = '>';
			*buf = '\0';
		}
		else if(len >= 1) {
			*buf = '\0';
		}
	}

	// Copy as much as possible:
	i = len;
	while(--i > 0 && *str)
		*buf++ = *str++;
	*buf = '\0';

	// Mark incomplete copy:
	if(*str && len >= 4) {
		buf[-3] = '.';
		buf[-2] = '.';
		buf[-1] = '.';
	}
}

//-----------------------------------------------------------------------------
// str_cat: Secure string concatenation.
//-----------------------------------------------------------------------------

void	str_cat(register char* buf, int len, register str_t str)
{
	// Avoid crashes:
	if(buf == 0)
		return;

	// Find end, decrement length:
	while(*buf && len > 0) {
		buf++;
		len--;
	}

	// Impossible: Buffer contents longer than buffer length:
	if(*buf)
		return;

	// Call str_cpy:
	str_cpy(buf, len, str);
}

//-----------------------------------------------------------------------------
// str_eq: Compare two strings for equality.
//-----------------------------------------------------------------------------

bool	str_eq(register str_t str1, register str_t str2)
{
	// Avoid crashes:
	if(str1 == STR_NULL)
		return(str2 == STR_NULL);
	else if(str2 == STR_NULL)
		return(false);

	// Compare strings:
	while(*str1 && *str1==*str2) {
		str1++;
		str2++;
	}
	return(*str1==*str2);
}

//-----------------------------------------------------------------------------
// str_contains: Check whether character appears in string.
//-----------------------------------------------------------------------------

bool	str_contains(register char c, register str_t str)
{
	// Avoid crashes:
	if(str == 0)
		return(false);

	// Search character:
	while(*str && *str != c)
		str++;

	// Was the character found?
	return(*str == c);
}

//-----------------------------------------------------------------------------
// str_indent: Returns string of spaces of a given length (at most 75):
//-----------------------------------------------------------------------------

str_t str_indent(int n)
{
	// String consisting of 75 spaces:
	static const str_t s =
"                                                                           ";

	// Negative lengths might occur when the field is too short:
	if(n < 0)
		return "";

	// The maximal length is 75. If more is requested, we return 75 spaces:
	if(n > 75)
		return s;

	// Normal case. n is between 0 and 75 inclusive:
	return s + (75 - n);
}

//-----------------------------------------------------------------------------
// str_error: System Error Message for an Error Number.
//-----------------------------------------------------------------------------

void	str_error(int err_no, char *buf, int buf_size)
{

#ifdef VER_WINDOWS
	if(strerror_s(buf, buf_size, err_no) != 0)
		str_cpy(buf, buf_size, "Unknown error");
#else
	str_t msg = strerror(err_no);
	if(msg)
		str_cpy(buf, buf_size, msg);
	else
		str_cpy(buf, buf_size, "Unknown error");

	// Does not work with GNU-C, but it is an XSI-compliant function:
	// if(strerror_r(err_no, buf, buf_size) != 0)
	//	str_cpy(buf, buf_size, "Unknown error");
#endif

}


//-----------------------------------------------------------------------------
// str_uint: Convert unsigned int to text. Returns number of characters.
//-----------------------------------------------------------------------------

int str_uint(char *buf, int buf_len, unsigned int n)
{
	char stack[STR_INT_LEN];
	int i = 0;

	// Check parameters:
	if(buf == 0)
		return -1;
	if(buf_len < 2)
		return -2;

	// Handle special case of 0:
	if(n == 0) {
		buf[0] = '0';
		buf[1] = '\0';
		return 1;
	}

	// Produce digits from the end:
	while(n > 0 && i < STR_INT_LEN && i < buf_len) {
		stack[i++] = "0123456789"[n % 10];
		n = n / 10;
	}

	// Handle overflow:
	if(n != 0) {
		buf[0] = '*';
		buf[1] = '\0';
		return 1;
	}

	// Copy digits to buffer:
	int j = 0;
	while(i > 0 && j < buf_len)
		buf[j++] = stack[--i];

	// Mark end of string, handle overflow:
	if(j < buf_len)
		buf[j] = '\0';
	else {
		buf[0] = '*';
		buf[1] = '\0';
		return 1;
	}

	// Ok, return number of characters written:
	return j;
}

//-----------------------------------------------------------------------------
// str_int: Convert int to text. Returns number of characters.
//-----------------------------------------------------------------------------

int str_int(char *buf, int buf_len, int n)
{
	// Check parameters:
	if(buf == 0)
		return -1;
	if(buf_len < 2)
		return -2;

	// Use conversion for unsigned ints:
	if(n >= 0)
		return str_uint(buf, buf_len, static_cast<unsigned int>(n));
	else {
		// Make number positive:
		n = -n; // Sorry. Does not work for minimal number.
		if(n < 0) // At least we check for this case.
			return -3;

		// Treat case of a buffer of size 2 (certainly too small):
		if(buf_len == 2) {
			buf[0] = '*';
			buf[1] = '\0';
			return 1;
		}

		// Now write the sign and call str_uint:
		buf[0] = '-';
		return str_uint(buf+1, buf_len-1,
			static_cast<unsigned int>(n));
	}
}


//-----------------------------------------------------------------------------
// str_fill: Returns string of spaces to make integer a given width.
//-----------------------------------------------------------------------------

str_t str_fill(int n, int field_width)
{
	// Determine print width of n in decimal notation:
	int real_width = 1;
	if(n < 0) {
		real_width++;
		n = -n;
		if(n < 0) {
			// Minimal integer
			n = n + 1;
			n = -n;
		}
	}
	while(n >= 10) {
		n /= 10;
		real_width++;
		if(real_width >= 10)
			break;
	}

	// Determine width of indent string:
	int indent_width = field_width - real_width;
	if(indent_width < 0)
		indent_width = 0;

	// Return string:
	return str_indent(indent_width);
}

//-----------------------------------------------------------------------------
// str_ptr: Convert pointer to text (hex number). Returns number of characters.
//-----------------------------------------------------------------------------

int str_ptr(char *buf, int buf_len, const void *ptr)
{
	// Check parameters:
	if(buf == 0)
		return -1;
	if(buf_len < 2)
		return -2;

	// Preparation for overflows:
	int n = buf_len;
	char *p = buf;
	bool overflow = false;

	// Determine the number of digits:
	VER_INT64 addr = reinterpret_cast<VER_INT64>(ptr);
	int digits = 16;
	if(addr >> 32 == 0)
		digits = 8;

	// Hex conversion:
	for(int shift = (digits - 1) * 4; shift >= 0; shift -= 4) {
		if(n-- > 0)
			*p++ = "0123456789ABCDEF"[(addr >> shift) & 0xF];
		else
			overflow = true;
	}
	*p = '\0';

	// Clearly mark buffer overflow:
	if(overflow) {
		for(int i = 0; i < buf_len; i++)
			buf[i] = '*';
		buf[buf_len-1] = '\0';
		return buf_len-1;
	}

	// Otherwise the number of characters written is digits:
	return digits;
}

