// ============================================================================
// Project:	Deductive Database
// Filename:	strmem.cpp
// Purpose:	String Memory Pool (Permits to Save/Copy Strings in own Memory)
// Last Change:	07.09.2015
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2015 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"
#include "str.h"
#include "check.h"
#ifdef	VER_DUMP
#include "err.h"
#endif

#include "strmem.h"

//=============================================================================
// Constructor and Destructor:
//=============================================================================


//-----------------------------------------------------------------------------
// Constructor:
//-----------------------------------------------------------------------------

strmem_c::strmem_c(mpool_t mem_pool)
{
	// Set Magic number:
	CHECK_CODE(Magic = STRMEM_MAGIC);

	// Initialize attributes (which are not nested objects):
	MemPage    = PAGE_NULL;
	FreePtr    = static_cast<char *>(0);
	FreeBytes  = 0;
	NumPages   = 0;
	NumStrings = 0;
	MemPool    = mem_pool;
}

//-----------------------------------------------------------------------------
// Destructor:
//-----------------------------------------------------------------------------

strmem_c::~strmem_c()
{
	// Check this object:
	CHECK_VALID("strmem_c::destructor");

	// Make this object invalid:
	CHECK_CODE(Magic = 0);
}


//=============================================================================
// Public Methods:
//=============================================================================


//-----------------------------------------------------------------------------
// insert: Insert a string into this string memory pool, return address.
//-----------------------------------------------------------------------------


str_t strmem_c::insert(str_t start, str_t end)
{
	// Check this object:
	CHECK_VALID("strmem_c::insert");

	// Check parameters:
	CHECK(start != STR_NULL, "strmem_c::insert: start is NULL");
	CHECK(end != STR_NULL, "strmem_c::insert: end is NULL");
	CHECK(start < end, "strmem_c::insert: start not before end");

	// Get new memory page if current page has not enough space:
	int length = end - start;	// end points just behind string
	if(length >= FreeBytes) {	// We need additional null byte
		// Allocate memory page, handle failure:
		page_t page = MemPool->alloc_page();
		if(page == PAGE_NULL)
			return STR_NULL;
		NumPages++;

		// Make new page available:
		MemPage = page;
		FreePtr = reinterpret_cast<char *>(MemPage);
		FreeBytes = VER_PAGESIZE;
	}

	// Copy string:
	CHECK(length < FreeBytes, "strmem_c::insert: String too long");
	str_t copied_str = FreePtr;
	for(str_t p = start; p < end; )
		*FreePtr++ = *p++;
	*FreePtr++ = '\0';
	FreeBytes -= (length + 1);

	// For statistics:
	NumStrings++;

	// Return address of string copy:
	return copied_str;
}


//=============================================================================
// Debugging Support:
//=============================================================================


//-----------------------------------------------------------------------------
// check: Check the integrity of this object, return error message or NULL.
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
str_t	strmem_c::check() const
{
	// Check magic number:
	if(Magic != STRMEM_MAGIC)
		return "wrong magic number";

	// FreePtr must point into MemPage:
	if(MemPage != PAGE_NULL &&
		FreePtr < reinterpret_cast<char*>(MemPage))
		return "FreePtr outside bounds (too small)";
	if(MemPage != PAGE_NULL &&
		FreePtr > reinterpret_cast<char*>(MemPage) + VER_PAGESIZE)
		return "FreePtr outside bounds (too large)";

	// FreeBytes can be computed from FreePtr:
	if(MemPage == PAGE_NULL && FreeBytes != 0)
		return "MemPage is null, but FreeBytes != 0";
	if(MemPage != PAGE_NULL &&
		FreeBytes != VER_PAGESIZE -
				(FreePtr - reinterpret_cast<char*>(MemPage)))
		return "FreeBytes has wrong value (inconsistent with FreePtr)";

	// Ok:
	return 0;
}

#endif

//-----------------------------------------------------------------------------
// dump: Show data structure.
//-----------------------------------------------------------------------------

#ifdef	VER_DUMP

void strmem_c::dump(str_t headline) const
{
	// Headline:
	if(headline == STR_NULL)
		headline = "String Table";
	err_c::dump_begin(headline);

	// Basic information:
	err_c::dump_str("MemPool Name: '");
	err_c::dump_str(MemPool->name());
	err_c::dump_str("'");
	err_c::dump_nl();
	err_c::dump_str("NumStrings:    ");
	err_c::dump_int(NumStrings);
	err_c::dump_nl();
	err_c::dump_str("NumPages:      ");
	err_c::dump_int(NumPages);
	err_c::dump_str(" [");
	err_c::dump_int((NumPages * VER_PAGESIZE)/1024);
	err_c::dump_str(" KB]");
	err_c::dump_nl();
	err_c::dump_str("MemPage:     ");
	err_c::dump_ptr(MemPage);
	err_c::dump_str(" [Current Memory Page for Copying Strings]");
	err_c::dump_nl();
	err_c::dump_str("FreePtr:     ");
	err_c::dump_ptr(FreePtr);
	err_c::dump_str(" [corresponds to Index: ");
	int ptr_i = static_cast<int>(FreePtr-reinterpret_cast<char*>(MemPage));
	err_c::dump_int(ptr_i);
	err_c::dump_str("]");
	err_c::dump_nl();
	err_c::dump_str("FreeBytes:   ");
	err_c::dump_int(FreeBytes);
	err_c::dump_str(" [Sum with FreePtr Index: ");
	err_c::dump_int(FreeBytes + ptr_i);
	err_c::dump_str("]");
	err_c::dump_nl();
	err_c::dump_str("Object Size: ");
	err_c::dump_uint(static_cast<unsigned int>(sizeof(strmem_c)));
	err_c::dump_str(" Bytes");
	err_c::dump_nl();
	err_c::dump_nl();

	// Display contents of current page:
	err_c::dump_str("Contents of current page:");
	bool first = 0;
	for(str_t p = reinterpret_cast<char*>(MemPage); p < FreePtr; p++) {
		if(first) {
			err_c::dump_str("  '");
			first = false;
		}
		if(*p == '\0') {
			err_c::dump_str("'");
			err_c::dump_nl();
			first = true;
		}
		else if(*p == '\n')
			err_c::dump_str("\\n");
		else if(*p == '\r')
			err_c::dump_str("\\r");
		else if(*p == '\t')
			err_c::dump_str("\\t");
		else if(*p == '\\')
			err_c::dump_str("\\\\");
		else
			err_c::dump_char(*p);
	}
	if(!first) {
			err_c::dump_str(" ** NO STRING TERMINATOR **");
			err_c::dump_nl();
	}

	// Done:
	err_c::dump_end();
}

#endif

