// ============================================================================
// Project:	Deductive Database
// Filename:	strmem.h
// Purpose:	String Memory Pool (Permits to Save/Copy Strings in own Memory)
// Last Change:	07.09.2015
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2015 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Include File Frame:
//=============================================================================

#ifndef	STRMEM_INCLUDED
#define STRMEM_INCLUDED

//=============================================================================
// Used Types and Macros:
//=============================================================================

#ifndef VER_INCLUDED
#include "ver.h"
#endif

#ifndef STR_INCLUDED
#include "str.h"
#endif

#ifndef CHECK_INCLUDED
#include "check.h"
#endif

#ifndef MPOOL_INCLUDED
#include "mpool.h"
#endif


//=============================================================================
// Private Constants:
//=============================================================================

//-----------------------------------------------------------------------------
// STRMEM_MAGIC: Magic number (identifies objects of this class).
//-----------------------------------------------------------------------------

static const long STRMEM_MAGIC = 0x534D450AL;	// 'SME\n'


//=============================================================================
// Class for String Memory Pool (Permits to Save/Copy Strings in OWn Memory)
//=============================================================================

class strmem_c {
	public:


//-----------------------------------------------------------------------------
// Constructor, Destructor:
//-----------------------------------------------------------------------------

	// Constructor:
	explicit strmem_c(mpool_t mem_pool);

	// Destructor:
	~strmem_c();

//-----------------------------------------------------------------------------
// (Object) Methods:
//-----------------------------------------------------------------------------

	// insert: Copy a string into this string memory pool, return address.
	str_t insert(str_t start, str_t end);

	// num_strings: Returns the number of strings stored in this pool.
	int num_strings() const
	{
		return NumStrings;
	}

	// num_pages: Returns the number of memory pages used.
	int num_pages() const
	{
		return NumPages;
	}

#ifdef	VER_DUMP
	// dump: Show data structure.
	void dump(str_t headline = STR_NULL) const;
#endif

//-----------------------------------------------------------------------------
// Debugging Support:
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
	// Integrity check:
	public:
	str_t	check() const;

	// Magic number (identifies objects of this class for debugging).
	private:
	long	Magic;		// Must be "STRMEM_MAGIC".
#endif

//-----------------------------------------------------------------------------
// Copy-constructor and assignment operator are not supported for this class:
//-----------------------------------------------------------------------------

	private:

	strmem_c(const strmem_c& obj);			// Not implemented
	strmem_c& operator=(const strmem_c& obj);	// Not implemented

//-----------------------------------------------------------------------------
// Private Object Members:
//-----------------------------------------------------------------------------

	private:

	// Memory Pool for Allocating Storage Pages:
	mpool_t MemPool;

	// Current Page for Storing String Data:
	page_t MemPage;

	// Pointer to next free byte in current page:
	char *FreePtr;

	// Number of free bytes in current page:
	int FreeBytes;

	// Number of strings in this string memory pool:
	int NumStrings;

	// Number pages used so far:
	int NumPages;

//=============================================================================
// End of Class:
//=============================================================================

};

//-----------------------------------------------------------------------------
// Define pointer type:
//-----------------------------------------------------------------------------

typedef strmem_c *strmem_t;

//-----------------------------------------------------------------------------
// Define null pointer:
//-----------------------------------------------------------------------------

#define STRMEM_NULL (static_cast<strmem_t>(0))

//=============================================================================
// End of Include File:
//=============================================================================

#endif

