// ============================================================================
// Project:	Deductive Database
// Filename:	test.cpp
// Purpose:	List of all Tests/Benchmarks
// Last Change:	21.10.2016
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2016 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"
#include "str.h"
#include "err.h"
#include "check.h"
#include "perf.h"
#include "out.h"

// Domain data structures:
#include "strtab.h"
#include "atomtab.h"

// Relation data structures:
#include "bind.h"
#include "rel.h"
#include "list_2.h"
#include "set_1.h"
#include "set_1e.h"
#include "rel_ii.h"
#include "rel_n_n.h"

// Test of data structures for relations:
#include "rtest.h"
#include "test_set1.h"

// Loader Classes:
#include "load.h"
#include "load_dblp.h"
#include "load_tc.h"

// Benchmark classes:
#include "bench_dblp_1.h"
#include "bench_dblp_2.h"
#include "bench_dblp_3.h"
#include "bench_j1axy_1.h"
#include "bench_j1axy_2.h"
#include "bench_j1axy_3.h"
#include "bench_j1axy_4.h"
#include "bench_j1axy_5.h"
#include "bench_j1axy_6.h"
#include "bench_j1axy_7.h"
#include "bench_j1axy_8.h"
#include "bench_j1axy_9.h"
#include "bench_tcff_1.h"
#include "bench_tcff_2.h"
#include "bench_tcff_3.h"
#include "bench_tcff_4.h"
#include "bench_tcbf_1.h"
#include "bench_tcbf_2.h"
#include "bench_tcbf_4.h"
#include "bench_wine.h"

// This class:
#include "test.h"


//=============================================================================
// Constructor and Destructor:
//=============================================================================

//-----------------------------------------------------------------------------
// Constructor:
//-----------------------------------------------------------------------------

test_c::test_c(str_t test_id, bool (*test_fun)(test_t test),
			int test_arg, str_t test_name)
{
	// Check parameter:
	CHECK(test_id != STR_NULL, "test_c::constructor: test_id is NULL");
	CHECK(test_arg >= 0, "test_c::constructor: test_arg is negative");
	CHECK(test_fun != 0, "test_c::constructor: test_fun is NULL");
	CHECK(test_name != STR_NULL, "test_c::constructor: test_name is NULL");

	// Initialize attributes:
	Id        = test_id;
	Arg       = test_arg;
	Fun       = test_fun;
	Name      = test_name;
	Done      = false;
	Ok        = false;
	Warn      = false;
	TotalTime = -1;
	LoadTime  = -1;
	EvalTime  = -1;

	// Set Magic number:
	CHECK_CODE(Magic = TEST_MAGIC);
}

//-----------------------------------------------------------------------------
// Destructor:
//-----------------------------------------------------------------------------

test_c::~test_c()
{
	// Check this object:
	CHECK_VALID("test_c::destructor");

	// Make this object invalid:
	CHECK_CODE(Magic = 0);
}


//=============================================================================
// Exported Static Methods:
//=============================================================================


//-----------------------------------------------------------------------------
// get_test: Get test with a given ID.
//-----------------------------------------------------------------------------

test_t test_c::get_test(str_t test_id)
{
	// Check parameter:
	CHECK(test_id != STR_NULL, "test_c::get_test: test_id is NULL");

	// Search the test:
	for(test_t test = Tests; test != TEST_NULL && test->Id[0] != '\0' &&
		(test->Id[0] != '.' || test->Id[1] != '\0'); test++) {
		if(str_eq(test->Id, test_id))
			return test;
	}
	return TEST_NULL;
}


//-----------------------------------------------------------------------------
// print_list: Print list of all tests.
//-----------------------------------------------------------------------------

void test_c::print_list()
{
	// Loop over all tests:
	for(test_t test = Tests; test != TEST_NULL && test->Id[0] != '\0' &&
		(test->Id[0] != '.' || test->Id[1] != '\0'); test++) {
		out_c::help_test(test->Id, test->Name);
	}
}

//=============================================================================
// Exported (Object) Methods:
//=============================================================================

//-----------------------------------------------------------------------------
// run: Execute this test.
//-----------------------------------------------------------------------------


bool test_c::run()
{
	// Check this object:
	CHECK_VALID("test_c::run");

	// Print headline:
	out_c::verbose_nl();
	out_c::verbose_headline(Id);
	out_c::verbose_nl();

	// Reset attributes:
	Ok        = true;
	Warn      = false;
	TotalTime = -1;
	LoadTime  = -1;
	EvalTime  = -1;

	// Measure the total execution time:
	perf_c timer;
	timer.start();

	// Run the test:
	bool ok = (this->Fun)(this);

	// Show total execution time:
	timer.stop();
	long realtime = timer.realtime_msec();
	long cputime  = timer.cputime_msec();
	out_c::verbose_nl();
	out_c::verbose_time("Total execution time:  ", realtime, cputime);
	TotalTime = max_time(realtime, cputime);

	// Show the result:
	if(ok && !Ok) {
		out_c::verbose_bad_ok(Id);
		Warn = true;
	}
	else if(ok)
		out_c::verbose_success(Id);
	else
		out_c::verbose_failure(Id);

	// Some extra space after the test output:
	out_c::verbose_nl();

	// Remember Result:
	Done = true;
	Ok   = ok;

	// Print summary line:
	str_t result = "OK";
	if(Warn)
		result = "Warning";
	if(!Ok)
		result = "Failed";
	out_c::summary(Id, LoadTime, EvalTime, TotalTime, result);

	// Return result:
	return ok;
}

//=============================================================================
// Debugging Support:
//=============================================================================

//-----------------------------------------------------------------------------
// check: Check the integrity of this object, return error message or NULL.
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
str_t	test_c::check() const
{
	// Check magic number:
	if(Magic != TEST_MAGIC)
		return "wrong magic number";

	// Check attributes:
	if(Id == STR_NULL)
		CHECK_ERR("id is NULL");
	if(Arg < 0)
		CHECK_ERR("Arg is negative");
	if(Fun == 0)
		CHECK_ERR("Fun is NULL");
	if(Name == STR_NULL)
		CHECK_ERR("Name is NULL");

	// Ok:
	return 0;
}
#endif


//-----------------------------------------------------------------------------
// dump: Show data structure.
//-----------------------------------------------------------------------------

#ifdef	VER_DUMP
void	test_c::dump(str_t headline) const
{
	// Insert default headline:
	if(headline == STR_NULL)
		headline = "Test Object";

	// Headline:
	err_c::dump_begin(headline);

	// Dump Id:
	err_c::dump_str("Id:   '");
	err_c::dump_str(Id);
	err_c::dump_str("'");
	err_c::dump_nl();

	// Dump Argument:
	err_c::dump_str("Arg:   ");
	err_c::dump_str(Id);
	err_c::dump_nl();

	// Dump Function:
	//err_c::dump_str("Fun:   ");
	//err_c::dump_ptr(reinterpret_cast<void *>(Fun));
	//err_c::dump_nl();
	// ISO C++ forbids casting between pointer-to-function and
	// pointer-to-object (Warning of g++ with -pedantic)

	// Dump Name:
	err_c::dump_str("Name: '");
	err_c::dump_str(Name);
	err_c::dump_str("'");
	err_c::dump_nl();

	// Done:
	err_c::dump_end();
}
#endif

//=============================================================================
// Auxiliary Functions for Performing the Tests:
//=============================================================================

//-----------------------------------------------------------------------------
// run_bench: Run a Benchmark.
//-----------------------------------------------------------------------------

bool test_c::run_bench(bench_t bench)
{
	// Check this object:
	CHECK_VALID("test_c::run_bench");

	// Check parameter:
	CHECK_PAR(bench, "test_c::run_bench");

	// Print benchmark name in verbose mode:
	out_c::verbose_benchmark(bench->name());

	// Get file name selected by parameter:
	str_t file_name = bench->file(Arg);
	if(file_name == STR_NULL) {
		err_c::invalid_test_arg(Id, Arg);
		return false;
	}

	// Print input file name in verbose mode:
	out_c::verbose_input(file_name);

	// Declare timers:
	perf_c timer_load;
	perf_c timer_eval;

	// Load the data:
	out_c::verbose_msg("Loading data ...\n");
	timer_load.start();
	bool load_ok = bench->load(file_name);
	timer_load.stop();
	if(!load_ok) {
		test_failed("Error occurred while loading data");
		return false;
	}
	long load_realtime = timer_load.realtime_msec();
	long load_cputime  = timer_load.cputime_msec();
	out_c::verbose_msg("Loading data finished. ");
	out_c::verbose_time("Time: ", load_realtime, load_cputime);
	LoadTime = max_time(load_realtime, load_cputime);

	// Do the bottom-up evaluation:
	out_c::verbose_msg("Bottom-up evaluation runs ...\n");
	timer_eval.start();
	long eval_result = bench->eval();
	timer_eval.stop();
	if(eval_result < 0) {
		test_failed("Error in bottom-up evaluation");
		return false;
	}
	long eval_realtime = timer_eval.realtime_msec();
	long eval_cputime  = timer_eval.cputime_msec();
	out_c::verbose_msg("Bottom-up evaluation finished. ");
	out_c::verbose_time("Time: ", eval_realtime, eval_cputime);
	EvalTime = max_time(eval_realtime, eval_cputime);

	// Show number of computed answer tuples:
	out_c::verbose_result(eval_result);

	// Check the result if the expected result is known:
	long expected_result = bench->result(Arg);
	if(expected_result < 0) {
		out_c::verbose_cannot_check();
	}
	else if(eval_result != expected_result) {
			wrong_result("eval()", eval_result, expected_result);
			return false;
	}
	else {
		out_c::verbose_bench_ok();
	}

	// Ok:
	out_c::mem_pools_tab();
	return true;
}


//-----------------------------------------------------------------------------
// run_bench_static: Run a Benchmark defined as a static method.
//-----------------------------------------------------------------------------

bool test_c::run_bench_static(str_t name,
		long (*fun)(str_t file, perf_t timer_load, perf_t timer_eval),
		str_t file_name, long expected_result)
{
	// Check this object:
	CHECK_VALID("test_c::run_bench_static");

	// Check parameters:
	CHECK(name != STR_NULL, "test_c::run_bench_static: name is NULL");
	CHECK(fun != 0, "test_c::run_bench_static: function is NULL");

	// Print benchmark name in verbose mode:
	out_c::verbose_benchmark(name);

	// Check and show file name:
	if(file_name == STR_NULL) {
		err_c::invalid_test_arg(Id, Arg);
		return false;
	}
	out_c::verbose_input(file_name);

	// Declare timers:
	perf_c timer_load;
	perf_c timer_eval;

	// Do the bottom-up evaluation:
	out_c::verbose_msg("Running ...\n");
	long result = (*fun)(file_name, &timer_load, &timer_eval);
	if(result < 0) {
		test_failed(
			"Benchmark execution failed: Result is negative.\n");
		return false;
	}
	else
		out_c::verbose_msg("Benchmark execution finished.\n");

	// Show runtime:
	long load_realtime = timer_load.realtime_msec();
	long load_cputime  = timer_load.cputime_msec();
	long eval_realtime = timer_eval.realtime_msec();
	long eval_cputime  = timer_eval.cputime_msec();
	out_c::verbose_time("Loading Time:          ",
				load_realtime, load_cputime);
	LoadTime = max_time(load_realtime, load_cputime);
	out_c::verbose_time("Bottom-Up Eval Time:   ",
				eval_realtime, eval_cputime);
	EvalTime = max_time(eval_realtime, eval_cputime);

	// Show number of computed answer tuples:
	out_c::verbose_result(result);

	// Check the result if the expected result is known:
	if(expected_result < 0) {
		out_c::verbose_cannot_check();
	}
	else if(result != expected_result) {
			wrong_result("eval()", result, expected_result);
			return false;
	}
	else {
		out_c::verbose_bench_ok();
	}

	// Ok:
	return true;
}

//-----------------------------------------------------------------------------
// run_rtest: Run Test of a Data Structure for Storing Relations.
//-----------------------------------------------------------------------------

bool test_c::run_rtest(rtest_t rtest, rel_t rel)
{
	// Check this object:
	CHECK_VALID("test_c::run_rtest");

	// Check parameter:
	CHECK_PAR(rtest, "test_c::run_rtest (rtest)");
	CHECK_PAR(rel, "test_c::run_rtest (rel)");

	// Print benchmark name in verbose mode:
	// out_c::verbose_benchmark(bench->name());

	// Get file name selected by parameter:
	// str_t file_name = bench->file(Arg);
	// if(file_name == STR_NULL) {
	//	err_c::invalid_test_arg(Id, Arg);
	//	return false;
	//}

	// Print input file name in verbose mode:
	//out_c::verbose_input(file_name);

	// Check that there was no memory error in the initialization:
	if(rel->mem_err()) {
		test_failed(
			"Not enough memory for initializing data structure");
		return false;
	}

	// Declare timers:
	perf_c timer_insert;
	perf_c timer_lookup;

	// Insert the data (this is treated like a data load):
	out_c::verbose_msg("Inserting data ...\n");
	timer_insert.start();
	int insert_err = rtest->insert();
	timer_insert.stop();
	if(rel->mem_err()) {
		test_failed("Memory error while inserting data");
		return false;
	}
	if(insert_err != 0) {
		test_failed("Error occurred while inserting data");
		return false;
	}
	if(rel->num_rows() != rtest->num_rows()) {
		test_failed("Relation size differs from inserted rows");
		return false;
	}
	long insert_realtime = timer_insert.realtime_msec();
	long insert_cputime  = timer_insert.cputime_msec();
	out_c::verbose_msg("Inserting data finished. ");
	out_c::verbose_time("Time: ", insert_realtime, insert_cputime);
	LoadTime = max_time(insert_realtime, insert_cputime);
	out_c::verbose_nl();

	// Look up the data (this is treated like the bottom-up evaluation):
	out_c::verbose_msg("Looking up data ...\n");
	timer_lookup.start();
	int lookup_err = rtest->lookup();
	timer_lookup.stop();
	if(rel->mem_err()) {
		test_failed("Memory error while looking up data");
		return false;
	}
	if(lookup_err != 0) {
		test_failed("Error detected while looking up data");
		return false;
	}
	if(rel->num_rows() != rtest->num_rows()) {
		test_failed("Relation size differs from inserted rows");
		return false;
	}
	long lookup_realtime = timer_lookup.realtime_msec();
	long lookup_cputime  = timer_lookup.cputime_msec();
	out_c::verbose_msg("Lookup of data finished. ");
	out_c::verbose_time("Time: ", lookup_realtime, lookup_cputime);
	EvalTime = max_time(lookup_realtime, lookup_cputime);
	out_c::verbose_nl();

	// Print statistics:
	int num_rows     = rtest->num_rows();
	out_c::verbose_num_rows(num_rows);
	int bound_args   = rel->bound_args();
	int free_args    = rel->free_args();
	out_c::verbose_num_cols(bound_args, free_args);
	str_t impl_name  = rel->impl_name();
	out_c::verbose_implementation(impl_name);
	int hash_size    = rtest->size_value(RELSIZE_HASHTAB_SIZE);
	if(hash_size > 0) {
		out_c::verbose_hashtab_size(hash_size);
		int entries_used = rtest->size_value(RELSIZE_ENTRIES_USED);
		if(entries_used >= 0)
			out_c::verbose_hashtab_usage(hash_size, entries_used);
		if(entries_used > 0)
			out_c::verbose_hashtab_chain(num_rows, entries_used);
	}
	int bucket_size  = rtest->size_value(RELSIZE_BUCKET_SIZE);
	int bucket_bytes = rtest->size_value(RELSIZE_BUCKET_BYTES);
	if(bucket_size >= 0)
			out_c::verbose_bucket_size(bucket_size, bucket_bytes);
	int total_buckets = rtest->size_value(RELSIZE_TOTAL_BUCKETS);
	if(total_buckets >= 0)
			out_c::verbose_total_buckets(total_buckets);
	int overflow_buckets = rtest->size_value(RELSIZE_OVERFLOW_BUCKETS);
	if(overflow_buckets >= 0)
			out_c::verbose_overflow_buckets(overflow_buckets);
	if(total_buckets >= 0)
			out_c::verbose_rows_per_bucket(num_rows, total_buckets,
								bucket_size);
	int num_pages = rel->num_pages();
	out_c::verbose_mem_pages(num_pages);
	int obj_size  = rtest->size_value(RELSIZE_OBJ_SIZE);
	if(obj_size >= 0)
		out_c::verbose_obj_size(obj_size);
		out_c::verbose_bytes_per_row(num_pages, obj_size, num_rows);
	//else {
	//	out_c::verbose_bench_ok();
	//}

	// Show memory pools:
	out_c::mem_pools_tab();

	// Ok:
	return true;
}

//-----------------------------------------------------------------------------
// Function to call if argument is not used:
//-----------------------------------------------------------------------------

void test_c::arg_unused()
{
	// Check this object:
	CHECK_VALID("test_c::arg_unused");

	// This test has no argument, the only valid value is 0:
	if(Arg != 0)
		err_c::test_has_no_arg(Id);
}


//-----------------------------------------------------------------------------
// test_failed: An error occurred during the test.
//-----------------------------------------------------------------------------

void test_c::test_failed(str_t msg)
{
	// Check this object:
	CHECK_VALID("test_c::test_failed");

	// Check parameter:
	CHECK(msg != STR_NULL, "test_c::test_failed: msg is NULL");

	// Produce output:
	out_c::verbose_test_failed(msg);

	// Remember failure:
	Ok = false;
}

//-----------------------------------------------------------------------------
// wrong_result: A function called in the test returned a wrong result.
//-----------------------------------------------------------------------------

void test_c::wrong_result(str_t fun, long result, long expected_result)
{
	// Check this object:
	CHECK_VALID("test_c::wrong_result");

	// Check parameter:
	CHECK(fun != STR_NULL, "test_c::test_failed: fun is NULL");

	// Produce output:
	out_c::verbose_wrong_result(fun, result, expected_result);

	// Remember failure:
	Ok = false;
}


//-----------------------------------------------------------------------------
// max_time: Return the larger of realtime+cputime, check plausibility.
//-----------------------------------------------------------------------------

long test_c::max_time(long realtime, long cputime) {

	// The normal case is that realtime is slightly more than CPU-time:
	if(realtime >= cputime) {
		// 10ms absolute difference or 5% relative difference are ok,
		// else print warning:
		if(realtime > cputime + 10 &&
			realtime > cputime + (cputime / 20)) {
			out_c::verbose_real_gt_cpu(realtime, cputime);
			Warn = true;
		}
		return realtime;
	}

	// CPU-time is more than realtime:
	else {
		// Print warning if difference is more than 10ms:
		if(cputime > realtime + 10) {
			out_c::verbose_cpu_gt_real(realtime, cputime);
			Warn = true;
		}
		return realtime;
	}
}


//=============================================================================
// Test of Data Structures for Relations (Static Methods):
//=============================================================================


//-----------------------------------------------------------------------------
// Test the set with one int column, implemented as dynamic hash table:
//-----------------------------------------------------------------------------

bool test_c::set_1d(test_t test)
{
	set_1_c rel("test_rel");
	test_set1_c<set_1_t> rtest(&rel, test->Arg);
	bool ok = test->run_rtest(&rtest, &rel);
	if(out_c::show_dump())
		rtest.dump();
	return ok;
}

//-----------------------------------------------------------------------------
// Test the set with one int column, implemented as extendible hash table:
//-----------------------------------------------------------------------------

bool test_c::set_1e(test_t test)
{
	set_1e_c rel("test_rel");
	test_set1_c<set_1e_t> rtest(&rel, test->Arg);
	bool ok = test->run_rtest(&rtest, &rel);
	if(out_c::show_dump())
		rtest.dump();
	return ok;
}

//=============================================================================
// All Benchmark Functions (Static Methods):
//=============================================================================


//-----------------------------------------------------------------------------
// Do the "DBLP" Benchmark, Version 1:
//-----------------------------------------------------------------------------

bool test_c::bench_dblp_1(test_t test)
{
	str_t bench_name = bench_dblp_1_c::name();
	str_t file_name = bench_dblp_1_c::file(test->Arg);
	long expected_result = bench_dblp_1_c::result(test->Arg);
	bool ok = test->run_bench_static(bench_name, &(bench_dblp_1_c::run),
					file_name, expected_result);
	return ok;
}

//-----------------------------------------------------------------------------
// Do the "DBLP" Benchmark, Version 2 (with Cursor Interface):
//-----------------------------------------------------------------------------

bool test_c::bench_dblp_2(test_t test)
{
	bench_dblp_2_c bench;
	bool ok = test->run_bench(&bench);
	if(out_c::show_dump())
		bench.dump();
	return ok;
}

//-----------------------------------------------------------------------------
// Do the "DBLP" Benchmark, Version 3 (with Separate Objects):
//-----------------------------------------------------------------------------

bool test_c::bench_dblp_3(test_t test)
{
	bench_dblp_3_c bench;
	bool ok = test->run_bench(&bench);
	if(out_c::show_dump())
		bench.dump();
	return ok;
}

//-----------------------------------------------------------------------------
// Do the "Join 1" Benchmark for a(X,Y), Ver. 1
//-----------------------------------------------------------------------------

bool test_c::bench_j1axy_1(test_t test)
{
	bench_j1axy_1_c bench;
	bool ok = test->run_bench(&bench);
	if(out_c::show_dump())
		bench.dump();
	return ok;
}

//-----------------------------------------------------------------------------
// Do the "Join 1" Benchmark for a(X,Y), Ver. 2
//-----------------------------------------------------------------------------

bool test_c::bench_j1axy_2(test_t test)
{
	bench_j1axy_2_c bench;
	bool ok = test->run_bench(&bench);
	if(out_c::show_dump())
		bench.dump();
	return ok;
}

//-----------------------------------------------------------------------------
// Do the "Join 1" Benchmark for a(X,Y), Ver. 3
//-----------------------------------------------------------------------------

bool test_c::bench_j1axy_3(test_t test)
{
	str_t bench_name = bench_j1axy_3_c::name();
	str_t file_name = bench_j1axy_3_c::file(test->Arg);
	long expected_result = bench_j1axy_3_c::result(test->Arg);
	bool ok = test->run_bench_static(bench_name, &(bench_j1axy_3_c::run),
					file_name, expected_result);
	return ok;
}

//-----------------------------------------------------------------------------
// Do the "Join 1" Benchmark for a(X,Y), Ver. 4
//-----------------------------------------------------------------------------

bool test_c::bench_j1axy_4(test_t test)
{
	bench_j1axy_4_c bench;
	bool ok = test->run_bench(&bench);
	if(out_c::show_dump())
		bench.dump();
	return ok;
}

//-----------------------------------------------------------------------------
// Do the "Join 1" Benchmark for a(X,Y), Ver. 5
//-----------------------------------------------------------------------------

bool test_c::bench_j1axy_5(test_t test)
{
	bench_j1axy_5_c bench;
	bool ok = test->run_bench(&bench);
	if(out_c::show_dump())
		bench.dump();
	return ok;
}

//-----------------------------------------------------------------------------
// Do the "Join 1" Benchmark for a(X,Y), Ver. 6
//-----------------------------------------------------------------------------

bool test_c::bench_j1axy_6(test_t test)
{
	bench_j1axy_6_c bench;
	bool ok = test->run_bench(&bench);
	if(out_c::show_dump())
		bench.dump();
	return ok;
}

//-----------------------------------------------------------------------------
// Do the "Join 1" Benchmark for a(X,Y), Ver. 7
//-----------------------------------------------------------------------------

bool test_c::bench_j1axy_7(test_t test)
{
	bench_j1axy_7_c bench;
	bool ok = test->run_bench(&bench);
	if(out_c::show_dump())
		bench.dump();
	return ok;
}

//-----------------------------------------------------------------------------
// Do the "Join 1" Benchmark for a(X,Y), Ver. 8
//-----------------------------------------------------------------------------

bool test_c::bench_j1axy_8(test_t test)
{
	bench_j1axy_8_c bench;
	bool ok = test->run_bench(&bench);
	if(out_c::show_dump())
		bench.dump();
	return ok;
}

//-----------------------------------------------------------------------------
// Do the "Join 1" Benchmark for a(X,Y), Ver. 9
//-----------------------------------------------------------------------------

bool test_c::bench_j1axy_9(test_t test)
{
	bench_j1axy_9_c bench;
	bool ok = test->run_bench(&bench);
	if(out_c::show_dump())
		bench.dump();
	return ok;
}

//-----------------------------------------------------------------------------
// Do the "tc(X,Y)" Benchmark, Version 1 (with static function, local vars):
//-----------------------------------------------------------------------------

bool test_c::bench_tcff_1(test_t test)
{
	str_t bench_name = bench_tcff_1_c::name();
	str_t file_name = bench_tcff_1_c::file(test->Arg);
	long expected_result = bench_tcff_1_c::result(test->Arg);
	bool ok = test->run_bench_static(bench_name, &(bench_tcff_1_c::run),
					file_name, expected_result);
	return ok;
}

//-----------------------------------------------------------------------------
// Do the "tc(X, Y)" Benchmark, Version 2 (with Cursor Interface):
//-----------------------------------------------------------------------------

bool test_c::bench_tcff_2(test_t test)
{
	bench_tcff_2_c bench;
	bool ok = test->run_bench(&bench);
	if(out_c::show_dump())
		bench.dump();
	return ok;
}

//-----------------------------------------------------------------------------
// Do the "tc(X, Y)" Benchmark, Version 3 (Seminaive Evaluation):
//-----------------------------------------------------------------------------

bool test_c::bench_tcff_3(test_t test)
{
	bench_tcff_3_c bench;
	bool ok = test->run_bench(&bench);
	if(out_c::show_dump())
		bench.dump();
	return ok;
}

//-----------------------------------------------------------------------------
// Do the "tc(X, Y)" Benchmark, Version 4 (Recursive Functions):
//-----------------------------------------------------------------------------

bool test_c::bench_tcff_4(test_t test)
{
	bench_tcff_4_c bench;
	bool ok = test->run_bench(&bench);
	if(out_c::show_dump())
		bench.dump();
	return ok;
}

//-----------------------------------------------------------------------------
// Do the "tc(1,X)" Benchmark, Version 1:
//-----------------------------------------------------------------------------

bool test_c::bench_tcbf_1(test_t test)
{
	str_t bench_name = bench_tcbf_1_c::name();
	str_t file_name = bench_tcbf_1_c::file(test->Arg);
	long expected_result = bench_tcbf_1_c::result(test->Arg);
	bool ok = test->run_bench_static(bench_name, &(bench_tcbf_1_c::run),
					file_name, expected_result);
	return ok;
}

//-----------------------------------------------------------------------------
// Do the "tc(1,X)" Benchmark, Version 2:
//-----------------------------------------------------------------------------

bool test_c::bench_tcbf_2(test_t test)
{
	bench_tcbf_2_c bench;
	bool ok = test->run_bench(&bench);
	if(out_c::show_dump())
		bench.dump();
	return ok;
}

//-----------------------------------------------------------------------------
// Do the "tc(1,X)" Benchmark, Version 4:
//-----------------------------------------------------------------------------

bool test_c::bench_tcbf_4(test_t test)
{
	bench_tcbf_4_c bench;
	bool ok = test->run_bench(&bench);
	if(out_c::show_dump())
		bench.dump();
	return ok;
}

//-----------------------------------------------------------------------------
// Do the "Wine Ontology" Benchmark:
//-----------------------------------------------------------------------------

bool test_c::bench_wine(test_t test)
{
#ifdef VER_WINE_ENABLED
	bench_wine_c bench;
	bool ok = test->run_bench(&bench);
	if(out_c::show_dump())
		bench.dump();
	return ok;
#else
	// Use test object to avoid warning:
	test->Ok = false;

	// Print error message: This version is compiled without Wine Ontology:
	err_c::wine_bench_disabled();
	return false;
#endif
}


//-----------------------------------------------------------------------------
// This is the function of the last entry in the test table.
//-----------------------------------------------------------------------------

bool test_c::end_marker(test_t test)
{
	CHECK_IMPOSSIBLE("test_c::end_marker should never be called");
	test->arg_unused(); // This is to avoid warnings about unused args.
	return false;
}


//=============================================================================
// Data Load Tests (Partial Benchmarks):
//=============================================================================


//-----------------------------------------------------------------------------
// Load DBLP data file:
//-----------------------------------------------------------------------------

bool test_c::load_dblp(test_t test)
{
	// Argument is not used for this test:
	test->arg_unused();

	// File with DBLP data:
	str_t testfile = VER_FILE_DBLP;

	// Print verbose message:
	out_c::verbose_msg("Loading '");
	out_c::verbose_msg(testfile);
	out_c::verbose_msg("' ...\n");

	// --------------------------------------------------------------------
	// Domains:
	// --------------------------------------------------------------------

	// Make string table for first argument:
	strtab_c doc_ids("String Table for Arg.1");

	// Make atom table for second argument:
	atomtab_c attrs("Atom Table for Arg.2");

	// Construct atoms which are relevant for selection condition:
	int title  = attrs.put("title");
	int year   = attrs.put("year");
	int author = attrs.put("author");
	int month  = attrs.put("month");

	// Make string table for third argument:
	strtab_c str_values("String Table for Arg.3");

	// --------------------------------------------------------------------
	// Relations for EDB body literals:
	// --------------------------------------------------------------------

	list_2_c  att_title("ATT_TITLE");
	rel_n_n_c att_year("ATT_YEAR");
	rel_n_n_c att_author("ATT_AUTHOR");
	rel_n_n_c att_month("ATT_MONTH");

	// --------------------------------------------------------------------
	// Load data file, fill domains and EDB relations:
	// --------------------------------------------------------------------

	// Load data file:
	bool ok = load_dblp_c::load(testfile,
			&doc_ids, &attrs, &str_values,
			title, year, author, month,
			&att_title, &att_year, &att_author, &att_month);
	if(!ok) {
		test->test_failed("Load failed");
		return false;
	}

	// Print memory pools if requested:
	out_c::verbose_nl();
	out_c::mem_pools_tab();

	// Print relation statistics if requested:
	if(out_c::show_rel()) {
		out_c::rel_size_headline();
		out_c::rel_size_nonstd("ATT_TITLE", att_title.num_rows(),
					att_title.num_pages());
		out_c::rel_size_nonstd("ATT_YEAR", att_year.num_rows(),
					att_year.num_pages());
		out_c::rel_size_nonstd("ATT_AUTHOR", att_author.num_rows(),
					att_author.num_pages());
		out_c::rel_size_nonstd("ATT_MONTH", att_month.num_rows(),
					att_month.num_pages());
		out_c::rel_size_strtab(&doc_ids);
		out_c::rel_size_atomtab(&attrs);
		out_c::rel_size_strtab(&str_values);
		out_c::rel_size_done();
	}

	// Ok:
	return true;
}

//-----------------------------------------------------------------------------
// Load transitive closure data file:
//-----------------------------------------------------------------------------

bool test_c::load_tc(test_t test)
{
	// Argument is not used for this test:
	test->arg_unused();

	// Test data file:
	str_t testfile = VER_FILE_TC_1M_D2_CYC;

	// Print verbose message:
	out_c::verbose_msg("Loading '");
	out_c::verbose_msg(testfile);
	out_c::verbose_msg("' ...\n");

	// Relations for EDB body literals:
	rel_ii_c  par_ff("par_ff");
	rel_n_n_c par_fb("par_fb");

	// Load data file, fill EDB relations:
	bool ok = load_tc_c::load(testfile, &par_ff, &par_fb);
	if(!ok) {
		test->test_failed("Load failed.");
		return false;
	}

	// Print memory pools if requested:
	out_c::verbose_nl();
	out_c::mem_pools_tab();

	// Print relation statistics if requested:
	if(out_c::show_rel()) {
		out_c::rel_size_headline();
		out_c::rel_size_nonstd("par_ff", par_ff.num_rows(),
						par_ff.num_pages());
		out_c::rel_size(&par_fb);
		out_c::rel_size_done();
	}

	// Ok:
	return true;
}

//-----------------------------------------------------------------------------
// Load transitive closure data file, new version with general loader:
//-----------------------------------------------------------------------------

bool test_c::load_tc_gen(test_t test)
{
	// Argument is not used for this test:
	test->arg_unused();

	// Test data file:
	str_t testfile = VER_FILE_TC_1M_D2_CYC;

	// Print verbose message:
	out_c::verbose_msg("Loading '");
	out_c::verbose_msg(testfile);
	out_c::verbose_msg("' with general loader ...\n");

	// Measure time for loading data file:
	perf_c timer;
	timer.start();

	// Relations for EDB body literals:
	list_2_c  par_ff("par_ff");
	rel_n_n_c par_fb("par_fb");

	// Loader object (will contain data about predicates):
	load_c loader;

	// Predicate (input to loader):
	pred_c par_p("par", ARGTYPE_INT, ARGTYPE_INT);
	par_p.add_rel(&par_ff);
	par_p.add_rel(&par_fb, BIND_FB);

	// Add predicate definition to loader:
	loader.add(&par_p);

	// Load data file:
	bool ok = loader.load(testfile);
	if(!ok) {
		test->test_failed("Load failed.");
		return false;
	}

	// Print memory allocation statistics:
	out_c::verbose_nl();
	out_c::mem_pools_tab();

	// Print relation statistics if requested:
	if(out_c::show_rel()) {
		out_c::rel_size_headline();
		out_c::rel_size(&par_ff);
		out_c::rel_size(&par_fb);
		out_c::rel_size_done();
	}

	// Ok:
	return true;
}


//=============================================================================
// List of all tests:
//=============================================================================


test_c test_c::Tests[] = {
	test_c("dblp_1_test", test_c::bench_dblp_1, 0,
		"DBLP Bench. [V1: Handwritten loader, RelLocalVars] TestFile"),
	test_c("dblp_1", test_c::bench_dblp_1, 1,
		"DBLP Bench. [V1: Handwritten loader, Rel in Local Vars]"),
	test_c("dblp_1_big", test_c::bench_dblp_1, 2,
		"DBLP Bench. [V1: Handwritten loader, Rel in Local Vars] 3GB"),
	test_c("dblp_2_test", test_c::bench_dblp_2, 0,
		"DBLP Bench. [V2: Cursor Interface, Rel+Var->Attr] TestFile"),
	test_c("dblp_2", test_c::bench_dblp_2, 1,
		"DBLP Bench. [V2: Cursor Interface, Rel+Var->Attr]"),
	test_c("dblp_2_big", test_c::bench_dblp_2, 2,
		"DBLP Bench. [V2: Cursor Interface, Rel+Var->Attr] 3GB"),
	test_c("dblp_3_test", test_c::bench_dblp_3, 0,
		"DBLP Bench. [V3: Separate Objects] TestFile"),
	test_c("dblp_3", test_c::bench_dblp_3, 1,
		"DBLP Bench. [V3: Separate Objects]"),
	test_c("dblp_3_big", test_c::bench_dblp_3, 2,
		"DBLP Bench. [V3: Separate Objects] 3GB"),

	test_c("j1axy_1", test_c::bench_j1axy_1, 0,
		"Join1, a(X,Y) Bench. [V1: Rel in Attr, no dupcheck] test"),
	test_c("j1axy_1_10k", test_c::bench_j1axy_1, 1,
		"Join1, a(X,Y) Bench. [V1: Rel in Attr, no dupcheck] 10K*5"),
	test_c("j1axy_1_50k", test_c::bench_j1axy_1, 2,
		"Join1, a(X,Y) Bench. [V1: Rel in Attr, no dupcheck] 50K*5"),
	test_c("j1axy_1_250k", test_c::bench_j1axy_1, 3,
		"Join1, a(X,Y) Bench. [V1: Rel in Attr, no dupcheck] 250K*5"),

	test_c("j1axy_2", test_c::bench_j1axy_2, 0,
		"Join1, a(X,Y) Bench. [V2: RelAttr, no goto, no dupch] test"),
	test_c("j1axy_2_10k", test_c::bench_j1axy_2, 1,
		"Join1, a(X,Y) Bench. [V2: RelAttr, no goto, no dupch] 10K*5"),
	test_c("j1axy_2_50k", test_c::bench_j1axy_2, 2,
		"Join1, a(X,Y) Bench. [V2: RelAttr, no goto, no dupch] 50K*5"),
	test_c("j1axy_2_250k", test_c::bench_j1axy_2, 3,
		"Join1, a(X,Y) Bench. [V2: RelAttr, no goto, no dupch] 250K*5"),

	test_c("j1axy_3", test_c::bench_j1axy_3, 0,
		"Join1, a(X,Y) Bench. [V3: Rel:LocalVar, static, NoDC] test"),
	test_c("j1axy_3_10k", test_c::bench_j1axy_3, 1,
		"Join1, a(X,Y) Bench. [V3: Rel:LocalVar, static, NoDC] 10K*5"),
	test_c("j1axy_3_50k", test_c::bench_j1axy_3, 2,
		"Join1, a(X,Y) Bench. [V3: Rel:LocalVar, static, NoDC] 50K*5"),
	test_c("j1axy_3_250k", test_c::bench_j1axy_3, 3,
		"Join1, a(X,Y) Bench. [V3: Rel:LocalVar, static, NoDC] 250K*5"),

	test_c("j1axy_4", test_c::bench_j1axy_4, 0,
		"Join1, a(X,Y) Bench. [V4: RelAttr, dupchk: b1,b2,c1] test"),
	test_c("j1axy_4_10k", test_c::bench_j1axy_4, 1,
		"Join1, a(X,Y) Bench. [V4: RelAttr, dupchk: b1,b2,c1] 10K*5"),
	test_c("j1axy_4_50k", test_c::bench_j1axy_4, 2,
		"Join1, a(X,Y) Bench. [V4: RelAttr, dupchk: b1,b2,c1] 50K*5"),
	test_c("j1axy_4_250k", test_c::bench_j1axy_4, 3,
		"Join1, a(X,Y) Bench. [V4: RelAttr, dupchk: b1,b2,c1] 250K*5"),

	test_c("j1axy_5", test_c::bench_j1axy_5, 0,
		"Join1, a(X,Y) Bench. [V5: RelAttr, -goto, fullDupCh] test"),
	test_c("j1axy_5_10k", test_c::bench_j1axy_5, 1,
		"Join1, a(X,Y) Bench. [V5: RelAttr, -goto, fullDupCh] 10K*5"),
	test_c("j1axy_5_50k", test_c::bench_j1axy_5, 2,
		"Join1, a(X,Y) Bench. [V5: RelAttr, -goto, fullDupCh] 50K*5"),
	test_c("j1axy_5_250k", test_c::bench_j1axy_5, 3,
		"Join1, a(X,Y) Bench. [V5: RelAttr, -goto, fullDupCh] 250K*5"),

	test_c("j1axy_6", test_c::bench_j1axy_6, 0,
		"Join1, a(X,Y) Bench. [V6: no goto, bitmap-dupcheck], test"),
	test_c("j1axy_6_10k", test_c::bench_j1axy_6, 1,
		"Join1, a(X,Y) Bench. [V6: no goto, bitmap-dupcheck], 10K*5"),
	test_c("j1axy_6_50k", test_c::bench_j1axy_6, 2,
		"Join1, a(X,Y) Bench. [V6: no goto, bitmap-dupcheck], 50K*5"),
	test_c("j1axy_6_250k", test_c::bench_j1axy_6, 3,
		"Join1, a(X,Y) Bench. [V6: no goto, bitmap-dupcheck], 250K*5"),

	test_c("j1axy_7", test_c::bench_j1axy_7, 0,
		"Join1, a(X,Y) Bench. [V7: goto, bitmap-dupC, cursor], test"),
	test_c("j1axy_7_10k", test_c::bench_j1axy_7, 1,
		"Join1, a(X,Y) Bench. [V7: goto, bitmap-dupC, cursor], 10K*5"),
	test_c("j1axy_7_50k", test_c::bench_j1axy_7, 2,
		"Join1, a(X,Y) Bench. [V7: goto, bitmap-dupC, cursor], 50K*5"),
	test_c("j1axy_7_250k", test_c::bench_j1axy_7, 3,
		"Join1, a(X,Y) Bench. [V7: goto, bitmap-dupC, cursor], 250K*5"),

	test_c("j1axy_8", test_c::bench_j1axy_8, 0,
		"Join1, a(X,Y) Bench. [V8: functions, bitmap-dupC], test"),
	test_c("j1axy_8_10k", test_c::bench_j1axy_8, 1,
		"Join1, a(X,Y) Bench. [V8: functions, bitmap-dupC], 10K*5"),
	test_c("j1axy_8_50k", test_c::bench_j1axy_8, 2,
		"Join1, a(X,Y) Bench. [V8: functions, bitmap-dupC], 50K*5"),
	test_c("j1axy_8_250k", test_c::bench_j1axy_8, 3,
		"Join1, a(X,Y) Bench. [V8: functions, bitmap-dupC], 250K*5"),

	test_c("j1axy_9", test_c::bench_j1axy_9, 0,
		"Join1, a(X,Y) Bench. [V9: fun without args, bitmap], test"),
	test_c("j1axy_9_10k", test_c::bench_j1axy_9, 1,
		"Join1, a(X,Y) Bench. [V9: fun without args, bitmap], 10K*5"),
	test_c("j1axy_9_50k", test_c::bench_j1axy_9, 2,
		"Join1, a(X,Y) Bench. [V9: fun without args, bitmap], 50K*5"),
	test_c("j1axy_9_250k", test_c::bench_j1axy_9, 3,
		"Join1, a(X,Y) Bench. [V9: fun without args, bitmap], 250K*5"),

	test_c("tcff_1_test", test_c::bench_tcff_1, 0,
		"tc(X,Y) Bench. [V1: RelLocalVar], Small Test File"),
	test_c("tcff_1_f1", test_c::bench_tcff_1, 1,
		"tc(X,Y) Bench. [V1: RelLocalVar], File 1: 50K-nocyc"),
	test_c("tcff_1_f2", test_c::bench_tcff_1, 2,
		"tc(X,Y) Bench. [V1: RelLocalVar], File 2: 50K-cyc"),
	test_c("tcff_1_f3", test_c::bench_tcff_1, 3,
		"tc(X,Y) Bench. [V1: RelLocalVar], File 3: 250K-nocyc"),
	test_c("tcff_1_f4", test_c::bench_tcff_1, 4,
		"tc(X,Y) Bench. [V1: RelLocalVar], File 4: 250K-cyc"),
	test_c("tcff_1_f5", test_c::bench_tcff_1, 5,
		"tc(X,Y) Bench. [V1: RelLocalVar], File 5: 500K-nocyc"),
	test_c("tcff_1_f6", test_c::bench_tcff_1, 6,
		"tc(X,Y) Bench. [V1: RelLocalVar], File 6: 500K-cyc"),
	test_c("tcff_1_f7", test_c::bench_tcff_1, 7,
		"tc(X,Y) Bench. [V1: RelLocalVar], File 7: 2k-500K-nocyc"),
	test_c("tcff_1_f8", test_c::bench_tcff_1, 8,
		"tc(X,Y) Bench. [V1: RelLocalVar], File 8: 2k-500K-cyc"),
	test_c("tcff_1_f9", test_c::bench_tcff_1, 9,
		"tc(X,Y) Bench. [V1: RelLocalVar], File 9: 2k-1M-nocyc"),
	test_c("tcff_1_f10", test_c::bench_tcff_1, 10,
		"tc(X,Y) Bench. [V1: RelLocalVar], File 10: 2k-1M-cyc"),

	test_c("tcff_2_test", test_c::bench_tcff_2, 0,
		"tc(X,Y) Bench. [V2: Cursor Interface], Small Test File"),
	test_c("tcff_2_f1", test_c::bench_tcff_2, 1,
		"tc(X,Y) Bench. [V2: Cursor Interface], File 1: 50K-nocyc"),
	test_c("tcff_2_f2", test_c::bench_tcff_2, 2,
		"tc(X,Y) Bench. [V2: Cursor Interface], File 2: 50K-cyc"),
	test_c("tcff_2_f3", test_c::bench_tcff_2, 3,
		"tc(X,Y) Bench. [V2: Cursor Interface], File 3: 250K-nocyc"),
	test_c("tcff_2_f4", test_c::bench_tcff_2, 4,
		"tc(X,Y) Bench. [V2: Cursor Interface], File 4: 250K-cyc"),
	test_c("tcff_2_f5", test_c::bench_tcff_2, 5,
		"tc(X,Y) Bench. [V2: Cursor Interface], File 5: 500K-nocyc"),
	test_c("tcff_2_f6", test_c::bench_tcff_2, 6,
		"tc(X,Y) Bench. [V2: Cursor Interface], File 6: 500K-cyc"),
	test_c("tcff_2_f7", test_c::bench_tcff_2, 7,
		"tc(X,Y) Bench. [V2: Cursor Interface], File 7: 2k-500K-nocyc"),
	test_c("tcff_2_f8", test_c::bench_tcff_2, 8,
		"tc(X,Y) Bench. [V2: Cursor Interface], File 8: 2k-500K-cyc"),
	test_c("tcff_2_f9", test_c::bench_tcff_2, 9,
		"tc(X,Y) Bench. [V2: Cursor Interface], File 9: 2k-1M-nocyc"),
	test_c("tcff_2_f10", test_c::bench_tcff_2, 10,
		"tc(X,Y) Bench. [V2: Cursor Interface], File 10: 2k-1M-cyc"),

	test_c("tcff_3_test", test_c::bench_tcff_3, 0,
		"tc(X,Y) Bench. [V3: Seminaive Eval.], Small Test File"),
	test_c("tcff_3_f1", test_c::bench_tcff_3, 1,
		"tc(X,Y) Bench. [V3: Seminaive Eval.], File 1: 50K-nocyc"),
	test_c("tcff_3_f2", test_c::bench_tcff_3, 2,
		"tc(X,Y) Bench. [V3: Seminaive Eval.], File 2: 50K-cyc"),
	test_c("tcff_3_f3", test_c::bench_tcff_3, 3,
		"tc(X,Y) Bench. [V3: Seminaive Eval.], File 3: 250K-nocyc"),
	test_c("tcff_3_f4", test_c::bench_tcff_3, 4,
		"tc(X,Y) Bench. [V3: Seminaive Eval.], File 4: 250K-cyc"),
	test_c("tcff_3_f5", test_c::bench_tcff_3, 5,
		"tc(X,Y) Bench. [V3: Seminaive Eval.], File 5: 500K-nocyc"),
	test_c("tcff_3_f6", test_c::bench_tcff_3, 6,
		"tc(X,Y) Bench. [V3: Seminaive Eval.], File 6: 500K-cyc"),
	test_c("tcff_3_f7", test_c::bench_tcff_3, 7,
		"tc(X,Y) Bench. [V3: Seminaive Eval.], File 7: 2k-500K-nocyc"),
	test_c("tcff_3_f8", test_c::bench_tcff_3, 8,
		"tc(X,Y) Bench. [V3: Seminaive Eval.], File 8: 2k-500K-cyc"),
	test_c("tcff_3_f9", test_c::bench_tcff_3, 9,
		"tc(X,Y) Bench. [V3: Seminaive Eval.], File 9: 2k-1M-nocyc"),
	test_c("tcff_3_f10", test_c::bench_tcff_3, 10,
		"tc(X,Y) Bench. [V3: Seminaive Eval.], File 10: 2k-1M-cyc"),

	test_c("tcff_4_test", test_c::bench_tcff_4, 0,
		"tc(X,Y) Bench. [V4: Functions], Small Test File"),
	test_c("tcff_4_f1", test_c::bench_tcff_4, 1,
		"tc(X,Y) Bench. [V4: Functions], File 1: 50K-nocyc"),
	test_c("tcff_4_f2", test_c::bench_tcff_4, 2,
		"tc(X,Y) Bench. [V4: Functions], File 2: 50K-cyc"),
	test_c("tcff_4_f3", test_c::bench_tcff_4, 3,
		"tc(X,Y) Bench. [V4: Functions], File 3: 250K-nocyc"),
	test_c("tcff_4_f4", test_c::bench_tcff_4, 4,
		"tc(X,Y) Bench. [V4: Functions], File 4: 250K-cyc"),
	test_c("tcff_4_f5", test_c::bench_tcff_4, 5,
		"tc(X,Y) Bench. [V4: Functions], File 5: 500K-nocyc"),
	test_c("tcff_4_f6", test_c::bench_tcff_4, 6,
		"tc(X,Y) Bench. [V4: Functions], File 6: 500K-cyc"),
	test_c("tcff_4_f7", test_c::bench_tcff_4, 7,
		"tc(X,Y) Bench. [V4: Functions], File 7: 2k-500K-nocyc"),
	test_c("tcff_4_f8", test_c::bench_tcff_4, 8,
		"tc(X,Y) Bench. [V4: Functions], File 8: 2k-500K-cyc"),
	test_c("tcff_4_f9", test_c::bench_tcff_4, 9,
		"tc(X,Y) Bench. [V4: Functions], File 9: 2k-1M-nocyc"),
	test_c("tcff_4_f10", test_c::bench_tcff_4, 10,
		"tc(X,Y) Bench. [V4: Functions], File 10: 2k-1M-cyc"),

	test_c("tcbf_1_test", test_c::bench_tcbf_1, 0,
		"tc(1,X) Bench. [V1: RelLocalVar], Small Test File"),
	test_c("tcbf_1_f1", test_c::bench_tcbf_1, 1,
		"tc(1,X) Bench. [V1: RelLocalVar], File 1: 50K-nocyc"),
	test_c("tcbf_1_f2", test_c::bench_tcbf_1, 2,
		"tc(1,X) Bench. [V1: RelLocalVar], File 2: 50K-cyc"),
	test_c("tcbf_1_f3", test_c::bench_tcbf_1, 3,
		"tc(1,X) Bench. [V1: RelLocalVar], File 3: 250K-nocyc"),
	test_c("tcbf_1_f4", test_c::bench_tcbf_1, 4,
		"tc(1,X) Bench. [V1: RelLocalVar], File 4: 250K-cyc"),
	test_c("tcbf_1_f5", test_c::bench_tcbf_1, 5,
		"tc(1,X) Bench. [V1: RelLocalVar], File 5: 500K-nocyc"),
	test_c("tcbf_1_f6", test_c::bench_tcbf_1, 6,
		"tc(1,X) Bench. [V1: RelLocalVar], File 6: 500K-cyc"),
	test_c("tcbf_1_f7", test_c::bench_tcbf_1, 7,
		"tc(1,X) Bench. [V1: RelLocalVar], File 7: 2k-500K-nocyc"),
	test_c("tcbf_1_f8", test_c::bench_tcbf_1, 8,
		"tc(1,X) Bench. [V1: RelLocalVar], File 8: 2k-500K-cyc"),
	test_c("tcbf_1_f9", test_c::bench_tcbf_1, 9,
		"tc(1,X) Bench. [V1: RelLocalVar], File 9: 2k-1M-nocyc"),
	test_c("tcbf_1_f10", test_c::bench_tcbf_1, 10,
		"tc(1,X) Bench. [V1: RelLocalVar], File 10: 2k-1M-cyc"),

	test_c("tcbf_2_test", test_c::bench_tcbf_2, 0,
		"tc(1,X) Bench. [V2: Cursor, RelAttr], Small Test File"),
	test_c("tcbf_2_f1", test_c::bench_tcbf_2, 1,
		"tc(1,X) Bench. [V2: Cursor, RelAttr], File 1: 50K-nocyc"),
	test_c("tcbf_2_f2", test_c::bench_tcbf_2, 2,
		"tc(1,X) Bench. [V2: Cursor, RelAttr], File 2: 50K-cyc"),
	test_c("tcbf_2_f3", test_c::bench_tcbf_2, 3,
		"tc(1,X) Bench. [V2: Cursor, RelAttr], File 3: 250K-nocyc"),
	test_c("tcbf_2_f4", test_c::bench_tcbf_2, 4,
		"tc(1,X) Bench. [V2: Cursor, RelAttr], File 4: 250K-cyc"),
	test_c("tcbf_2_f5", test_c::bench_tcbf_2, 5,
		"tc(1,X) Bench. [V2: Cursor, RelAttr], File 5: 500K-nocyc"),
	test_c("tcbf_2_f6", test_c::bench_tcbf_2, 6,
		"tc(1,X) Bench. [V2: Cursor, RelAttr], File 6: 500K-cyc"),
	test_c("tcbf_2_f7", test_c::bench_tcbf_2, 7,
		"tc(1,X) Bench. [V2: Cursor, RelAttr], File 7: 2k-500K-nocyc"),
	test_c("tcbf_2_f8", test_c::bench_tcbf_2, 8,
		"tc(1,X) Bench. [V2: Cursor, RelAttr], File 8: 2k-500K-cyc"),
	test_c("tcbf_2_f9", test_c::bench_tcbf_2, 9,
		"tc(1,X) Bench. [V2: Cursor, RelAttr], File 9: 2k-1M-nocyc"),
	test_c("tcbf_2_f10", test_c::bench_tcbf_2, 10,
		"tc(1,X) Bench. [V2: Cursor, RelAttr], File 10: 2k-1M-cyc"),

	test_c("tcbf_4_test", test_c::bench_tcbf_4, 0,
		"tc(1,X) Bench. [V4: Procedures], Small Test File"),
	test_c("tcbf_4_f1", test_c::bench_tcbf_4, 1,
		"tc(1,X) Bench. [V4: Procedures], File 1: 50K-nocyc"),
	test_c("tcbf_4_f2", test_c::bench_tcbf_4, 2,
		"tc(1,X) Bench. [V4: Procedures], File 2: 50K-cyc"),
	test_c("tcbf_4_f3", test_c::bench_tcbf_4, 3,
		"tc(1,X) Bench. [V4: Procedures], File 3: 250K-nocyc"),
	test_c("tcbf_4_f4", test_c::bench_tcbf_4, 4,
		"tc(1,X) Bench. [V4: Procedures], File 4: 250K-cyc"),
	test_c("tcbf_4_f5", test_c::bench_tcbf_4, 5,
		"tc(1,X) Bench. [V4: Procedures], File 5: 500K-nocyc"),
	test_c("tcbf_4_f6", test_c::bench_tcbf_4, 6,
		"tc(1,X) Bench. [V4: Procedures], File 6: 500K-cyc"),
	test_c("tcbf_4_f7", test_c::bench_tcbf_4, 7,
		"tc(1,X) Bench. [V4: Procedures], File 7: 2k-500K-nocyc"),
	test_c("tcbf_4_f8", test_c::bench_tcbf_4, 8,
		"tc(1,X) Bench. [V4: Procedures], File 8: 2k-500K-cyc"),
	test_c("tcbf_4_f9", test_c::bench_tcbf_4, 9,
		"tc(1,X) Bench. [V4: Procedures], File 9: 2k-1M-nocyc"),
	test_c("tcbf_4_f10", test_c::bench_tcbf_4, 10,
		"tc(1,X) Bench. [V4: Procedures], File 10: 2k-1M-cyc"),

	test_c("wine", test_c::bench_wine, 0,
		"Wine Ontology Benchmark"),

	test_c("LoadDBLP", test_c::load_dblp, 0,
		"Load DBLP data file (with hand-crafted loader)"),
	test_c("LoadTC", test_c::load_tc, 0,
		"Load transitive closure data file (with handcrafted loader)"),
	test_c("LoadTCGen", test_c::load_tc_gen, 0,
		"Load transitive closure data file (with general loader)"),

	test_c("set_1e_test", test_c::set_1e, 0,
		"Set with 1 column, extendible Hash table, 10 rows seq."),
	test_c("set_1e_s", test_c::set_1e, 1,
		"Set with 1 column, extendible Hash table, 1 mio rows seq."),
	test_c("set_1e_m", test_c::set_1e, 2,
		"Set with 1 column, extendible Hash table, 1 mio rows mix."),

	test_c("set_1d_test", test_c::set_1d, 0,
		"Set with 1 column, dynamic Hash table, 10 rows seq."),
	test_c("set_1d_s", test_c::set_1d, 1,
		"Set with 1 column, dynamic Hash table, 1 mio rows seq."),
	test_c("set_1d_m", test_c::set_1d, 2,
		"Set with 1 column, dynamic Hash table, 1 mio rows mix."),

	// This must be the last entry in the table:
	test_c(".", test_c::end_marker, 1, "End Marker - No real test"),
};

